﻿// Copyright (c) 2015 NEC Corporation
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy of this
// software and associated documentation files (the "Software"), to deal in the Software
// without restriction, including without limitation the rights to use, copy, modify, merge,
// publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons
// to whom the Software is furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in all copies or
// substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
// INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
// PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE
// FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
// OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
// DEALINGS IN THE SOFTWARE.
using System;
using System.Diagnostics;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.IO;
using ICSharpCode.NRefactory.CSharp;
using System.Reflection;
using ICSharpCode.NCMCSParser.Properties;


namespace ICSharpCode.NCMCSParser
{
    /**
     * 意味解析(Visitor)クラス.
     * C#Parserのパーサ結果（ノード）を元に関数や条件式を解析する。
     */
    public class NCMCSVisitor : DepthFirstAstVisitor<object, object>
    {
        bool flgOutsideFunction = false;
        bool flgFunctionBlock = false;
        bool flgBranchCondition = false;
        private List<String> scopenames = new List<String>();

        private Stack<TokenData> ppIfStack = new Stack<TokenData>();
        private List<string> targetCodeList = new List<string>();


        private List<int> commentLineList = null;


        /**
         * Debugビルド用に解析対象コードを読み込む。
         */
        [Conditional("DEBUG")]
        void loadTargetCode(object data)
        {
            string filename = ((NCMStructData)data).parse_filepath;
            using (StreamReader sr = new StreamReader(filename))
            {
                int i = 0;
                while (sr.EndOfStream == false)
                {
                    targetCodeList.Add(sr.ReadLine().Replace("\t", "    "));
                    i++;
                }
            }
        }

        /**
         * 行単位で測定情報を出力する（"--debug2"のみ有効）
         */
        void debugPrintLineDetail(NCMStructData.NCMFuncData fncdata, object data)
        {
            const string LOG_HEADER = "LINE,ELOC,CLOC,SKIP, CND,WHIL, FOR,  IF,SWCH,CASE,DFLT, AND,NEST";
            const int MAX_LENGTH = 70;
            string MAX_FORMAT = string.Format("{{0,-{0}}}", MAX_LENGTH);

            int startKey = 0;
            int endKey = 0;
            foreach (var item in fncdata.counterList)
            {
                if (startKey == 0)
                {
                    startKey = item.Key;
                }

                if (endKey < item.Key)
                {
                    endKey = item.Key;
                }

            }

            string line = "----";
            for (int i = 0; i < LOG_HEADER.Split(',').Length - 1; i++)
            {
                line += ",----";
            }
            string padding = (targetCodeList.Count != 0) ? " ".PadRight(MAX_LENGTH) + "," : string.Empty;


            ((NCMStructData)data).dbugbufPrintLn(2, padding + LOG_HEADER); //* MSG_LOG_ID_CS_00000041

            string statement = string.Empty;
            for (int i = startKey; i <= endKey; i++)
            {
                if (targetCodeList.Count > 0)
                {

                    if (targetCodeList.Count > i - 1)
                    {
                        statement = string.Format(MAX_FORMAT, targetCodeList[i - 1]);
                    }
                    else
                    {
                        statement = string.Empty;
                    }

                    if (statement.Length > MAX_LENGTH)
                    {
                        statement = statement.Substring(0, MAX_LENGTH);
                    }

                    statement += ",";
                }

                if (fncdata.counterList.ContainsKey(i))
                {
                    ((NCMStructData)data).dbugbufPrintLn(2, "" +
                        statement +
                        string.Format("{0,4},", i) + fncdata.counterList[i]); //* MSG_LOG_ID_CS_00000042
                }
                else
                {
                    ((NCMStructData)data).dbugbufPrintLn(2, "" +
                        statement +
                        string.Format("{0,4},", i)); //* MSG_LOG_ID_CS_00000043
                }
            }

            ((NCMStructData)data).dbugbufPrintLn(2, padding + line); //* MSG_LOG_ID_CS_00000044
            ((NCMStructData)data).dbugbufPrintLn(2,
                string.Format(padding + "{0,4},{1,4},{2,4},{3,4},{4,4},{5,4},{6,4},{7,4},{8,4},{9,4},{10,4},{11,4},{12,4}",
                    fncdata.BodyEndToken.Line - fncdata.FuncStartToken.Line + 1,
                    fncdata.cntEloc,
                    fncdata.cntCloc,
                    fncdata.cntSkip,
                    fncdata.cntCndOps,
                    fncdata.cntLoopsDoWhile,
                    fncdata.cntLoopsFor,
                    fncdata.cntConditionalIf,
                    fncdata.cntSelectionSwitch,
                    fncdata.cntSelectionCase,
                    fncdata.cntSelectionDefault,
                    fncdata.cntLogicalAnd,
                    fncdata.cntNests
                )); //* MSG_LOG_ID_CS_00000045
            ((NCMStructData)data).dbugbufPrintLn(2, ""); //* MSG_LOG_ID_CS_00000046
        }

        /**
         * 解析対象となる行がDictionaryに存在しない場合は追加する
         */
        void addTargetLine(ref Dictionary<int, NCMStructData.NCMFuncData.NCMLineData> dict, int line)
        {
            if (dict == null || line < 0)
            {
                return;
            }


            if (!dict.ContainsKey(line))
            {
                NCMStructData.NCMFuncData.NCMLineData nc = new NCMStructData.NCMFuncData.NCMLineData();
                dict.Add(line, nc);
            }
        }

        /**
         * トークン系ビジターをもとに、カウントを行う。
         */
        void CountToken(TokenData td, object data)
        {
            ICSharpCode.NCMCSParser.NCMStructData.NCMFuncData fncdata = null;
            fncdata = ((NCMStructData)data).functionGetData();


            if (fncdata == null || fncdata.counterList == null)
            {
                return;
            }

            if (td == null || data == null)
            {
                return;
            }


            if (td.StartLocation.IsEmpty || td.EndLocation.IsEmpty ||
                td.StartLocation.Column == 0 || td.EndLocation.Column == 0)
            {
                ((NCMStructData)data).dbugbufPrintLn(2, Resources.MSG_LOG_ID_CS_00000047_01 + td.ToString()); //* MSG_LOG_ID_CS_00000047
                return;
            }



            string[] rows = new string[0];
            int rows_count = 0;
            if (td.Token != null && td.Token.StartsWith("@"))
            {
                rows = td.Token.Split(new string[] { Environment.NewLine }, StringSplitOptions.None);
                rows_count = rows.Length - 1;
            }
            int startLine = td.StartLocation.Line;
            int endLine = td.EndLocation.Line + rows_count;


            for (int currentLine = startLine; currentLine <= endLine; currentLine++)
            {
                ((NCMStructData)data).dbugbufPrintLn(2, Resources.MSG_LOG_ID_CS_00000048_01 + td.ToString()); //* MSG_LOG_ID_CS_00000048


                addTargetLine(ref fncdata.counterList, currentLine);


                switch (td.RoleName)
                {
                    case "Comment":
                        if (td.ItemType.Equals(CommentType.InactiveCode.ToString()))
                        {

                            fncdata.counterList[currentLine].isInactiveCode = true;
                        }
                        else
                        {
                            fncdata.counterList[currentLine].cntCloc++;
                        }
                        break;

                    case "PreProcessorDirective":
                        if (fncdata.counterList[currentLine].isInactiveCode)
                        {
                            fncdata.counterList[currentLine].isInactiveCode = false;
                        }
                        fncdata.counterList[currentLine].cntEloc++;


                        switch (td.PpType)
                        {
                            case PreProcessorDirectiveType.If:

                                markInactiveCode(td, data);

                                ppIfStack.Push(td);
                                break;
                            case PreProcessorDirectiveType.Elif:
                            case PreProcessorDirectiveType.Else:

                                markInactiveCode(td, data);

                                if (ppIfStack.Count > 0)
                                {
                                    ppIfStack.Pop();
                                }
                                ppIfStack.Push(td);
                                break;
                            case PreProcessorDirectiveType.Endif:

                                markInactiveCode(td, data);

                                if (ppIfStack.Count > 0)
                                {
                                    ppIfStack.Pop();
                                }
                                break;
                        }
                        break;
                    case "LBrace":

                        fncdata.cntNestsLevel++;
                        if (fncdata.cntNests < fncdata.cntNestsLevel)
                        {
                            fncdata.cntNests = fncdata.cntNestsLevel;


                            fncdata.counterList[currentLine].cntNests = (ushort)fncdata.cntNestsLevel;
                        }
                        fncdata.counterList[currentLine].cntEloc++;
                        break;
                    case "RBrace":

                        fncdata.cntNestsLevel--;
                        fncdata.counterList[currentLine].cntEloc++;
                        break;
                    case "Comma":
                    case "Statement":
                    default:
                        fncdata.counterList[currentLine].cntEloc++;
                        break;
                }

            }
        }

        /**
         * プリプロセッサディレクティブ内の無効行をマークする。
         */
        void markInactiveCode(TokenData td, object data)
        {
            ICSharpCode.NCMCSParser.NCMStructData.NCMFuncData fncdata = null;
            fncdata = ((NCMStructData)data).functionGetData();

            if (fncdata == null || ppIfStack.Count == 0)
            {
                return;
            }



            if (!this.ppIfStack.Peek().PpTake)
            {

                for (int i = ppIfStack.Peek().StartLocation.Line + 1; i < td.StartLocation.Line; i++)
                {
                    addTargetLine(ref fncdata.counterList, i);


                    fncdata.counterList[i].cntEloc = 0;
                    fncdata.counterList[i].isInactiveCode = true;
                }
            }
        }

        /**
         * カウント用リストの作成
         */
        bool preLineCounter(object data)
        {
            ICSharpCode.NCMCSParser.NCMStructData.NCMFuncData fncdata = null;
            fncdata = ((NCMStructData)data).functionGetData();


            if (fncdata != null && fncdata.counterList == null)
            {
                fncdata.counterList = new Dictionary<int, NCMStructData.NCMFuncData.NCMLineData>();
                return true;
            }
            return false;
        }


        /**
         * メソッド系ビジター毎にラインカウントする
         */
        void mainLineCounter(NCMStructData.NCMFuncData fncdata, object data)
        {
            if (fncdata == null)
            {
                return;
            }


            foreach (var item in fncdata.counterList)
            {

                if (item.Value.cntEloc != 0)
                {
                    fncdata.cntEloc++;
                }
                else
                {
                    if (item.Value.cntCloc != 0)
                    {
                        fncdata.cntCloc++;
                    }

                    if (item.Value.isInactiveCode)
                    {
                        fncdata.cntSkip++;
                    }
                }
            }

            debugPrintLineDetail(fncdata, data);


            fncdata.counterList.Clear();
            fncdata.counterList = null;

        }

        #region FUNC用クラス名リスト

        /**
         * 関数名リストから関数名を取得.
         */
        public String getScopeNames()
        {
            String txt = "";
            foreach (String fnm in scopenames)
                txt += fnm + ".";

            if (txt.Length > 1)
                txt = txt.Remove(txt.Length - 1);

            return txt;
        }

        /**
         * 関数名リストの最後尾に関数名を追加.
         */
        private void addScopeList(String name)
        {
            scopenames.Add(name);
        }

        /**
         * 関数名リストの最後尾を削除.
         */
        private void removeScopeList()
        {
            if (scopenames.Count > 0)
            {
                scopenames.RemoveAt(scopenames.Count - 1);
            }
        }


        /**
         * 逐語的識別子(@プレフィックス)が指定されている場合、snmに"@"を付加する。
         */
        string GetVerbatimName(string snm, Identifier identifier)
        {
            return (identifier == null) ? snm : (identifier.IsVerbatim) ? "@" + snm : snm;
        }

        /**
         * 型名を取得。逐語的識別子(@プレフィックス)付きに対応。
         */
        string GetChildByTypeName<T>(T typenode) where T : AstType
        {
            if (typenode == null)
                return string.Empty;
            if (typenode.FirstChild == null)
            {
                return typenode.ToString();
            }

            Identifier an = null;
            bool beforeTarget = false;
            StringBuilder res = new StringBuilder();
            for (var cur = typenode.FirstChild; cur != null; cur = cur.NextSibling)
            {
                switch (cur.Role.ToString())
                {
                    case "Identifier":
                        res.Append((beforeTarget) ? "::" : "");
                        an = ((Identifier)cur);
                        res.Append((an.IsVerbatim) ? "@" + an.Name : an.Name);
                        beforeTarget = false;
                        break;
                    case "Target":
                    case "TypeArgument":
                        if (cur.GetType().Name == "SimpleType" && ((SimpleType)cur).IdentifierToken != null)
                        {
                            if (((SimpleType)cur).TypeArguments.Count() <= 0)
                            {
                                res.Append((beforeTarget) ? "::" : "");
                                an = ((SimpleType)cur).IdentifierToken;
                                res.Append((an.IsVerbatim) ? "@" + an.Name : an.Name);
                            }
                            else
                            {

                                res.Append(GetChildByTypeName((SimpleType)cur));
                            }
                            beforeTarget = (cur.Role.ToString() == "Target") ? true : false;
                        }
                        else
                        {
                            res.Append((beforeTarget) ? "::" : "");
                            res.Append(cur.ToString());
                            beforeTarget = false;
                        }
                        break;
                    default:
                        res.Append(GetTokenNodeWords(cur));
                        beforeTarget = false;
                        break;
                }
            }
            return res.ToString();
        }

        /**
         * 指定されたCSharpTokenNodeの文字列を返却する。
         * CSharpTokenNode以外が指定された場合は、node.ToString()値を返却する。
         */
        string GetTokenNodeWords<T>(T node) where T : AstNode
        {
            string res = string.Empty;
            if (node.GetType().Name == "CSharpTokenNode")
            {
                switch (node.Role.ToString())
                {
                    case "Pointer":
                        res = "*";
                        break;
                    case "Colon":
                        res = ":";
                        break;
                    case "Nullable":
                        res = "?";
                        break;
                    case "LChevron":
                        res = "<";
                        break;
                    case "RChevron":
                        res = ">";
                        break;
                    case "Comma":
                        res = ",";
                        break;
                    case "Dot":
                        res = ".";
                        break;
                    case "LBracket":
                        res = "[";
                        break;
                    case "RBracket":
                        res = "]";
                        break;
                    case "LPar":
                        res = "(";
                        break;
                    case "RPar":
                        res = ")";
                        break;
                    case "LBrace":
                        res = "{";
                        break;
                    case "RBrace":
                        res = "}";
                        break;
                    case "Semicolon":
                        res = ";";
                        break;
                    case "Assign":
                        res = "=";
                        break;
                    default:
                        res = node.ToString();
                        break;
                }
            }
            else
            {
                res = node.ToString();
            }
            return res;
        }


        #endregion

        #region 最上位ノード（CompilationUnit）
        /**
         * 最上位ノード.
         */
        public override object VisitCompilationUnit(CompilationUnit unit, object data)
        {
            ((NCMStructData)data).dbugbufPrintLn(3, Resources.MSG_LOG_ID_CS_00000049_01); //* MSG_LOG_ID_CS_00000049


            loadTargetCode(data);

            object obNode = VisitChildren(unit, data);
            ((NCMStructData)data).dbugbufPrintLn(3, Resources.MSG_LOG_ID_CS_00000050_01); //* MSG_LOG_ID_CS_00000050

            return obNode;
        }
        #endregion

        #region クラス系ノード（Class,Struct,Delegate）

        /**
         * クラスノード。名前を関数名リストに登録する.
         */
        public override object VisitTypeDeclaration(TypeDeclaration typeDeclaration, object data)
        {
            ((NCMStructData)data).dbugbufPrintLn(3, Resources.MSG_LOG_ID_CS_00000051_01); //* MSG_LOG_ID_CS_00000051
            if (!flgOutsideFunction && typeDeclaration.ClassType != ClassType.Enum)
            {
                String sname = typeDeclaration.Name;


                sname = GetVerbatimName(sname, typeDeclaration.NameToken);


                addScopeList(sname);
            }

            object obNode = VisitChildren(typeDeclaration, data);

            if (!flgOutsideFunction && typeDeclaration.ClassType != ClassType.Enum)
            {
                removeScopeList();
            }

            ((NCMStructData)data).dbugbufPrintLn(3, Resources.MSG_LOG_ID_CS_00000052_01); //* MSG_LOG_ID_CS_00000052
            return obNode;
        }

        /**
         * イベントアクセサーノード。名前を関数名リストに登録する.
         */
        public override object VisitCustomEventDeclaration(CustomEventDeclaration eventDeclaration, object data)
        {
            ((NCMStructData)data).dbugbufPrintLn(3, Resources.MSG_LOG_ID_CS_00000053_01); //* MSG_LOG_ID_CS_00000053

            if (!flgOutsideFunction)
            {
                String sname = eventDeclaration.Name;


                sname = GetVerbatimName(sname, eventDeclaration.NameToken);


                if (eventDeclaration.PrivateImplementationType != AstType.Null)
                {

                    AstNode pIType = eventDeclaration.PrivateImplementationType;
                    sname = pIType.ToString() + "." + sname;
                }
                addScopeList(sname);


                commentLineList = new List<int>();

            }

            object obNode = VisitChildren(eventDeclaration, data);
            if (!flgOutsideFunction)
            {
                removeScopeList();


                commentLineList.Clear();
                commentLineList = null;

            }
            ((NCMStructData)data).dbugbufPrintLn(3, Resources.MSG_LOG_ID_CS_00000054_01); //* MSG_LOG_ID_CS_00000054
            return obNode;
        }

        /**
         * インデクサノード。名前(this)を関数名リストに登録する.
         */
        public override object VisitIndexerDeclaration(IndexerDeclaration indexerDeclaration, object data)
        {

            ((NCMStructData)data).dbugbufPrintLn(3, Resources.MSG_LOG_ID_CS_00000055_01); //* MSG_LOG_ID_CS_00000055

            if (!flgOutsideFunction)
            {

                String sname = "this";
                if (indexerDeclaration.Name.Length > 0)
                {
                    sname = indexerDeclaration.Name;


                    sname = GetVerbatimName(sname, indexerDeclaration.NameToken);

                }
                if (indexerDeclaration.PrivateImplementationType != AstType.Null)
                {

                    AstNode pIType = indexerDeclaration.PrivateImplementationType;
                    sname = pIType.ToString() + "." + sname;
                }
                addScopeList(sname);


                commentLineList = new List<int>();

            }

            object obNode = VisitChildren(indexerDeclaration, data);

            if (!flgOutsideFunction)
            {
                removeScopeList();


                commentLineList.Clear();
                commentLineList = null;

            }

            ((NCMStructData)data).dbugbufPrintLn(3, Resources.MSG_LOG_ID_CS_00000056_01); //* MSG_LOG_ID_CS_00000056
            return obNode;
        }

        /**
         * プロパティノード。名前を関数名リストに登録する.
         */
        public override object VisitPropertyDeclaration(PropertyDeclaration propertyDeclaration, object data)
        {
            ((NCMStructData)data).dbugbufPrintLn(3, Resources.MSG_LOG_ID_CS_00000057_01); //* MSG_LOG_ID_CS_00000057

            if (!flgOutsideFunction)
            {
                String sname = propertyDeclaration.Name;


                sname = GetVerbatimName(sname, propertyDeclaration.NameToken);



                if (propertyDeclaration.PrivateImplementationType != AstType.Null)
                {

                    AstNode pIType = propertyDeclaration.PrivateImplementationType;
                    sname = pIType.ToString() + "." + sname;
                }
                addScopeList(sname);
            }

            object obNode = VisitChildren(propertyDeclaration, data);

            if (!flgOutsideFunction)
            {
                removeScopeList();
            }

            ((NCMStructData)data).dbugbufPrintLn(3, Resources.MSG_LOG_ID_CS_00000058_01); //* MSG_LOG_ID_CS_00000058

            return obNode;
        }

        #endregion


        #region メソッド系ノード

        /**
         * パラメータの型名を取得。メソッド系処理で利用。
         */
        string GetParenthesis(IEnumerable<AstNode> list)
        {
            string res = "(";
            bool isFirst = true;
            if (list.Any())
            {
                foreach (AstNode node in list)
                {
                    if (isFirst)
                    {
                        isFirst = false;
                    }
                    else
                    {
                        res += ",";
                    }

                    if (((ParameterDeclaration)node).Type == AstType.Null)
                    {
                        res += " " + "__arglist" + " ";
                    }
                    else
                    {

                        res += " " + GetChildByTypeName(((ParameterDeclaration)node).Type) + " ";

                    }
                }
            }
            else
            {

                res += " ";
            }
            return res + ")";
        }

        /**
         * コンストラクタノード。関数処理（開始位置,BODY開始位置,終了位置等を設定）を実行。
         */
        public override object VisitConstructorDeclaration(ConstructorDeclaration constructorDeclaration, object data)
        {
            int iline = 0;
            bool flgStartFunc = false;
            ICSharpCode.NCMCSParser.NCMStructData.NCMFuncData fncdata = null;
            ((NCMStructData)data).dbugbufPrintLn(3, Resources.MSG_LOG_ID_CS_00000059_01); //* MSG_LOG_ID_CS_00000059

            if (!flgOutsideFunction
                && !constructorDeclaration.Body.IsNull)
            {
                flgOutsideFunction = true;
                flgStartFunc = true;
                flgFunctionBlock = false;
                iline = constructorDeclaration.StartLocation.Line;
                String sname = constructorDeclaration.IdentifierToken.Name;


                sname = GetVerbatimName(sname, constructorDeclaration.IdentifierToken);


                addScopeList(sname);

                fncdata = ((NCMStructData)data).functionNewData(constructorDeclaration.StartLocation);
                fncdata.nameFunc = this.getScopeNames();
                fncdata.funcParams = GetParenthesis(constructorDeclaration.Parameters);

                fncdata.FuncStartToken = constructorDeclaration.StartLocation;
                fncdata.cntFunctionBase++;




                ((NCMStructData)data).dbugbufPrintLn(2, Resources.MSG_LOG_ID_CS_00000060_01 + iline.ToString()
                    + Resources.MSG_LOG_ID_CS_00000060_02 + fncdata.nameFunc + Resources.MSG_LOG_ID_CS_00000060_03); //* MSG_LOG_ID_CS_00000060


                preLineCounter(data);
            }

            object obNode = VisitChildren(constructorDeclaration, data);

            if (flgStartFunc)
            {
                fncdata.BodyStartToken = constructorDeclaration.Body.StartLocation;
                fncdata.BodyEndToken = constructorDeclaration.Body.EndLocation;
                removeScopeList();
                flgOutsideFunction = false;

                mainLineCounter(fncdata, data);

                ((NCMStructData)data).dbugbufPrint(2, Resources.MSG_LOG_ID_CS_00000061_01
                    + iline.ToString()
                    + Resources.MSG_LOG_ID_CS_00000061_02 + fncdata.nameFunc); //* MSG_LOG_ID_CS_00000061
                ((NCMStructData)data).dbugbufPrintLn(2, "(LINE:"
                    + fncdata.FuncStartToken.Line.ToString()
                    + "-" + fncdata.BodyStartToken.Line.ToString()
                    + "-" + fncdata.BodyEndToken.Line.ToString()
                    + ",CYCLO:" + fncdata.cntCyclo.ToString()
                    + ",CYCLOWC:" + fncdata.cntCycloWC.ToString()
                    + ",BRANCHS:" + fncdata.cntBranchs.ToString()
                    + ",CNDOPS:" + fncdata.cntCndOps.ToString()
                    + ",FBase:" + fncdata.cntFunctionBase.ToString()
                    + ",DoWhile:" + fncdata.cntLoopsDoWhile.ToString()
                    + ",For:" + fncdata.cntLoopsFor.ToString()
                    + ",If:" + fncdata.cntConditionalIf.ToString()
                    + ",Switch:" + fncdata.cntSelectionSwitch.ToString()
                    + ",Case:" + fncdata.cntSelectionCase.ToString()
                    + ",Default:" + fncdata.cntSelectionDefault.ToString()
                    + ",And/Or:" + fncdata.cntLogicalAnd.ToString()
                    + ")"); //* MSG_LOG_ID_CS_00000062
            }

            ((NCMStructData)data).dbugbufPrintLn(3, Resources.MSG_LOG_ID_CS_00000063_01); //* MSG_LOG_ID_CS_00000063

            return obNode;
        }

        /**
         * デストラクターノード。関数処理（開始位置,BODY開始位置,終了位置等を設定）やを実行。
         * ※デストラクターはパラメーター、アクセシビリティ修飾子の指定は行えない。
         */
        public override object VisitDestructorDeclaration(DestructorDeclaration destructorDeclaration, object data)
        {
            int iline = 0;
            bool flgStartFunc = false;
            ICSharpCode.NCMCSParser.NCMStructData.NCMFuncData fncdata = null;
            ((NCMStructData)data).dbugbufPrintLn(3, Resources.MSG_LOG_ID_CS_00000064_01); //* MSG_LOG_ID_CS_00000064

            if (!flgOutsideFunction
                && !destructorDeclaration.Body.IsNull)
            {
                flgOutsideFunction = true;
                flgStartFunc = true;
                flgFunctionBlock = false;
                iline = destructorDeclaration.StartLocation.Line;


                TypeDeclaration type = destructorDeclaration.Parent as TypeDeclaration;


                String sname = string.Empty;
                if (type != null)
                {
                    sname = GetVerbatimName(type.Name, type.NameToken);
                }
                else
                {
                    sname = GetVerbatimName(destructorDeclaration.Name, destructorDeclaration.IdentifierToken);
                }


                if (destructorDeclaration.TildeToken.Role == DestructorDeclaration.TildeRole)
                {
                    sname = "~" + sname;
                }
                addScopeList(sname);

                fncdata = ((NCMStructData)data).functionNewData(destructorDeclaration.StartLocation);
                fncdata.nameFunc = this.getScopeNames();
                fncdata.funcParams = "( )";

                fncdata.FuncStartToken = destructorDeclaration.StartLocation;
                fncdata.cntFunctionBase++;




                ((NCMStructData)data).dbugbufPrintLn(2, Resources.MSG_LOG_ID_CS_00000065_01 + iline.ToString()
                    + Resources.MSG_LOG_ID_CS_00000065_02 + fncdata.nameFunc + Resources.MSG_LOG_ID_CS_00000065_03); //* MSG_LOG_ID_CS_00000065


                preLineCounter(data);
            }

            object obNode = VisitChildren(destructorDeclaration, data);

            if (flgStartFunc)
            {
                fncdata.BodyStartToken = destructorDeclaration.Body.StartLocation;
                fncdata.BodyEndToken = destructorDeclaration.Body.EndLocation;
                removeScopeList();
                flgOutsideFunction = false;

                mainLineCounter(fncdata, data);

                ((NCMStructData)data).dbugbufPrint(2, Resources.MSG_LOG_ID_CS_00000066_01
                    + iline.ToString()
                    + Resources.MSG_LOG_ID_CS_00000066_02 + fncdata.nameFunc); //* MSG_LOG_ID_CS_00000066
                ((NCMStructData)data).dbugbufPrintLn(2, "(LINE:"
                    + fncdata.FuncStartToken.Line.ToString()
                    + "-" + fncdata.BodyStartToken.Line.ToString()
                    + "-" + fncdata.BodyEndToken.Line.ToString()
                    + ",CYCLO:" + fncdata.cntCyclo.ToString()
                    + ",CYCLOWC:" + fncdata.cntCycloWC.ToString()
                    + ",BRANCHS:" + fncdata.cntBranchs.ToString()
                    + ",CNDOPS:" + fncdata.cntCndOps.ToString()
                    + ",FBase:" + fncdata.cntFunctionBase.ToString()
                    + ",DoWhile:" + fncdata.cntLoopsDoWhile.ToString()
                    + ",For:" + fncdata.cntLoopsFor.ToString()
                    + ",If:" + fncdata.cntConditionalIf.ToString()
                    + ",Switch:" + fncdata.cntSelectionSwitch.ToString()
                    + ",Case:" + fncdata.cntSelectionCase.ToString()
                    + ",Default:" + fncdata.cntSelectionDefault.ToString()
                    + ",And/Or:" + fncdata.cntLogicalAnd.ToString()
                    + ")"); //* MSG_LOG_ID_CS_00000067
            }

            ((NCMStructData)data).dbugbufPrintLn(3, Resources.MSG_LOG_ID_CS_00000068_01); //* MSG_LOG_ID_CS_00000068

            return obNode;
        }


        /**
         * メソッドノード。関数処理（開始位置,BODY開始位置,終了位置等を設定）やを実行。
         */
        public override object VisitMethodDeclaration(MethodDeclaration methodDeclaration, object data)
        {
            int iline = 0;
            bool flgStartFunc = false;
            ICSharpCode.NCMCSParser.NCMStructData.NCMFuncData fncdata = null;
            ((NCMStructData)data).dbugbufPrintLn(3, Resources.MSG_LOG_ID_CS_00000069_01); //* MSG_LOG_ID_CS_00000069

            if (!flgOutsideFunction
                && !methodDeclaration.Body.IsNull)
            {
                flgOutsideFunction = true;
                flgStartFunc = true;
                flgFunctionBlock = false;
                iline = methodDeclaration.StartLocation.Line;
                String sname = methodDeclaration.Name;


                sname = GetVerbatimName(sname, methodDeclaration.NameToken);


                if (methodDeclaration.PrivateImplementationType != AstType.Null)
                {

                    AstNode pIType = methodDeclaration.PrivateImplementationType;
                    sname = pIType.ToString() + "." + sname;
                }

                addScopeList(sname);

                fncdata = ((NCMStructData)data).functionNewData(methodDeclaration.StartLocation);
                fncdata.nameFunc = this.getScopeNames();
                fncdata.funcParams = GetParenthesis(methodDeclaration.Parameters);

                fncdata.FuncStartToken = methodDeclaration.StartLocation;
                fncdata.cntFunctionBase++;




                ((NCMStructData)data).dbugbufPrintLn(2, Resources.MSG_LOG_ID_CS_00000070_01 + iline.ToString()
                    + Resources.MSG_LOG_ID_CS_00000070_02 + fncdata.nameFunc + Resources.MSG_LOG_ID_CS_00000070_03); //* MSG_LOG_ID_CS_00000070


                preLineCounter(data);
            }

            object obNode = VisitChildren(methodDeclaration, data);

            if (flgStartFunc)
            {
                fncdata.BodyStartToken = methodDeclaration.Body.StartLocation;
                fncdata.BodyEndToken = methodDeclaration.Body.EndLocation;
                removeScopeList();
                flgOutsideFunction = false;

                mainLineCounter(fncdata, data);

                ((NCMStructData)data).dbugbufPrint(2, Resources.MSG_LOG_ID_CS_00000071_01
                    + iline.ToString()
                    + Resources.MSG_LOG_ID_CS_00000071_02 + fncdata.nameFunc); //* MSG_LOG_ID_CS_00000071
                ((NCMStructData)data).dbugbufPrintLn(2, "(LINE:"
                    + fncdata.FuncStartToken.Line.ToString()
                    + "-" + fncdata.BodyStartToken.Line.ToString()
                    + "-" + fncdata.BodyEndToken.Line.ToString()
                    + ",CYCLO:" + fncdata.cntCyclo.ToString()
                    + ",CYCLOWC:" + fncdata.cntCycloWC.ToString()
                    + ",BRANCHS:" + fncdata.cntBranchs.ToString()
                    + ",CNDOPS:" + fncdata.cntCndOps.ToString()
                    + ",FBase:" + fncdata.cntFunctionBase.ToString()
                    + ",DoWhile:" + fncdata.cntLoopsDoWhile.ToString()
                    + ",For:" + fncdata.cntLoopsFor.ToString()
                    + ",If:" + fncdata.cntConditionalIf.ToString()
                    + ",Switch:" + fncdata.cntSelectionSwitch.ToString()
                    + ",Case:" + fncdata.cntSelectionCase.ToString()
                    + ",Default:" + fncdata.cntSelectionDefault.ToString()
                    + ",And/Or:" + fncdata.cntLogicalAnd.ToString()
                    + ")"); //* MSG_LOG_ID_CS_00000072
            }

            ((NCMStructData)data).dbugbufPrintLn(3, Resources.MSG_LOG_ID_CS_00000073_01); //* MSG_LOG_ID_CS_00000073
            return obNode;
        }

        /**
         * アクセサノード。関数処理（開始位置,BODY開始位置,終了位置等を設定）やを実行。
         */
        public override object VisitAccessor(Accessor accessor, object data)
        {
            int iline = 0;
            bool flgStartFunc = false;
            ICSharpCode.NCMCSParser.NCMStructData.NCMFuncData fncdata = null;
            ((NCMStructData)data).dbugbufPrintLn(3, Resources.MSG_LOG_ID_CS_00000074_01); //* MSG_LOG_ID_CS_00000074

#if !DEBUG_AUTO_PROPRETY
            if (!flgOutsideFunction)
#else
            
            if (!flgOutsideFunction
                && !accessor.Body.IsNull)
#endif
            {
                flgOutsideFunction = true;
                flgStartFunc = true;
                flgFunctionBlock = false;
                iline = accessor.StartLocation.Line;
                String sname = string.Empty;
                if (accessor.Role == PropertyDeclaration.GetterRole)
                {
                    sname = "get";
                }
                else if (accessor.Role == PropertyDeclaration.SetterRole)
                {
                    sname = "set";
                }
                else if (accessor.Role == CustomEventDeclaration.AddAccessorRole)
                {
                    sname = "add";
                }
                else if (accessor.Role == CustomEventDeclaration.RemoveAccessorRole)
                {
                    sname = "remove";
                }
                addScopeList(sname);

                fncdata = ((NCMStructData)data).functionNewData(accessor.StartLocation);
                fncdata.nameFunc = this.getScopeNames();
                fncdata.funcParams = string.Empty;

#if DEBUG_ACCESSOR_ATTRIBUTES
                int attributeCount = accessor.Attributes.Count;
                fncdata.FuncStartToken =
                    new NRefactory.TextLocation(accessor.StartLocation.Line + attributeCount, accessor.StartLocation.Column);
                fncdata.cntEloc = -1 * attributeCount;
#else
                fncdata.FuncStartToken = accessor.StartLocation;
#endif

                fncdata.cntFunctionBase++;




                ((NCMStructData)data).dbugbufPrintLn(2, Resources.MSG_LOG_ID_CS_00000075_01 + iline.ToString()
                    + Resources.MSG_LOG_ID_CS_00000075_02 + fncdata.nameFunc + Resources.MSG_LOG_ID_CS_00000075_03); //* MSG_LOG_ID_CS_00000075


                preLineCounter(data);
            }

            object obNode = VisitChildren(accessor, data);

            if (flgStartFunc)
            {
                fncdata.BodyStartToken = (accessor.Body.StartLocation.IsEmpty) ? accessor.StartLocation : accessor.Body.StartLocation;
                fncdata.BodyEndToken = (accessor.Body.EndLocation.IsEmpty) ? accessor.EndLocation : accessor.Body.EndLocation;


                if (commentLineList != null)
                {
                    foreach (int currentLine in commentLineList)
                    {
                        if (currentLine < fncdata.BodyStartToken.Line ||
                            currentLine > fncdata.BodyEndToken.Line)
                        {
                            continue;
                        }

                        if (!fncdata.counterList.ContainsKey(currentLine))
                        {
                            addTargetLine(ref fncdata.counterList, currentLine);
                            fncdata.counterList[currentLine].cntCloc++;
                        }
                    }
                }


                removeScopeList();
                flgOutsideFunction = false;

                mainLineCounter(fncdata, data);

                ((NCMStructData)data).dbugbufPrint(2, Resources.MSG_LOG_ID_CS_00000076_01
                    + iline.ToString()
                    + Resources.MSG_LOG_ID_CS_00000076_02 + fncdata.nameFunc); //* MSG_LOG_ID_CS_00000076
                ((NCMStructData)data).dbugbufPrintLn(2, "(LINE:"
                    + fncdata.FuncStartToken.Line.ToString()
                    + "-" + fncdata.BodyStartToken.Line.ToString()
                    + "-" + fncdata.BodyEndToken.Line.ToString()
                    + ",CYCLO:" + fncdata.cntCyclo.ToString()
                    + ",CYCLOWC:" + fncdata.cntCycloWC.ToString()
                    + ",BRANCHS:" + fncdata.cntBranchs.ToString()
                    + ",CNDOPS:" + fncdata.cntCndOps.ToString()
                    + ",FBase:" + fncdata.cntFunctionBase.ToString()
                    + ",DoWhile:" + fncdata.cntLoopsDoWhile.ToString()
                    + ",For:" + fncdata.cntLoopsFor.ToString()
                    + ",If:" + fncdata.cntConditionalIf.ToString()
                    + ",Switch:" + fncdata.cntSelectionSwitch.ToString()
                    + ",Case:" + fncdata.cntSelectionCase.ToString()
                    + ",Default:" + fncdata.cntSelectionDefault.ToString()
                    + ",And/Or:" + fncdata.cntLogicalAnd.ToString()
                    + ")"); //* MSG_LOG_ID_CS_00000077
            }

            ((NCMStructData)data).dbugbufPrintLn(3, Resources.MSG_LOG_ID_CS_00000078_01); //* MSG_LOG_ID_CS_00000078

            return obNode;
        }

        /**
         * オペレータノード。関数処理（開始位置,BODY開始位置,終了位置等を設定）やを実行。
         */
        public override object VisitOperatorDeclaration(OperatorDeclaration operatorDeclaration, object data)
        {
            int iline = 0;
            bool flgStartFunc = false;
            ICSharpCode.NCMCSParser.NCMStructData.NCMFuncData fncdata = null;
            ((NCMStructData)data).dbugbufPrintLn(3, Resources.MSG_LOG_ID_CS_00000079_01); //* MSG_LOG_ID_CS_00000079

            if (!flgOutsideFunction
                && !operatorDeclaration.Body.IsNull)
            {
                flgOutsideFunction = true;
                flgStartFunc = true;
                flgFunctionBlock = false;
                iline = operatorDeclaration.StartLocation.Line;
                String sname = "operator";
                if (operatorDeclaration.OperatorType == OperatorType.Explicit
                    || operatorDeclaration.OperatorType == OperatorType.Implicit)
                {
                    sname += " " + operatorDeclaration.ReturnType.ToString();
                }
                else
                {
                    sname += " " + OperatorDeclaration.GetToken(operatorDeclaration.OperatorType);
                }
                addScopeList(sname);

                fncdata = ((NCMStructData)data).functionNewData(operatorDeclaration.StartLocation);
                fncdata.nameFunc = this.getScopeNames();
                fncdata.funcParams = GetParenthesis(operatorDeclaration.Parameters);

                fncdata.FuncStartToken = operatorDeclaration.StartLocation;
                fncdata.cntFunctionBase++;

                ((NCMStructData)data).dbugbufPrintLn(2, Resources.MSG_LOG_ID_CS_00000080_01 + iline.ToString()
                    + Resources.MSG_LOG_ID_CS_00000080_02 + fncdata.nameFunc + Resources.MSG_LOG_ID_CS_00000080_03); //* MSG_LOG_ID_CS_00000080


                preLineCounter(data);
            }

            object obNode = VisitChildren(operatorDeclaration, data);

            if (flgStartFunc)
            {
                fncdata.BodyStartToken = operatorDeclaration.Body.StartLocation;
                fncdata.BodyEndToken = operatorDeclaration.Body.EndLocation;
                removeScopeList();
                flgOutsideFunction = false;

                mainLineCounter(fncdata, data);

                ((NCMStructData)data).dbugbufPrint(2, Resources.MSG_LOG_ID_CS_00000081_01
                    + iline.ToString()
                    + Resources.MSG_LOG_ID_CS_00000081_02 + fncdata.nameFunc); //* MSG_LOG_ID_CS_00000081
                ((NCMStructData)data).dbugbufPrintLn(2, "(LINE:"
                    + fncdata.FuncStartToken.Line.ToString()
                    + "-" + fncdata.BodyStartToken.Line.ToString()
                    + "-" + fncdata.BodyEndToken.Line.ToString()
                    + ",CYCLO:" + fncdata.cntCyclo.ToString()
                    + ",CYCLOWC:" + fncdata.cntCycloWC.ToString()
                    + ",BRANCHS:" + fncdata.cntBranchs.ToString()
                    + ",CNDOPS:" + fncdata.cntCndOps.ToString()
                    + ",FBase:" + fncdata.cntFunctionBase.ToString()
                    + ",DoWhile:" + fncdata.cntLoopsDoWhile.ToString()
                    + ",For:" + fncdata.cntLoopsFor.ToString()
                    + ",If:" + fncdata.cntConditionalIf.ToString()
                    + ",Switch:" + fncdata.cntSelectionSwitch.ToString()
                    + ",Case:" + fncdata.cntSelectionCase.ToString()
                    + ",Default:" + fncdata.cntSelectionDefault.ToString()
                    + ",And/Or:" + fncdata.cntLogicalAnd.ToString()
                    + ")"); //* MSG_LOG_ID_CS_00000082
            }

            ((NCMStructData)data).dbugbufPrintLn(3, Resources.MSG_LOG_ID_CS_00000083_01); //* MSG_LOG_ID_CS_00000083

            return obNode;
        }


        #endregion

        #region メトリクスカウント用 Override

        /**
         * ブロック数を測定。
         */
        public override object VisitBlockStatement(BlockStatement blockStatement, object data)
        {
            ICSharpCode.NCMCSParser.NCMStructData.NCMFuncData fncdata = null;
            ((NCMStructData)data).dbugbufPrintLn(3, Resources.MSG_LOG_ID_CS_00000084_01); //* MSG_LOG_ID_CS_00000084

            if (flgOutsideFunction)
            {
                fncdata = ((NCMStructData)data).functionGetData();
                if (fncdata != null)
                {

                    if (!flgFunctionBlock)
                    {

                        fncdata.cntCyclo++;
                        fncdata.cntCycloWC++;
                        fncdata.cntBranchs++;
                        flgFunctionBlock = true;
                    }

                    ((NCMStructData)data).dbugbufPrintLn(2, Resources.MSG_LOG_ID_CS_00000085_01 + blockStatement.Role.ToString()
                    + Resources.MSG_LOG_ID_CS_00000085_02 + blockStatement.StartLocation.Line.ToString()
                    + Resources.MSG_LOG_ID_CS_00000085_03 + blockStatement.StartLocation.Column.ToString()
                    + Resources.MSG_LOG_ID_CS_00000085_04); //* MSG_LOG_ID_CS_00000085
                }
                else
                {
                    ((NCMStructData)data).dbugbufPrintLn(2, Resources.MSG_LOG_ID_CS_00000086_01); //* MSG_LOG_ID_CS_00000086
                }
            }

            object obNode = VisitChildren(blockStatement, data);

            ((NCMStructData)data).dbugbufPrintLn(3, Resources.MSG_LOG_ID_CS_00000087_01); //* MSG_LOG_ID_CS_00000087

            return obNode;
        }

        /**
         * ForEachノード。CYCLO,BRANCHSのカウント。
         */
        public override object VisitForeachStatement(ForeachStatement foreachStatement, object data)
        {
            ((NCMStructData)data).dbugbufPrintLn(3, Resources.MSG_LOG_ID_CS_00000088_01); //* MSG_LOG_ID_CS_00000088
            if (flgOutsideFunction && !flgBranchCondition)
            {
                ICSharpCode.NCMCSParser.NCMStructData.NCMFuncData fncdata = null;
                fncdata = ((NCMStructData)data).functionGetData();
                if (fncdata != null)
                {
                    fncdata.cntLoopsFor++;
                    fncdata.cntCyclo++;
                    fncdata.cntCycloWC++;
                    fncdata.cntBranchs++;
                    ((NCMStructData)data).dbugbufPrintLn(2, Resources.MSG_LOG_ID_CS_00000089_01 + foreachStatement.Role.ToString()
                    + Resources.MSG_LOG_ID_CS_00000089_02 + foreachStatement.StartLocation.Line.ToString()
                    + Resources.MSG_LOG_ID_CS_00000089_03 + foreachStatement.StartLocation.Column.ToString()
                    + Resources.MSG_LOG_ID_CS_00000089_04); //* MSG_LOG_ID_CS_00000089


                    addTargetLine(ref fncdata.counterList, foreachStatement.StartLocation.Line);
                    fncdata.counterList[foreachStatement.StartLocation.Line].cntLoopsFor++;
                }
                else
                {
                    ((NCMStructData)data).dbugbufPrintLn(2, Resources.MSG_LOG_ID_CS_00000090_01); //* MSG_LOG_ID_CS_00000090
                }
            }

            object obNode = VisitChildren(foreachStatement, data);

            ((NCMStructData)data).dbugbufPrintLn(3, Resources.MSG_LOG_ID_CS_00000091_01); //* MSG_LOG_ID_CS_00000091

            return obNode;
        }

        /**
         * Forステートメントノード。CYCLO,BRANCHSのカウント。
         */
        public override object VisitForStatement(ForStatement forStatement, object data)
        {
            ((NCMStructData)data).dbugbufPrintLn(3, Resources.MSG_LOG_ID_CS_00000092_01); //* MSG_LOG_ID_CS_00000092
            if (flgOutsideFunction && !flgBranchCondition)
            {
                ICSharpCode.NCMCSParser.NCMStructData.NCMFuncData fncdata = null;
                fncdata = ((NCMStructData)data).functionGetData();
                if (fncdata != null)
                {
                    fncdata.cntLoopsFor++;
                    fncdata.cntCyclo++;
                    fncdata.cntCycloWC++;
                    fncdata.cntBranchs++;
                    ((NCMStructData)data).dbugbufPrintLn(2, Resources.MSG_LOG_ID_CS_00000093_01 + forStatement.Role.ToString()
                    + Resources.MSG_LOG_ID_CS_00000093_02 + forStatement.StartLocation.Line.ToString()
                    + Resources.MSG_LOG_ID_CS_00000093_03 + forStatement.StartLocation.Column.ToString()
                    + Resources.MSG_LOG_ID_CS_00000093_04); //* MSG_LOG_ID_CS_00000093


                    flgBranchCondition = true;
                    forStatement.Condition.AcceptVisitor(this, data);
                    flgBranchCondition = false;


                    addTargetLine(ref fncdata.counterList, forStatement.StartLocation.Line);
                    fncdata.counterList[forStatement.StartLocation.Line].cntLoopsFor++;
                }
                else
                {
                    ((NCMStructData)data).dbugbufPrintLn(2, Resources.MSG_LOG_ID_CS_00000094_01); //* MSG_LOG_ID_CS_00000094
                }
            }

            object obNode = VisitChildren(forStatement, data);

            ((NCMStructData)data).dbugbufPrintLn(3, Resources.MSG_LOG_ID_CS_00000095_01); //* MSG_LOG_ID_CS_00000095

            return obNode;
        }

        /**
         * Switchステートメントノード。CYCLOのカウント。
         */
        public override object VisitSwitchStatement(SwitchStatement switchStatement, object data)
        {
            ((NCMStructData)data).dbugbufPrintLn(3, Resources.MSG_LOG_ID_CS_00000096_01); //* MSG_LOG_ID_CS_00000096

            ICSharpCode.NCMCSParser.NCMStructData.NCMFuncData fncdata = null;
            if (flgOutsideFunction && !flgBranchCondition)
            {

                fncdata = ((NCMStructData)data).functionGetData();
                if (fncdata != null)
                {
                    fncdata.cntSelectionSwitch++;
                    fncdata.cntCyclo++;

                    ((NCMStructData)data).dbugbufPrintLn(2, Resources.MSG_LOG_ID_CS_00000097_01 + switchStatement.Role.ToString()
                    + Resources.MSG_LOG_ID_CS_00000097_02 + switchStatement.StartLocation.Line.ToString()
                    + Resources.MSG_LOG_ID_CS_00000097_03 + switchStatement.StartLocation.Column.ToString()
                    + Resources.MSG_LOG_ID_CS_00000097_04); //* MSG_LOG_ID_CS_00000097


                    flgBranchCondition = true;
                    switchStatement.Expression.AcceptVisitor(this, data);
                    flgBranchCondition = false;


                    addTargetLine(ref fncdata.counterList, switchStatement.StartLocation.Line);
                    fncdata.counterList[switchStatement.StartLocation.Line].cntSelectionSwitch++;
                }
                else
                {
                    ((NCMStructData)data).dbugbufPrintLn(2, Resources.MSG_LOG_ID_CS_00000098_01); //* MSG_LOG_ID_CS_00000098
                }
            }

            object obNode = VisitChildren(switchStatement, data);

            ((NCMStructData)data).dbugbufPrintLn(3, Resources.MSG_LOG_ID_CS_00000099_01); //* MSG_LOG_ID_CS_00000099

            return obNode;
        }

        /**
         * Caseラベルノード。BRANCHSのカウント。
         */
        public override object VisitCaseLabel(CaseLabel caseLabel, object data)
        {

            ((NCMStructData)data).dbugbufPrintLn(3, Resources.MSG_LOG_ID_CS_00000100_01); //* MSG_LOG_ID_CS_00000100
            if (flgOutsideFunction && !flgBranchCondition)
            {
                ICSharpCode.NCMCSParser.NCMStructData.NCMFuncData fncdata = null;
                fncdata = ((NCMStructData)data).functionGetData();
                if (fncdata != null)
                {

                    if (caseLabel.Expression.IsNull)
                    {

                        fncdata.cntSelectionDefault++;
                    }
                    else
                    {

                        fncdata.cntSelectionCase++;
                        fncdata.cntCycloWC++;
                        fncdata.cntBranchs++;


                        flgBranchCondition = true;
                        caseLabel.Expression.AcceptVisitor(this, data);
                        flgBranchCondition = false;
                    }
                    ((NCMStructData)data).dbugbufPrintLn(2, Resources.MSG_LOG_ID_CS_00000101_01 + caseLabel.Role.ToString()
                    + Resources.MSG_LOG_ID_CS_00000101_02 + caseLabel.StartLocation.Line.ToString()
                    + Resources.MSG_LOG_ID_CS_00000101_03 + caseLabel.StartLocation.Column.ToString()
                    + Resources.MSG_LOG_ID_CS_00000101_04); //* MSG_LOG_ID_CS_00000101


                    addTargetLine(ref fncdata.counterList, caseLabel.StartLocation.Line);
                    if (caseLabel.Expression.IsNull)
                    {
                        fncdata.counterList[caseLabel.StartLocation.Line].cntSelectionDefault++;
                    }
                    else
                    {
                        fncdata.counterList[caseLabel.StartLocation.Line].cntSelectionCase++;
                    }
                }
                else
                {
                    ((NCMStructData)data).dbugbufPrintLn(2, Resources.MSG_LOG_ID_CS_00000102_01); //* MSG_LOG_ID_CS_00000102
                }
            }


            object obNode = VisitChildren(caseLabel, data);

            ((NCMStructData)data).dbugbufPrintLn(3, Resources.MSG_LOG_ID_CS_00000103_01); //* MSG_LOG_ID_CS_00000103

            return obNode;
        }

        /**
         * Doループステートメントノード。CYCLO,BRANCHSのカウント。
         */
        public override object VisitDoWhileStatement(DoWhileStatement doWhileStatement, object data)
        {

            ((NCMStructData)data).dbugbufPrintLn(3, Resources.MSG_LOG_ID_CS_00000104_01); //* MSG_LOG_ID_CS_00000104
            if (flgOutsideFunction && !flgBranchCondition)
            {
                ICSharpCode.NCMCSParser.NCMStructData.NCMFuncData fncdata = null;
                fncdata = ((NCMStructData)data).functionGetData();
                if (fncdata != null)
                {
                    fncdata.cntLoopsDoWhile++;
                    fncdata.cntCyclo++;
                    fncdata.cntCycloWC++;
                    fncdata.cntBranchs++;
                    ((NCMStructData)data).dbugbufPrintLn(2, Resources.MSG_LOG_ID_CS_00000105_01 + doWhileStatement.Role.ToString()
                    + Resources.MSG_LOG_ID_CS_00000105_02 + doWhileStatement.StartLocation.Line.ToString()
                    + Resources.MSG_LOG_ID_CS_00000105_03 + doWhileStatement.StartLocation.Column.ToString()
                    + Resources.MSG_LOG_ID_CS_00000105_04); //* MSG_LOG_ID_CS_00000105


                    flgBranchCondition = true;
                    doWhileStatement.Condition.AcceptVisitor(this, data);
                    flgBranchCondition = false;


                    addTargetLine(ref fncdata.counterList, doWhileStatement.StartLocation.Line);
                    fncdata.counterList[doWhileStatement.StartLocation.Line].cntLoopsDoWhile++;
                }
                else
                {
                    ((NCMStructData)data).dbugbufPrintLn(2, Resources.MSG_LOG_ID_CS_00000106_01); //* MSG_LOG_ID_CS_00000106
                }
            }

            object obNode = VisitChildren(doWhileStatement, data);

            ((NCMStructData)data).dbugbufPrintLn(3, Resources.MSG_LOG_ID_CS_00000107_01); //* MSG_LOG_ID_CS_00000107

            return obNode;
        }

        /**
         * Whileループステートメントノード。CYCLO,BRANCHSのカウント。
         */
        public override object VisitWhileStatement(WhileStatement whileStatement, object data)
        {

            ((NCMStructData)data).dbugbufPrintLn(3, Resources.MSG_LOG_ID_CS_00000108_01); //* MSG_LOG_ID_CS_00000108
            if (flgOutsideFunction && !flgBranchCondition)
            {
                ICSharpCode.NCMCSParser.NCMStructData.NCMFuncData fncdata = null;
                fncdata = ((NCMStructData)data).functionGetData();
                if (fncdata != null)
                {
                    fncdata.cntLoopsDoWhile++;
                    fncdata.cntCyclo++;
                    fncdata.cntCycloWC++;
                    fncdata.cntBranchs++;
                    ((NCMStructData)data).dbugbufPrintLn(2, Resources.MSG_LOG_ID_CS_00000109_01 + whileStatement.Role.ToString()
                    + Resources.MSG_LOG_ID_CS_00000109_02 + whileStatement.StartLocation.Line.ToString()
                    + Resources.MSG_LOG_ID_CS_00000109_03 + whileStatement.StartLocation.Column.ToString()
                    + Resources.MSG_LOG_ID_CS_00000109_04); //* MSG_LOG_ID_CS_00000109


                    flgBranchCondition = true;
                    whileStatement.Condition.AcceptVisitor(this, data);
                    flgBranchCondition = false;


                    addTargetLine(ref fncdata.counterList, whileStatement.StartLocation.Line);
                    fncdata.counterList[whileStatement.StartLocation.Line].cntLoopsDoWhile++;
                }
                else
                {
                    ((NCMStructData)data).dbugbufPrintLn(2, Resources.MSG_LOG_ID_CS_00000110_01); //* MSG_LOG_ID_CS_00000110
                }
            }

            object obNode = VisitChildren(whileStatement, data);

            ((NCMStructData)data).dbugbufPrintLn(3, Resources.MSG_LOG_ID_CS_00000111_01); //* MSG_LOG_ID_CS_00000111

            return obNode;
        }

        /**
         * ifステートメントノード。CYCLO,BRANCHSのカウント。
         */
        public override object VisitIfElseStatement(IfElseStatement ifElseStatement, object data)
        {
            ((NCMStructData)data).dbugbufPrintLn(3, Resources.MSG_LOG_ID_CS_00000112_01); //* MSG_LOG_ID_CS_00000112
            if (flgOutsideFunction && !flgBranchCondition)
            {
                ICSharpCode.NCMCSParser.NCMStructData.NCMFuncData fncdata = null;
                fncdata = ((NCMStructData)data).functionGetData();
                if (fncdata != null)
                {
                    fncdata.cntConditionalIf++;
                    fncdata.cntCyclo++;
                    fncdata.cntCycloWC++;
                    fncdata.cntBranchs++;
                    ((NCMStructData)data).dbugbufPrintLn(2, Resources.MSG_LOG_ID_CS_00000113_01 + ifElseStatement.Role.ToString()
                    + Resources.MSG_LOG_ID_CS_00000113_02 + ifElseStatement.StartLocation.Line.ToString()
                    + Resources.MSG_LOG_ID_CS_00000113_03 + ifElseStatement.StartLocation.Column.ToString()
                    + Resources.MSG_LOG_ID_CS_00000113_04); //* MSG_LOG_ID_CS_00000113


                    flgBranchCondition = true;
                    ifElseStatement.Condition.AcceptVisitor(this, data);
                    flgBranchCondition = false;


                    addTargetLine(ref fncdata.counterList, ifElseStatement.StartLocation.Line);
                    fncdata.counterList[ifElseStatement.StartLocation.Line].cntConditionalIf++;
                }
                else
                {
                    ((NCMStructData)data).dbugbufPrintLn(2, Resources.MSG_LOG_ID_CS_00000114_01); //* MSG_LOG_ID_CS_00000114
                }
            }

            object obNode = VisitChildren(ifElseStatement, data);

            ((NCMStructData)data).dbugbufPrintLn(3, Resources.MSG_LOG_ID_CS_00000115_01); //* MSG_LOG_ID_CS_00000115

            return obNode;
        }

        /**
         * 演算式ノード。BRANCHSの（&& や ||）カウント。
         */
        public override object VisitBinaryOperatorExpression(BinaryOperatorExpression binaryOperatorExpression, object data)
        {
            ((NCMStructData)data).dbugbufPrintLn(3, Resources.MSG_LOG_ID_CS_00000116_01); //* MSG_LOG_ID_CS_00000116
            if (flgOutsideFunction && flgBranchCondition)
            {
                ICSharpCode.NCMCSParser.NCMStructData.NCMFuncData fncdata = null;
                fncdata = ((NCMStructData)data).functionGetData();
                if (fncdata != null)
                {
                    switch (binaryOperatorExpression.Operator)
                    {
                        case BinaryOperatorType.ConditionalAnd:
                        case BinaryOperatorType.ConditionalOr:

                            fncdata.cntLogicalAnd++;
                            fncdata.cntBranchs++;
                            ((NCMStructData)data).dbugbufPrintLn(2, Resources.MSG_LOG_ID_CS_00000117_01 + binaryOperatorExpression.Role.ToString()
                            + Resources.MSG_LOG_ID_CS_00000117_02 + binaryOperatorExpression.StartLocation.Line.ToString()
                            + Resources.MSG_LOG_ID_CS_00000117_03 + binaryOperatorExpression.StartLocation.Column.ToString()
                            + Resources.MSG_LOG_ID_CS_00000117_04); //* MSG_LOG_ID_CS_00000117


                            addTargetLine(ref fncdata.counterList, binaryOperatorExpression.StartLocation.Line);
                            fncdata.counterList[binaryOperatorExpression.StartLocation.Line].cntLogicalAnd++;

                            break;
                        default:

                            break;
                    }

                }
                else
                {
                    ((NCMStructData)data).dbugbufPrintLn(2, Resources.MSG_LOG_ID_CS_00000118_01); //* MSG_LOG_ID_CS_00000118
                }
            }

            object obNode = VisitChildren(binaryOperatorExpression, data);

            ((NCMStructData)data).dbugbufPrintLn(3, Resources.MSG_LOG_ID_CS_00000119_01); //* MSG_LOG_ID_CS_00000119

            return obNode;
        }

        /**
         * 三項演算式ノード。CNDOPSのカウント。
         */
        public override object VisitConditionalExpression(ConditionalExpression conditionalExpression, object data)
        {
            ((NCMStructData)data).dbugbufPrintLn(3, Resources.MSG_LOG_ID_CS_00000120_01); //* MSG_LOG_ID_CS_00000120

            if (flgOutsideFunction && !flgBranchCondition)
            {

                ICSharpCode.NCMCSParser.NCMStructData.NCMFuncData fncdata = null;
                fncdata = ((NCMStructData)data).functionGetData();
                if (fncdata != null)
                {
                    fncdata.cntInlinedIf++;
                    fncdata.cntCndOps++;
                    ((NCMStructData)data).dbugbufPrintLn(2, Resources.MSG_LOG_ID_CS_00000121_01
                    + conditionalExpression.Condition.ToString() + Resources.MSG_LOG_ID_CS_00000121_02
                    + conditionalExpression.TrueExpression.ToString() + Resources.MSG_LOG_ID_CS_00000121_03
                    + conditionalExpression.FalseExpression.ToString()
                    + Resources.MSG_LOG_ID_CS_00000121_04 + conditionalExpression.StartLocation.Line.ToString()
                    + Resources.MSG_LOG_ID_CS_00000121_05 + conditionalExpression.StartLocation.Column.ToString()
                    + Resources.MSG_LOG_ID_CS_00000121_06); //* MSG_LOG_ID_CS_00000121








                    addTargetLine(ref fncdata.counterList, conditionalExpression.StartLocation.Line);
                    fncdata.counterList[conditionalExpression.StartLocation.Line].cntCndOps++;
                }
                else
                {
                    ((NCMStructData)data).dbugbufPrintLn(2, Resources.MSG_LOG_ID_CS_00000122_01); //* MSG_LOG_ID_CS_00000122
                }
            }

            object obNode = VisitChildren(conditionalExpression, data);

            ((NCMStructData)data).dbugbufPrintLn(3, Resources.MSG_LOG_ID_CS_00000123_01); //* MSG_LOG_ID_CS_00000123

            return obNode;
        }


        public override object VisitArrayCreateExpression(ArrayCreateExpression arrayCreateExpression, object data)
        {
            ((NCMStructData)data).dbugbufPrintLn(3, Resources.MSG_LOG_ID_CS_00000124_01); //* MSG_LOG_ID_CS_00000124

            if (flgOutsideFunction)
            {

                ICSharpCode.NCMCSParser.NCMStructData.NCMFuncData fncdata = null;
                fncdata = ((NCMStructData)data).functionGetData();
            }

            object obNode = VisitChildren(arrayCreateExpression, data);

            ((NCMStructData)data).dbugbufPrintLn(3, Resources.MSG_LOG_ID_CS_00000125_01); //* MSG_LOG_ID_CS_00000125

            return obNode;
        }

        public override object VisitNullReferenceExpression(NullReferenceExpression nullReferenceExpression, object data)
        {
            ((NCMStructData)data).dbugbufPrintLn(3, Resources.MSG_LOG_ID_CS_00000126_01); //* MSG_LOG_ID_CS_00000126
            object obNode = VisitChildren(nullReferenceExpression, data);

            if (flgOutsideFunction)
            {
                TokenData td = new TokenData("", nullReferenceExpression.NodeType.ToString(),
                                                nullReferenceExpression.Role.ToString(),
                                                nullReferenceExpression.StartLocation,
                                                nullReferenceExpression.EndLocation);

                CountToken(td, data);
            }

            ((NCMStructData)data).dbugbufPrintLn(3, Resources.MSG_LOG_ID_CS_00000127_01); //* MSG_LOG_ID_CS_00000127

            return obNode;
        }

        #endregion

        #region Statement系ノード
        public override object VisitEmptyStatement(EmptyStatement emptyStatement, object data)
        {
            TokenData td = new TokenData("", emptyStatement.NodeType.ToString(), emptyStatement.Role.ToString(),
                emptyStatement.StartLocation, emptyStatement.EndLocation);


            CountToken(td, data);

            object obNode = VisitChildren(emptyStatement, data);
            return obNode;
        }
        #endregion

        #region Token系ノード

        public override object VisitCSharpTokenNode(CSharpTokenNode token, object data)
        {
            TokenData td = new TokenData("", token.NodeType.ToString(),
                                token.Role.ToString(), token.StartLocation, token.EndLocation);


            CountToken(td, data);
            object obNode = VisitChildren(token, data);
            return obNode;
        }

        public override object VisitPrimitiveType(PrimitiveType primitiveType, object data)
        {
            TokenData td = new TokenData(primitiveType.Keyword, primitiveType.NodeType.ToString(),
                               primitiveType.Role.ToString(), primitiveType.StartLocation, primitiveType.EndLocation);
            td.RoleName = primitiveType.Role.ToString();


            CountToken(td, data);

            object obNode = VisitChildren(primitiveType, data);
            return obNode;
        }
        public override object VisitIdentifier(Identifier identifier, object data)
        {
            TokenData td = new TokenData(identifier.Name, identifier.NodeType.ToString(),
                              identifier.Role.ToString(), identifier.Parent.Role.ToString(),
                              identifier.StartLocation, identifier.EndLocation);


            CountToken(td, data);

            object obNode = VisitChildren(identifier, data);
            return obNode;
        }
        public override object VisitComment(Comment comment, object data)
        {
            TokenData td = new TokenData(comment.Content, comment.NodeType.ToString(),
                               comment.Role.ToString(), comment.StartLocation, comment.EndLocation);
            td.ItemType = comment.CommentType.ToString();



            if (commentLineList != null)
            {
                for (int i = comment.StartLocation.Line; i <= comment.EndLocation.Line; i++)
                {
                    commentLineList.Add(i);
                }
            }



            CountToken(td, data);

            object obNode = VisitChildren(comment, data);
            return obNode;
        }
        public override object VisitPrimitiveExpression(PrimitiveExpression primitiveExpression, object data)
        {

            TokenData td = new TokenData(primitiveExpression.LiteralValue, primitiveExpression.NodeType.ToString(),
                               primitiveExpression.Role.ToString(), primitiveExpression.StartLocation, primitiveExpression.EndLocation);



            CountToken(td, data);

            object obNode = VisitChildren(primitiveExpression, data);
            return obNode;
        }

        public override object VisitPreProcessorDirective(PreProcessorDirective preProcessorDirective, object data)
        {

            TokenData td = new TokenData(preProcessorDirective.Argument, preProcessorDirective.NodeType.ToString(),
                               preProcessorDirective.Role.ToString(), preProcessorDirective.StartLocation, preProcessorDirective.EndLocation);
            td.ItemType = preProcessorDirective.Type.ToString();

            td.PpTake = preProcessorDirective.Take;
            td.PpType = preProcessorDirective.Type;


            CountToken(td, data);

            object obNode = VisitChildren(preProcessorDirective, data);
            return obNode;
        }

        #endregion

    }
}
