// created by jay 0.7 (c) 1998 Axel.Schreiner@informatik.uni-osnabrueck.de

///  #line 2 "cs-parser.jay"
//
// cs-parser.jay: The Parser for the C# compiler
//
// Authors: Miguel de Icaza (miguel@gnome.org)
//          Ravi Pratap     (ravi@ximian.com)
//          Marek Safar	    (marek.safar@gmail.com)
//
// Dual Licensed under the terms of the GNU GPL and the MIT X11 license
//
// (C) 2001 Ximian, Inc (http://www.ximian.com)
// (C) 2004-2011 Novell, Inc
// Copyright 2011 Xamarin Inc.
//
// TODO:
//   (1) Figure out why error productions dont work.  `type-declaration' is a
//       great spot to put an `error' because you can reproduce it with this input:
//	 "public X { }"
//
/* Modified by NEC Corporation in 2015. */

using System.Text;
using System.IO;
using System;
using System.Collections.Generic;

namespace Mono.CSharp
{
	/// <summary>
	///    The C# Parser
	/// </summary>
	public class CSharpParser
	{
		[Flags]
		enum ParameterModifierType
		{
			Ref		= 1 << 1,
			Out		= 1 << 2,
			This	= 1 << 3,
			Params	= 1 << 4,
			Arglist	= 1 << 5,
			DefaultValue = 1 << 6,
			
			All = Ref | Out | This | Params | Arglist | DefaultValue
		}
		
		static readonly object ModifierNone = 0;
	
		NamespaceContainer  current_namespace;
		TypeContainer   current_container;
		DeclSpace	current_class;
		PropertyBase current_property;
		EventProperty current_event;
		EventField current_event_field;
		FieldBase current_field;
	
		/// <summary>
		///   Current block is used to add statements as we find
		///   them.  
		/// </summary>
		Block      current_block;
		
		BlockVariableDeclaration current_variable;

		Delegate   current_delegate;
		
		AnonymousMethodExpression current_anonymous_method;

		/// <summary>
		///   This is used by the unary_expression code to resolve
		///   a name against a parameter.  
		/// </summary>
		
		// FIXME: This is very ugly and it's very hard to reset it correctly
		// on all places, especially when some parameters are autogenerated.
		ParametersCompiled current_local_parameters;

		bool parsing_anonymous_method;
		
		bool async_block;

		///
		/// An out-of-band stack.
		///
		static Stack<object> oob_stack;

		///
		/// Controls the verbosity of the errors produced by the parser
		///
		static public int yacc_verbose_flag;

		/// 
		/// Used by the interactive shell, flags whether EOF was reached
		/// and an error was produced
		///
		public bool UnexpectedEOF;

		///
		/// The current file.
		///
		readonly CompilationSourceFile file;

		///
		/// Temporary Xml documentation cache.
		/// For enum types, we need one more temporary store.
		///
		string tmpComment;
		string enumTypeComment;
	       		
		/// Current attribute target
		string current_attr_target;
		
		ParameterModifierType valid_param_mod;
		
		bool default_parameter_used;

		/// When using the interactive parser, this holds the
		/// resulting expression
		public Class InteractiveResult;

		//
		// Keeps track of global data changes to undo on parser error
		//
		public Undo undo;
		
		Stack<Linq.QueryBlock> linq_clause_blocks;

		ModuleContainer module;
		
		readonly CompilerContext compiler;
		readonly LanguageVersion lang_version;
		readonly bool doc_support;
		readonly CompilerSettings settings;
		readonly Report report;
		
		//
		// Instead of allocating carrier array everytime we
		// share the bucket for very common constructs which can never
		// be recursive
		//
		static List<Parameter> parameters_bucket = new List<Parameter> (6);
		
		//
		// Full AST support members
		//
		LocationsBag lbag;
		UsingsBag ubag;
		List<Tuple<Modifiers, Location>> mod_locations;
		Location parameterModifierLocation, savedLocation, savedOpenLocation, savedCloseLocation;
		Location savedAttrParenOpenLocation, savedAttrParenCloseLocation, savedOperatorLocation;
		Stack<List<Location>> locationListStack = new Stack<List<Location>> (); // used for type parameters
		List<Location> attributeCommas = new List<Location> ();
		List<Location> attributeArgumentCommas = new List<Location> ();
		List<Location> parameterListCommas = new List<Location> ();
///  #line default

  /** error output stream.
      It should be changeable.
    */
  public System.IO.TextWriter ErrorOutput = System.Console.Out;

  /** simplified error message.
      @see <a href="#yyerror(java.lang.String, java.lang.String[])">yyerror</a>
    */
  public void yyerror (string message) {
    yyerror(message, null);
  }

  /* An EOF token */
  public int eof_token;

  /** (syntax) error message.
      Can be overwritten to control message format.
      @param message text to be displayed.
      @param expected vector of acceptable tokens, if available.
    */
  public void yyerror (string message, string[] expected) {
    if ((yacc_verbose_flag > 0) && (expected != null) && (expected.Length  > 0)) {
      ErrorOutput.Write (message+", expecting");
      for (int n = 0; n < expected.Length; ++ n)
        ErrorOutput.Write (" "+expected[n]);
        ErrorOutput.WriteLine ();
    } else
      ErrorOutput.WriteLine (message);
  }

  /** debugging support, requires the package jay.yydebug.
      Set to null to suppress debugging messages.
    */
//t  internal yydebug.yyDebug debug;

  protected const int yyFinal = 7;
//t // Put this array into a separate class so it is only initialized if debugging is actually used
//t // Use MarshalByRefObject to disable inlining
//t class YYRules : MarshalByRefObject {
//t  public static readonly string [] yyRule = {
//t    "$accept : compilation_unit",
//t    "compilation_unit : outer_declaration opt_EOF",
//t    "$$1 :",
//t    "compilation_unit : interactive_parsing $$1 opt_EOF",
//t    "compilation_unit : documentation_parsing",
//t    "outer_declaration : opt_extern_alias_directives opt_using_directives",
//t    "outer_declaration : opt_extern_alias_directives opt_using_directives namespace_or_type_declarations opt_attributes",
//t    "outer_declaration : opt_extern_alias_directives opt_using_directives attribute_sections",
//t    "outer_declaration : error",
//t    "opt_EOF :",
//t    "opt_EOF : EOF",
//t    "extern_alias_directives : extern_alias_directive",
//t    "extern_alias_directives : extern_alias_directives extern_alias_directive",
//t    "extern_alias_directive : EXTERN_ALIAS IDENTIFIER IDENTIFIER SEMICOLON",
//t    "extern_alias_directive : EXTERN_ALIAS error",
//t    "using_directives : using_directive",
//t    "using_directives : using_directives using_directive",
//t    "using_directive : using_alias_directive",
//t    "using_directive : using_namespace_directive",
//t    "using_alias_directive : USING IDENTIFIER ASSIGN namespace_or_type_name SEMICOLON",
//t    "using_alias_directive : USING error",
//t    "using_namespace_directive : USING namespace_name SEMICOLON",
//t    "$$2 :",
//t    "$$3 :",
//t    "namespace_declaration : opt_attributes NAMESPACE qualified_identifier $$2 OPEN_BRACE $$3 opt_extern_alias_directives opt_using_directives opt_namespace_or_type_declarations CLOSE_BRACE opt_semicolon",
//t    "qualified_identifier : IDENTIFIER",
//t    "qualified_identifier : qualified_identifier DOT IDENTIFIER",
//t    "qualified_identifier : error",
//t    "opt_semicolon :",
//t    "opt_semicolon : SEMICOLON",
//t    "opt_comma :",
//t    "opt_comma : COMMA",
//t    "namespace_name : namespace_or_type_name",
//t    "opt_using_directives :",
//t    "opt_using_directives : using_directives",
//t    "opt_extern_alias_directives :",
//t    "opt_extern_alias_directives : extern_alias_directives",
//t    "opt_namespace_or_type_declarations :",
//t    "opt_namespace_or_type_declarations : namespace_or_type_declarations",
//t    "namespace_or_type_declarations : namespace_or_type_declaration",
//t    "namespace_or_type_declarations : namespace_or_type_declarations namespace_or_type_declaration",
//t    "namespace_or_type_declaration : type_declaration",
//t    "namespace_or_type_declaration : namespace_declaration",
//t    "type_declaration : class_declaration",
//t    "type_declaration : struct_declaration",
//t    "type_declaration : interface_declaration",
//t    "type_declaration : enum_declaration",
//t    "type_declaration : delegate_declaration",
//t    "opt_attributes :",
//t    "opt_attributes : attribute_sections",
//t    "attribute_sections : attribute_section",
//t    "attribute_sections : attribute_sections attribute_section",
//t    "$$4 :",
//t    "attribute_section : OPEN_BRACKET $$4 attribute_section_cont",
//t    "$$5 :",
//t    "attribute_section_cont : attribute_target COLON $$5 attribute_list opt_comma CLOSE_BRACKET",
//t    "attribute_section_cont : attribute_list opt_comma CLOSE_BRACKET",
//t    "attribute_target : IDENTIFIER",
//t    "attribute_target : EVENT",
//t    "attribute_target : RETURN",
//t    "attribute_target : error",
//t    "attribute_list : attribute",
//t    "attribute_list : attribute_list COMMA attribute",
//t    "$$6 :",
//t    "attribute : attribute_name $$6 opt_attribute_arguments",
//t    "attribute_name : namespace_or_type_name",
//t    "opt_attribute_arguments :",
//t    "opt_attribute_arguments : OPEN_PARENS attribute_arguments CLOSE_PARENS",
//t    "attribute_arguments :",
//t    "attribute_arguments : positional_or_named_argument",
//t    "attribute_arguments : named_attribute_argument",
//t    "attribute_arguments : attribute_arguments COMMA positional_or_named_argument",
//t    "attribute_arguments : attribute_arguments COMMA named_attribute_argument",
//t    "positional_or_named_argument : expression",
//t    "positional_or_named_argument : named_argument",
//t    "$$7 :",
//t    "named_attribute_argument : IDENTIFIER ASSIGN $$7 expression",
//t    "named_argument : identifier_inside_body COLON opt_named_modifier expression",
//t    "opt_named_modifier :",
//t    "opt_named_modifier : REF",
//t    "opt_named_modifier : OUT",
//t    "opt_class_member_declarations :",
//t    "opt_class_member_declarations : class_member_declarations",
//t    "class_member_declarations : class_member_declaration",
//t    "class_member_declarations : class_member_declarations class_member_declaration",
//t    "class_member_declaration : constant_declaration",
//t    "class_member_declaration : field_declaration",
//t    "class_member_declaration : method_declaration",
//t    "class_member_declaration : property_declaration",
//t    "class_member_declaration : event_declaration",
//t    "class_member_declaration : indexer_declaration",
//t    "class_member_declaration : operator_declaration",
//t    "class_member_declaration : constructor_declaration",
//t    "class_member_declaration : destructor_declaration",
//t    "class_member_declaration : type_declaration",
//t    "class_member_declaration : error",
//t    "$$8 :",
//t    "$$9 :",
//t    "$$10 :",
//t    "$$11 :",
//t    "$$12 :",
//t    "struct_declaration : opt_attributes opt_modifiers opt_partial STRUCT $$8 type_declaration_name $$9 opt_class_base opt_type_parameter_constraints_clauses $$10 OPEN_BRACE $$11 opt_class_member_declarations CLOSE_BRACE $$12 opt_semicolon",
//t    "struct_declaration : opt_attributes opt_modifiers opt_partial STRUCT error",
//t    "$$13 :",
//t    "constant_declaration : opt_attributes opt_modifiers CONST type IDENTIFIER $$13 constant_initializer opt_constant_declarators SEMICOLON",
//t    "opt_constant_declarators :",
//t    "opt_constant_declarators : constant_declarators",
//t    "constant_declarators : constant_declarator",
//t    "constant_declarators : constant_declarators constant_declarator",
//t    "constant_declarator : COMMA IDENTIFIER constant_initializer",
//t    "$$14 :",
//t    "constant_initializer : ASSIGN $$14 constant_initializer_expr",
//t    "constant_initializer : error",
//t    "constant_initializer_expr : constant_expression",
//t    "constant_initializer_expr : array_initializer",
//t    "$$15 :",
//t    "field_declaration : opt_attributes opt_modifiers member_type IDENTIFIER $$15 opt_field_initializer opt_field_declarators SEMICOLON",
//t    "$$16 :",
//t    "field_declaration : opt_attributes opt_modifiers FIXED simple_type IDENTIFIER $$16 fixed_field_size opt_fixed_field_declarators SEMICOLON",
//t    "field_declaration : opt_attributes opt_modifiers FIXED simple_type error SEMICOLON",
//t    "opt_field_initializer :",
//t    "$$17 :",
//t    "opt_field_initializer : ASSIGN $$17 variable_initializer",
//t    "opt_field_declarators :",
//t    "opt_field_declarators : field_declarators",
//t    "field_declarators : field_declarator",
//t    "field_declarators : field_declarators field_declarator",
//t    "field_declarator : COMMA IDENTIFIER",
//t    "$$18 :",
//t    "field_declarator : COMMA IDENTIFIER ASSIGN $$18 variable_initializer",
//t    "opt_fixed_field_declarators :",
//t    "opt_fixed_field_declarators : fixed_field_declarators",
//t    "fixed_field_declarators : fixed_field_declarator",
//t    "fixed_field_declarators : fixed_field_declarators fixed_field_declarator",
//t    "fixed_field_declarator : COMMA IDENTIFIER fixed_field_size",
//t    "$$19 :",
//t    "fixed_field_size : OPEN_BRACKET $$19 expression CLOSE_BRACKET",
//t    "fixed_field_size : OPEN_BRACKET error",
//t    "variable_initializer : expression",
//t    "variable_initializer : array_initializer",
//t    "variable_initializer : error",
//t    "$$20 :",
//t    "method_declaration : method_header $$20 method_body",
//t    "$$21 :",
//t    "$$22 :",
//t    "method_header : opt_attributes opt_modifiers member_type method_declaration_name OPEN_PARENS $$21 opt_formal_parameter_list CLOSE_PARENS $$22 opt_type_parameter_constraints_clauses",
//t    "$$23 :",
//t    "$$24 :",
//t    "$$25 :",
//t    "method_header : opt_attributes opt_modifiers PARTIAL VOID $$23 method_declaration_name OPEN_PARENS $$24 opt_formal_parameter_list CLOSE_PARENS $$25 opt_type_parameter_constraints_clauses",
//t    "method_header : opt_attributes opt_modifiers member_type modifiers method_declaration_name OPEN_PARENS opt_formal_parameter_list CLOSE_PARENS",
//t    "method_body : block",
//t    "method_body : SEMICOLON",
//t    "opt_formal_parameter_list :",
//t    "opt_formal_parameter_list : formal_parameter_list",
//t    "formal_parameter_list : fixed_parameters",
//t    "formal_parameter_list : fixed_parameters COMMA parameter_array",
//t    "formal_parameter_list : fixed_parameters COMMA arglist_modifier",
//t    "formal_parameter_list : parameter_array COMMA error",
//t    "formal_parameter_list : fixed_parameters COMMA parameter_array COMMA error",
//t    "formal_parameter_list : arglist_modifier COMMA error",
//t    "formal_parameter_list : fixed_parameters COMMA ARGLIST COMMA error",
//t    "formal_parameter_list : parameter_array",
//t    "formal_parameter_list : arglist_modifier",
//t    "formal_parameter_list : error",
//t    "fixed_parameters : fixed_parameter",
//t    "fixed_parameters : fixed_parameters COMMA fixed_parameter",
//t    "fixed_parameter : opt_attributes opt_parameter_modifier parameter_type IDENTIFIER",
//t    "fixed_parameter : opt_attributes opt_parameter_modifier parameter_type IDENTIFIER OPEN_BRACKET CLOSE_BRACKET",
//t    "fixed_parameter : opt_attributes opt_parameter_modifier parameter_type error",
//t    "$$26 :",
//t    "fixed_parameter : opt_attributes opt_parameter_modifier parameter_type IDENTIFIER ASSIGN $$26 constant_expression",
//t    "opt_parameter_modifier :",
//t    "opt_parameter_modifier : parameter_modifiers",
//t    "parameter_modifiers : parameter_modifier",
//t    "parameter_modifiers : parameter_modifiers parameter_modifier",
//t    "parameter_modifier : REF",
//t    "parameter_modifier : OUT",
//t    "parameter_modifier : THIS",
//t    "parameter_array : opt_attributes params_modifier type IDENTIFIER",
//t    "parameter_array : opt_attributes params_modifier type IDENTIFIER ASSIGN constant_expression",
//t    "parameter_array : opt_attributes params_modifier type error",
//t    "params_modifier : PARAMS",
//t    "params_modifier : PARAMS parameter_modifier",
//t    "params_modifier : PARAMS params_modifier",
//t    "arglist_modifier : ARGLIST",
//t    "$$27 :",
//t    "$$28 :",
//t    "$$29 :",
//t    "property_declaration : opt_attributes opt_modifiers member_type member_declaration_name $$27 OPEN_BRACE $$28 accessor_declarations $$29 CLOSE_BRACE",
//t    "$$30 :",
//t    "$$31 :",
//t    "$$32 :",
//t    "indexer_declaration : opt_attributes opt_modifiers member_type indexer_declaration_name OPEN_BRACKET $$30 opt_formal_parameter_list CLOSE_BRACKET OPEN_BRACE $$31 accessor_declarations $$32 CLOSE_BRACE",
//t    "accessor_declarations : get_accessor_declaration",
//t    "accessor_declarations : get_accessor_declaration accessor_declarations",
//t    "accessor_declarations : set_accessor_declaration",
//t    "accessor_declarations : set_accessor_declaration accessor_declarations",
//t    "accessor_declarations : error",
//t    "$$33 :",
//t    "get_accessor_declaration : opt_attributes opt_modifiers GET $$33 accessor_body",
//t    "$$34 :",
//t    "set_accessor_declaration : opt_attributes opt_modifiers SET $$34 accessor_body",
//t    "accessor_body : block",
//t    "accessor_body : SEMICOLON",
//t    "accessor_body : error",
//t    "$$35 :",
//t    "$$36 :",
//t    "$$37 :",
//t    "$$38 :",
//t    "interface_declaration : opt_attributes opt_modifiers opt_partial INTERFACE $$35 type_declaration_name $$36 opt_class_base opt_type_parameter_constraints_clauses $$37 OPEN_BRACE opt_interface_member_declarations CLOSE_BRACE $$38 opt_semicolon",
//t    "interface_declaration : opt_attributes opt_modifiers opt_partial INTERFACE error",
//t    "opt_interface_member_declarations :",
//t    "opt_interface_member_declarations : interface_member_declarations",
//t    "interface_member_declarations : interface_member_declaration",
//t    "interface_member_declarations : interface_member_declarations interface_member_declaration",
//t    "interface_member_declaration : constant_declaration",
//t    "interface_member_declaration : field_declaration",
//t    "interface_member_declaration : method_declaration",
//t    "interface_member_declaration : property_declaration",
//t    "interface_member_declaration : event_declaration",
//t    "interface_member_declaration : indexer_declaration",
//t    "interface_member_declaration : operator_declaration",
//t    "interface_member_declaration : constructor_declaration",
//t    "interface_member_declaration : type_declaration",
//t    "$$39 :",
//t    "operator_declaration : opt_attributes opt_modifiers operator_declarator $$39 operator_body",
//t    "operator_body : block",
//t    "operator_body : SEMICOLON",
//t    "operator_type : type_expression_or_array",
//t    "operator_type : VOID",
//t    "$$40 :",
//t    "operator_declarator : operator_type OPERATOR overloadable_operator OPEN_PARENS $$40 opt_formal_parameter_list CLOSE_PARENS",
//t    "operator_declarator : conversion_operator_declarator",
//t    "overloadable_operator : BANG",
//t    "overloadable_operator : TILDE",
//t    "overloadable_operator : OP_INC",
//t    "overloadable_operator : OP_DEC",
//t    "overloadable_operator : TRUE",
//t    "overloadable_operator : FALSE",
//t    "overloadable_operator : PLUS",
//t    "overloadable_operator : MINUS",
//t    "overloadable_operator : STAR",
//t    "overloadable_operator : DIV",
//t    "overloadable_operator : PERCENT",
//t    "overloadable_operator : BITWISE_AND",
//t    "overloadable_operator : BITWISE_OR",
//t    "overloadable_operator : CARRET",
//t    "overloadable_operator : OP_SHIFT_LEFT",
//t    "overloadable_operator : OP_SHIFT_RIGHT",
//t    "overloadable_operator : OP_EQ",
//t    "overloadable_operator : OP_NE",
//t    "overloadable_operator : OP_GT",
//t    "overloadable_operator : OP_LT",
//t    "overloadable_operator : OP_GE",
//t    "overloadable_operator : OP_LE",
//t    "$$41 :",
//t    "conversion_operator_declarator : IMPLICIT OPERATOR type OPEN_PARENS $$41 opt_formal_parameter_list CLOSE_PARENS",
//t    "$$42 :",
//t    "conversion_operator_declarator : EXPLICIT OPERATOR type OPEN_PARENS $$42 opt_formal_parameter_list CLOSE_PARENS",
//t    "conversion_operator_declarator : IMPLICIT error",
//t    "conversion_operator_declarator : EXPLICIT error",
//t    "constructor_declaration : constructor_declarator constructor_body",
//t    "$$43 :",
//t    "$$44 :",
//t    "constructor_declarator : opt_attributes opt_modifiers IDENTIFIER $$43 OPEN_PARENS opt_formal_parameter_list CLOSE_PARENS $$44 opt_constructor_initializer",
//t    "constructor_body : block_prepared",
//t    "constructor_body : SEMICOLON",
//t    "opt_constructor_initializer :",
//t    "opt_constructor_initializer : constructor_initializer",
//t    "$$45 :",
//t    "constructor_initializer : COLON BASE OPEN_PARENS $$45 opt_argument_list CLOSE_PARENS",
//t    "$$46 :",
//t    "constructor_initializer : COLON THIS OPEN_PARENS $$46 opt_argument_list CLOSE_PARENS",
//t    "constructor_initializer : error",
//t    "$$47 :",
//t    "destructor_declaration : opt_attributes opt_modifiers TILDE $$47 IDENTIFIER OPEN_PARENS CLOSE_PARENS method_body",
//t    "$$48 :",
//t    "event_declaration : opt_attributes opt_modifiers EVENT type member_declaration_name $$48 opt_event_initializer opt_event_declarators SEMICOLON",
//t    "$$49 :",
//t    "$$50 :",
//t    "event_declaration : opt_attributes opt_modifiers EVENT type member_declaration_name OPEN_BRACE $$49 event_accessor_declarations $$50 CLOSE_BRACE",
//t    "opt_event_initializer :",
//t    "$$51 :",
//t    "opt_event_initializer : ASSIGN $$51 event_variable_initializer",
//t    "opt_event_declarators :",
//t    "opt_event_declarators : event_declarators",
//t    "event_declarators : event_declarator",
//t    "event_declarators : event_declarators event_declarator",
//t    "event_declarator : COMMA IDENTIFIER",
//t    "$$52 :",
//t    "event_declarator : COMMA IDENTIFIER ASSIGN $$52 event_variable_initializer",
//t    "$$53 :",
//t    "event_variable_initializer : $$53 variable_initializer",
//t    "event_accessor_declarations : add_accessor_declaration remove_accessor_declaration",
//t    "event_accessor_declarations : remove_accessor_declaration add_accessor_declaration",
//t    "event_accessor_declarations : add_accessor_declaration",
//t    "event_accessor_declarations : remove_accessor_declaration",
//t    "event_accessor_declarations : error",
//t    "$$54 :",
//t    "add_accessor_declaration : opt_attributes opt_modifiers ADD $$54 event_accessor_block",
//t    "$$55 :",
//t    "remove_accessor_declaration : opt_attributes opt_modifiers REMOVE $$55 event_accessor_block",
//t    "event_accessor_block : opt_semicolon",
//t    "event_accessor_block : block",
//t    "$$56 :",
//t    "$$57 :",
//t    "$$58 :",
//t    "enum_declaration : opt_attributes opt_modifiers ENUM type_declaration_name opt_enum_base $$56 OPEN_BRACE $$57 opt_enum_member_declarations $$58 CLOSE_BRACE opt_semicolon",
//t    "opt_enum_base :",
//t    "opt_enum_base : COLON type",
//t    "opt_enum_base : COLON error",
//t    "opt_enum_member_declarations :",
//t    "opt_enum_member_declarations : enum_member_declarations",
//t    "opt_enum_member_declarations : enum_member_declarations COMMA",
//t    "enum_member_declarations : enum_member_declaration",
//t    "enum_member_declarations : enum_member_declarations COMMA enum_member_declaration",
//t    "enum_member_declaration : opt_attributes IDENTIFIER",
//t    "$$59 :",
//t    "enum_member_declaration : opt_attributes IDENTIFIER $$59 ASSIGN constant_expression",
//t    "$$60 :",
//t    "$$61 :",
//t    "$$62 :",
//t    "delegate_declaration : opt_attributes opt_modifiers DELEGATE member_type type_declaration_name OPEN_PARENS $$60 opt_formal_parameter_list CLOSE_PARENS $$61 opt_type_parameter_constraints_clauses $$62 SEMICOLON",
//t    "opt_nullable :",
//t    "opt_nullable : INTERR_NULLABLE",
//t    "namespace_or_type_name : member_name",
//t    "namespace_or_type_name : qualified_alias_member IDENTIFIER opt_type_argument_list",
//t    "member_name : type_name",
//t    "member_name : namespace_or_type_name DOT IDENTIFIER opt_type_argument_list",
//t    "type_name : IDENTIFIER opt_type_argument_list",
//t    "opt_type_argument_list :",
//t    "opt_type_argument_list : OP_GENERICS_LT type_arguments OP_GENERICS_GT",
//t    "opt_type_argument_list : OP_GENERICS_LT error",
//t    "type_arguments : type",
//t    "type_arguments : type_arguments COMMA type",
//t    "$$63 :",
//t    "type_declaration_name : IDENTIFIER $$63 opt_type_parameter_list",
//t    "member_declaration_name : method_declaration_name",
//t    "method_declaration_name : type_declaration_name",
//t    "method_declaration_name : explicit_interface IDENTIFIER opt_type_parameter_list",
//t    "indexer_declaration_name : THIS",
//t    "indexer_declaration_name : explicit_interface THIS",
//t    "explicit_interface : IDENTIFIER opt_type_argument_list DOT",
//t    "explicit_interface : qualified_alias_member IDENTIFIER opt_type_argument_list DOT",
//t    "explicit_interface : explicit_interface IDENTIFIER opt_type_argument_list DOT",
//t    "opt_type_parameter_list :",
//t    "opt_type_parameter_list : OP_GENERICS_LT_DECL type_parameters OP_GENERICS_GT",
//t    "type_parameters : type_parameter",
//t    "type_parameters : type_parameters COMMA type_parameter",
//t    "type_parameter : opt_attributes opt_type_parameter_variance IDENTIFIER",
//t    "type_parameter : error",
//t    "type_and_void : type_expression_or_array",
//t    "type_and_void : VOID",
//t    "member_type : type_and_void",
//t    "type : type_expression_or_array",
//t    "type : VOID",
//t    "simple_type : type_expression",
//t    "simple_type : VOID",
//t    "parameter_type : type_expression_or_array",
//t    "parameter_type : VOID",
//t    "type_expression_or_array : type_expression",
//t    "type_expression_or_array : type_expression rank_specifiers",
//t    "type_expression : namespace_or_type_name opt_nullable",
//t    "type_expression : namespace_or_type_name pointer_stars",
//t    "type_expression : builtin_types opt_nullable",
//t    "type_expression : builtin_types pointer_stars",
//t    "type_expression : VOID pointer_stars",
//t    "type_list : base_type_name",
//t    "type_list : type_list COMMA base_type_name",
//t    "base_type_name : type",
//t    "base_type_name : error",
//t    "builtin_types : OBJECT",
//t    "builtin_types : STRING",
//t    "builtin_types : BOOL",
//t    "builtin_types : DECIMAL",
//t    "builtin_types : FLOAT",
//t    "builtin_types : DOUBLE",
//t    "builtin_types : integral_type",
//t    "integral_type : SBYTE",
//t    "integral_type : BYTE",
//t    "integral_type : SHORT",
//t    "integral_type : USHORT",
//t    "integral_type : INT",
//t    "integral_type : UINT",
//t    "integral_type : LONG",
//t    "integral_type : ULONG",
//t    "integral_type : CHAR",
//t    "primary_expression : primary_expression_or_type",
//t    "primary_expression : literal",
//t    "primary_expression : array_creation_expression",
//t    "primary_expression : parenthesized_expression",
//t    "primary_expression : default_value_expression",
//t    "primary_expression : invocation_expression",
//t    "primary_expression : element_access",
//t    "primary_expression : this_access",
//t    "primary_expression : base_access",
//t    "primary_expression : post_increment_expression",
//t    "primary_expression : post_decrement_expression",
//t    "primary_expression : object_or_delegate_creation_expression",
//t    "primary_expression : anonymous_type_expression",
//t    "primary_expression : typeof_expression",
//t    "primary_expression : sizeof_expression",
//t    "primary_expression : checked_expression",
//t    "primary_expression : unchecked_expression",
//t    "primary_expression : pointer_member_access",
//t    "primary_expression : anonymous_method_expression",
//t    "primary_expression : undocumented_expressions",
//t    "primary_expression_or_type : IDENTIFIER opt_type_argument_list",
//t    "primary_expression_or_type : IDENTIFIER GENERATE_COMPLETION",
//t    "primary_expression_or_type : member_access",
//t    "literal : boolean_literal",
//t    "literal : LITERAL",
//t    "literal : NULL",
//t    "boolean_literal : TRUE",
//t    "boolean_literal : FALSE",
//t    "open_parens_any : OPEN_PARENS",
//t    "open_parens_any : OPEN_PARENS_CAST",
//t    "close_parens : CLOSE_PARENS",
//t    "close_parens : COMPLETE_COMPLETION",
//t    "parenthesized_expression : OPEN_PARENS expression CLOSE_PARENS",
//t    "parenthesized_expression : OPEN_PARENS expression COMPLETE_COMPLETION",
//t    "member_access : primary_expression DOT IDENTIFIER opt_type_argument_list",
//t    "member_access : builtin_types DOT IDENTIFIER opt_type_argument_list",
//t    "member_access : BASE DOT IDENTIFIER opt_type_argument_list",
//t    "member_access : qualified_alias_member IDENTIFIER opt_type_argument_list",
//t    "member_access : primary_expression DOT GENERATE_COMPLETION",
//t    "member_access : primary_expression DOT IDENTIFIER GENERATE_COMPLETION",
//t    "member_access : builtin_types DOT GENERATE_COMPLETION",
//t    "member_access : builtin_types DOT IDENTIFIER GENERATE_COMPLETION",
//t    "invocation_expression : primary_expression open_parens_any opt_argument_list close_parens",
//t    "opt_object_or_collection_initializer :",
//t    "opt_object_or_collection_initializer : object_or_collection_initializer",
//t    "object_or_collection_initializer : OPEN_BRACE opt_member_initializer_list close_brace_or_complete_completion",
//t    "object_or_collection_initializer : OPEN_BRACE member_initializer_list COMMA CLOSE_BRACE",
//t    "opt_member_initializer_list :",
//t    "opt_member_initializer_list : member_initializer_list",
//t    "member_initializer_list : member_initializer",
//t    "member_initializer_list : member_initializer_list COMMA member_initializer",
//t    "member_initializer_list : member_initializer_list error",
//t    "member_initializer : IDENTIFIER ASSIGN initializer_value",
//t    "member_initializer : GENERATE_COMPLETION",
//t    "member_initializer : non_assignment_expression opt_COMPLETE_COMPLETION",
//t    "member_initializer : OPEN_BRACE expression_list CLOSE_BRACE",
//t    "member_initializer : OPEN_BRACE CLOSE_BRACE",
//t    "initializer_value : expression",
//t    "initializer_value : object_or_collection_initializer",
//t    "opt_argument_list :",
//t    "opt_argument_list : argument_list",
//t    "argument_list : argument_or_named_argument",
//t    "argument_list : argument_list COMMA argument",
//t    "argument_list : argument_list COMMA named_argument",
//t    "argument_list : argument_list COMMA",
//t    "argument_list : COMMA error",
//t    "argument : expression",
//t    "argument : non_simple_argument",
//t    "argument_or_named_argument : argument",
//t    "argument_or_named_argument : named_argument",
//t    "non_simple_argument : REF variable_reference",
//t    "non_simple_argument : OUT variable_reference",
//t    "non_simple_argument : ARGLIST OPEN_PARENS argument_list CLOSE_PARENS",
//t    "non_simple_argument : ARGLIST OPEN_PARENS CLOSE_PARENS",
//t    "variable_reference : expression",
//t    "element_access : primary_expression OPEN_BRACKET_EXPR expression_list_arguments CLOSE_BRACKET",
//t    "element_access : primary_expression OPEN_BRACKET_EXPR expression_list_arguments error",
//t    "element_access : primary_expression OPEN_BRACKET_EXPR error",
//t    "expression_list : expression",
//t    "expression_list : expression_list COMMA expression",
//t    "expression_list : expression_list error",
//t    "expression_list_arguments : expression_list_argument",
//t    "expression_list_arguments : expression_list_arguments COMMA expression_list_argument",
//t    "expression_list_argument : expression",
//t    "expression_list_argument : named_argument",
//t    "this_access : THIS",
//t    "base_access : BASE OPEN_BRACKET_EXPR expression_list_arguments CLOSE_BRACKET",
//t    "base_access : BASE OPEN_BRACKET error",
//t    "post_increment_expression : primary_expression OP_INC",
//t    "post_decrement_expression : primary_expression OP_DEC",
//t    "object_or_delegate_creation_expression : NEW new_expr_type open_parens_any opt_argument_list CLOSE_PARENS opt_object_or_collection_initializer",
//t    "object_or_delegate_creation_expression : NEW new_expr_type object_or_collection_initializer",
//t    "array_creation_expression : NEW new_expr_type OPEN_BRACKET_EXPR expression_list CLOSE_BRACKET opt_rank_specifier opt_array_initializer",
//t    "array_creation_expression : NEW new_expr_type rank_specifiers opt_array_initializer",
//t    "array_creation_expression : NEW rank_specifier array_initializer",
//t    "array_creation_expression : NEW new_expr_type OPEN_BRACKET CLOSE_BRACKET OPEN_BRACKET_EXPR error CLOSE_BRACKET",
//t    "array_creation_expression : NEW new_expr_type error",
//t    "$$64 :",
//t    "new_expr_type : $$64 simple_type",
//t    "anonymous_type_expression : NEW OPEN_BRACE anonymous_type_parameters_opt_comma CLOSE_BRACE",
//t    "anonymous_type_parameters_opt_comma : anonymous_type_parameters_opt",
//t    "anonymous_type_parameters_opt_comma : anonymous_type_parameters COMMA",
//t    "anonymous_type_parameters_opt :",
//t    "anonymous_type_parameters_opt : anonymous_type_parameters",
//t    "anonymous_type_parameters : anonymous_type_parameter",
//t    "anonymous_type_parameters : anonymous_type_parameters COMMA anonymous_type_parameter",
//t    "anonymous_type_parameter : IDENTIFIER ASSIGN variable_initializer",
//t    "anonymous_type_parameter : IDENTIFIER",
//t    "anonymous_type_parameter : member_access",
//t    "anonymous_type_parameter : error",
//t    "opt_rank_specifier :",
//t    "opt_rank_specifier : rank_specifiers",
//t    "rank_specifiers : rank_specifier",
//t    "rank_specifiers : rank_specifier rank_specifiers",
//t    "rank_specifier : OPEN_BRACKET CLOSE_BRACKET",
//t    "rank_specifier : OPEN_BRACKET dim_separators CLOSE_BRACKET",
//t    "dim_separators : COMMA",
//t    "dim_separators : dim_separators COMMA",
//t    "opt_array_initializer :",
//t    "opt_array_initializer : array_initializer",
//t    "array_initializer : OPEN_BRACE CLOSE_BRACE",
//t    "array_initializer : OPEN_BRACE variable_initializer_list opt_comma CLOSE_BRACE",
//t    "variable_initializer_list : variable_initializer",
//t    "variable_initializer_list : variable_initializer_list COMMA variable_initializer",
//t    "$$65 :",
//t    "typeof_expression : TYPEOF $$65 open_parens_any typeof_type_expression CLOSE_PARENS",
//t    "typeof_type_expression : type_and_void",
//t    "typeof_type_expression : unbound_type_name",
//t    "typeof_type_expression : error",
//t    "unbound_type_name : identifier_inside_body generic_dimension",
//t    "unbound_type_name : qualified_alias_member identifier_inside_body generic_dimension",
//t    "unbound_type_name : unbound_type_name DOT identifier_inside_body",
//t    "unbound_type_name : unbound_type_name DOT identifier_inside_body generic_dimension",
//t    "unbound_type_name : namespace_or_type_name DOT identifier_inside_body generic_dimension",
//t    "generic_dimension : GENERIC_DIMENSION",
//t    "qualified_alias_member : IDENTIFIER DOUBLE_COLON",
//t    "sizeof_expression : SIZEOF open_parens_any type CLOSE_PARENS",
//t    "checked_expression : CHECKED open_parens_any expression CLOSE_PARENS",
//t    "unchecked_expression : UNCHECKED open_parens_any expression CLOSE_PARENS",
//t    "pointer_member_access : primary_expression OP_PTR IDENTIFIER opt_type_argument_list",
//t    "$$66 :",
//t    "anonymous_method_expression : DELEGATE opt_anonymous_method_signature $$66 block",
//t    "$$67 :",
//t    "anonymous_method_expression : ASYNC DELEGATE opt_anonymous_method_signature $$67 block",
//t    "opt_anonymous_method_signature :",
//t    "opt_anonymous_method_signature : anonymous_method_signature",
//t    "$$68 :",
//t    "anonymous_method_signature : OPEN_PARENS $$68 opt_formal_parameter_list CLOSE_PARENS",
//t    "default_value_expression : DEFAULT open_parens_any type CLOSE_PARENS",
//t    "unary_expression : primary_expression",
//t    "unary_expression : BANG prefixed_unary_expression",
//t    "unary_expression : TILDE prefixed_unary_expression",
//t    "unary_expression : OPEN_PARENS_CAST type CLOSE_PARENS prefixed_unary_expression",
//t    "unary_expression : AWAIT prefixed_unary_expression",
//t    "prefixed_unary_expression : unary_expression",
//t    "prefixed_unary_expression : PLUS prefixed_unary_expression",
//t    "prefixed_unary_expression : MINUS prefixed_unary_expression",
//t    "prefixed_unary_expression : OP_INC prefixed_unary_expression",
//t    "prefixed_unary_expression : OP_DEC prefixed_unary_expression",
//t    "prefixed_unary_expression : STAR prefixed_unary_expression",
//t    "prefixed_unary_expression : BITWISE_AND prefixed_unary_expression",
//t    "multiplicative_expression : prefixed_unary_expression",
//t    "multiplicative_expression : multiplicative_expression STAR prefixed_unary_expression",
//t    "multiplicative_expression : multiplicative_expression DIV prefixed_unary_expression",
//t    "multiplicative_expression : multiplicative_expression PERCENT prefixed_unary_expression",
//t    "additive_expression : multiplicative_expression",
//t    "additive_expression : additive_expression PLUS multiplicative_expression",
//t    "additive_expression : additive_expression MINUS multiplicative_expression",
//t    "additive_expression : parenthesized_expression MINUS multiplicative_expression",
//t    "additive_expression : additive_expression AS type",
//t    "additive_expression : additive_expression IS type",
//t    "shift_expression : additive_expression",
//t    "shift_expression : shift_expression OP_SHIFT_LEFT additive_expression",
//t    "shift_expression : shift_expression OP_SHIFT_RIGHT additive_expression",
//t    "relational_expression : shift_expression",
//t    "relational_expression : relational_expression OP_LT shift_expression",
//t    "relational_expression : relational_expression OP_GT shift_expression",
//t    "relational_expression : relational_expression OP_LE shift_expression",
//t    "relational_expression : relational_expression OP_GE shift_expression",
//t    "equality_expression : relational_expression",
//t    "equality_expression : equality_expression OP_EQ relational_expression",
//t    "equality_expression : equality_expression OP_NE relational_expression",
//t    "and_expression : equality_expression",
//t    "and_expression : and_expression BITWISE_AND equality_expression",
//t    "exclusive_or_expression : and_expression",
//t    "exclusive_or_expression : exclusive_or_expression CARRET and_expression",
//t    "inclusive_or_expression : exclusive_or_expression",
//t    "inclusive_or_expression : inclusive_or_expression BITWISE_OR exclusive_or_expression",
//t    "conditional_and_expression : inclusive_or_expression",
//t    "conditional_and_expression : conditional_and_expression OP_AND inclusive_or_expression",
//t    "conditional_or_expression : conditional_and_expression",
//t    "conditional_or_expression : conditional_or_expression OP_OR conditional_and_expression",
//t    "null_coalescing_expression : conditional_or_expression",
//t    "null_coalescing_expression : conditional_or_expression OP_COALESCING null_coalescing_expression",
//t    "conditional_expression : null_coalescing_expression",
//t    "conditional_expression : null_coalescing_expression INTERR expression COLON expression",
//t    "assignment_expression : prefixed_unary_expression ASSIGN expression",
//t    "assignment_expression : prefixed_unary_expression OP_MULT_ASSIGN expression",
//t    "assignment_expression : prefixed_unary_expression OP_DIV_ASSIGN expression",
//t    "assignment_expression : prefixed_unary_expression OP_MOD_ASSIGN expression",
//t    "assignment_expression : prefixed_unary_expression OP_ADD_ASSIGN expression",
//t    "assignment_expression : prefixed_unary_expression OP_SUB_ASSIGN expression",
//t    "assignment_expression : prefixed_unary_expression OP_SHIFT_LEFT_ASSIGN expression",
//t    "assignment_expression : prefixed_unary_expression OP_SHIFT_RIGHT_ASSIGN expression",
//t    "assignment_expression : prefixed_unary_expression OP_AND_ASSIGN expression",
//t    "assignment_expression : prefixed_unary_expression OP_OR_ASSIGN expression",
//t    "assignment_expression : prefixed_unary_expression OP_XOR_ASSIGN expression",
//t    "lambda_parameter_list : lambda_parameter",
//t    "lambda_parameter_list : lambda_parameter_list COMMA lambda_parameter",
//t    "lambda_parameter : parameter_modifier parameter_type identifier_inside_body",
//t    "lambda_parameter : parameter_type identifier_inside_body",
//t    "lambda_parameter : IDENTIFIER",
//t    "opt_lambda_parameter_list :",
//t    "opt_lambda_parameter_list : lambda_parameter_list",
//t    "lambda_expression_body : lambda_expression_body_simple",
//t    "lambda_expression_body : block",
//t    "$$69 :",
//t    "lambda_expression_body_simple : $$69 expression_or_error",
//t    "expression_or_error : expression",
//t    "expression_or_error : error",
//t    "$$70 :",
//t    "lambda_expression : IDENTIFIER ARROW $$70 lambda_expression_body",
//t    "$$71 :",
//t    "lambda_expression : ASYNC identifier_inside_body ARROW $$71 lambda_expression_body",
//t    "$$72 :",
//t    "$$73 :",
//t    "lambda_expression : OPEN_PARENS_LAMBDA $$72 opt_lambda_parameter_list CLOSE_PARENS ARROW $$73 lambda_expression_body",
//t    "$$74 :",
//t    "$$75 :",
//t    "lambda_expression : ASYNC OPEN_PARENS_LAMBDA $$74 opt_lambda_parameter_list CLOSE_PARENS ARROW $$75 lambda_expression_body",
//t    "expression : assignment_expression",
//t    "expression : non_assignment_expression",
//t    "non_assignment_expression : conditional_expression",
//t    "non_assignment_expression : lambda_expression",
//t    "non_assignment_expression : query_expression",
//t    "non_assignment_expression : ARGLIST",
//t    "undocumented_expressions : REFVALUE OPEN_PARENS non_assignment_expression COMMA type CLOSE_PARENS",
//t    "undocumented_expressions : REFTYPE open_parens_any expression CLOSE_PARENS",
//t    "undocumented_expressions : MAKEREF open_parens_any expression CLOSE_PARENS",
//t    "constant_expression : expression",
//t    "boolean_expression : expression",
//t    "$$76 :",
//t    "$$77 :",
//t    "$$78 :",
//t    "$$79 :",
//t    "class_declaration : opt_attributes opt_modifiers opt_partial CLASS $$76 type_declaration_name $$77 opt_class_base opt_type_parameter_constraints_clauses $$78 OPEN_BRACE opt_class_member_declarations CLOSE_BRACE $$79 opt_semicolon",
//t    "opt_partial :",
//t    "opt_partial : PARTIAL",
//t    "opt_modifiers :",
//t    "opt_modifiers : modifiers",
//t    "modifiers : modifier",
//t    "modifiers : modifiers modifier",
//t    "modifier : NEW",
//t    "modifier : PUBLIC",
//t    "modifier : PROTECTED",
//t    "modifier : INTERNAL",
//t    "modifier : PRIVATE",
//t    "modifier : ABSTRACT",
//t    "modifier : SEALED",
//t    "modifier : STATIC",
//t    "modifier : READONLY",
//t    "modifier : VIRTUAL",
//t    "modifier : OVERRIDE",
//t    "modifier : EXTERN",
//t    "modifier : VOLATILE",
//t    "modifier : UNSAFE",
//t    "modifier : ASYNC",
//t    "opt_class_base :",
//t    "opt_class_base : COLON type_list",
//t    "opt_type_parameter_constraints_clauses :",
//t    "opt_type_parameter_constraints_clauses : type_parameter_constraints_clauses",
//t    "opt_type_parameter_constraints_clauses : error",
//t    "type_parameter_constraints_clauses : type_parameter_constraints_clause",
//t    "type_parameter_constraints_clauses : type_parameter_constraints_clauses type_parameter_constraints_clause",
//t    "type_parameter_constraints_clause : WHERE IDENTIFIER COLON type_parameter_constraints",
//t    "type_parameter_constraints : type_parameter_constraint",
//t    "type_parameter_constraints : type_parameter_constraints COMMA type_parameter_constraint",
//t    "type_parameter_constraint : type",
//t    "type_parameter_constraint : NEW OPEN_PARENS CLOSE_PARENS",
//t    "type_parameter_constraint : CLASS",
//t    "type_parameter_constraint : STRUCT",
//t    "opt_type_parameter_variance :",
//t    "opt_type_parameter_variance : type_parameter_variance",
//t    "type_parameter_variance : OUT",
//t    "type_parameter_variance : IN",
//t    "$$80 :",
//t    "block : OPEN_BRACE $$80 opt_statement_list block_end",
//t    "block_end : CLOSE_BRACE",
//t    "block_end : COMPLETE_COMPLETION",
//t    "$$81 :",
//t    "block_prepared : OPEN_BRACE $$81 opt_statement_list CLOSE_BRACE",
//t    "opt_statement_list :",
//t    "opt_statement_list : statement_list",
//t    "statement_list : statement",
//t    "statement_list : statement_list statement",
//t    "statement : block_variable_declaration",
//t    "statement : valid_declaration_statement",
//t    "statement : labeled_statement",
//t    "statement : IDENTIFIER error",
//t    "statement : error",
//t    "interactive_statement_list : interactive_statement",
//t    "interactive_statement_list : interactive_statement_list interactive_statement",
//t    "interactive_statement : block_variable_declaration",
//t    "interactive_statement : interactive_valid_declaration_statement",
//t    "interactive_statement : labeled_statement",
//t    "valid_declaration_statement : block",
//t    "valid_declaration_statement : empty_statement",
//t    "valid_declaration_statement : expression_statement",
//t    "valid_declaration_statement : selection_statement",
//t    "valid_declaration_statement : iteration_statement",
//t    "valid_declaration_statement : jump_statement",
//t    "valid_declaration_statement : try_statement",
//t    "valid_declaration_statement : checked_statement",
//t    "valid_declaration_statement : unchecked_statement",
//t    "valid_declaration_statement : lock_statement",
//t    "valid_declaration_statement : using_statement",
//t    "valid_declaration_statement : unsafe_statement",
//t    "valid_declaration_statement : fixed_statement",
//t    "interactive_valid_declaration_statement : block",
//t    "interactive_valid_declaration_statement : empty_statement",
//t    "interactive_valid_declaration_statement : interactive_expression_statement",
//t    "interactive_valid_declaration_statement : selection_statement",
//t    "interactive_valid_declaration_statement : iteration_statement",
//t    "interactive_valid_declaration_statement : jump_statement",
//t    "interactive_valid_declaration_statement : try_statement",
//t    "interactive_valid_declaration_statement : checked_statement",
//t    "interactive_valid_declaration_statement : unchecked_statement",
//t    "interactive_valid_declaration_statement : lock_statement",
//t    "interactive_valid_declaration_statement : using_statement",
//t    "interactive_valid_declaration_statement : unsafe_statement",
//t    "interactive_valid_declaration_statement : fixed_statement",
//t    "embedded_statement : valid_declaration_statement",
//t    "embedded_statement : block_variable_declaration",
//t    "embedded_statement : labeled_statement",
//t    "embedded_statement : error",
//t    "empty_statement : SEMICOLON",
//t    "$$82 :",
//t    "labeled_statement : identifier_inside_body COLON $$82 statement",
//t    "variable_type : variable_type_simple",
//t    "variable_type : variable_type_simple rank_specifiers",
//t    "variable_type_simple : primary_expression_or_type opt_nullable",
//t    "variable_type_simple : primary_expression_or_type pointer_stars",
//t    "variable_type_simple : builtin_types opt_nullable",
//t    "variable_type_simple : builtin_types pointer_stars",
//t    "variable_type_simple : VOID pointer_stars",
//t    "variable_type_simple : VOID",
//t    "pointer_stars : pointer_star",
//t    "pointer_stars : pointer_star pointer_stars",
//t    "pointer_star : STAR",
//t    "identifier_inside_body : IDENTIFIER",
//t    "identifier_inside_body : AWAIT",
//t    "$$83 :",
//t    "block_variable_declaration : variable_type identifier_inside_body $$83 opt_local_variable_initializer opt_variable_declarators SEMICOLON",
//t    "$$84 :",
//t    "block_variable_declaration : CONST variable_type identifier_inside_body $$84 const_variable_initializer opt_const_declarators SEMICOLON",
//t    "opt_local_variable_initializer :",
//t    "opt_local_variable_initializer : ASSIGN block_variable_initializer",
//t    "opt_local_variable_initializer : ASSIGN error",
//t    "opt_local_variable_initializer : error",
//t    "opt_variable_declarators :",
//t    "opt_variable_declarators : variable_declarators",
//t    "opt_using_or_fixed_variable_declarators :",
//t    "opt_using_or_fixed_variable_declarators : variable_declarators",
//t    "variable_declarators : variable_declarator",
//t    "variable_declarators : variable_declarators variable_declarator",
//t    "variable_declarator : COMMA identifier_inside_body",
//t    "variable_declarator : COMMA identifier_inside_body ASSIGN block_variable_initializer",
//t    "const_variable_initializer :",
//t    "const_variable_initializer : ASSIGN constant_initializer_expr",
//t    "opt_const_declarators :",
//t    "opt_const_declarators : const_declarators",
//t    "const_declarators : const_declarator",
//t    "const_declarators : const_declarators const_declarator",
//t    "const_declarator : COMMA identifier_inside_body ASSIGN constant_initializer_expr",
//t    "block_variable_initializer : variable_initializer",
//t    "block_variable_initializer : STACKALLOC simple_type OPEN_BRACKET_EXPR expression CLOSE_BRACKET",
//t    "block_variable_initializer : STACKALLOC simple_type",
//t    "expression_statement : statement_expression SEMICOLON",
//t    "expression_statement : statement_expression COMPLETE_COMPLETION",
//t    "interactive_expression_statement : interactive_statement_expression SEMICOLON",
//t    "interactive_expression_statement : interactive_statement_expression COMPLETE_COMPLETION",
//t    "statement_expression : expression",
//t    "interactive_statement_expression : expression",
//t    "interactive_statement_expression : error",
//t    "selection_statement : if_statement",
//t    "selection_statement : switch_statement",
//t    "if_statement : IF open_parens_any boolean_expression CLOSE_PARENS embedded_statement",
//t    "if_statement : IF open_parens_any boolean_expression CLOSE_PARENS embedded_statement ELSE embedded_statement",
//t    "if_statement : IF open_parens_any boolean_expression error",
//t    "$$85 :",
//t    "switch_statement : SWITCH open_parens_any expression CLOSE_PARENS OPEN_BRACE $$85 opt_switch_sections CLOSE_BRACE",
//t    "opt_switch_sections :",
//t    "opt_switch_sections : switch_sections",
//t    "switch_sections : switch_section",
//t    "switch_sections : switch_sections switch_section",
//t    "switch_sections : error",
//t    "$$86 :",
//t    "switch_section : switch_labels $$86 statement_list",
//t    "switch_labels : switch_label",
//t    "switch_labels : switch_labels switch_label",
//t    "switch_label : CASE constant_expression COLON",
//t    "switch_label : DEFAULT_COLON",
//t    "iteration_statement : while_statement",
//t    "iteration_statement : do_statement",
//t    "iteration_statement : for_statement",
//t    "iteration_statement : foreach_statement",
//t    "while_statement : WHILE open_parens_any boolean_expression CLOSE_PARENS embedded_statement",
//t    "do_statement : DO embedded_statement WHILE open_parens_any boolean_expression CLOSE_PARENS SEMICOLON",
//t    "do_statement : DO embedded_statement error",
//t    "do_statement : DO embedded_statement WHILE open_parens_any boolean_expression error",
//t    "$$87 :",
//t    "for_statement : FOR open_parens_any $$87 for_statement_cont",
//t    "$$88 :",
//t    "$$89 :",
//t    "$$90 :",
//t    "for_statement_cont : opt_for_initializer SEMICOLON $$88 opt_for_condition SEMICOLON $$89 opt_for_iterator CLOSE_PARENS $$90 embedded_statement",
//t    "for_statement_cont : error",
//t    "opt_for_initializer :",
//t    "opt_for_initializer : for_initializer",
//t    "$$91 :",
//t    "for_initializer : variable_type identifier_inside_body $$91 opt_local_variable_initializer opt_variable_declarators",
//t    "for_initializer : statement_expression_list",
//t    "opt_for_condition :",
//t    "opt_for_condition : boolean_expression",
//t    "opt_for_iterator :",
//t    "opt_for_iterator : for_iterator",
//t    "for_iterator : statement_expression_list",
//t    "statement_expression_list : statement_expression",
//t    "statement_expression_list : statement_expression_list COMMA statement_expression",
//t    "foreach_statement : FOREACH open_parens_any type IN expression CLOSE_PARENS",
//t    "$$92 :",
//t    "foreach_statement : FOREACH open_parens_any type identifier_inside_body IN expression CLOSE_PARENS $$92 embedded_statement",
//t    "foreach_statement : FOREACH open_parens_any type identifier_inside_body error",
//t    "foreach_statement : FOREACH open_parens_any type error",
//t    "jump_statement : break_statement",
//t    "jump_statement : continue_statement",
//t    "jump_statement : goto_statement",
//t    "jump_statement : return_statement",
//t    "jump_statement : throw_statement",
//t    "jump_statement : yield_statement",
//t    "break_statement : BREAK SEMICOLON",
//t    "continue_statement : CONTINUE SEMICOLON",
//t    "goto_statement : GOTO identifier_inside_body SEMICOLON",
//t    "goto_statement : GOTO CASE constant_expression SEMICOLON",
//t    "goto_statement : GOTO DEFAULT SEMICOLON",
//t    "return_statement : RETURN opt_expression SEMICOLON",
//t    "throw_statement : THROW opt_expression SEMICOLON",
//t    "yield_statement : identifier_inside_body RETURN opt_expression SEMICOLON",
//t    "yield_statement : identifier_inside_body BREAK SEMICOLON",
//t    "opt_expression :",
//t    "opt_expression : expression",
//t    "try_statement : TRY block catch_clauses",
//t    "try_statement : TRY block FINALLY block",
//t    "try_statement : TRY block catch_clauses FINALLY block",
//t    "try_statement : TRY block error",
//t    "catch_clauses : catch_clause",
//t    "catch_clauses : catch_clauses catch_clause",
//t    "opt_identifier :",
//t    "opt_identifier : identifier_inside_body",
//t    "catch_clause : CATCH block",
//t    "$$93 :",
//t    "catch_clause : CATCH open_parens_any type opt_identifier CLOSE_PARENS $$93 block_prepared",
//t    "catch_clause : CATCH open_parens_any error",
//t    "checked_statement : CHECKED block",
//t    "unchecked_statement : UNCHECKED block",
//t    "$$94 :",
//t    "unsafe_statement : UNSAFE $$94 block",
//t    "lock_statement : LOCK open_parens_any expression CLOSE_PARENS embedded_statement",
//t    "$$95 :",
//t    "$$96 :",
//t    "fixed_statement : FIXED open_parens_any variable_type identifier_inside_body $$95 using_or_fixed_variable_initializer opt_using_or_fixed_variable_declarators CLOSE_PARENS $$96 embedded_statement",
//t    "$$97 :",
//t    "$$98 :",
//t    "using_statement : USING open_parens_any variable_type identifier_inside_body $$97 using_initialization CLOSE_PARENS $$98 embedded_statement",
//t    "using_statement : USING open_parens_any expression CLOSE_PARENS embedded_statement",
//t    "using_initialization : using_or_fixed_variable_initializer opt_using_or_fixed_variable_declarators",
//t    "using_initialization : error",
//t    "using_or_fixed_variable_initializer :",
//t    "using_or_fixed_variable_initializer : ASSIGN variable_initializer",
//t    "query_expression : first_from_clause query_body",
//t    "query_expression : nested_from_clause query_body",
//t    "query_expression : first_from_clause COMPLETE_COMPLETION",
//t    "query_expression : nested_from_clause COMPLETE_COMPLETION",
//t    "first_from_clause : FROM_FIRST identifier_inside_body IN expression",
//t    "first_from_clause : FROM_FIRST type identifier_inside_body IN expression",
//t    "nested_from_clause : FROM identifier_inside_body IN expression",
//t    "nested_from_clause : FROM type identifier_inside_body IN expression",
//t    "$$99 :",
//t    "from_clause : FROM identifier_inside_body IN $$99 expression",
//t    "$$100 :",
//t    "from_clause : FROM type identifier_inside_body IN $$100 expression",
//t    "query_body : opt_query_body_clauses select_or_group_clause opt_query_continuation",
//t    "query_body : opt_query_body_clauses COMPLETE_COMPLETION",
//t    "query_body : error",
//t    "$$101 :",
//t    "select_or_group_clause : SELECT $$101 expression",
//t    "$$102 :",
//t    "$$103 :",
//t    "select_or_group_clause : GROUP $$102 expression $$103 BY expression",
//t    "opt_query_body_clauses :",
//t    "opt_query_body_clauses : query_body_clauses",
//t    "query_body_clauses : query_body_clause",
//t    "query_body_clauses : query_body_clauses query_body_clause",
//t    "query_body_clause : from_clause",
//t    "query_body_clause : let_clause",
//t    "query_body_clause : where_clause",
//t    "query_body_clause : join_clause",
//t    "query_body_clause : orderby_clause",
//t    "$$104 :",
//t    "let_clause : LET identifier_inside_body ASSIGN $$104 expression",
//t    "$$105 :",
//t    "where_clause : WHERE $$105 expression",
//t    "$$106 :",
//t    "$$107 :",
//t    "$$108 :",
//t    "join_clause : JOIN identifier_inside_body IN $$106 expression ON $$107 expression EQUALS $$108 expression opt_join_into",
//t    "$$109 :",
//t    "$$110 :",
//t    "$$111 :",
//t    "join_clause : JOIN type identifier_inside_body IN $$109 expression ON $$110 expression EQUALS $$111 expression opt_join_into",
//t    "opt_join_into :",
//t    "opt_join_into : INTO identifier_inside_body",
//t    "$$112 :",
//t    "orderby_clause : ORDERBY $$112 orderings",
//t    "orderings : order_by",
//t    "$$113 :",
//t    "orderings : order_by COMMA $$113 orderings_then_by",
//t    "orderings_then_by : then_by",
//t    "$$114 :",
//t    "orderings_then_by : orderings_then_by COMMA $$114 then_by",
//t    "order_by : expression",
//t    "order_by : expression ASCENDING",
//t    "order_by : expression DESCENDING",
//t    "then_by : expression",
//t    "then_by : expression ASCENDING",
//t    "then_by : expression DESCENDING",
//t    "opt_query_continuation :",
//t    "$$115 :",
//t    "opt_query_continuation : INTO identifier_inside_body $$115 query_body",
//t    "interactive_parsing : EVAL_STATEMENT_PARSER EOF",
//t    "interactive_parsing : EVAL_USING_DECLARATIONS_UNIT_PARSER using_directives opt_COMPLETE_COMPLETION",
//t    "$$116 :",
//t    "interactive_parsing : EVAL_STATEMENT_PARSER $$116 interactive_statement_list opt_COMPLETE_COMPLETION",
//t    "interactive_parsing : EVAL_COMPILATION_UNIT_PARSER interactive_compilation_unit",
//t    "interactive_compilation_unit : opt_extern_alias_directives opt_using_directives",
//t    "interactive_compilation_unit : opt_extern_alias_directives opt_using_directives namespace_or_type_declarations",
//t    "opt_COMPLETE_COMPLETION :",
//t    "opt_COMPLETE_COMPLETION : COMPLETE_COMPLETION",
//t    "close_brace_or_complete_completion : CLOSE_BRACE",
//t    "close_brace_or_complete_completion : COMPLETE_COMPLETION",
//t    "documentation_parsing : DOC_SEE doc_cref",
//t    "doc_cref : doc_type_declaration_name opt_doc_method_sig",
//t    "doc_cref : builtin_types opt_doc_method_sig",
//t    "doc_cref : builtin_types DOT IDENTIFIER opt_doc_method_sig",
//t    "doc_cref : doc_type_declaration_name DOT THIS",
//t    "$$117 :",
//t    "doc_cref : doc_type_declaration_name DOT THIS OPEN_BRACKET $$117 opt_doc_parameters CLOSE_BRACKET",
//t    "doc_cref : EXPLICIT OPERATOR type opt_doc_method_sig",
//t    "doc_cref : IMPLICIT OPERATOR type opt_doc_method_sig",
//t    "doc_cref : OPERATOR overloadable_operator opt_doc_method_sig",
//t    "doc_type_declaration_name : type_declaration_name",
//t    "doc_type_declaration_name : doc_type_declaration_name DOT type_declaration_name",
//t    "opt_doc_method_sig :",
//t    "$$118 :",
//t    "opt_doc_method_sig : OPEN_PARENS $$118 opt_doc_parameters CLOSE_PARENS",
//t    "opt_doc_parameters :",
//t    "opt_doc_parameters : doc_parameters",
//t    "doc_parameters : doc_parameter",
//t    "doc_parameters : doc_parameters COMMA doc_parameter",
//t    "doc_parameter : opt_parameter_modifier parameter_type",
//t  };
//t public static string getRule (int index) {
//t    return yyRule [index];
//t }
//t}
  protected static readonly string [] yyNames = {    
    "end-of-file",null,null,null,null,null,null,null,null,null,null,null,
    null,null,null,null,null,null,null,null,null,null,null,null,null,null,
    null,null,null,null,null,null,null,null,null,null,null,null,null,null,
    null,null,null,null,null,null,null,null,null,null,null,null,null,null,
    null,null,null,null,null,null,null,null,null,null,null,null,null,null,
    null,null,null,null,null,null,null,null,null,null,null,null,null,null,
    null,null,null,null,null,null,null,null,null,null,null,null,null,null,
    null,null,null,null,null,null,null,null,null,null,null,null,null,null,
    null,null,null,null,null,null,null,null,null,null,null,null,null,null,
    null,null,null,null,null,null,null,null,null,null,null,null,null,null,
    null,null,null,null,null,null,null,null,null,null,null,null,null,null,
    null,null,null,null,null,null,null,null,null,null,null,null,null,null,
    null,null,null,null,null,null,null,null,null,null,null,null,null,null,
    null,null,null,null,null,null,null,null,null,null,null,null,null,null,
    null,null,null,null,null,null,null,null,null,null,null,null,null,null,
    null,null,null,null,null,null,null,null,null,null,null,null,null,null,
    null,null,null,null,null,null,null,null,null,null,null,null,null,null,
    null,null,null,null,null,null,null,null,null,null,null,null,null,null,
    null,null,null,null,null,null,null,"EOF","NONE","ERROR",
    "FIRST_KEYWORD","ABSTRACT","AS","ADD","BASE","BOOL","BREAK","BYTE",
    "CASE","CATCH","CHAR","CHECKED","CLASS","CONST","CONTINUE","DECIMAL",
    "DEFAULT","DELEGATE","DO","DOUBLE","ELSE","ENUM","EVENT","EXPLICIT",
    "EXTERN","FALSE","FINALLY","FIXED","FLOAT","FOR","FOREACH","GOTO",
    "IF","IMPLICIT","IN","INT","INTERFACE","INTERNAL","IS","LOCK","LONG",
    "NAMESPACE","NEW","NULL","OBJECT","OPERATOR","OUT","OVERRIDE",
    "PARAMS","PRIVATE","PROTECTED","PUBLIC","READONLY","REF","RETURN",
    "REMOVE","SBYTE","SEALED","SHORT","SIZEOF","STACKALLOC","STATIC",
    "STRING","STRUCT","SWITCH","THIS","THROW","TRUE","TRY","TYPEOF",
    "UINT","ULONG","UNCHECKED","UNSAFE","USHORT","USING","VIRTUAL","VOID",
    "VOLATILE","WHERE","WHILE","ARGLIST","PARTIAL","ARROW","FROM",
    "FROM_FIRST","JOIN","ON","EQUALS","SELECT","GROUP","BY","LET",
    "ORDERBY","ASCENDING","DESCENDING","INTO","INTERR_NULLABLE",
    "EXTERN_ALIAS","REFVALUE","REFTYPE","MAKEREF","ASYNC","AWAIT","GET",
    "SET","LAST_KEYWORD","OPEN_BRACE","CLOSE_BRACE","OPEN_BRACKET",
    "CLOSE_BRACKET","OPEN_PARENS","CLOSE_PARENS","DOT","COMMA","COLON",
    "SEMICOLON","TILDE","PLUS","MINUS","BANG","ASSIGN","OP_LT","OP_GT",
    "BITWISE_AND","BITWISE_OR","STAR","PERCENT","DIV","CARRET","INTERR",
    "DOUBLE_COLON","OP_INC","OP_DEC","OP_SHIFT_LEFT","OP_SHIFT_RIGHT",
    "OP_LE","OP_GE","OP_EQ","OP_NE","OP_AND","OP_OR","OP_MULT_ASSIGN",
    "OP_DIV_ASSIGN","OP_MOD_ASSIGN","OP_ADD_ASSIGN","OP_SUB_ASSIGN",
    "OP_SHIFT_LEFT_ASSIGN","OP_SHIFT_RIGHT_ASSIGN","OP_AND_ASSIGN",
    "OP_XOR_ASSIGN","OP_OR_ASSIGN","OP_PTR","OP_COALESCING",
    "OP_GENERICS_LT","OP_GENERICS_LT_DECL","OP_GENERICS_GT","LITERAL",
    "IDENTIFIER","OPEN_PARENS_LAMBDA","OPEN_PARENS_CAST",
    "GENERIC_DIMENSION","DEFAULT_COLON","OPEN_BRACKET_EXPR",
    "EVAL_STATEMENT_PARSER","EVAL_COMPILATION_UNIT_PARSER",
    "EVAL_USING_DECLARATIONS_UNIT_PARSER","DOC_SEE","GENERATE_COMPLETION",
    "COMPLETE_COMPLETION","UMINUS",
  };

  /** index-checked interface to yyNames[].
      @param token single character or %token value.
      @return token name or [illegal] or [unknown].
    */
//t  public static string yyname (int token) {
//t    if ((token < 0) || (token > yyNames.Length)) return "[illegal]";
//t    string name;
//t    if ((name = yyNames[token]) != null) return name;
//t    return "[unknown]";
//t  }

  int yyExpectingState;
  /** computes list of expected tokens on error by tracing the tables.
      @param state for which to compute the list.
      @return list of token names.
    */
  protected int [] yyExpectingTokens (int state){
    int token, n, len = 0;
    bool[] ok = new bool[yyNames.Length];
    if ((n = yySindex[state]) != 0)
      for (token = n < 0 ? -n : 0;
           (token < yyNames.Length) && (n+token < yyTable.Length); ++ token)
        if (yyCheck[n+token] == token && !ok[token] && yyNames[token] != null) {
          ++ len;
          ok[token] = true;
        }
    if ((n = yyRindex[state]) != 0)
      for (token = n < 0 ? -n : 0;
           (token < yyNames.Length) && (n+token < yyTable.Length); ++ token)
        if (yyCheck[n+token] == token && !ok[token] && yyNames[token] != null) {
          ++ len;
          ok[token] = true;
        }
    int [] result = new int [len];
    for (n = token = 0; n < len;  ++ token)
      if (ok[token]) result[n++] = token;
    return result;
  }
  protected string[] yyExpecting (int state) {
    int [] tokens = yyExpectingTokens (state);
    string [] result = new string[tokens.Length];
    for (int n = 0; n < tokens.Length;  n++)
      result[n++] = yyNames[tokens [n]];
    return result;
  }

  /** the generated parser, with debugging messages.
      Maintains a state and a value stack, currently with fixed maximum size.
      @param yyLex scanner.
      @param yydebug debug message writer implementing yyDebug, or null.
      @return result of the last reduction, if any.
      @throws yyException on irrecoverable parse error.
    */
  internal Object yyparse (yyParser.yyInput yyLex, Object yyd)
				 {
//t    this.debug = (yydebug.yyDebug)yyd;
    return yyparse(yyLex);
  }

  /** initial size and increment of the state/value stack [default 256].
      This is not final so that it can be overwritten outside of invocations
      of yyparse().
    */
  protected int yyMax;

  /** executed at the beginning of a reduce action.
      Used as $$ = yyDefault($1), prior to the user-specified action, if any.
      Can be overwritten to provide deep copy, etc.
      @param first value for $1, or null.
      @return first.
    */
  protected Object yyDefault (Object first) {
    return first;
  }

	static int[] global_yyStates;
	static object[] global_yyVals;
	protected bool use_global_stacks;
	object[] yyVals;					// value stack
	object yyVal;						// value stack ptr
	int yyToken;						// current input
	int yyTop;

  /** the generated parser.
      Maintains a state and a value stack, currently with fixed maximum size.
      @param yyLex scanner.
      @return result of the last reduction, if any.
      @throws yyException on irrecoverable parse error.
    */
  internal Object yyparse (yyParser.yyInput yyLex)
  {
    if (yyMax <= 0) yyMax = 256;		// initial size
    int yyState = 0;                   // state stack ptr
    int [] yyStates;               	// state stack 
    yyVal = null;
    yyToken = -1;
    int yyErrorFlag = 0;				// #tks to shift
	if (use_global_stacks && global_yyStates != null) {
		yyVals = global_yyVals;
		yyStates = global_yyStates;
   } else {
		yyVals = new object [yyMax];
		yyStates = new int [yyMax];
		if (use_global_stacks) {
			global_yyVals = yyVals;
			global_yyStates = yyStates;
		}
	}

    /*yyLoop:*/ for (yyTop = 0;; ++ yyTop) {
      if (yyTop >= yyStates.Length) {			// dynamically increase
        global::System.Array.Resize (ref yyStates, yyStates.Length+yyMax);
        global::System.Array.Resize (ref yyVals, yyVals.Length+yyMax);
      }
      yyStates[yyTop] = yyState;
      yyVals[yyTop] = yyVal;
//t      if (debug != null) debug.push(yyState, yyVal);

      /*yyDiscarded:*/ while (true) {	// discarding a token does not change stack
        int yyN;
        if ((yyN = yyDefRed[yyState]) == 0) {	// else [default] reduce (yyN)
          if (yyToken < 0) {
            yyToken = yyLex.advance() ? yyLex.token() : 0;
//t            if (debug != null)
//t              debug.lex(yyState, yyToken, yyname(yyToken), yyLex.value());
          }
          if ((yyN = yySindex[yyState]) != 0 && ((yyN += yyToken) >= 0)
              && (yyN < yyTable.Length) && (yyCheck[yyN] == yyToken)) {
//t            if (debug != null)
//t              debug.shift(yyState, yyTable[yyN], yyErrorFlag-1);
            yyState = yyTable[yyN];		// shift to yyN
            yyVal = yyLex.value();
            yyToken = -1;
            if (yyErrorFlag > 0) -- yyErrorFlag;
            goto continue_yyLoop;
          }
          if ((yyN = yyRindex[yyState]) != 0 && (yyN += yyToken) >= 0
              && yyN < yyTable.Length && yyCheck[yyN] == yyToken)
            yyN = yyTable[yyN];			// reduce (yyN)
          else
            switch (yyErrorFlag) {
  
            case 0:
              yyExpectingState = yyState;
              // yyerror(String.Format ("syntax error, got token `{0}'", yyname (yyToken)), yyExpecting(yyState));
//t              if (debug != null) debug.error("syntax error");
              if (yyToken == 0 /*eof*/ || yyToken == eof_token) throw new yyParser.yyUnexpectedEof ();
              goto case 1;
            case 1: case 2:
              yyErrorFlag = 3;
              do {
                if ((yyN = yySindex[yyStates[yyTop]]) != 0
                    && (yyN += Token.yyErrorCode) >= 0 && yyN < yyTable.Length
                    && yyCheck[yyN] == Token.yyErrorCode) {
//t                  if (debug != null)
//t                    debug.shift(yyStates[yyTop], yyTable[yyN], 3);
                  yyState = yyTable[yyN];
                  yyVal = yyLex.value();
                  goto continue_yyLoop;
                }
//t                if (debug != null) debug.pop(yyStates[yyTop]);
              } while (-- yyTop >= 0);
//t              if (debug != null) debug.reject();
              throw new yyParser.yyException("irrecoverable syntax error");
  
            case 3:
              if (yyToken == 0) {
//t                if (debug != null) debug.reject();
                throw new yyParser.yyException("irrecoverable syntax error at end-of-file");
              }
//t              if (debug != null)
//t                debug.discard(yyState, yyToken, yyname(yyToken),
//t  							yyLex.value());
              yyToken = -1;
              goto continue_yyDiscarded;		// leave stack alone
            }
        }
        int yyV = yyTop + 1-yyLen[yyN];
//t        if (debug != null)
//t          debug.reduce(yyState, yyStates[yyV-1], yyN, YYRules.getRule (yyN), yyLen[yyN]);
        yyVal = yyV > yyTop ? null : yyVals[yyV]; // yyVal = yyDefault(yyV > yyTop ? null : yyVals[yyV]);
        switch (yyN) {
case 1:
///  #line 388 "cs-parser.jay"
  {
		Lexer.check_incorrect_doc_comment ();
	  }
  break;
case 2:
///  #line 389 "cs-parser.jay"
  { Lexer.CompleteOnEOF = false; }
  break;
case 6:
  case_6();
  break;
case 7:
///  #line 406 "cs-parser.jay"
  {
		module.AddAttributes ((Attributes) yyVals[0+yyTop], current_namespace);
	  }
  break;
case 8:
  case_8();
  break;
case 13:
  case_13();
  break;
case 14:
///  #line 444 "cs-parser.jay"
  {
	  	syntax_error (GetLocation (yyVals[-1+yyTop]), "`alias' expected");   /* TODO: better*/
	  }
  break;
case 17:
  case_17();
  break;
case 18:
  case_18();
  break;
case 19:
  case_19();
  break;
case 20:
  case_20();
  break;
case 21:
  case_21();
  break;
case 22:
  case_22();
  break;
case 23:
  case_23();
  break;
case 24:
  case_24();
  break;
case 25:
  case_25();
  break;
case 26:
  case_26();
  break;
case 27:
  case_27();
  break;
case 32:
  case_32();
  break;
case 41:
  case_41();
  break;
case 42:
///  #line 633 "cs-parser.jay"
  {
		current_namespace.DeclarationFound = true;
	  }
  break;
case 50:
  case_50();
  break;
case 51:
  case_51();
  break;
case 52:
  case_52();
  break;
case 53:
  case_53();
  break;
case 54:
  case_54();
  break;
case 55:
  case_55();
  break;
case 56:
  case_56();
  break;
case 57:
  case_57();
  break;
case 58:
///  #line 742 "cs-parser.jay"
  { yyVal = "event"; savedCloseLocation = GetLocation (yyVals[0+yyTop]); }
  break;
case 59:
///  #line 743 "cs-parser.jay"
  { yyVal = "return"; savedCloseLocation = GetLocation (yyVals[0+yyTop]); }
  break;
case 60:
  case_60();
  break;
case 61:
///  #line 760 "cs-parser.jay"
  {
		yyVal = new List<Attribute> (4) { (Attribute) yyVals[0+yyTop] };
	  }
  break;
case 62:
  case_62();
  break;
case 63:
///  #line 775 "cs-parser.jay"
  {
		++lexer.parsing_block;
	  }
  break;
case 64:
  case_64();
  break;
case 66:
///  #line 802 "cs-parser.jay"
  { yyVal = null; }
  break;
case 67:
  case_67();
  break;
case 68:
///  #line 813 "cs-parser.jay"
  { yyVal = null; }
  break;
case 69:
  case_69();
  break;
case 70:
  case_70();
  break;
case 71:
  case_71();
  break;
case 72:
  case_72();
  break;
case 73:
///  #line 857 "cs-parser.jay"
  {
	  	yyVal = new Argument ((Expression) yyVals[0+yyTop]);
	  }
  break;
case 75:
///  #line 865 "cs-parser.jay"
  {
		++lexer.parsing_block;
	  }
  break;
case 76:
  case_76();
  break;
case 77:
  case_77();
  break;
case 78:
///  #line 891 "cs-parser.jay"
  { yyVal = null; }
  break;
case 79:
///  #line 895 "cs-parser.jay"
  { 
		yyVal = Argument.AType.Ref;
	  }
  break;
case 80:
///  #line 899 "cs-parser.jay"
  { 
		yyVal = Argument.AType.Out;
	  }
  break;
case 83:
///  #line 911 "cs-parser.jay"
  {
		lexer.parsing_modifiers = true;
	  }
  break;
case 84:
///  #line 915 "cs-parser.jay"
  {
		lexer.parsing_modifiers = true;
	  }
  break;
case 95:
  case_95();
  break;
case 96:
///  #line 945 "cs-parser.jay"
  {
		lexer.ConstraintsParsing = true;
	  }
  break;
case 97:
  case_97();
  break;
case 98:
  case_98();
  break;
case 99:
  case_99();
  break;
case 100:
  case_100();
  break;
case 101:
  case_101();
  break;
case 102:
///  #line 986 "cs-parser.jay"
  {
		Error_SyntaxError (yyToken);
	  }
  break;
case 103:
  case_103();
  break;
case 104:
  case_104();
  break;
case 107:
///  #line 1027 "cs-parser.jay"
  {
		current_field.AddDeclarator ((FieldDeclarator) yyVals[0+yyTop]);
	  }
  break;
case 108:
///  #line 1031 "cs-parser.jay"
  {
		current_field.AddDeclarator ((FieldDeclarator) yyVals[0+yyTop]);
	  }
  break;
case 109:
  case_109();
  break;
case 110:
///  #line 1047 "cs-parser.jay"
  {
		++lexer.parsing_block;
	  }
  break;
case 111:
  case_111();
  break;
case 112:
  case_112();
  break;
case 115:
  case_115();
  break;
case 116:
  case_116();
  break;
case 117:
  case_117();
  break;
case 118:
  case_118();
  break;
case 119:
///  #line 1126 "cs-parser.jay"
  {
		report.Error (1641, GetLocation (yyVals[-1+yyTop]), "A fixed size buffer field must have the array size specifier after the field name");
	  }
  break;
case 121:
  case_121();
  break;
case 122:
  case_122();
  break;
case 125:
///  #line 1156 "cs-parser.jay"
  {
		current_field.AddDeclarator ((FieldDeclarator) yyVals[0+yyTop]);
	  }
  break;
case 126:
///  #line 1160 "cs-parser.jay"
  {
		current_field.AddDeclarator ((FieldDeclarator) yyVals[0+yyTop]);
	  }
  break;
case 127:
  case_127();
  break;
case 128:
///  #line 1173 "cs-parser.jay"
  {
		++lexer.parsing_block;
	  }
  break;
case 129:
  case_129();
  break;
case 132:
///  #line 1192 "cs-parser.jay"
  {
		current_field.AddDeclarator ((FieldDeclarator) yyVals[0+yyTop]);
	  }
  break;
case 133:
///  #line 1196 "cs-parser.jay"
  {
		current_field.AddDeclarator ((FieldDeclarator) yyVals[0+yyTop]);
	  }
  break;
case 134:
  case_134();
  break;
case 135:
///  #line 1212 "cs-parser.jay"
  {
		++lexer.parsing_block;
	  }
  break;
case 136:
  case_136();
  break;
case 137:
  case_137();
  break;
case 140:
  case_140();
  break;
case 141:
  case_141();
  break;
case 142:
  case_142();
  break;
case 143:
///  #line 1283 "cs-parser.jay"
  {
		valid_param_mod = ParameterModifierType.All;
	  }
  break;
case 144:
///  #line 1287 "cs-parser.jay"
  {
		lexer.ConstraintsParsing = true;
	  }
  break;
case 145:
  case_145();
  break;
case 146:
///  #line 1321 "cs-parser.jay"
  {
		lexer.parsing_generic_declaration = true;
	  }
  break;
case 147:
  case_147();
  break;
case 148:
///  #line 1331 "cs-parser.jay"
  {
		lexer.ConstraintsParsing = true;
	  }
  break;
case 149:
  case_149();
  break;
case 150:
  case_150();
  break;
case 152:
///  #line 1389 "cs-parser.jay"
  { savedLocation = GetLocation (yyVals[0+yyTop]); yyVal = null; }
  break;
case 153:
///  #line 1393 "cs-parser.jay"
  { yyVal = ParametersCompiled.EmptyReadOnlyParameters; }
  break;
case 155:
  case_155();
  break;
case 156:
  case_156();
  break;
case 157:
  case_157();
  break;
case 158:
  case_158();
  break;
case 159:
  case_159();
  break;
case 160:
  case_160();
  break;
case 161:
  case_161();
  break;
case 162:
///  #line 1465 "cs-parser.jay"
  {
		yyVal = new ParametersCompiled (new Parameter[] { (Parameter) yyVals[0+yyTop] } );
	  }
  break;
case 163:
///  #line 1469 "cs-parser.jay"
  {
		yyVal = new ParametersCompiled (new Parameter [] { new ArglistParameter (GetLocation (yyVals[0+yyTop])) }, true);
	  }
  break;
case 164:
  case_164();
  break;
case 165:
  case_165();
  break;
case 166:
  case_166();
  break;
case 167:
  case_167();
  break;
case 168:
  case_168();
  break;
case 169:
  case_169();
  break;
case 170:
///  #line 1544 "cs-parser.jay"
  {
	  	++lexer.parsing_block;
	  }
  break;
case 171:
  case_171();
  break;
case 172:
///  #line 1585 "cs-parser.jay"
  { yyVal = Parameter.Modifier.NONE; }
  break;
case 174:
///  #line 1593 "cs-parser.jay"
  {
		yyVal = yyVals[0+yyTop];
	  }
  break;
case 175:
  case_175();
  break;
case 176:
  case_176();
  break;
case 177:
  case_177();
  break;
case 178:
  case_178();
  break;
case 179:
  case_179();
  break;
case 180:
  case_180();
  break;
case 181:
  case_181();
  break;
case 182:
  case_182();
  break;
case 183:
  case_183();
  break;
case 184:
///  #line 1686 "cs-parser.jay"
  {
		Error_DuplicateParameterModifier (GetLocation (yyVals[-1+yyTop]), Parameter.Modifier.PARAMS);
	  }
  break;
case 185:
  case_185();
  break;
case 186:
  case_186();
  break;
case 187:
  case_187();
  break;
case 188:
  case_188();
  break;
case 189:
  case_189();
  break;
case 190:
///  #line 1740 "cs-parser.jay"
  {
	  	valid_param_mod = ParameterModifierType.Params | ParameterModifierType.DefaultValue;
	  }
  break;
case 191:
  case_191();
  break;
case 192:
///  #line 1769 "cs-parser.jay"
  {
		lexer.PropertyParsing = false;
	  }
  break;
case 193:
  case_193();
  break;
case 198:
  case_198();
  break;
case 199:
  case_199();
  break;
case 200:
  case_200();
  break;
case 201:
  case_201();
  break;
case 202:
  case_202();
  break;
case 204:
  case_204();
  break;
case 205:
  case_205();
  break;
case 206:
///  #line 1918 "cs-parser.jay"
  {
		lexer.ConstraintsParsing = true;
	  }
  break;
case 207:
  case_207();
  break;
case 208:
  case_208();
  break;
case 209:
  case_209();
  break;
case 210:
  case_210();
  break;
case 211:
///  #line 1955 "cs-parser.jay"
  {
		Error_SyntaxError (yyToken);	  
	  }
  break;
case 214:
///  #line 1967 "cs-parser.jay"
  {
		lexer.parsing_modifiers = true;
	  }
  break;
case 215:
///  #line 1971 "cs-parser.jay"
  {
		lexer.parsing_modifiers = true;
	  }
  break;
case 216:
///  #line 1978 "cs-parser.jay"
  {
		report.Error (525, GetLocation (yyVals[0+yyTop]), "Interfaces cannot contain fields or constants");
	  }
  break;
case 217:
///  #line 1982 "cs-parser.jay"
  {
		report.Error (525, GetLocation (yyVals[0+yyTop]), "Interfaces cannot contain fields or constants");
	  }
  break;
case 222:
///  #line 1990 "cs-parser.jay"
  {
	  	report.Error (567, GetLocation (yyVals[0+yyTop]), "Interfaces cannot contain operators");
	  }
  break;
case 223:
///  #line 1994 "cs-parser.jay"
  {
	  	report.Error (526, GetLocation (yyVals[0+yyTop]), "Interfaces cannot contain contructors");
	  }
  break;
case 224:
///  #line 1998 "cs-parser.jay"
  {
	  	report.Error (524, GetLocation (yyVals[0+yyTop]), "Interfaces cannot declare classes, structs, interfaces, delegates, or enumerations");
	  }
  break;
case 225:
///  #line 2004 "cs-parser.jay"
  {
	  }
  break;
case 226:
  case_226();
  break;
case 228:
///  #line 2037 "cs-parser.jay"
  { savedLocation = GetLocation (yyVals[0+yyTop]); yyVal = null; }
  break;
case 230:
  case_230();
  break;
case 231:
///  #line 2053 "cs-parser.jay"
  {
		valid_param_mod = ParameterModifierType.DefaultValue;
	  }
  break;
case 232:
  case_232();
  break;
case 234:
///  #line 2099 "cs-parser.jay"
  { yyVal = Operator.OpType.LogicalNot; savedOperatorLocation = GetLocation (yyVals[0+yyTop]); }
  break;
case 235:
///  #line 2100 "cs-parser.jay"
  { yyVal = Operator.OpType.OnesComplement; savedOperatorLocation = GetLocation (yyVals[0+yyTop]); }
  break;
case 236:
///  #line 2101 "cs-parser.jay"
  { yyVal = Operator.OpType.Increment; savedOperatorLocation = GetLocation (yyVals[0+yyTop]); }
  break;
case 237:
///  #line 2102 "cs-parser.jay"
  { yyVal = Operator.OpType.Decrement; savedOperatorLocation = GetLocation (yyVals[0+yyTop]); }
  break;
case 238:
///  #line 2103 "cs-parser.jay"
  { yyVal = Operator.OpType.True; savedOperatorLocation = GetLocation (yyVals[0+yyTop]); }
  break;
case 239:
///  #line 2104 "cs-parser.jay"
  { yyVal = Operator.OpType.False; savedOperatorLocation = GetLocation (yyVals[0+yyTop]); }
  break;
case 240:
///  #line 2106 "cs-parser.jay"
  { yyVal = Operator.OpType.Addition; savedOperatorLocation = GetLocation (yyVals[0+yyTop]); }
  break;
case 241:
///  #line 2107 "cs-parser.jay"
  { yyVal = Operator.OpType.Subtraction; savedOperatorLocation = GetLocation (yyVals[0+yyTop]); }
  break;
case 242:
///  #line 2109 "cs-parser.jay"
  { yyVal = Operator.OpType.Multiply; savedOperatorLocation = GetLocation (yyVals[0+yyTop]); }
  break;
case 243:
///  #line 2110 "cs-parser.jay"
  {  yyVal = Operator.OpType.Division; savedOperatorLocation = GetLocation (yyVals[0+yyTop]); }
  break;
case 244:
///  #line 2111 "cs-parser.jay"
  { yyVal = Operator.OpType.Modulus; savedOperatorLocation = GetLocation (yyVals[0+yyTop]); }
  break;
case 245:
///  #line 2112 "cs-parser.jay"
  { yyVal = Operator.OpType.BitwiseAnd; savedOperatorLocation = GetLocation (yyVals[0+yyTop]); }
  break;
case 246:
///  #line 2113 "cs-parser.jay"
  { yyVal = Operator.OpType.BitwiseOr; savedOperatorLocation = GetLocation (yyVals[0+yyTop]); }
  break;
case 247:
///  #line 2114 "cs-parser.jay"
  { yyVal = Operator.OpType.ExclusiveOr; savedOperatorLocation = GetLocation (yyVals[0+yyTop]); }
  break;
case 248:
///  #line 2115 "cs-parser.jay"
  { yyVal = Operator.OpType.LeftShift; savedOperatorLocation = GetLocation (yyVals[0+yyTop]); }
  break;
case 249:
///  #line 2116 "cs-parser.jay"
  { yyVal = Operator.OpType.RightShift; savedOperatorLocation = GetLocation (yyVals[0+yyTop]); }
  break;
case 250:
///  #line 2117 "cs-parser.jay"
  { yyVal = Operator.OpType.Equality; savedOperatorLocation = GetLocation (yyVals[0+yyTop]); }
  break;
case 251:
///  #line 2118 "cs-parser.jay"
  { yyVal = Operator.OpType.Inequality; savedOperatorLocation = GetLocation (yyVals[0+yyTop]); }
  break;
case 252:
///  #line 2119 "cs-parser.jay"
  { yyVal = Operator.OpType.GreaterThan; savedOperatorLocation = GetLocation (yyVals[0+yyTop]); }
  break;
case 253:
///  #line 2120 "cs-parser.jay"
  { yyVal = Operator.OpType.LessThan; savedOperatorLocation = GetLocation (yyVals[0+yyTop]); }
  break;
case 254:
///  #line 2121 "cs-parser.jay"
  { yyVal = Operator.OpType.GreaterThanOrEqual; savedOperatorLocation = GetLocation (yyVals[0+yyTop]); }
  break;
case 255:
///  #line 2122 "cs-parser.jay"
  { yyVal = Operator.OpType.LessThanOrEqual; savedOperatorLocation = GetLocation (yyVals[0+yyTop]); }
  break;
case 256:
///  #line 2129 "cs-parser.jay"
  {
		valid_param_mod = ParameterModifierType.DefaultValue;
	  }
  break;
case 257:
  case_257();
  break;
case 258:
///  #line 2148 "cs-parser.jay"
  {
		valid_param_mod = ParameterModifierType.DefaultValue;
	  }
  break;
case 259:
  case_259();
  break;
case 260:
  case_260();
  break;
case 261:
  case_261();
  break;
case 262:
  case_262();
  break;
case 263:
  case_263();
  break;
case 264:
  case_264();
  break;
case 265:
  case_265();
  break;
case 267:
///  #line 2251 "cs-parser.jay"
  { current_block = null; yyVal = null; }
  break;
case 270:
///  #line 2263 "cs-parser.jay"
  {
		++lexer.parsing_block;
	  }
  break;
case 271:
  case_271();
  break;
case 272:
///  #line 2273 "cs-parser.jay"
  {
		++lexer.parsing_block;
	  }
  break;
case 273:
  case_273();
  break;
case 274:
  case_274();
  break;
case 275:
  case_275();
  break;
case 276:
  case_276();
  break;
case 277:
  case_277();
  break;
case 278:
  case_278();
  break;
case 279:
  case_279();
  break;
case 280:
  case_280();
  break;
case 281:
  case_281();
  break;
case 283:
///  #line 2382 "cs-parser.jay"
  {
	  	++lexer.parsing_block;
	  }
  break;
case 284:
  case_284();
  break;
case 287:
///  #line 2399 "cs-parser.jay"
  {
		current_event_field.AddDeclarator ((FieldDeclarator) yyVals[0+yyTop]);
	  }
  break;
case 288:
///  #line 2403 "cs-parser.jay"
  {
		current_event_field.AddDeclarator ((FieldDeclarator) yyVals[0+yyTop]);
	  }
  break;
case 289:
  case_289();
  break;
case 290:
///  #line 2416 "cs-parser.jay"
  {
		++lexer.parsing_block;
	  }
  break;
case 291:
  case_291();
  break;
case 292:
  case_292();
  break;
case 293:
///  #line 2441 "cs-parser.jay"
  {
		yyVal = yyVals[0+yyTop];
	  }
  break;
case 296:
  case_296();
  break;
case 297:
  case_297();
  break;
case 298:
  case_298();
  break;
case 299:
  case_299();
  break;
case 300:
  case_300();
  break;
case 301:
  case_301();
  break;
case 302:
  case_302();
  break;
case 303:
  case_303();
  break;
case 305:
  case_305();
  break;
case 306:
  case_306();
  break;
case 307:
  case_307();
  break;
case 308:
  case_308();
  break;
case 310:
  case_310();
  break;
case 311:
  case_311();
  break;
case 314:
///  #line 2604 "cs-parser.jay"
  {
		lbag.AppendToMember (current_class, GetLocation (yyVals[0+yyTop]));
	  }
  break;
case 316:
  case_316();
  break;
case 317:
  case_317();
  break;
case 318:
  case_318();
  break;
case 319:
  case_319();
  break;
case 320:
///  #line 2662 "cs-parser.jay"
  {
		valid_param_mod = ParameterModifierType.Ref | ParameterModifierType.Out | ParameterModifierType.Params | ParameterModifierType.DefaultValue;
	  }
  break;
case 321:
  case_321();
  break;
case 322:
///  #line 2684 "cs-parser.jay"
  {
		lexer.ConstraintsParsing = false;
	  }
  break;
case 323:
  case_323();
  break;
case 325:
  case_325();
  break;
case 327:
  case_327();
  break;
case 329:
  case_329();
  break;
case 330:
  case_330();
  break;
case 332:
  case_332();
  break;
case 333:
  case_333();
  break;
case 334:
  case_334();
  break;
case 335:
  case_335();
  break;
case 336:
///  #line 2789 "cs-parser.jay"
  {
		lexer.parsing_generic_declaration = true;
	  }
  break;
case 337:
  case_337();
  break;
case 338:
  case_338();
  break;
case 340:
  case_340();
  break;
case 341:
  case_341();
  break;
case 342:
  case_342();
  break;
case 343:
  case_343();
  break;
case 344:
  case_344();
  break;
case 345:
  case_345();
  break;
case 347:
  case_347();
  break;
case 348:
  case_348();
  break;
case 349:
  case_349();
  break;
case 350:
  case_350();
  break;
case 351:
  case_351();
  break;
case 353:
///  #line 2910 "cs-parser.jay"
  {
		yyVal = new TypeExpression (compiler.BuiltinTypes.Void, GetLocation (yyVals[0+yyTop]));
	  }
  break;
case 354:
///  #line 2917 "cs-parser.jay"
  {
		lexer.parsing_generic_declaration = true;
	  }
  break;
case 356:
  case_356();
  break;
case 358:
  case_358();
  break;
case 360:
  case_360();
  break;
case 362:
///  #line 2955 "cs-parser.jay"
  {
		yyVal = new ComposedCast ((FullNamedExpression) yyVals[-1+yyTop], (ComposedTypeSpecifier) yyVals[0+yyTop]);
	  }
  break;
case 363:
  case_363();
  break;
case 364:
///  #line 2975 "cs-parser.jay"
  {
		yyVal = new ComposedCast (((MemberName) yyVals[-1+yyTop]).GetTypeExpression (), (ComposedTypeSpecifier) yyVals[0+yyTop]);
	  }
  break;
case 365:
  case_365();
  break;
case 366:
///  #line 2984 "cs-parser.jay"
  {
		yyVal = new ComposedCast ((FullNamedExpression) yyVals[-1+yyTop], (ComposedTypeSpecifier) yyVals[0+yyTop]);
	  }
  break;
case 367:
///  #line 2988 "cs-parser.jay"
  {
		yyVal = new ComposedCast (new TypeExpression (compiler.BuiltinTypes.Void, GetLocation (yyVals[-1+yyTop])), (ComposedTypeSpecifier) yyVals[0+yyTop]);
	  }
  break;
case 368:
  case_368();
  break;
case 369:
  case_369();
  break;
case 370:
  case_370();
  break;
case 371:
  case_371();
  break;
case 372:
///  #line 3027 "cs-parser.jay"
  { yyVal = new TypeExpression (compiler.BuiltinTypes.Object, GetLocation (yyVals[0+yyTop])); }
  break;
case 373:
///  #line 3028 "cs-parser.jay"
  { yyVal = new TypeExpression (compiler.BuiltinTypes.String, GetLocation (yyVals[0+yyTop])); }
  break;
case 374:
///  #line 3029 "cs-parser.jay"
  { yyVal = new TypeExpression (compiler.BuiltinTypes.Bool, GetLocation (yyVals[0+yyTop])); }
  break;
case 375:
///  #line 3030 "cs-parser.jay"
  { yyVal = new TypeExpression (compiler.BuiltinTypes.Decimal, GetLocation (yyVals[0+yyTop])); }
  break;
case 376:
///  #line 3031 "cs-parser.jay"
  { yyVal = new TypeExpression (compiler.BuiltinTypes.Float, GetLocation (yyVals[0+yyTop])); }
  break;
case 377:
///  #line 3032 "cs-parser.jay"
  { yyVal = new TypeExpression (compiler.BuiltinTypes.Double, GetLocation (yyVals[0+yyTop])); }
  break;
case 379:
///  #line 3037 "cs-parser.jay"
  { yyVal = new TypeExpression (compiler.BuiltinTypes.SByte, GetLocation (yyVals[0+yyTop])); }
  break;
case 380:
///  #line 3038 "cs-parser.jay"
  { yyVal = new TypeExpression (compiler.BuiltinTypes.Byte, GetLocation (yyVals[0+yyTop])); }
  break;
case 381:
///  #line 3039 "cs-parser.jay"
  { yyVal = new TypeExpression (compiler.BuiltinTypes.Short, GetLocation (yyVals[0+yyTop])); }
  break;
case 382:
///  #line 3040 "cs-parser.jay"
  { yyVal = new TypeExpression (compiler.BuiltinTypes.UShort, GetLocation (yyVals[0+yyTop])); }
  break;
case 383:
///  #line 3041 "cs-parser.jay"
  { yyVal = new TypeExpression (compiler.BuiltinTypes.Int, GetLocation (yyVals[0+yyTop])); }
  break;
case 384:
///  #line 3042 "cs-parser.jay"
  { yyVal = new TypeExpression (compiler.BuiltinTypes.UInt, GetLocation (yyVals[0+yyTop])); }
  break;
case 385:
///  #line 3043 "cs-parser.jay"
  { yyVal = new TypeExpression (compiler.BuiltinTypes.Long, GetLocation (yyVals[0+yyTop])); }
  break;
case 386:
///  #line 3044 "cs-parser.jay"
  { yyVal = new TypeExpression (compiler.BuiltinTypes.ULong, GetLocation (yyVals[0+yyTop])); }
  break;
case 387:
///  #line 3045 "cs-parser.jay"
  { yyVal = new TypeExpression (compiler.BuiltinTypes.Char, GetLocation (yyVals[0+yyTop])); }
  break;
case 408:
  case_408();
  break;
case 409:
  case_409();
  break;
case 413:
///  #line 3092 "cs-parser.jay"
  { yyVal = new NullLiteral (GetLocation (yyVals[0+yyTop])); }
  break;
case 414:
///  #line 3096 "cs-parser.jay"
  { yyVal = new BoolLiteral (compiler.BuiltinTypes, true, GetLocation (yyVals[0+yyTop])); }
  break;
case 415:
///  #line 3097 "cs-parser.jay"
  { yyVal = new BoolLiteral (compiler.BuiltinTypes, false, GetLocation (yyVals[0+yyTop])); }
  break;
case 420:
  case_420();
  break;
case 421:
///  #line 3130 "cs-parser.jay"
  {
		yyVal = new ParenthesizedExpression ((Expression) yyVals[-1+yyTop]);
	  }
  break;
case 422:
  case_422();
  break;
case 423:
  case_423();
  break;
case 424:
  case_424();
  break;
case 425:
  case_425();
  break;
case 426:
///  #line 3165 "cs-parser.jay"
  {
		yyVal = new CompletionMemberAccess ((Expression) yyVals[-2+yyTop], null,GetLocation (yyVals[0+yyTop]));
	  }
  break;
case 427:
  case_427();
  break;
case 428:
///  #line 3173 "cs-parser.jay"
  {
		yyVal = new CompletionMemberAccess ((Expression) yyVals[-2+yyTop], null, lexer.Location);
	  }
  break;
case 429:
  case_429();
  break;
case 430:
  case_430();
  break;
case 431:
///  #line 3189 "cs-parser.jay"
  { yyVal = null; }
  break;
case 433:
  case_433();
  break;
case 434:
  case_434();
  break;
case 435:
///  #line 3212 "cs-parser.jay"
  { yyVal = null; }
  break;
case 436:
///  #line 3216 "cs-parser.jay"
  {
		yyVal = yyVals[0+yyTop];
	}
  break;
case 437:
  case_437();
  break;
case 438:
  case_438();
  break;
case 439:
  case_439();
  break;
case 440:
  case_440();
  break;
case 441:
///  #line 3249 "cs-parser.jay"
  {
		yyVal = new CompletionElementInitializer (null, GetLocation (yyVals[0+yyTop]));
	  }
  break;
case 442:
  case_442();
  break;
case 443:
  case_443();
  break;
case 444:
  case_444();
  break;
case 447:
///  #line 3277 "cs-parser.jay"
  { yyVal = null; }
  break;
case 449:
  case_449();
  break;
case 450:
  case_450();
  break;
case 451:
  case_451();
  break;
case 452:
  case_452();
  break;
case 453:
  case_453();
  break;
case 454:
///  #line 3329 "cs-parser.jay"
  {
		yyVal = new Argument ((Expression) yyVals[0+yyTop]);
	  }
  break;
case 458:
  case_458();
  break;
case 459:
  case_459();
  break;
case 460:
  case_460();
  break;
case 461:
  case_461();
  break;
case 463:
  case_463();
  break;
case 464:
///  #line 3374 "cs-parser.jay"
  {
		yyVal = new ElementAccess ((Expression) yyVals[-3+yyTop], (Arguments) yyVals[-1+yyTop], GetLocation (yyVals[-2+yyTop]));
	  }
  break;
case 465:
///  #line 3378 "cs-parser.jay"
  {
		yyVal = new ElementAccess ((Expression) yyVals[-2+yyTop], null, GetLocation (yyVals[-1+yyTop]));
	  }
  break;
case 466:
  case_466();
  break;
case 467:
  case_467();
  break;
case 468:
  case_468();
  break;
case 469:
  case_469();
  break;
case 470:
  case_470();
  break;
case 471:
///  #line 3424 "cs-parser.jay"
  {
	  	yyVal = new Argument ((Expression) yyVals[0+yyTop]);
	  }
  break;
case 473:
///  #line 3432 "cs-parser.jay"
  {
		yyVal = new This (GetLocation (yyVals[0+yyTop]));
	  }
  break;
case 474:
  case_474();
  break;
case 475:
  case_475();
  break;
case 476:
///  #line 3452 "cs-parser.jay"
  {
		yyVal = new UnaryMutator (UnaryMutator.Mode.PostIncrement, (Expression) yyVals[-1+yyTop], GetLocation (yyVals[0+yyTop]));
	  }
  break;
case 477:
///  #line 3459 "cs-parser.jay"
  {
		yyVal = new UnaryMutator (UnaryMutator.Mode.PostDecrement, (Expression) yyVals[-1+yyTop], GetLocation (yyVals[0+yyTop]));
	  }
  break;
case 478:
  case_478();
  break;
case 479:
  case_479();
  break;
case 480:
  case_480();
  break;
case 481:
  case_481();
  break;
case 482:
  case_482();
  break;
case 483:
  case_483();
  break;
case 484:
  case_484();
  break;
case 485:
///  #line 3525 "cs-parser.jay"
  {
		++lexer.parsing_type;
	  }
  break;
case 486:
  case_486();
  break;
case 487:
  case_487();
  break;
case 490:
///  #line 3552 "cs-parser.jay"
  { yyVal = null; }
  break;
case 492:
  case_492();
  break;
case 493:
  case_493();
  break;
case 494:
  case_494();
  break;
case 495:
  case_495();
  break;
case 496:
  case_496();
  break;
case 497:
  case_497();
  break;
case 501:
  case_501();
  break;
case 502:
  case_502();
  break;
case 503:
  case_503();
  break;
case 504:
///  #line 3630 "cs-parser.jay"
  {
		yyVal = 2;
	  }
  break;
case 505:
///  #line 3634 "cs-parser.jay"
  {
		yyVal = ((int) yyVals[-1+yyTop]) + 1;
	  }
  break;
case 506:
///  #line 3641 "cs-parser.jay"
  {
		yyVal = null;
	  }
  break;
case 507:
///  #line 3645 "cs-parser.jay"
  {
		yyVal = yyVals[0+yyTop];
	  }
  break;
case 508:
  case_508();
  break;
case 509:
  case_509();
  break;
case 510:
  case_510();
  break;
case 511:
  case_511();
  break;
case 512:
///  #line 3689 "cs-parser.jay"
  {
	  	lexer.TypeOfParsing = true;
	  }
  break;
case 513:
  case_513();
  break;
case 516:
  case_516();
  break;
case 517:
  case_517();
  break;
case 518:
  case_518();
  break;
case 519:
  case_519();
  break;
case 520:
  case_520();
  break;
case 521:
  case_521();
  break;
case 522:
  case_522();
  break;
case 523:
  case_523();
  break;
case 524:
  case_524();
  break;
case 525:
  case_525();
  break;
case 526:
  case_526();
  break;
case 527:
  case_527();
  break;
case 528:
///  #line 3809 "cs-parser.jay"
  {
		start_anonymous (false, (ParametersCompiled) yyVals[0+yyTop], false, GetLocation (yyVals[-1+yyTop]));
	  }
  break;
case 529:
  case_529();
  break;
case 530:
///  #line 3822 "cs-parser.jay"
  {
		start_anonymous (false, (ParametersCompiled) yyVals[0+yyTop], true, GetLocation (yyVals[-2+yyTop]));
	  }
  break;
case 531:
  case_531();
  break;
case 532:
///  #line 3839 "cs-parser.jay"
  {
		yyVal = ParametersCompiled.Undefined;
	  }
  break;
case 534:
///  #line 3847 "cs-parser.jay"
  {
	  	valid_param_mod = ParameterModifierType.Ref | ParameterModifierType.Out;
	  }
  break;
case 535:
  case_535();
  break;
case 536:
  case_536();
  break;
case 538:
///  #line 3873 "cs-parser.jay"
  {
		yyVal = new Unary (Unary.Operator.LogicalNot, (Expression) yyVals[0+yyTop], GetLocation (yyVals[-1+yyTop]));
	  }
  break;
case 539:
///  #line 3877 "cs-parser.jay"
  {
		yyVal = new Unary (Unary.Operator.OnesComplement, (Expression) yyVals[0+yyTop], GetLocation (yyVals[-1+yyTop]));
	  }
  break;
case 540:
  case_540();
  break;
case 541:
  case_541();
  break;
case 543:
///  #line 3905 "cs-parser.jay"
  { 
	  	yyVal = new Unary (Unary.Operator.UnaryPlus, (Expression) yyVals[0+yyTop], GetLocation (yyVals[-1+yyTop]));
	  }
  break;
case 544:
///  #line 3909 "cs-parser.jay"
  { 
		yyVal = new Unary (Unary.Operator.UnaryNegation, (Expression) yyVals[0+yyTop], GetLocation (yyVals[-1+yyTop]));
	  }
  break;
case 545:
///  #line 3913 "cs-parser.jay"
  {
		yyVal = new UnaryMutator (UnaryMutator.Mode.PreIncrement, (Expression) yyVals[0+yyTop], GetLocation (yyVals[-1+yyTop]));
	  }
  break;
case 546:
///  #line 3917 "cs-parser.jay"
  {
		yyVal = new UnaryMutator (UnaryMutator.Mode.PreDecrement, (Expression) yyVals[0+yyTop], GetLocation (yyVals[-1+yyTop]));
	  }
  break;
case 547:
///  #line 3921 "cs-parser.jay"
  {
		yyVal = new Indirection ((Expression) yyVals[0+yyTop], GetLocation (yyVals[-1+yyTop]));
	  }
  break;
case 548:
///  #line 3925 "cs-parser.jay"
  {
		yyVal = new Unary (Unary.Operator.AddressOf, (Expression) yyVals[0+yyTop], GetLocation (yyVals[-1+yyTop]));
	  }
  break;
case 550:
  case_550();
  break;
case 551:
  case_551();
  break;
case 552:
  case_552();
  break;
case 554:
  case_554();
  break;
case 555:
///  #line 3957 "cs-parser.jay"
  {
		yyVal = new Binary (Binary.Operator.Subtraction, (Expression) yyVals[-2+yyTop], (Expression) yyVals[0+yyTop], GetLocation (yyVals[-1+yyTop]));
	  }
  break;
case 556:
  case_556();
  break;
case 557:
///  #line 3966 "cs-parser.jay"
  {
		yyVal = new As ((Expression) yyVals[-2+yyTop], (Expression) yyVals[0+yyTop], GetLocation (yyVals[-1+yyTop]));
	  }
  break;
case 558:
///  #line 3970 "cs-parser.jay"
  {
		yyVal = new Is ((Expression) yyVals[-2+yyTop], (Expression) yyVals[0+yyTop], GetLocation (yyVals[-1+yyTop]));
	  }
  break;
case 560:
  case_560();
  break;
case 561:
  case_561();
  break;
case 563:
  case_563();
  break;
case 564:
  case_564();
  break;
case 565:
  case_565();
  break;
case 566:
  case_566();
  break;
case 568:
  case_568();
  break;
case 569:
  case_569();
  break;
case 571:
  case_571();
  break;
case 573:
  case_573();
  break;
case 575:
  case_575();
  break;
case 577:
  case_577();
  break;
case 579:
  case_579();
  break;
case 581:
  case_581();
  break;
case 583:
  case_583();
  break;
case 584:
///  #line 4094 "cs-parser.jay"
  {
		yyVal = new SimpleAssign ((Expression) yyVals[-2+yyTop], (Expression) yyVals[0+yyTop], GetLocation (yyVals[-1+yyTop]));
	  }
  break;
case 585:
  case_585();
  break;
case 586:
  case_586();
  break;
case 587:
  case_587();
  break;
case 588:
  case_588();
  break;
case 589:
  case_589();
  break;
case 590:
  case_590();
  break;
case 591:
  case_591();
  break;
case 592:
  case_592();
  break;
case 593:
  case_593();
  break;
case 594:
  case_594();
  break;
case 595:
  case_595();
  break;
case 596:
  case_596();
  break;
case 597:
  case_597();
  break;
case 598:
  case_598();
  break;
case 599:
  case_599();
  break;
case 600:
///  #line 4191 "cs-parser.jay"
  { yyVal = ParametersCompiled.EmptyReadOnlyParameters; }
  break;
case 601:
  case_601();
  break;
case 604:
///  #line 4207 "cs-parser.jay"
  {
		start_block (lexer.Location);
	  }
  break;
case 605:
  case_605();
  break;
case 607:
  case_607();
  break;
case 608:
  case_608();
  break;
case 609:
  case_609();
  break;
case 610:
  case_610();
  break;
case 611:
  case_611();
  break;
case 612:
///  #line 4252 "cs-parser.jay"
  {
	  	valid_param_mod = ParameterModifierType.Ref | ParameterModifierType.Out;
	  }
  break;
case 613:
  case_613();
  break;
case 614:
  case_614();
  break;
case 615:
///  #line 4266 "cs-parser.jay"
  {
	  	valid_param_mod = ParameterModifierType.Ref | ParameterModifierType.Out;	  
	  }
  break;
case 616:
  case_616();
  break;
case 617:
  case_617();
  break;
case 623:
///  #line 4291 "cs-parser.jay"
  {
		yyVal = new ArglistAccess (GetLocation (yyVals[0+yyTop]));
	  }
  break;
case 624:
  case_624();
  break;
case 625:
  case_625();
  break;
case 626:
  case_626();
  break;
case 628:
///  #line 4320 "cs-parser.jay"
  {
		yyVal = new BooleanExpression ((Expression) yyVals[0+yyTop]);
	  }
  break;
case 629:
///  #line 4333 "cs-parser.jay"
  {
		lexer.ConstraintsParsing = true;
	  }
  break;
case 630:
  case_630();
  break;
case 631:
  case_631();
  break;
case 632:
  case_632();
  break;
case 633:
  case_633();
  break;
case 634:
///  #line 4376 "cs-parser.jay"
  { yyVal = null; }
  break;
case 635:
///  #line 4378 "cs-parser.jay"
  { yyVal = yyVals[0+yyTop]; StoreModifierLocation (Modifiers.PARTIAL, GetLocation (yyVals[0+yyTop])); }
  break;
case 636:
  case_636();
  break;
case 637:
///  #line 4391 "cs-parser.jay"
  {
		lexer.parsing_modifiers = false;		
	  }
  break;
case 639:
  case_639();
  break;
case 640:
  case_640();
  break;
case 641:
  case_641();
  break;
case 642:
  case_642();
  break;
case 643:
  case_643();
  break;
case 644:
  case_644();
  break;
case 645:
  case_645();
  break;
case 646:
  case_646();
  break;
case 647:
  case_647();
  break;
case 648:
  case_648();
  break;
case 649:
  case_649();
  break;
case 650:
  case_650();
  break;
case 651:
  case_651();
  break;
case 652:
  case_652();
  break;
case 653:
  case_653();
  break;
case 654:
  case_654();
  break;
case 656:
  case_656();
  break;
case 658:
///  #line 4511 "cs-parser.jay"
  {
		yyVal = yyVals[0+yyTop];
	  }
  break;
case 659:
  case_659();
  break;
case 660:
  case_660();
  break;
case 661:
  case_661();
  break;
case 662:
  case_662();
  break;
case 663:
  case_663();
  break;
case 664:
  case_664();
  break;
case 665:
  case_665();
  break;
case 666:
  case_666();
  break;
case 667:
///  #line 4602 "cs-parser.jay"
  {
		yyVal = new SpecialContraintExpr (SpecialConstraint.Class, GetLocation (yyVals[0+yyTop]));
	  }
  break;
case 668:
///  #line 4606 "cs-parser.jay"
  {
		yyVal = new SpecialContraintExpr (SpecialConstraint.Struct, GetLocation (yyVals[0+yyTop]));
	  }
  break;
case 669:
///  #line 4613 "cs-parser.jay"
  {
		yyVal = Variance.None;
	  }
  break;
case 670:
  case_670();
  break;
case 671:
  case_671();
  break;
case 672:
  case_672();
  break;
case 673:
  case_673();
  break;
case 674:
///  #line 4658 "cs-parser.jay"
  {
		yyVal = yyVals[0+yyTop];
	  }
  break;
case 675:
  case_675();
  break;
case 676:
  case_676();
  break;
case 677:
  case_677();
  break;
case 678:
  case_678();
  break;
case 683:
///  #line 4702 "cs-parser.jay"
  {
		current_block.AddStatement ((Statement) yyVals[0+yyTop]);
	  }
  break;
case 684:
///  #line 4706 "cs-parser.jay"
  {
		current_block.AddStatement ((Statement) yyVals[0+yyTop]);
	  }
  break;
case 686:
  case_686();
  break;
case 687:
  case_687();
  break;
case 690:
///  #line 4740 "cs-parser.jay"
  {
		current_block.AddStatement ((Statement) yyVals[0+yyTop]);
	  }
  break;
case 691:
///  #line 4744 "cs-parser.jay"
  {
		current_block.AddStatement ((Statement) yyVals[0+yyTop]);
	  }
  break;
case 720:
  case_720();
  break;
case 721:
  case_721();
  break;
case 722:
  case_722();
  break;
case 723:
  case_723();
  break;
case 724:
  case_724();
  break;
case 727:
  case_727();
  break;
case 728:
  case_728();
  break;
case 729:
  case_729();
  break;
case 730:
  case_730();
  break;
case 731:
///  #line 4888 "cs-parser.jay"
  {
		yyVal = new ComposedCast ((FullNamedExpression) yyVals[-1+yyTop], (ComposedTypeSpecifier) yyVals[0+yyTop]);
	  }
  break;
case 732:
///  #line 4892 "cs-parser.jay"
  {
		yyVal = new ComposedCast (new TypeExpression (compiler.BuiltinTypes.Void, GetLocation (yyVals[-1+yyTop])), (ComposedTypeSpecifier) yyVals[0+yyTop]);
	  }
  break;
case 733:
  case_733();
  break;
case 735:
  case_735();
  break;
case 736:
///  #line 4913 "cs-parser.jay"
  {
		yyVal = ComposedTypeSpecifier.CreatePointer (GetLocation (yyVals[0+yyTop]));
	  }
  break;
case 738:
  case_738();
  break;
case 739:
  case_739();
  break;
case 740:
  case_740();
  break;
case 741:
  case_741();
  break;
case 742:
  case_742();
  break;
case 744:
  case_744();
  break;
case 745:
  case_745();
  break;
case 746:
  case_746();
  break;
case 750:
  case_750();
  break;
case 753:
  case_753();
  break;
case 754:
  case_754();
  break;
case 755:
///  #line 5038 "cs-parser.jay"
  {
		report.Error (145, lexer.Location, "A const field requires a value to be provided");
	  }
  break;
case 756:
  case_756();
  break;
case 761:
  case_761();
  break;
case 763:
  case_763();
  break;
case 764:
  case_764();
  break;
case 765:
  case_765();
  break;
case 766:
///  #line 5088 "cs-parser.jay"
  { yyVal = yyVals[-1+yyTop]; }
  break;
case 767:
///  #line 5092 "cs-parser.jay"
  { yyVal = yyVals[-1+yyTop]; }
  break;
case 768:
///  #line 5093 "cs-parser.jay"
  { yyVal = yyVals[-1+yyTop]; }
  break;
case 769:
  case_769();
  break;
case 770:
  case_770();
  break;
case 771:
  case_771();
  break;
case 774:
  case_774();
  break;
case 775:
  case_775();
  break;
case 776:
  case_776();
  break;
case 777:
///  #line 5167 "cs-parser.jay"
  {
		start_block (GetLocation (yyVals[0+yyTop]));
	  }
  break;
case 778:
  case_778();
  break;
case 779:
  case_779();
  break;
case 781:
  case_781();
  break;
case 782:
  case_782();
  break;
case 783:
  case_783();
  break;
case 784:
///  #line 5211 "cs-parser.jay"
  {
		current_block = current_block.CreateSwitchBlock (lexer.Location);
	  }
  break;
case 785:
///  #line 5215 "cs-parser.jay"
  {
		yyVal = new SwitchSection ((List<SwitchLabel>) yyVals[-2+yyTop], current_block);
	  }
  break;
case 786:
  case_786();
  break;
case 787:
  case_787();
  break;
case 788:
  case_788();
  break;
case 789:
///  #line 5244 "cs-parser.jay"
  {
		yyVal = new SwitchLabel (null, GetLocation (yyVals[0+yyTop]));
	  }
  break;
case 794:
  case_794();
  break;
case 795:
  case_795();
  break;
case 796:
  case_796();
  break;
case 797:
  case_797();
  break;
case 798:
  case_798();
  break;
case 799:
///  #line 5298 "cs-parser.jay"
  {
		yyVal = yyVals[0+yyTop];
	  }
  break;
case 800:
///  #line 5306 "cs-parser.jay"
  {
		((For) yyVals[-2+yyTop]).Initializer = (Statement) yyVals[-1+yyTop];
	  }
  break;
case 801:
///  #line 5310 "cs-parser.jay"
  {
		((For) yyVals[-5+yyTop]).Condition = (BooleanExpression) yyVals[-1+yyTop];
	  }
  break;
case 802:
///  #line 5314 "cs-parser.jay"
  {
		((For) yyVals[-8+yyTop]).Iterator = (Statement) yyVals[-1+yyTop];
	  }
  break;
case 803:
  case_803();
  break;
case 804:
  case_804();
  break;
case 805:
///  #line 5334 "cs-parser.jay"
  { yyVal = new EmptyStatement (lexer.Location); }
  break;
case 807:
  case_807();
  break;
case 808:
  case_808();
  break;
case 810:
///  #line 5355 "cs-parser.jay"
  { yyVal = null; }
  break;
case 812:
///  #line 5360 "cs-parser.jay"
  { yyVal = new EmptyStatement (lexer.Location); }
  break;
case 816:
  case_816();
  break;
case 817:
  case_817();
  break;
case 818:
  case_818();
  break;
case 819:
  case_819();
  break;
case 820:
  case_820();
  break;
case 821:
  case_821();
  break;
case 828:
  case_828();
  break;
case 829:
  case_829();
  break;
case 830:
  case_830();
  break;
case 831:
  case_831();
  break;
case 832:
  case_832();
  break;
case 833:
  case_833();
  break;
case 834:
  case_834();
  break;
case 835:
  case_835();
  break;
case 836:
  case_836();
  break;
case 839:
///  #line 5536 "cs-parser.jay"
  {
		yyVal = new TryCatch ((Block) yyVals[-1+yyTop], (List<Catch>) yyVals[0+yyTop], GetLocation (yyVals[-2+yyTop]), false);
	  }
  break;
case 840:
  case_840();
  break;
case 841:
  case_841();
  break;
case 842:
  case_842();
  break;
case 843:
  case_843();
  break;
case 844:
  case_844();
  break;
case 847:
///  #line 5589 "cs-parser.jay"
  {
		yyVal = new Catch ((Block) yyVals[0+yyTop], GetLocation (yyVals[-1+yyTop]));
	  }
  break;
case 848:
  case_848();
  break;
case 849:
///  #line 5608 "cs-parser.jay"
  {
		yyVal = yyVals[-1+yyTop];
	  }
  break;
case 850:
  case_850();
  break;
case 851:
///  #line 5626 "cs-parser.jay"
  {
		yyVal = new Checked ((Block) yyVals[0+yyTop], GetLocation (yyVals[-1+yyTop]));
	  }
  break;
case 852:
///  #line 5633 "cs-parser.jay"
  {
		yyVal = new Unchecked ((Block) yyVals[0+yyTop], GetLocation (yyVals[-1+yyTop]));
	  }
  break;
case 853:
  case_853();
  break;
case 854:
///  #line 5643 "cs-parser.jay"
  {
		yyVal = new Unsafe ((Block) yyVals[0+yyTop], GetLocation (yyVals[-2+yyTop]));
	  }
  break;
case 855:
  case_855();
  break;
case 856:
  case_856();
  break;
case 857:
  case_857();
  break;
case 858:
  case_858();
  break;
case 859:
  case_859();
  break;
case 860:
  case_860();
  break;
case 861:
  case_861();
  break;
case 862:
  case_862();
  break;
case 864:
  case_864();
  break;
case 865:
///  #line 5735 "cs-parser.jay"
  {
		Error_MissingInitializer (lexer.Location);
	  }
  break;
case 866:
  case_866();
  break;
case 867:
  case_867();
  break;
case 868:
  case_868();
  break;
case 869:
  case_869();
  break;
case 870:
  case_870();
  break;
case 871:
  case_871();
  break;
case 872:
  case_872();
  break;
case 873:
  case_873();
  break;
case 874:
  case_874();
  break;
case 875:
///  #line 5836 "cs-parser.jay"
  {
		current_block = new Linq.QueryBlock (current_block, lexer.Location);
	  }
  break;
case 876:
  case_876();
  break;
case 877:
///  #line 5851 "cs-parser.jay"
  {
		current_block = new Linq.QueryBlock (current_block, lexer.Location);
	  }
  break;
case 878:
  case_878();
  break;
case 879:
  case_879();
  break;
case 881:
  case_881();
  break;
case 882:
///  #line 5896 "cs-parser.jay"
  {
	  	current_block = new Linq.QueryBlock (current_block, lexer.Location);
	  }
  break;
case 883:
  case_883();
  break;
case 884:
  case_884();
  break;
case 885:
  case_885();
  break;
case 886:
  case_886();
  break;
case 890:
  case_890();
  break;
case 896:
///  #line 5955 "cs-parser.jay"
  {
	  	current_block = new Linq.QueryBlock (current_block, lexer.Location);
	  }
  break;
case 897:
  case_897();
  break;
case 898:
///  #line 5974 "cs-parser.jay"
  {
	  	current_block = new Linq.QueryBlock (current_block, lexer.Location);
	  }
  break;
case 899:
  case_899();
  break;
case 900:
  case_900();
  break;
case 901:
  case_901();
  break;
case 902:
  case_902();
  break;
case 903:
  case_903();
  break;
case 904:
  case_904();
  break;
case 905:
  case_905();
  break;
case 906:
  case_906();
  break;
case 907:
  case_907();
  break;
case 909:
///  #line 6118 "cs-parser.jay"
  {
		yyVal = yyVals[0+yyTop];
	  }
  break;
case 910:
///  #line 6125 "cs-parser.jay"
  {
		current_block = new Linq.QueryBlock (current_block, lexer.Location);
	  }
  break;
case 911:
  case_911();
  break;
case 913:
  case_913();
  break;
case 914:
  case_914();
  break;
case 916:
  case_916();
  break;
case 917:
  case_917();
  break;
case 918:
///  #line 6171 "cs-parser.jay"
  {
		yyVal = new Linq.OrderByAscending ((Linq.QueryBlock) current_block, (Expression)yyVals[0+yyTop]);	
	  }
  break;
case 919:
  case_919();
  break;
case 920:
  case_920();
  break;
case 921:
///  #line 6188 "cs-parser.jay"
  {
		yyVal = new Linq.ThenByAscending ((Linq.QueryBlock) current_block, (Expression)yyVals[0+yyTop]);	
	  }
  break;
case 922:
  case_922();
  break;
case 923:
  case_923();
  break;
case 925:
  case_925();
  break;
case 926:
  case_926();
  break;
case 929:
  case_929();
  break;
case 930:
  case_930();
  break;
case 938:
///  #line 6312 "cs-parser.jay"
  {
		module.DocumentationBuilder.ParsedName = (MemberName) yyVals[0+yyTop];
	  }
  break;
case 939:
///  #line 6319 "cs-parser.jay"
  {
		module.DocumentationBuilder.ParsedParameters = (List<DocumentationParameter>)yyVals[0+yyTop];
	  }
  break;
case 940:
  case_940();
  break;
case 941:
  case_941();
  break;
case 942:
///  #line 6336 "cs-parser.jay"
  {
		yyVal = new MemberName ((MemberName) yyVals[-2+yyTop], new MemberName (MemberCache.IndexerNameAlias));
	  }
  break;
case 943:
///  #line 6340 "cs-parser.jay"
  {
		valid_param_mod = ParameterModifierType.Ref | ParameterModifierType.Out;
	  }
  break;
case 944:
  case_944();
  break;
case 945:
  case_945();
  break;
case 946:
  case_946();
  break;
case 947:
  case_947();
  break;
case 949:
///  #line 6376 "cs-parser.jay"
  {
		yyVal = new MemberName (((MemberName) yyVals[-2+yyTop]), (MemberName) yyVals[0+yyTop]);
	  }
  break;
case 951:
///  #line 6384 "cs-parser.jay"
  {
		valid_param_mod = ParameterModifierType.Ref | ParameterModifierType.Out;
	  }
  break;
case 952:
///  #line 6388 "cs-parser.jay"
  {
		yyVal = yyVals[-1+yyTop];
	  }
  break;
case 953:
///  #line 6395 "cs-parser.jay"
  {
		yyVal = new List<DocumentationParameter> (0);
	  }
  break;
case 955:
  case_955();
  break;
case 956:
  case_956();
  break;
case 957:
  case_957();
  break;
///  #line default
        }
        yyTop -= yyLen[yyN];
        yyState = yyStates[yyTop];
        int yyM = yyLhs[yyN];
        if (yyState == 0 && yyM == 0) {
//t          if (debug != null) debug.shift(0, yyFinal);
          yyState = yyFinal;
          if (yyToken < 0) {
            yyToken = yyLex.advance() ? yyLex.token() : 0;
//t            if (debug != null)
//t               debug.lex(yyState, yyToken,yyname(yyToken), yyLex.value());
          }
          if (yyToken == 0) {
//t            if (debug != null) debug.accept(yyVal);
            return yyVal;
          }
          goto continue_yyLoop;
        }
        if (((yyN = yyGindex[yyM]) != 0) && ((yyN += yyState) >= 0)
            && (yyN < yyTable.Length) && (yyCheck[yyN] == yyState))
          yyState = yyTable[yyN];
        else
          yyState = yyDgoto[yyM];
//t        if (debug != null) debug.shift(yyStates[yyTop], yyState);
	 goto continue_yyLoop;
      continue_yyDiscarded: ;	// implements the named-loop continue: 'continue yyDiscarded'
      }
    continue_yyLoop: ;		// implements the named-loop continue: 'continue yyLoop'
    }
  }

/*
 All more than 3 lines long rules are wrapped into a method
*/
void case_6()
///  #line 396 "cs-parser.jay"
{
		if (yyVals[0+yyTop] != null) {
			Attributes attrs = (Attributes) yyVals[0+yyTop];
			report.Error (1730, attrs.Attrs [0].Location,
				"Assembly and module attributes must precede all other elements except using clauses and extern alias declarations");
		}
	  }

void case_8()
///  #line 408 "cs-parser.jay"
{
		if (yyToken == Token.EXTERN_ALIAS)
			report.Error (439, lexer.Location, "An extern alias declaration must precede all other elements");
		else
			Error_SyntaxError (yyToken);
	  }

void case_13()
///  #line 428 "cs-parser.jay"
{
		var lt = (Tokenizer.LocatedToken) yyVals[-2+yyTop];
		string s = lt.Value;
		if (s != "alias"){
			syntax_error (lt.Location, "`alias' expected");
		} else if (lang_version == LanguageVersion.ISO_1) {
			FeatureIsNotAvailable (lt.Location, "external alias");
		} else {
			lt = (Tokenizer.LocatedToken) yyVals[-1+yyTop]; 
			current_namespace.AddUsingExternalAlias (lt.Value, lt.Location, report);
			ubag.AddExternAlias  (GetLocation (yyVals[-3+yyTop]), GetLocation (yyVals[-2+yyTop]), lt, GetLocation (yyVals[0+yyTop]));
		}
	  }

void case_17()
///  #line 454 "cs-parser.jay"
{
		if (doc_support)
			Lexer.doc_state = XmlCommentState.Allowed;
	  }

void case_18()
///  #line 459 "cs-parser.jay"
{
		if (doc_support)
			Lexer.doc_state = XmlCommentState.Allowed;
	  }

void case_19()
///  #line 467 "cs-parser.jay"
{
		var lt = (Tokenizer.LocatedToken) yyVals[-3+yyTop];
		if (lang_version != LanguageVersion.ISO_1 && lt.Value == "global") {
			report.Warning (440, 2, lt.Location,
			 "An alias named `global' will not be used when resolving `global::'. The global namespace will be used instead");
		}

		current_namespace.AddUsingAlias (lt.Value, (MemberName) yyVals[-1+yyTop], GetLocation (yyVals[-4+yyTop]));
		ubag.AddUsingAlias (GetLocation (yyVals[-4+yyTop]), lt, GetLocation (yyVals[-2+yyTop]), (MemberName) yyVals[-1+yyTop], GetLocation (yyVals[0+yyTop]));
		  }

void case_20()
///  #line 478 "cs-parser.jay"
{
		Error_SyntaxError (yyToken);
		yyVal = null;
	  }

void case_21()
///  #line 486 "cs-parser.jay"
{
		current_namespace.AddUsing ((MemberName) yyVals[-1+yyTop], GetLocation (yyVals[-2+yyTop]));
		ubag.AddUsing (GetLocation (yyVals[-2+yyTop]), (MemberName) yyVals[-1+yyTop], GetLocation (yyVals[0+yyTop]));
	  }

void case_22()
///  #line 499 "cs-parser.jay"
{
		Attributes attrs = (Attributes) yyVals[-2+yyTop];
		MemberName name = (MemberName) yyVals[0+yyTop];
		if (attrs != null) {
			bool valid_global_attrs = true;
			if ((current_namespace.DeclarationFound || current_namespace != file.NamespaceContainer)) {
				valid_global_attrs = false;
			} else {
				foreach (var a in attrs.Attrs) {
					if (a.ExplicitTarget == "assembly" || a.ExplicitTarget == "module")
						continue;
						
					valid_global_attrs = false;
					break;
				}
			}
			
			if (!valid_global_attrs)
				report.Error (1671, name.Location, "A namespace declaration cannot have modifiers or attributes");
		}
	
		module.AddAttributes (attrs, current_namespace);
		
		current_namespace = new NamespaceContainer (name, module, current_namespace, file);
		module.AddTypesContainer (current_namespace);
		current_class = current_namespace.SlaveDeclSpace;
		current_container = current_class.PartialContainer;
		
		ubag.DeclareNamespace (GetLocation (yyVals[-1+yyTop]), name);
	  }

void case_23()
///  #line 530 "cs-parser.jay"
{
		if (doc_support)
			Lexer.doc_state = XmlCommentState.Allowed;
		ubag.OpenNamespace (GetLocation (yyVals[0+yyTop]));
	  }

void case_24()
///  #line 536 "cs-parser.jay"
{
		current_namespace = current_namespace.Parent;
		current_class = current_namespace.SlaveDeclSpace;
		current_container = current_class.PartialContainer;
		ubag.CloseNamespace (GetLocation (yyVals[-1+yyTop]));
		if (yyVals[0+yyTop] != null)
			ubag.EndNamespace (GetLocation (yyVals[0+yyTop]));
		else
			ubag.EndNamespace ();
	  }

void case_25()
///  #line 550 "cs-parser.jay"
{
		var lt = (Tokenizer.LocatedToken) yyVals[0+yyTop];
		yyVal = new MemberName (lt.Value, lt.Location);
	  }

void case_26()
///  #line 555 "cs-parser.jay"
{
		var lt = (Tokenizer.LocatedToken) yyVals[0+yyTop];
		yyVal = new MemberName ((MemberName) yyVals[-2+yyTop], lt.Value, lt.Location) {
			DotLocation = GetLocation (yyVals[-1+yyTop])
		};
	  }

void case_27()
///  #line 562 "cs-parser.jay"
{
		Error_SyntaxError (yyToken);
		yyVal = new MemberName ("<invalid>", lexer.Location);
	  }

void case_32()
///  #line 580 "cs-parser.jay"
{
		MemberName name = (MemberName) yyVals[0+yyTop];

		if (name.TypeArguments != null)
			syntax_error (lexer.Location, "namespace name expected");

		yyVal = name;
	  }

void case_41()
///  #line 612 "cs-parser.jay"
{
		if (yyVals[0+yyTop] != null) {
			TypeContainer ds = (TypeContainer)yyVals[0+yyTop];

			if ((ds.ModFlags & (Modifiers.PRIVATE | Modifiers.PROTECTED)) != 0){
				report.Error (1527, ds.Location, 
				"Namespace elements cannot be explicitly declared as private, protected or protected internal");
			}

			/* Here is a trick, for explicit attributes we don't know where they belong to until*/
			/* we parse succeeding declaration hence we parse them as normal and re-attach them*/
			/* when we know whether they are global (assembly:, module:) or local (type:).*/
			if (ds.OptAttributes != null) {
				ds.OptAttributes.ConvertGlobalAttributes (ds, current_namespace, !current_namespace.DeclarationFound && current_namespace == file.NamespaceContainer);
			}
		}
		current_namespace.DeclarationFound = true;
	  }

void case_50()
///  #line 662 "cs-parser.jay"
{
		var sect = (List<Attribute>) yyVals[0+yyTop];
		yyVal = new Attributes (sect);
		if (locationListStack.Count > 0)
			lbag.AddLocation (sect, locationListStack.Pop ());
		if (attributeCommas.Count > 0) {
			lbag.AppendTo (sect, attributeCommas);
			attributeCommas.Clear ();
		}
	  }

void case_51()
///  #line 673 "cs-parser.jay"
{
		Attributes attrs = yyVals[-1+yyTop] as Attributes;
		var sect = (List<Attribute>) yyVals[0+yyTop];
		
		if (locationListStack.Count > 0)
			lbag.AddLocation (sect, locationListStack.Pop ());
		if (attrs == null)
			attrs = new Attributes (sect);
		else
			attrs.AddAttributes (sect);
		yyVal = attrs;
	  }

void case_52()
///  #line 689 "cs-parser.jay"
{
		lexer.parsing_attribute_section = true;
		savedOpenLocation = GetLocation (yyVals[0+yyTop]);
	  }

void case_53()
///  #line 694 "cs-parser.jay"
{
		lexer.parsing_attribute_section = false;
		yyVal = yyVals[0+yyTop];
	  }

void case_54()
///  #line 702 "cs-parser.jay"
{
		current_attr_target = (string) yyVals[-1+yyTop];
		if (current_attr_target == "assembly" || current_attr_target == "module") {
			Lexer.check_incorrect_doc_comment ();
		}
	  }

void case_55()
///  #line 709 "cs-parser.jay"
{
		/* when attribute target is invalid*/
		if (current_attr_target == string.Empty)
			yyVal = new List<Attribute> (0);
		else
			yyVal = yyVals[-2+yyTop];
	  
		current_attr_target = null;
		lexer.parsing_attribute_section = false;
		if (yyVals[-1+yyTop] != null) {
			locationListStack.Push (new List<Location>(new [] { savedOpenLocation, savedCloseLocation, GetLocation (yyVals[-4+yyTop]), GetLocation (yyVals[-1+yyTop]), GetLocation (yyVals[0+yyTop]) }));
		} else {
			locationListStack.Push (new List<Location>(new [] { savedOpenLocation, savedCloseLocation, GetLocation (yyVals[-4+yyTop]), GetLocation (yyVals[0+yyTop]) }));
		}
 	  }

void case_56()
///  #line 725 "cs-parser.jay"
{
		yyVal = yyVals[-2+yyTop];
		if (yyVals[-1+yyTop] != null) {
			locationListStack.Push (new List<Location>(new [] { savedOpenLocation, GetLocation (yyVals[-1+yyTop]), GetLocation (yyVals[0+yyTop]) }));
		} else {
			locationListStack.Push (new List<Location>(new [] { savedOpenLocation, GetLocation (yyVals[0+yyTop]) }));
		}
	  }

void case_57()
///  #line 737 "cs-parser.jay"
{
		var lt = (Tokenizer.LocatedToken) yyVals[0+yyTop];
		yyVal = CheckAttributeTarget (lt.Value, lt.Location);
		savedCloseLocation = GetLocation (yyVals[0+yyTop]);
	  }

void case_60()
///  #line 745 "cs-parser.jay"
{
		if (yyToken == Token.IDENTIFIER) {
			Error_SyntaxError (yyToken);
			yyVal = null;
		} else {
			string name = GetTokenName (yyToken);
			yyVal = CheckAttributeTarget (name, GetLocation (yyVals[0+yyTop]));
		}
	  }

void case_62()
///  #line 762 "cs-parser.jay"
{
		var attrs = (List<Attribute>) yyVals[-2+yyTop];
		attrs.Add ((Attribute) yyVals[0+yyTop]);
		attributeCommas.Add (GetLocation (yyVals[-1+yyTop]));

		yyVal = attrs;
	  }

void case_64()
///  #line 777 "cs-parser.jay"
{
		--lexer.parsing_block;
		MemberName mname = (MemberName) yyVals[-2+yyTop];
		if (mname.IsGeneric) {
			report.Error (404, lexer.Location,
				      "'<' unexpected: attributes cannot be generic");
		}

		Arguments [] arguments = (Arguments []) yyVals[0+yyTop];
		ATypeNameExpression expr = mname.GetTypeExpression ();
		yyVal = new Attribute (current_attr_target, expr, arguments, mname.Location, lexer.IsEscapedIdentifier (mname));
		if (arguments != null) {
			attributeArgumentCommas.Insert (0, savedAttrParenOpenLocation);
			attributeArgumentCommas.Add (savedAttrParenCloseLocation);
			lbag.AddLocation (yyVal, attributeArgumentCommas);
			attributeArgumentCommas.Clear ();
		}
	  }

void case_67()
///  #line 804 "cs-parser.jay"
{
		savedAttrParenOpenLocation = GetLocation (yyVals[-2+yyTop]);
		savedAttrParenCloseLocation = GetLocation (yyVals[0+yyTop]);
		yyVal = yyVals[-1+yyTop];
	  }

void case_69()
///  #line 815 "cs-parser.jay"
{
	  	Arguments a = new Arguments (4);
		a.Add ((Argument) yyVals[0+yyTop]);
		yyVal = new Arguments [] { a, null };
	  }

void case_70()
///  #line 821 "cs-parser.jay"
{
	  	Arguments a = new Arguments (4);
		a.Add ((Argument) yyVals[0+yyTop]);  
		yyVal = new Arguments [] { null, a };
	  }

void case_71()
///  #line 827 "cs-parser.jay"
{
		Arguments[] o = (Arguments[]) yyVals[-2+yyTop];
		if (o [1] != null) {
			report.Error (1016, ((Argument) yyVals[0+yyTop]).Expr.Location, "Named attribute arguments must appear after the positional arguments");
			o [0] = new Arguments (4);
		}
		
		Arguments args = ((Arguments) o [0]);
		if (args.Count > 0 && !(yyVals[0+yyTop] is NamedArgument) && args [args.Count - 1] is NamedArgument)
			Error_NamedArgumentExpected ((NamedArgument) args [args.Count - 1]);
		
		args.Add ((Argument) yyVals[0+yyTop]);
		attributeArgumentCommas.Add (GetLocation (yyVals[-1+yyTop]));
	  }

void case_72()
///  #line 842 "cs-parser.jay"
{
		Arguments[] o = (Arguments[]) yyVals[-2+yyTop];
		if (o [1] == null) {
			o [1] = new Arguments (4);
		}

		((Arguments) o [1]).Add ((Argument) yyVals[0+yyTop]);
		attributeArgumentCommas.Add (GetLocation (yyVals[-1+yyTop]));
	  }

void case_76()
///  #line 867 "cs-parser.jay"
{
	  	--lexer.parsing_block;
	  	var lt = (Tokenizer.LocatedToken) yyVals[-3+yyTop];
		yyVal = new NamedArgument (lt.Value, lt.Location, (Expression) yyVals[0+yyTop]);	  
		lbag.AddLocation (yyVal, GetLocation(yyVals[-2+yyTop]));
	  }

void case_77()
///  #line 877 "cs-parser.jay"
{
		if (lang_version <= LanguageVersion.V_3)
			FeatureIsNotAvailable (GetLocation (yyVals[-3+yyTop]), "named argument");
			
		/* Avoid boxing in common case (no modifier)*/
		var arg_mod = yyVals[-1+yyTop] == null ? Argument.AType.None : (Argument.AType) yyVals[-1+yyTop];
			
		var lt = (Tokenizer.LocatedToken) yyVals[-3+yyTop];
		yyVal = new NamedArgument (lt.Value, lt.Location, (Expression) yyVals[0+yyTop], arg_mod);
		lbag.AddLocation (yyVal, GetLocation(yyVals[-2+yyTop]));
	  }

void case_95()
///  #line 930 "cs-parser.jay"
{
		report.Error (1519, lexer.Location, "Unexpected symbol `{0}' in class, struct, or interface member declaration",
			GetSymbolName (yyToken));
		yyVal = null;
		lexer.parsing_generic_declaration = false;
	  }

void case_97()
///  #line 947 "cs-parser.jay"
{ 
		MemberName name = MakeName ((MemberName) yyVals[0+yyTop]);
		push_current_class (new Struct (current_namespace, current_class, name, (Modifiers) yyVals[-4+yyTop], (Attributes) yyVals[-5+yyTop]), yyVals[-3+yyTop]);
		lbag.AddMember (current_class, GetModifierLocations (), GetLocation (yyVals[-2+yyTop]));
	  }

void case_98()
///  #line 954 "cs-parser.jay"
{
		lexer.ConstraintsParsing = false;

		current_class.SetParameterInfo ((List<Constraints>) yyVals[0+yyTop]);

		if (doc_support)
			current_container.DocComment = Lexer.consume_doc_comment ();

		
		lexer.parsing_modifiers = true;
	  }

void case_99()
///  #line 966 "cs-parser.jay"
{
		if (doc_support)
			Lexer.doc_state = XmlCommentState.Allowed;
	  }

void case_100()
///  #line 971 "cs-parser.jay"
{
		lbag.AppendToMember (current_class, GetLocation (yyVals[-3+yyTop]), GetLocation (yyVals[-1+yyTop]));
		--lexer.parsing_declaration;
		if (doc_support)
			Lexer.doc_state = XmlCommentState.Allowed;
	  }

void case_101()
///  #line 978 "cs-parser.jay"
{
		if (yyVals[-1+yyTop] != null) 
			current_class.OptionalSemicolon = GetLocation (yyVals[-1+yyTop]);
		yyVal = pop_current_class ();
	  }

void case_103()
///  #line 993 "cs-parser.jay"
{
		var lt = (Tokenizer.LocatedToken) yyVals[0+yyTop];
		var mod = (Modifiers) yyVals[-3+yyTop];
		current_field = new Const (current_class, (FullNamedExpression) yyVals[-1+yyTop], mod, new MemberName (lt.Value, lt.Location), (Attributes) yyVals[-4+yyTop]);
		current_container.AddConstant ((Const) current_field);
		
		if ((mod & Modifiers.STATIC) != 0) {
			report.Error (504, current_field.Location, "The constant `{0}' cannot be marked static", current_field.GetSignatureForError ());
		}
		
		yyVal = current_field;
	  }

void case_104()
///  #line 1006 "cs-parser.jay"
{
		if (doc_support) {
			current_field.DocComment = Lexer.consume_doc_comment ();
			Lexer.doc_state = XmlCommentState.Allowed;
		}
		
		current_field.Initializer = (ConstInitializer) yyVals[-2+yyTop];
		lbag.AddMember (current_field, GetModifierLocations (), GetLocation (yyVals[-6+yyTop]), GetLocation (yyVals[0+yyTop]));
		current_field = null;
	  }

void case_109()
///  #line 1036 "cs-parser.jay"
{
		var lt = (Tokenizer.LocatedToken) yyVals[-1+yyTop];
	  	yyVal = new FieldDeclarator (new SimpleMemberName (lt.Value, lt.Location), (ConstInitializer) yyVals[0+yyTop]);
	  	lbag.AddLocation (yyVal, GetLocation (yyVals[-2+yyTop]));
	  }

void case_111()
///  #line 1049 "cs-parser.jay"
{
		--lexer.parsing_block;
		yyVal = new ConstInitializer (current_field, (Expression) yyVals[0+yyTop], GetLocation (yyVals[-2+yyTop]));
	  	lbag.AddLocation (yyVal, GetLocation (yyVals[-2+yyTop]));
	  }

void case_112()
///  #line 1055 "cs-parser.jay"
{
		report.Error (145, lexer.Location, "A const field requires a value to be provided");
		yyVal = null;
	  }

void case_115()
///  #line 1070 "cs-parser.jay"
{
	  	lexer.parsing_generic_declaration = false;

		FullNamedExpression type = (FullNamedExpression) yyVals[-1+yyTop];
		if (type.Type != null && type.Type.Kind == MemberKind.Void)
			report.Error (670, GetLocation (yyVals[-1+yyTop]), "Fields cannot have void type");
			
		var lt = (Tokenizer.LocatedToken) yyVals[0+yyTop];
		current_field = new Field (current_class, type, (Modifiers) yyVals[-2+yyTop], new MemberName (lt.Value, lt.Location), (Attributes) yyVals[-3+yyTop]);
		current_container.AddField (current_field);
		yyVal = current_field;
	  }

void case_116()
///  #line 1085 "cs-parser.jay"
{ 
		if (doc_support) {
			current_field.DocComment = Lexer.consume_doc_comment ();
			Lexer.doc_state = XmlCommentState.Allowed;
		}
			
		lbag.AddMember (current_field, GetModifierLocations (), GetLocation (yyVals[0+yyTop]));
		yyVal = current_field;
		current_field = null;
	  }

void case_117()
///  #line 1098 "cs-parser.jay"
{ 
		if (lang_version < LanguageVersion.ISO_2)
			FeatureIsNotAvailable (GetLocation (yyVals[-2+yyTop]), "fixed size buffers");

		var lt = (Tokenizer.LocatedToken) yyVals[0+yyTop];
		current_field = new FixedField (current_class, (FullNamedExpression) yyVals[-1+yyTop], (Modifiers) yyVals[-3+yyTop],
			new MemberName (lt.Value, lt.Location), (Attributes) yyVals[-4+yyTop]);
			
		current_container.AddField (current_field);
	  }

void case_118()
///  #line 1109 "cs-parser.jay"
{
		if (doc_support) {
			current_field.DocComment = Lexer.consume_doc_comment ();
			Lexer.doc_state = XmlCommentState.Allowed;
	    }

		current_field.Initializer = (ConstInitializer) yyVals[-2+yyTop];	    
		lbag.AddMember (current_field, GetModifierLocations (), GetLocation (yyVals[-6+yyTop]), GetLocation (yyVals[0+yyTop]));
		yyVal = current_field;
	    current_field = null;
	  }

void case_121()
///  #line 1132 "cs-parser.jay"
{
	  	++lexer.parsing_block;
		current_local_parameters = ParametersCompiled.EmptyReadOnlyParameters;
	  	start_block (GetLocation (yyVals[0+yyTop]));
	  }

void case_122()
///  #line 1138 "cs-parser.jay"
{
	  	--lexer.parsing_block;
		current_field.Initializer = (Expression) yyVals[0+yyTop];
		lbag.AppendToMember (current_field, GetLocation (yyVals[-2+yyTop]));
		end_block (lexer.Location);
		current_local_parameters = null;
	  }

void case_127()
///  #line 1165 "cs-parser.jay"
{
		var lt = (Tokenizer.LocatedToken) yyVals[0+yyTop];
	  	yyVal = new FieldDeclarator (new SimpleMemberName (lt.Value, lt.Location), null);
	  	lbag.AddLocation (yyVal, GetLocation (yyVals[-1+yyTop]));
	  }

void case_129()
///  #line 1175 "cs-parser.jay"
{
		--lexer.parsing_block;
		var lt = (Tokenizer.LocatedToken) yyVals[-3+yyTop];	  
	  	yyVal = new FieldDeclarator (new SimpleMemberName (lt.Value, lt.Location), (Expression) yyVals[0+yyTop]);
	  	lbag.AddLocation (yyVal, GetLocation (yyVals[-4+yyTop]), GetLocation (yyVals[-2+yyTop]));
	  }

void case_134()
///  #line 1201 "cs-parser.jay"
{
		var lt = (Tokenizer.LocatedToken) yyVals[-1+yyTop];	  
		yyVal = new FieldDeclarator (new SimpleMemberName (lt.Value, lt.Location), (ConstInitializer) yyVals[0+yyTop]);
		lbag.AddLocation (yyVal, GetLocation (yyVals[-2+yyTop]));
	  }

void case_136()
///  #line 1214 "cs-parser.jay"
{
		--lexer.parsing_block;
		yyVal = new ConstInitializer (current_field, (Expression) yyVals[-1+yyTop], GetLocation (yyVals[-3+yyTop]));
		lbag.AddLocation (yyVal, GetLocation (yyVals[0+yyTop]));
	  }

void case_137()
///  #line 1220 "cs-parser.jay"
{
		report.Error (443, lexer.Location, "Value or constant expected");
		yyVal = null;
	  }

void case_140()
///  #line 1230 "cs-parser.jay"
{
		/* It has to be here for the parent to safely restore artificial block*/
	  	Error_SyntaxError (yyToken);
	  	yyVal = null;
	  }

void case_141()
///  #line 1239 "cs-parser.jay"
{
		if (doc_support)
			Lexer.doc_state = XmlCommentState.NotAllowed;

		/* Add it early in the case of body being eof for full ast*/
		Method m = (Method) yyVals[0+yyTop];
		async_block = (m.ModFlags & Modifiers.ASYNC) != 0;
		current_container.AddMethod (m);
	  }

void case_142()
///  #line 1249 "cs-parser.jay"
{
		Method method = (Method) yyVals[-2+yyTop];
		method.Block = (ToplevelBlock) yyVals[0+yyTop];
		async_block = false;
		
		if (method.Block == null) {
			lbag.AppendToMember (method, savedLocation); /* semicolon*/
			method.ParameterInfo.CheckParameters (method);

			if ((method.ModFlags & Modifiers.ASYNC) != 0) {
				report.Error (1994, method.Location, "`{0}': The async modifier can only be used with methods that have a body",
					method.GetSignatureForError ());
			}
		} else {
			if (current_container.Kind == MemberKind.Interface) {
				report.Error (531, method.Location, "`{0}': interface members cannot have a definition",
					method.GetSignatureForError ());
			}
		}

		current_local_parameters = null;

		if (doc_support)
			Lexer.doc_state = XmlCommentState.Allowed;
	  }

void case_145()
///  #line 1289 "cs-parser.jay"
{
		lexer.ConstraintsParsing = false;
		valid_param_mod = 0;
		MemberName name = (MemberName) yyVals[-6+yyTop];
		current_local_parameters = (ParametersCompiled) yyVals[-3+yyTop];

		GenericMethod generic = null;
		if (name.TypeArguments != null) {
			generic = new GenericMethod (current_namespace, current_class, name,
						     (FullNamedExpression) yyVals[-7+yyTop], current_local_parameters);

			generic.SetParameterInfo ((List<Constraints>) yyVals[0+yyTop]);
		} else if (yyVals[0+yyTop] != null) {
			report.Error (80, GetLocation (yyVals[0+yyTop]),
				"Constraints are not allowed on non-generic declarations");
		}

		var method = Method.Create (current_class, generic, (FullNamedExpression) yyVals[-7+yyTop], (Modifiers) yyVals[-8+yyTop],
				     name, current_local_parameters, (Attributes) yyVals[-9+yyTop], yyVals[0+yyTop] != null);
				     
		if (doc_support)
			method.DocComment = Lexer.consume_doc_comment ();

		lbag.AddMember (method, GetModifierLocations (), GetLocation (yyVals[-5+yyTop]), GetLocation (yyVals[-2+yyTop]));
		yyVal = method;
	  }

void case_147()
///  #line 1324 "cs-parser.jay"
{
		lexer.parsing_generic_declaration = false;
	  	valid_param_mod = ParameterModifierType.All;
	  }

void case_149()
///  #line 1333 "cs-parser.jay"
{
		lexer.ConstraintsParsing = false;
		valid_param_mod = 0;

		MemberName name = (MemberName) yyVals[-6+yyTop];
		current_local_parameters = (ParametersCompiled) yyVals[-3+yyTop];

		if (yyVals[-1+yyTop] != null && name.TypeArguments == null)
			report.Error (80, lexer.Location,
				      "Constraints are not allowed on non-generic declarations");

		GenericMethod generic = null;
		if (name.TypeArguments != null) {
			generic = new GenericMethod (current_namespace, current_class, name,
				new TypeExpression (compiler.BuiltinTypes.Void, GetLocation (yyVals[-8+yyTop])),
				current_local_parameters);

			generic.SetParameterInfo ((List<Constraints>) yyVals[-1+yyTop]);
		}

		var modifiers = (Modifiers) yyVals[-10+yyTop];
		modifiers |= Modifiers.PARTIAL;

		var method = Method.Create (current_class, generic, new TypeExpression (compiler.BuiltinTypes.Void, GetLocation (yyVals[-8+yyTop])),
				     modifiers, name, current_local_parameters, (Attributes) yyVals[-11+yyTop], yyVals[-1+yyTop] != null);

		if (doc_support)
			method.DocComment = Lexer.consume_doc_comment ();

		StoreModifierLocation (Modifiers.PARTIAL, GetLocation (yyVals[-9+yyTop]));
		lbag.AddMember (method, mod_locations, GetLocation (yyVals[-5+yyTop]), GetLocation (yyVals[-2+yyTop]));
		yyVal = method;
	  }

void case_150()
///  #line 1370 "cs-parser.jay"
{
		MemberName name = (MemberName) yyVals[-3+yyTop];
		report.Error (1585, name.Location, 
			"Member modifier `{0}' must precede the member type and name", ModifiersExtensions.Name ((Modifiers) yyVals[-4+yyTop]));

		var method = Method.Create (current_class, null, (FullNamedExpression) yyVals[-5+yyTop],
					    0, name, (ParametersCompiled) yyVals[-1+yyTop], (Attributes) yyVals[-7+yyTop], false);

		current_local_parameters = (ParametersCompiled) yyVals[-1+yyTop];

		if (doc_support)
			method.DocComment = Lexer.consume_doc_comment ();

		yyVal = method;
	  }

void case_155()
///  #line 1399 "cs-parser.jay"
{
		var pars_list = (List<Parameter>) yyVals[0+yyTop];
	  	yyVal = new ParametersCompiled (pars_list.ToArray ());
	  	lbag.AddLocation (yyVal, parameterListCommas);
	  }

void case_156()
///  #line 1405 "cs-parser.jay"
{
		var pars_list = (List<Parameter>) yyVals[-2+yyTop];
		pars_list.Add ((Parameter) yyVals[0+yyTop]);
		parameterListCommas.Add (GetLocation (yyVals[-1+yyTop]));
		
		yyVal = new ParametersCompiled (pars_list.ToArray ()); 
	  	lbag.AddLocation (yyVal, parameterListCommas);
	  }

void case_157()
///  #line 1414 "cs-parser.jay"
{
		var pars_list = (List<Parameter>) yyVals[-2+yyTop];
		pars_list.Add (new ArglistParameter (GetLocation (yyVals[0+yyTop])));
		parameterListCommas.Add (GetLocation (yyVals[-1+yyTop]));
		
		yyVal = new ParametersCompiled (pars_list.ToArray (), true);
	  	lbag.AddLocation (yyVal, parameterListCommas);
	  }

void case_158()
///  #line 1423 "cs-parser.jay"
{
		if (yyVals[-2+yyTop] != null)
			report.Error (231, ((Parameter) yyVals[-2+yyTop]).Location, "A params parameter must be the last parameter in a formal parameter list");

		yyVal = new ParametersCompiled (new Parameter[] { (Parameter) yyVals[-2+yyTop] } );			
	  	lbag.AddLocation (yyVal, parameterListCommas);
	  }

void case_159()
///  #line 1431 "cs-parser.jay"
{
		if (yyVals[-2+yyTop] != null)
			report.Error (231, ((Parameter) yyVals[-2+yyTop]).Location, "A params parameter must be the last parameter in a formal parameter list");

		var pars_list = (List<Parameter>) yyVals[-4+yyTop];
		pars_list.Add (new ArglistParameter (GetLocation (yyVals[-2+yyTop])));
		parameterListCommas.Add (GetLocation (yyVals[-3+yyTop]));
		parameterListCommas.Add (GetLocation (yyVals[-1+yyTop]));
		
		yyVal = new ParametersCompiled (pars_list.ToArray (), true);
	  	lbag.AddLocation (yyVal, parameterListCommas);
	  }

void case_160()
///  #line 1444 "cs-parser.jay"
{
		report.Error (257, GetLocation (yyVals[-2+yyTop]), "An __arglist parameter must be the last parameter in a formal parameter list");

		yyVal = new ParametersCompiled (new Parameter [] { new ArglistParameter (GetLocation (yyVals[-2+yyTop])) }, true);
	  	lbag.AddLocation (yyVal, parameterListCommas);
	  }

void case_161()
///  #line 1451 "cs-parser.jay"
{
		report.Error (257, GetLocation (yyVals[-2+yyTop]), "An __arglist parameter must be the last parameter in a formal parameter list");

		var pars_list = (List<Parameter>) yyVals[-4+yyTop];
		pars_list.Add (new ArglistParameter (GetLocation (yyVals[-2+yyTop])));
		parameterListCommas.Add (GetLocation (yyVals[-3+yyTop]));
		parameterListCommas.Add (GetLocation (yyVals[-1+yyTop]));

		yyVal = new ParametersCompiled (pars_list.ToArray (), true);
	  	lbag.AddLocation (yyVal, parameterListCommas);
	  }

void case_164()
///  #line 1471 "cs-parser.jay"
{
		Error_SyntaxError (yyToken);
		yyVal = ParametersCompiled.EmptyReadOnlyParameters;
	  }

void case_165()
///  #line 1479 "cs-parser.jay"
{
		parameters_bucket.Clear ();
		Parameter p = (Parameter) yyVals[0+yyTop];
		parameters_bucket.Add (p);
		parameterListCommas.Clear ();
		default_parameter_used = p.HasDefaultValue;
		yyVal = parameters_bucket;
	  }

void case_166()
///  #line 1488 "cs-parser.jay"
{
		var pars = (List<Parameter>) yyVals[-2+yyTop];
		Parameter p = (Parameter) yyVals[0+yyTop];
		if (p != null) {
			if (p.HasExtensionMethodModifier)
				report.Error (1100, p.Location, "The parameter modifier `this' can only be used on the first parameter");
			else if (!p.HasDefaultValue && default_parameter_used)
				report.Error (1737, p.Location, "Optional parameter cannot precede required parameters");

			default_parameter_used |= p.HasDefaultValue;
			pars.Add (p);
			
			parameterListCommas.Add (GetLocation (yyVals[-1+yyTop]));
		}
		
		yyVal = yyVals[-2+yyTop];
	  }

void case_167()
///  #line 1512 "cs-parser.jay"
{
		var lt = (Tokenizer.LocatedToken) yyVals[0+yyTop];
		yyVal = new Parameter ((FullNamedExpression) yyVals[-1+yyTop], lt.Value, (Parameter.Modifier) yyVals[-2+yyTop], (Attributes) yyVals[-3+yyTop], lt.Location);
		lbag.AddLocation (yyVal, parameterModifierLocation);
	  }

void case_168()
///  #line 1521 "cs-parser.jay"
{
		var lt = (Tokenizer.LocatedToken) yyVals[-2+yyTop];
		report.Error (1552, lt.Location, "Array type specifier, [], must appear before parameter name");
		yyVal = new Parameter ((FullNamedExpression) yyVals[-3+yyTop], lt.Value, (Parameter.Modifier) yyVals[-4+yyTop], (Attributes) yyVals[-5+yyTop], lt.Location);
		lbag.AddLocation (yyVal, parameterModifierLocation);
	  }

void case_169()
///  #line 1531 "cs-parser.jay"
{
		Error_SyntaxError (yyToken);	  
	  	Location l = GetLocation (yyVals[0+yyTop]);
		yyVal = new Parameter ((FullNamedExpression) yyVals[-1+yyTop], null, (Parameter.Modifier) yyVals[-2+yyTop], (Attributes) yyVals[-3+yyTop], l);
		lbag.AddLocation (yyVal, parameterModifierLocation);
	  }

void case_171()
///  #line 1546 "cs-parser.jay"
{
	  	--lexer.parsing_block;
		if (lang_version <= LanguageVersion.V_3) {
			FeatureIsNotAvailable (GetLocation (yyVals[-2+yyTop]), "optional parameter");
		}
		
		Parameter.Modifier mod = (Parameter.Modifier) yyVals[-5+yyTop];
		if (mod != Parameter.Modifier.NONE) {
			switch (mod) {
			case Parameter.Modifier.REF:
			case Parameter.Modifier.OUT:
				report.Error (1741, GetLocation (yyVals[-5+yyTop]), "Cannot specify a default value for the `{0}' parameter",
					Parameter.GetModifierSignature (mod));
				break;
				
			case Parameter.Modifier.This:
				report.Error (1743, GetLocation (yyVals[-5+yyTop]), "Cannot specify a default value for the `{0}' parameter",
					Parameter.GetModifierSignature (mod));
				break;
			default:
				throw new NotImplementedException (mod.ToString ());
			}
				
			mod = Parameter.Modifier.NONE;
		}
		
		if ((valid_param_mod & ParameterModifierType.DefaultValue) == 0)
			report.Error (1065, GetLocation (yyVals[-2+yyTop]), "Optional parameter is not valid in this context");
		
		var lt = (Tokenizer.LocatedToken) yyVals[-3+yyTop];
		yyVal = new Parameter ((FullNamedExpression) yyVals[-4+yyTop], lt.Value, mod, (Attributes) yyVals[-6+yyTop], lt.Location);
		lbag.AddLocation (yyVal, parameterModifierLocation, GetLocation (yyVals[-2+yyTop])); /* parameterModifierLocation should be ignored when mod == NONE*/
		
		if (yyVals[0+yyTop] != null)
			((Parameter) yyVal).DefaultValue = new DefaultParameterValueExpression ((Expression) yyVals[0+yyTop]);
	  }

void case_175()
///  #line 1595 "cs-parser.jay"
{
		Parameter.Modifier p2 = (Parameter.Modifier)yyVals[0+yyTop];
  		Parameter.Modifier mod = (Parameter.Modifier)yyVals[-1+yyTop] | p2;
  		if (((Parameter.Modifier)yyVals[-1+yyTop] & p2) == p2) {
  			Error_DuplicateParameterModifier (lexer.Location, p2);
  		} else {
	  		switch (mod & ~Parameter.Modifier.This) {
  				case Parameter.Modifier.REF:
					report.Error (1101, lexer.Location, "The parameter modifiers `this' and `ref' cannot be used altogether");
  					break;
   				case Parameter.Modifier.OUT:
					report.Error (1102, lexer.Location, "The parameter modifiers `this' and `out' cannot be used altogether");
  					break;
  				default:
 					report.Error (1108, lexer.Location, "A parameter cannot have specified more than one modifier");
 					break;
 			}
  		}
  		yyVal = mod;
	  }

void case_176()
///  #line 1619 "cs-parser.jay"
{
	  	if ((valid_param_mod & ParameterModifierType.Ref) == 0)
	  		Error_ParameterModifierNotValid ("ref", GetLocation (yyVals[0+yyTop]));
		parameterModifierLocation = GetLocation (yyVals[0+yyTop]);
	  	yyVal = Parameter.Modifier.REF;
	  }

void case_177()
///  #line 1626 "cs-parser.jay"
{
	  	if ((valid_param_mod & ParameterModifierType.Out) == 0)
	  		Error_ParameterModifierNotValid ("out", GetLocation (yyVals[0+yyTop]));
		parameterModifierLocation = GetLocation (yyVals[0+yyTop]);
	  	yyVal = Parameter.Modifier.OUT;
	  }

void case_178()
///  #line 1633 "cs-parser.jay"
{
		if ((valid_param_mod & ParameterModifierType.This) == 0)
	  		Error_ParameterModifierNotValid ("this", GetLocation (yyVals[0+yyTop]));

	  	if (lang_version <= LanguageVersion.ISO_2)
	  		FeatureIsNotAvailable (GetLocation (yyVals[0+yyTop]), "extension methods");
		parameterModifierLocation = GetLocation (yyVals[0+yyTop]);
		yyVal = Parameter.Modifier.This;
	  }

void case_179()
///  #line 1646 "cs-parser.jay"
{
		var lt = (Tokenizer.LocatedToken) yyVals[0+yyTop];
		yyVal = new ParamsParameter ((FullNamedExpression) yyVals[-1+yyTop], lt.Value, (Attributes) yyVals[-3+yyTop], lt.Location);
		lbag.AddLocation (yyVal, savedLocation);
	  }

void case_180()
///  #line 1652 "cs-parser.jay"
{
		report.Error (1751, GetLocation (yyVals[-4+yyTop]), "Cannot specify a default value for a parameter array");
		
		var lt = (Tokenizer.LocatedToken) yyVals[-2+yyTop];
		yyVal = new ParamsParameter ((FullNamedExpression) yyVals[-3+yyTop], lt.Value, (Attributes) yyVals[-5+yyTop], lt.Location);		
		lbag.AddLocation (yyVal, savedLocation);
	  }

void case_181()
///  #line 1660 "cs-parser.jay"
{
		Error_SyntaxError (yyToken);
		yyVal = null;
	  }

void case_182()
///  #line 1668 "cs-parser.jay"
{
		if ((valid_param_mod & ParameterModifierType.Params) == 0)
			report.Error (1670, (GetLocation (yyVals[0+yyTop])), "The `params' modifier is not allowed in current context");
		savedLocation = GetLocation (yyVals[0+yyTop]);
	  }

void case_183()
///  #line 1674 "cs-parser.jay"
{
		Parameter.Modifier mod = (Parameter.Modifier)yyVals[0+yyTop];
		if ((mod & Parameter.Modifier.This) != 0) {
			report.Error (1104, GetLocation (yyVals[-1+yyTop]), "The parameter modifiers `this' and `params' cannot be used altogether");
		} else {
			report.Error (1611, GetLocation (yyVals[-1+yyTop]), "The params parameter cannot be declared as ref or out");
		}	  
		savedLocation = GetLocation (yyVals[-1+yyTop]);
	  }

void case_185()
///  #line 1691 "cs-parser.jay"
{
	  	if ((valid_param_mod & ParameterModifierType.Arglist) == 0)
	  		report.Error (1669, GetLocation (yyVals[0+yyTop]), "__arglist is not valid in this context");
	  }

void case_186()
///  #line 1702 "cs-parser.jay"
{
		if (doc_support)
			tmpComment = Lexer.consume_doc_comment ();
	  }

void case_187()
///  #line 1707 "cs-parser.jay"
{
		var type = (FullNamedExpression) yyVals[-3+yyTop];
		current_property = new Property (current_class, type, (Modifiers) yyVals[-4+yyTop],
			(MemberName) yyVals[-2+yyTop], (Attributes) yyVals[-5+yyTop]);
			
		if (type.Type != null && type.Type.Kind == MemberKind.Void)
			report.Error (547, GetLocation (yyVals[-3+yyTop]), "`{0}': property or indexer cannot have void type", current_property.GetSignatureForError ());					
			
		current_container.AddProperty ((Property)current_property);
		lbag.AddMember (current_property, GetModifierLocations (), GetLocation (yyVals[0+yyTop]));
		
		lexer.PropertyParsing = true;
	  }

void case_188()
///  #line 1721 "cs-parser.jay"
{
		lexer.PropertyParsing = false;
		
		if (doc_support)
			current_property.DocComment = ConsumeStoredComment ();				
	  }

void case_189()
///  #line 1728 "cs-parser.jay"
{
		lbag.AppendToMember (current_property, GetLocation (yyVals[0+yyTop]));
		current_property = null;
	  }

void case_191()
///  #line 1742 "cs-parser.jay"
{
		valid_param_mod = 0;
		var type = (FullNamedExpression) yyVals[-6+yyTop];
		Indexer indexer = new Indexer (current_class, type, (MemberName) yyVals[-5+yyTop], (Modifiers) yyVals[-7+yyTop], (ParametersCompiled) yyVals[-2+yyTop], (Attributes) yyVals[-8+yyTop]);
			
		current_property = indexer;

  		current_container.AddIndexer (indexer);
		lbag.AddMember (current_property, GetModifierLocations (), GetLocation (yyVals[-4+yyTop]), GetLocation (yyVals[-1+yyTop]), GetLocation (yyVals[0+yyTop]));
  		
		if (type.Type != null && type.Type.Kind == MemberKind.Void)
			report.Error (620, GetLocation (yyVals[-6+yyTop]), "`{0}': indexer return type cannot be `void'", indexer.GetSignatureForError ());  		

		if (indexer.ParameterInfo.IsEmpty) {
			report.Error (1551, GetLocation (yyVals[-4+yyTop]), "Indexers must have at least one parameter");
		}

		if (doc_support) {
			tmpComment = Lexer.consume_doc_comment ();
			Lexer.doc_state = XmlCommentState.Allowed;
		}

		lexer.PropertyParsing = true;
	  }

void case_193()
///  #line 1771 "cs-parser.jay"
{
		if (current_property.AccessorFirst != null && current_property.AccessorFirst.Block == null)
			((Indexer) current_property).ParameterInfo.CheckParameters (current_property);
	  
		if (doc_support)
			current_property.DocComment = ConsumeStoredComment ();
			
		lbag.AppendToMember (current_property, GetLocation (yyVals[-1+yyTop]));
		current_property = null;		
	  }

void case_198()
///  #line 1790 "cs-parser.jay"
{
	  	if (yyToken == Token.CLOSE_BRACE) {
	  		report.Error (548, lexer.Location, "`{0}': property or indexer must have at least one accessor", current_property.GetSignatureForError ());
		} else {
			if (yyToken == Token.SEMICOLON)
				report.Error (1597, lexer.Location, "Semicolon after method or accessor block is not valid");
			else
				report.Error (1014, GetLocation (yyVals[0+yyTop]), "A get or set accessor expected");
		}
	  }

void case_199()
///  #line 1804 "cs-parser.jay"
{
		if (yyVals[-1+yyTop] != ModifierNone && lang_version == LanguageVersion.ISO_1) {
			FeatureIsNotAvailable (GetLocation (yyVals[-1+yyTop]), "access modifiers on properties");
		}
	  
		if (current_property.Get != null) {
			report.Error (1007, GetLocation (yyVals[0+yyTop]), "Property accessor already defined");
		}
		
		if (current_property is Indexer) {
			current_property.Get = new Indexer.GetIndexerMethod (current_property, (Modifiers) yyVals[-1+yyTop], ((Indexer)current_property).ParameterInfo.Clone (),
				(Attributes) yyVals[-2+yyTop], GetLocation (yyVals[0+yyTop]));
		} else {
			current_property.Get = new Property.GetMethod (current_property,
				(Modifiers) yyVals[-1+yyTop], (Attributes) yyVals[-2+yyTop], GetLocation (yyVals[0+yyTop]));
		}	
	  
		current_local_parameters = current_property.Get.ParameterInfo;	  
		lexer.PropertyParsing = false;
	  }

void case_200()
///  #line 1825 "cs-parser.jay"
{
	  	if (yyVals[0+yyTop] != null) {
	  		current_property.Get.Block = (ToplevelBlock) yyVals[0+yyTop];			
	  	
			if (current_container.Kind == MemberKind.Interface) {
				report.Error (531, current_property.Get.Block.StartLocation,
					"`{0}': interface members cannot have a definition", current_property.Get.GetSignatureForError ());
			}
			lbag.AddMember (current_property.Get, GetModifierLocations ());
		} else {
			lbag.AddMember (current_property.Get, GetModifierLocations (), savedLocation);
		}
	  
		current_local_parameters = null;
		lexer.PropertyParsing = true;

		if (doc_support)
			if (Lexer.doc_state == XmlCommentState.Error)
				Lexer.doc_state = XmlCommentState.NotAllowed;
	  }

void case_201()
///  #line 1849 "cs-parser.jay"
{
		if (yyVals[-1+yyTop] != ModifierNone && lang_version == LanguageVersion.ISO_1) {
			FeatureIsNotAvailable (GetLocation (yyVals[-1+yyTop]), "access modifiers on properties");
		}
		
		if (current_property.Set != null) {
			report.Error (1007, GetLocation (yyVals[0+yyTop]), "Property accessor already defined");
		}
	  
		if (current_property is Indexer) {
			current_property.Set = new Indexer.SetIndexerMethod (current_property, (Modifiers) yyVals[-1+yyTop],
				ParametersCompiled.MergeGenerated (compiler,
				((Indexer)current_property).ParameterInfo, true, new Parameter (
					current_property.TypeExpression, "value", Parameter.Modifier.NONE, null, GetLocation (yyVals[0+yyTop])),
					null),
				(Attributes) yyVals[-2+yyTop], GetLocation (yyVals[0+yyTop]));
		} else {
			current_property.Set = new Property.SetMethod (current_property, (Modifiers) yyVals[-1+yyTop], 
				ParametersCompiled.CreateImplicitParameter (current_property.TypeExpression, GetLocation (yyVals[0+yyTop])),
				(Attributes) yyVals[-2+yyTop], GetLocation (yyVals[0+yyTop]));
		}
		
		current_local_parameters = current_property.Set.ParameterInfo;	
		lexer.PropertyParsing = false;
	  }

void case_202()
///  #line 1875 "cs-parser.jay"
{
		if (yyVals[0+yyTop] != null) {		
			current_property.Set.Block = (ToplevelBlock) yyVals[0+yyTop];
		
			if (current_container.Kind == MemberKind.Interface) {
				report.Error (531, current_property.Set.Block.StartLocation,
					"`{0}': interface members cannot have a definition", current_property.Set.GetSignatureForError ());
			}
			lbag.AddMember (current_property.Set, GetModifierLocations ());
		} else {
			lbag.AddMember (current_property.Set, GetModifierLocations (), savedLocation);
		}
		
		current_local_parameters = null;
		lexer.PropertyParsing = true;

		if (doc_support
			&& Lexer.doc_state == XmlCommentState.Error)
			Lexer.doc_state = XmlCommentState.NotAllowed;
	  }

void case_204()
///  #line 1900 "cs-parser.jay"
{
		savedLocation = GetLocation (yyVals[0+yyTop]);
	  	yyVal = null;
	  }

void case_205()
///  #line 1905 "cs-parser.jay"
{
	  	Error_SyntaxError (1043, yyToken, "Invalid accessor body");
	  	yyVal = null;
	  }

void case_207()
///  #line 1920 "cs-parser.jay"
{
		MemberName name = MakeName ((MemberName) yyVals[0+yyTop]);
		push_current_class (new Interface (current_namespace, current_class, name, (Modifiers) yyVals[-4+yyTop], (Attributes) yyVals[-5+yyTop]), yyVals[-3+yyTop]);
		lbag.AddMember (current_class, GetModifierLocations (), GetLocation (yyVals[-2+yyTop]));		
	  }

void case_208()
///  #line 1927 "cs-parser.jay"
{
		lexer.ConstraintsParsing = false;

		current_class.SetParameterInfo ((List<Constraints>) yyVals[0+yyTop]);

		if (doc_support) {
			current_container.DocComment = Lexer.consume_doc_comment ();
			Lexer.doc_state = XmlCommentState.Allowed;
		}
		
		lexer.parsing_modifiers = true;
	  }

void case_209()
///  #line 1940 "cs-parser.jay"
{
		--lexer.parsing_declaration;	  
		if (doc_support)
			Lexer.doc_state = XmlCommentState.Allowed;
	  }

void case_210()
///  #line 1946 "cs-parser.jay"
{
		if (yyVals[0+yyTop] != null) 
			current_class.OptionalSemicolon = GetLocation (yyVals[0+yyTop]);
		lbag.AppendToMember (current_class, GetLocation (yyVals[-4+yyTop]), GetLocation (yyVals[-2+yyTop]));
		yyVal = pop_current_class ();
	  }

void case_226()
///  #line 2006 "cs-parser.jay"
{
		OperatorDeclaration decl = (OperatorDeclaration) yyVals[-2+yyTop];
		if (decl != null) {
			Operator op = new Operator (
				current_class, decl.optype, decl.ret_type, (Modifiers) yyVals[-3+yyTop], 
				current_local_parameters,
				(ToplevelBlock) yyVals[0+yyTop], (Attributes) yyVals[-4+yyTop], decl.location);
				
			if (op.Block == null)
				op.ParameterInfo.CheckParameters (op);

			if (doc_support) {
				op.DocComment = tmpComment;
				Lexer.doc_state = XmlCommentState.Allowed;
			}

			/* Note again, checking is done in semantic analysis*/
			current_container.AddOperator (op);

			lbag.AddMember (op, GetModifierLocations (), lbag.GetLocations (decl));
			if (yyVals[0+yyTop] == null) { /* Semicolon*/
				lbag.AppendTo (op, savedLocation); 
			}
		}
		
		current_local_parameters = null;
	  }

void case_230()
///  #line 2043 "cs-parser.jay"
{
		report.Error (590, GetLocation (yyVals[0+yyTop]), "User-defined operators cannot return void");
		yyVal = new TypeExpression (compiler.BuiltinTypes.Void, GetLocation (yyVals[0+yyTop]));
	  }

void case_232()
///  #line 2055 "cs-parser.jay"
{
		valid_param_mod = 0;

		Location loc = GetLocation (yyVals[-5+yyTop]);
		Operator.OpType op = (Operator.OpType) yyVals[-4+yyTop];
		current_local_parameters = (ParametersCompiled)yyVals[-1+yyTop];
		
		int p_count = current_local_parameters.Count;
		if (p_count == 1) {
			if (op == Operator.OpType.Addition)
				op = Operator.OpType.UnaryPlus;
			else if (op == Operator.OpType.Subtraction)
				op = Operator.OpType.UnaryNegation;
		}
		
		if (IsUnaryOperator (op)) {
			if (p_count == 2) {
				report.Error (1020, loc, "Overloadable binary operator expected");
			} else if (p_count != 1) {
				report.Error (1535, loc, "Overloaded unary operator `{0}' takes one parameter",
					Operator.GetName (op));
			}
		} else {
			if (p_count > 2) {
				report.Error (1534, loc, "Overloaded binary operator `{0}' takes two parameters",
					Operator.GetName (op));
			} else if (p_count != 2) {
				report.Error (1019, loc, "Overloadable unary operator expected");
			}
		}
		
		if (doc_support) {
			tmpComment = Lexer.consume_doc_comment ();
			Lexer.doc_state = XmlCommentState.NotAllowed;
		}

		yyVal = new OperatorDeclaration (op, (FullNamedExpression) yyVals[-6+yyTop], loc);
		lbag.AddLocation (yyVal, GetLocation (yyVals[-5+yyTop]), savedOperatorLocation, GetLocation (yyVals[-3+yyTop]), GetLocation (yyVals[0+yyTop]));
	  }

void case_257()
///  #line 2131 "cs-parser.jay"
{
		valid_param_mod = 0;

		Location loc = GetLocation (yyVals[-5+yyTop]);
		current_local_parameters = (ParametersCompiled)yyVals[-1+yyTop];  
		  
		if (doc_support) {
			tmpComment = Lexer.consume_doc_comment ();
			Lexer.doc_state = XmlCommentState.NotAllowed;
		}

		yyVal = new OperatorDeclaration (Operator.OpType.Implicit, (FullNamedExpression) yyVals[-4+yyTop], loc);
		lbag.AddLocation (yyVal, GetLocation (yyVals[-6+yyTop]), GetLocation (yyVals[-5+yyTop]), GetLocation (yyVals[-3+yyTop]), GetLocation (yyVals[0+yyTop]));
	  }

void case_259()
///  #line 2150 "cs-parser.jay"
{
		valid_param_mod = 0;
		
		Location loc = GetLocation (yyVals[-5+yyTop]);
		current_local_parameters = (ParametersCompiled)yyVals[-1+yyTop];  
		  
		if (doc_support) {
			tmpComment = Lexer.consume_doc_comment ();
			Lexer.doc_state = XmlCommentState.NotAllowed;
		}

		yyVal = new OperatorDeclaration (Operator.OpType.Explicit, (FullNamedExpression) yyVals[-4+yyTop], loc);
		lbag.AddLocation (yyVal, GetLocation (yyVals[-6+yyTop]), GetLocation (yyVals[-5+yyTop]), GetLocation (yyVals[-3+yyTop]), GetLocation (yyVals[0+yyTop]));
	  }

void case_260()
///  #line 2165 "cs-parser.jay"
{
	  	Error_SyntaxError (yyToken);
		current_local_parameters = ParametersCompiled.EmptyReadOnlyParameters;
		yyVal = new OperatorDeclaration (Operator.OpType.Implicit, null, GetLocation (yyVals[-1+yyTop]));
	  }

void case_261()
///  #line 2171 "cs-parser.jay"
{
	  	Error_SyntaxError (yyToken);
		current_local_parameters = ParametersCompiled.EmptyReadOnlyParameters;
	  	yyVal = new OperatorDeclaration (Operator.OpType.Explicit, null, GetLocation (yyVals[-1+yyTop]));
	  }

void case_262()
///  #line 2181 "cs-parser.jay"
{ 
		Constructor c = (Constructor) yyVals[-1+yyTop];
		c.Block = (ToplevelBlock) yyVals[0+yyTop];
		
		if (doc_support)
			c.DocComment = ConsumeStoredComment ();

		current_container.AddConstructor (c);

		current_local_parameters = null;
		if (doc_support)
			Lexer.doc_state = XmlCommentState.Allowed;
	  }

void case_263()
///  #line 2200 "cs-parser.jay"
{
		if (doc_support) {
			tmpComment = Lexer.consume_doc_comment ();
			Lexer.doc_state = XmlCommentState.Allowed;
		}
		
		valid_param_mod = ParameterModifierType.All;
	  }

void case_264()
///  #line 2209 "cs-parser.jay"
{
		valid_param_mod = 0;
		current_local_parameters = (ParametersCompiled) yyVals[-1+yyTop];  
		
		/**/
		/* start block here, so possible anonymous methods inside*/
		/* constructor initializer can get correct parent block*/
		/**/
	  	start_block (lexer.Location);
	  }

void case_265()
///  #line 2220 "cs-parser.jay"
{
		var lt = (Tokenizer.LocatedToken) yyVals[-6+yyTop];
		var mods = (Modifiers) yyVals[-7+yyTop];
		ConstructorInitializer ci = (ConstructorInitializer) yyVals[0+yyTop];

		Constructor c = new Constructor (current_class, lt.Value, mods,
			(Attributes) yyVals[-8+yyTop], current_local_parameters, ci, lt.Location);
		
		if (lt.Value != current_container.MemberName.Name) {
			report.Error (1520, c.Location, "Class, struct, or interface method must have a return type");
		} else if ((mods & Modifiers.STATIC) != 0) {
			if ((mods & Modifiers.AccessibilityMask) != 0){
				report.Error (515, c.Location,
					"`{0}': static constructor cannot have an access modifier",
					c.GetSignatureForError ());
			}
			if (ci != null) {
				report.Error (514, c.Location,
					"`{0}': static constructor cannot have an explicit `this' or `base' constructor call",
					c.GetSignatureForError ());
			
			}
		}
		
		lbag.AddMember (c, GetModifierLocations (), GetLocation (yyVals[-4+yyTop]), GetLocation (yyVals[-2+yyTop]));
		yyVal = c;
	  }

void case_271()
///  #line 2265 "cs-parser.jay"
{
	  	--lexer.parsing_block;
		yyVal = new ConstructorBaseInitializer ((Arguments) yyVals[-1+yyTop], GetLocation (yyVals[-4+yyTop]));
		lbag.AddLocation (yyVal, GetLocation (yyVals[-5+yyTop]), GetLocation (yyVals[-3+yyTop]), GetLocation (yyVals[0+yyTop]));
	  }

void case_273()
///  #line 2275 "cs-parser.jay"
{
	  	--lexer.parsing_block;
		yyVal = new ConstructorThisInitializer ((Arguments) yyVals[-1+yyTop], GetLocation (yyVals[-4+yyTop]));
		lbag.AddLocation (yyVal, GetLocation (yyVals[-5+yyTop]), GetLocation (yyVals[-3+yyTop]), GetLocation (yyVals[0+yyTop]));
	  }

void case_274()
///  #line 2281 "cs-parser.jay"
{
		Error_SyntaxError (yyToken);
		yyVal = null;
	  }

void case_275()
///  #line 2289 "cs-parser.jay"
{
		if (doc_support) {
			tmpComment = Lexer.consume_doc_comment ();
			Lexer.doc_state = XmlCommentState.NotAllowed;
		}
		
		current_local_parameters = ParametersCompiled.EmptyReadOnlyParameters;
	  }

void case_276()
///  #line 2298 "cs-parser.jay"
{
		var lt = (Tokenizer.LocatedToken) yyVals[-3+yyTop];
		if (lt.Value != current_container.MemberName.Name){
			report.Error (574, lt.Location, "Name of destructor must match name of class");
		} else if (current_container.Kind != MemberKind.Class){
			report.Error (575, lt.Location, "Only class types can contain destructor");
		}
		
		Destructor d = new Destructor (current_class, (Modifiers) yyVals[-6+yyTop],
			ParametersCompiled.EmptyReadOnlyParameters, (Attributes) yyVals[-7+yyTop], lt.Location);
		if (doc_support)
			d.DocComment = ConsumeStoredComment ();
		  
		d.Block = (ToplevelBlock) yyVals[0+yyTop];
		current_container.AddMethod (d);
		lbag.AddMember (d, GetModifierLocations (), GetLocation (yyVals[-5+yyTop]), GetLocation (yyVals[-2+yyTop]), GetLocation (yyVals[-1+yyTop]));

		current_local_parameters = null;
	  }

void case_277()
///  #line 2323 "cs-parser.jay"
{
		current_event_field = new EventField (current_class, (FullNamedExpression) yyVals[-1+yyTop], (Modifiers) yyVals[-3+yyTop], (MemberName) yyVals[0+yyTop], (Attributes) yyVals[-4+yyTop]);
		current_container.AddEvent (current_event_field);
		
		if (current_event_field.MemberName.Left != null) {
			report.Error (71, current_event_field.Location, "`{0}': An explicit interface implementation of an event must use property syntax",
			current_event_field.GetSignatureForError ());
		}
		
		yyVal = current_event_field;
	  }

void case_278()
///  #line 2337 "cs-parser.jay"
{
		if (doc_support) {
			current_event_field.DocComment = Lexer.consume_doc_comment ();
			Lexer.doc_state = XmlCommentState.Allowed;
		}
		
		lbag.AddMember (current_event_field, GetModifierLocations (), GetLocation (yyVals[-6+yyTop]), GetLocation (yyVals[0+yyTop]));
		current_event_field = null;
	  }

void case_279()
///  #line 2350 "cs-parser.jay"
{
		current_event = new EventProperty (current_class, (FullNamedExpression) yyVals[-2+yyTop], (Modifiers) yyVals[-4+yyTop], (MemberName) yyVals[-1+yyTop], (Attributes) yyVals[-5+yyTop]);
		current_container.AddEvent (current_event);
		lbag.AddMember (current_event, GetModifierLocations (), GetLocation (yyVals[-3+yyTop]), GetLocation (yyVals[0+yyTop]));
		
		lexer.EventParsing = true;
	  }

void case_280()
///  #line 2358 "cs-parser.jay"
{
		if (current_container.Kind == MemberKind.Interface)
			report.Error (69, GetLocation (yyVals[-2+yyTop]), "Event in interface cannot have add or remove accessors");
	  
		lexer.EventParsing = false;
	  }

void case_281()
///  #line 2365 "cs-parser.jay"
{
		if (doc_support) {
			current_event.DocComment = Lexer.consume_doc_comment ();
			Lexer.doc_state = XmlCommentState.Allowed;
		}
		
		lbag.AppendToMember (current_event, GetLocation (yyVals[-1+yyTop]));
	  	current_event = null;	
		current_local_parameters = null;
	  }

void case_284()
///  #line 2384 "cs-parser.jay"
{
	  	--lexer.parsing_block;
		current_event_field.Initializer = (Expression) yyVals[0+yyTop];
	  }

void case_289()
///  #line 2408 "cs-parser.jay"
{
		var lt = (Tokenizer.LocatedToken) yyVals[0+yyTop];
	  	yyVal = new FieldDeclarator (new SimpleMemberName (lt.Value, lt.Location), null);
	  	lbag.AddLocation (yyVal, GetLocation (yyVals[-1+yyTop]));
	  }

void case_291()
///  #line 2418 "cs-parser.jay"
{
		--lexer.parsing_block;
		var lt = (Tokenizer.LocatedToken) yyVals[-3+yyTop];	  
	  	yyVal = new FieldDeclarator (new SimpleMemberName (lt.Value, lt.Location), (Expression) yyVals[0+yyTop]);
	  	lbag.AddLocation (yyVal, GetLocation (yyVals[-4+yyTop]), GetLocation (yyVals[-2+yyTop]));
	  }

void case_292()
///  #line 2427 "cs-parser.jay"
{
		if (current_container.Kind == MemberKind.Interface) {
			report.Error (68, lexer.Location, "`{0}': event in interface cannot have an initializer",
				current_event_field.GetSignatureForError ());
		}
		
	  	if ((current_event_field.ModFlags & Modifiers.ABSTRACT) != 0) {
			report.Error (74, lexer.Location, "`{0}': abstract event cannot have an initializer",
				current_event_field.GetSignatureForError ());
	  	}		
	  }

void case_296()
///  #line 2448 "cs-parser.jay"
{
		report.Error (65, lexer.Location, "`{0}': event property must have both add and remove accessors",
			current_event.GetSignatureForError ());
	  }

void case_297()
///  #line 2453 "cs-parser.jay"
{
		report.Error (65, lexer.Location, "`{0}': event property must have both add and remove accessors",
			current_event.GetSignatureForError ());
	  }

void case_298()
///  #line 2458 "cs-parser.jay"
{ 
		report.Error (1055, GetLocation (yyVals[0+yyTop]), "An add or remove accessor expected");
		yyVal = null;
	  }

void case_299()
///  #line 2466 "cs-parser.jay"
{
	  	if (yyVals[-1+yyTop] != ModifierNone) {
			report.Error (1609, GetLocation (yyVals[-1+yyTop]), "Modifiers cannot be placed on event accessor declarations");
	  	}
	  	
	  	current_event.Add = new EventProperty.AddDelegateMethod (current_event, (Attributes) yyVals[-2+yyTop], GetLocation (yyVals[0+yyTop]));
		current_local_parameters = current_event.Add.ParameterInfo;
		
		lbag.AddMember (current_event.Add, GetModifierLocations ());
		lexer.EventParsing = false;		
	  }

void case_300()
///  #line 2478 "cs-parser.jay"
{
		lexer.EventParsing = true;
	  
	  	current_event.Add.Block = (ToplevelBlock) yyVals[0+yyTop];
		
		if (current_container.Kind == MemberKind.Interface) {
			report.Error (531, current_event.Add.Block.StartLocation,
				"`{0}': interface members cannot have a definition", current_event.Add.GetSignatureForError ());
		}
		
		current_local_parameters = null;
	  }

void case_301()
///  #line 2494 "cs-parser.jay"
{
	  	if (yyVals[-1+yyTop] != ModifierNone) {
			report.Error (1609, GetLocation (yyVals[-1+yyTop]), "Modifiers cannot be placed on event accessor declarations");
	  	}
	  	
	  	current_event.Remove = new EventProperty.RemoveDelegateMethod (current_event, (Attributes) yyVals[-2+yyTop], GetLocation (yyVals[0+yyTop]));
		current_local_parameters = current_event.Remove.ParameterInfo;

		lbag.AddMember (current_event.Remove, GetModifierLocations ());
		lexer.EventParsing = false;		
	  }

void case_302()
///  #line 2506 "cs-parser.jay"
{
		lexer.EventParsing = true;
	  
	  	current_event.Remove.Block = (ToplevelBlock) yyVals[0+yyTop];
		
		if (current_container.Kind == MemberKind.Interface) {
			report.Error (531, current_event.Remove.Block.StartLocation,
				"`{0}': interface members cannot have a definition", current_event.Remove.GetSignatureForError ());
		}
		
		current_local_parameters = null;
	  }

void case_303()
///  #line 2522 "cs-parser.jay"
{
		report.Error (73, lexer.Location, "An add or remove accessor must have a body");
		yyVal = null;
	  }

void case_305()
///  #line 2535 "cs-parser.jay"
{
		if (doc_support)
			enumTypeComment = Lexer.consume_doc_comment ();
	  }

void case_306()
///  #line 2540 "cs-parser.jay"
{
		if (doc_support)
			Lexer.doc_state = XmlCommentState.Allowed;

		MemberName name = (MemberName) yyVals[-3+yyTop];
		if (name.IsGeneric) {
			report.Error (1675, name.Location, "Enums cannot have type parameters");
		}
		
		push_current_class (new Enum (current_namespace, current_class, (TypeExpression) yyVals[-2+yyTop], (Modifiers) yyVals[-5+yyTop], MakeName (name), (Attributes) yyVals[-6+yyTop]), null);
		if (yyVals[-2+yyTop] != null) {
			lbag.AddMember (current_class, GetModifierLocations (), GetLocation (yyVals[-4+yyTop]), savedLocation, GetLocation (yyVals[0+yyTop]));
		} else {
			lbag.AddMember (current_class, GetModifierLocations (), GetLocation (yyVals[-4+yyTop]), GetLocation (yyVals[0+yyTop]));
		}
	  }

void case_307()
///  #line 2557 "cs-parser.jay"
{
	  	/* here will be evaluated after CLOSE_BLACE is consumed.*/
		if (doc_support)
			Lexer.doc_state = XmlCommentState.Allowed;
	  }

void case_308()
///  #line 2563 "cs-parser.jay"
{
		lbag.AppendToMember (current_class, GetLocation (yyVals[-1+yyTop]));
		if (yyVals[0+yyTop] != null) {
			current_class.OptionalSemicolon = GetLocation (yyVals[0+yyTop]);
			lbag.AppendToMember (current_class, GetLocation (yyVals[0+yyTop]));
		}
		if (doc_support)
			current_class.DocComment = enumTypeComment;
			
		--lexer.parsing_declaration;

		yyVal = pop_current_class ();
	  }

void case_310()
///  #line 2581 "cs-parser.jay"
{
	 	var te = yyVals[0+yyTop] as TypeExpression;
		if (te == null || !EnumSpec.IsValidUnderlyingType (te.Type)) {
			Enum.Error_1008 (GetLocation (yyVals[0+yyTop]), report);
			yyVal = null;
		} else {
			savedLocation = GetLocation (yyVals[-1+yyTop]);
			yyVal = yyVals[0+yyTop];
		}
	 }

void case_311()
///  #line 2592 "cs-parser.jay"
{
	 	Error_TypeExpected (GetLocation (yyVals[-1+yyTop]));
		yyVal = null;
	 }

void case_316()
///  #line 2610 "cs-parser.jay"
{
		lbag.AppendToMember (current_class, GetLocation (yyVals[-1+yyTop]));
	  	yyVal = yyVals[0+yyTop];
	  }

void case_317()
///  #line 2618 "cs-parser.jay"
{
		var lt = (Tokenizer.LocatedToken) yyVals[0+yyTop];
	  	var em = new EnumMember ((Enum) current_class, new MemberName (lt.Value, lt.Location), (Attributes) yyVals[-1+yyTop]);
	  	((Enum) current_class).AddEnumMember (em);

		if (doc_support) {
			em.DocComment = Lexer.consume_doc_comment ();
			Lexer.doc_state = XmlCommentState.Allowed;
		}

		yyVal = em;
	  }

void case_318()
///  #line 2631 "cs-parser.jay"
{
	  	++lexer.parsing_block;
		if (doc_support) {
			tmpComment = Lexer.consume_doc_comment ();
			Lexer.doc_state = XmlCommentState.NotAllowed;
		}
	  }

void case_319()
///  #line 2639 "cs-parser.jay"
{ 
		--lexer.parsing_block;
		
		var lt = (Tokenizer.LocatedToken) yyVals[-3+yyTop];
	  	var em = new EnumMember ((Enum) current_class, new MemberName (lt.Value, lt.Location), (Attributes) yyVals[-4+yyTop]);
	  	em.Initializer = new ConstInitializer (em, (Expression) yyVals[0+yyTop], GetLocation (yyVals[-1+yyTop]));
	  	((Enum) current_class).AddEnumMember (em);
		
		if (doc_support)
			em.DocComment = ConsumeStoredComment ();

		yyVal = em;
	  }

void case_321()
///  #line 2664 "cs-parser.jay"
{
		valid_param_mod = 0;

		MemberName name = MakeName ((MemberName) yyVals[-4+yyTop]);
		ParametersCompiled p = (ParametersCompiled) yyVals[-1+yyTop];

		Delegate del = new Delegate (current_namespace, current_class, (FullNamedExpression) yyVals[-5+yyTop],
					     (Modifiers) yyVals[-7+yyTop], name, p, (Attributes) yyVals[-8+yyTop]);
		p.CheckParameters (del);

		ubag.PushTypeDeclaration (del);
		ubag.PopTypeDeclaration ();

		current_container.AddDelegate (del);
		current_delegate = del;
		lexer.ConstraintsParsing = true;
	  }

void case_323()
///  #line 2686 "cs-parser.jay"
{
		if (doc_support) {
			current_delegate.DocComment = Lexer.consume_doc_comment ();
			Lexer.doc_state = XmlCommentState.Allowed;
		}
	  
		current_delegate.SetParameterInfo ((List<Constraints>) yyVals[-2+yyTop]);
		lbag.AddMember (current_delegate, GetModifierLocations (), GetLocation (yyVals[-10+yyTop]), GetLocation (yyVals[-7+yyTop]), GetLocation (yyVals[-4+yyTop]), GetLocation (yyVals[0+yyTop]));

		yyVal = current_delegate;

		current_delegate = null;
	  }

void case_325()
///  #line 2704 "cs-parser.jay"
{
		if (lang_version < LanguageVersion.ISO_2)
			FeatureIsNotAvailable (GetLocation (yyVals[0+yyTop]), "nullable types");
	  
	  	yyVal = ComposedTypeSpecifier.CreateNullable (GetLocation (yyVals[0+yyTop]));
	  }

void case_327()
///  #line 2715 "cs-parser.jay"
{
		var lt1 = (Tokenizer.LocatedToken) yyVals[-2+yyTop];
		var lt2 = (Tokenizer.LocatedToken) yyVals[-1+yyTop];
		
		yyVal = new MemberName (lt1.Value, lt2.Value, (TypeArguments) yyVals[0+yyTop], lt1.Location);
	  }

void case_329()
///  #line 2726 "cs-parser.jay"
{
		var lt = (Tokenizer.LocatedToken) yyVals[-1+yyTop];
		yyVal = new MemberName ((MemberName) yyVals[-3+yyTop], lt.Value, (TypeArguments) yyVals[0+yyTop], lt.Location) {
			DotLocation = GetLocation (yyVals[-2+yyTop])
		};
	  }

void case_330()
///  #line 2736 "cs-parser.jay"
{
		var lt = (Tokenizer.LocatedToken) yyVals[-1+yyTop];
		yyVal = new MemberName (lt.Value, (TypeArguments)yyVals[0+yyTop], lt.Location);	  
	  }

void case_332()
///  #line 2748 "cs-parser.jay"
{
		if (lang_version < LanguageVersion.ISO_2)
			FeatureIsNotAvailable (GetLocation (yyVals[-2+yyTop]), "generics");
		var list = locationListStack.Pop ();
		list.Add (GetLocation (yyVals[-2+yyTop]));
		list.Add (GetLocation (yyVals[-1+yyTop]));
		lbag.AddLocation (yyVals[-1+yyTop], list);
	
		yyVal = yyVals[-1+yyTop];;
	  }

void case_333()
///  #line 2759 "cs-parser.jay"
{
		Error_TypeExpected (lexer.Location);
		yyVal = new TypeArguments ();
	  }

void case_334()
///  #line 2767 "cs-parser.jay"
{
		TypeArguments type_args = new TypeArguments ();
		type_args.Add ((FullNamedExpression) yyVals[0+yyTop]);
		yyVal = type_args;
		locationListStack.Push (new List<Location> ());
	  }

void case_335()
///  #line 2774 "cs-parser.jay"
{
		TypeArguments type_args = (TypeArguments) yyVals[-2+yyTop];
		type_args.Add ((FullNamedExpression) yyVals[0+yyTop]);
		yyVal = type_args;
		locationListStack.Peek ().Add (GetLocation (yyVals[-1+yyTop]));
	  }

void case_337()
///  #line 2791 "cs-parser.jay"
{
		lexer.parsing_generic_declaration = false;
		var lt = (Tokenizer.LocatedToken) yyVals[-2+yyTop];
		yyVal = new MemberName (lt.Value, (TypeArguments)yyVals[0+yyTop], lt.Location);
	  }

void case_338()
///  #line 2800 "cs-parser.jay"
{
	  	MemberName mn = (MemberName)yyVals[0+yyTop];
	  	if (mn.TypeArguments != null)
	  		syntax_error (mn.Location, string.Format ("Member `{0}' cannot declare type arguments",
	  			mn.GetSignatureForError ()));
	  }

void case_340()
///  #line 2811 "cs-parser.jay"
{
		lexer.parsing_generic_declaration = false;	  
		var lt = (Tokenizer.LocatedToken) yyVals[-1+yyTop];
		yyVal = new MemberName ((MemberName) yyVals[-2+yyTop], lt.Value, (TypeArguments) yyVals[0+yyTop], lt.Location);
	  }

void case_341()
///  #line 2820 "cs-parser.jay"
{
		lexer.parsing_generic_declaration = false;	  
		yyVal = new MemberName (TypeContainer.DefaultIndexerName, GetLocation (yyVals[0+yyTop]));
	  }

void case_342()
///  #line 2825 "cs-parser.jay"
{
		lexer.parsing_generic_declaration = false;
		yyVal = new MemberName ((MemberName) yyVals[-1+yyTop], TypeContainer.DefaultIndexerName, null, GetLocation (yyVals[0+yyTop]));
	  }

void case_343()
///  #line 2833 "cs-parser.jay"
{
		var lt = (Tokenizer.LocatedToken) yyVals[-2+yyTop];
		yyVal = new MemberName (lt.Value, (TypeArguments) yyVals[-1+yyTop], lt.Location);
		lbag.AddLocation (yyVal, GetLocation (yyVals[0+yyTop]));
	  }

void case_344()
///  #line 2839 "cs-parser.jay"
{
		var lt1 = (Tokenizer.LocatedToken) yyVals[-3+yyTop];
		var lt2 = (Tokenizer.LocatedToken) yyVals[-2+yyTop];
		
		yyVal = new MemberName (lt1.Value, lt2.Value, (TypeArguments) yyVals[-1+yyTop], lt1.Location);
		lbag.AddLocation (yyVal, GetLocation (yyVals[0+yyTop]));
	  }

void case_345()
///  #line 2847 "cs-parser.jay"
{
		var lt = (Tokenizer.LocatedToken) yyVals[-2+yyTop];
		yyVal = new MemberName ((MemberName) yyVals[-3+yyTop], lt.Value, (TypeArguments) yyVals[-1+yyTop], lt.Location);
		lbag.AddLocation (yyVal, GetLocation (yyVals[0+yyTop]));
	  }

void case_347()
///  #line 2857 "cs-parser.jay"
{
		if (lang_version < LanguageVersion.ISO_2)
			FeatureIsNotAvailable (GetLocation (yyVals[-2+yyTop]), "generics");
	  
		yyVal = yyVals[-1+yyTop];
		lbag.AppendTo (yyVal, GetLocation (yyVals[-2+yyTop]), GetLocation (yyVals[0+yyTop]));
	  }

void case_348()
///  #line 2868 "cs-parser.jay"
{
		TypeArguments type_args = new TypeArguments ();
		type_args.Add ((FullNamedExpression)yyVals[0+yyTop]);
		yyVal = type_args;
	  }

void case_349()
///  #line 2874 "cs-parser.jay"
{
		TypeArguments type_args = (TypeArguments) yyVals[-2+yyTop];
		type_args.Add ((FullNamedExpression)yyVals[0+yyTop]);
		yyVal = type_args;
		lbag.AppendTo (type_args, GetLocation (yyVals[-1+yyTop]));
	  }

void case_350()
///  #line 2884 "cs-parser.jay"
{
		var lt = (Tokenizer.LocatedToken)yyVals[0+yyTop];
		var variance =  (Variance) yyVals[-1+yyTop];
		yyVal = new TypeParameterName (lt.Value, (Attributes)yyVals[-2+yyTop], variance, lt.Location);
		if (variance != Variance.None)
			lbag.AddLocation (yyVal, savedLocation);
  	  }

void case_351()
///  #line 2892 "cs-parser.jay"
{
  	  	if (GetTokenName (yyToken) == "type")
			report.Error (81, GetLocation (yyVals[0+yyTop]), "Type parameter declaration must be an identifier not a type");
		else
			Error_SyntaxError (yyToken);
			
  	  	yyVal = new TypeParameterName ("", null, lexer.Location);
  	  }

void case_356()
///  #line 2926 "cs-parser.jay"
{
	  	Expression.Error_VoidInvalidInTheContext (GetLocation (yyVals[0+yyTop]), report);
		yyVal = new TypeExpression (compiler.BuiltinTypes.Void, GetLocation (yyVals[0+yyTop]));
	  }

void case_358()
///  #line 2935 "cs-parser.jay"
{
	  	Expression.Error_VoidInvalidInTheContext (GetLocation (yyVals[0+yyTop]), report);
		yyVal = new TypeExpression (compiler.BuiltinTypes.Void, GetLocation (yyVals[0+yyTop]));
	  }

void case_360()
///  #line 2944 "cs-parser.jay"
{
	  	report.Error (1536, GetLocation (yyVals[0+yyTop]), "Invalid parameter type `void'");
		yyVal = new TypeExpression (compiler.BuiltinTypes.Void, GetLocation (yyVals[0+yyTop]));
	  }

void case_363()
///  #line 2960 "cs-parser.jay"
{
		MemberName name = (MemberName) yyVals[-1+yyTop];

		if (yyVals[0+yyTop] != null) {
			yyVal = new ComposedCast (name.GetTypeExpression (), (ComposedTypeSpecifier) yyVals[0+yyTop]);
		} else {
			if (name.Left == null && name.Name == "var")
				yyVal = new VarExpr (name.Location);
			else
				yyVal = name.GetTypeExpression ();
		}
	  }

void case_365()
///  #line 2977 "cs-parser.jay"
{
		if (yyVals[0+yyTop] != null)
			yyVal = new ComposedCast ((FullNamedExpression) yyVals[-1+yyTop], (ComposedTypeSpecifier) yyVals[0+yyTop]);
	  }

void case_368()
///  #line 2993 "cs-parser.jay"
{
		var types = new List<FullNamedExpression> (2);
		types.Add ((FullNamedExpression) yyVals[0+yyTop]);
		yyVal = types;
	  }

void case_369()
///  #line 2999 "cs-parser.jay"
{
		var types = (List<FullNamedExpression>) yyVals[-2+yyTop];
		types.Add ((FullNamedExpression) yyVals[0+yyTop]);
		lbag.AppendTo (types, GetLocation (yyVals[-1+yyTop]));
		yyVal = types;
	  }

void case_370()
///  #line 3009 "cs-parser.jay"
{
		if (yyVals[0+yyTop] is ComposedCast) {
			report.Error (1521, GetLocation (yyVals[0+yyTop]), "Invalid base type `{0}'", ((ComposedCast)yyVals[0+yyTop]).GetSignatureForError ());
		}
		yyVal = yyVals[0+yyTop];
	  }

void case_371()
///  #line 3016 "cs-parser.jay"
{
	  	Error_TypeExpected (lexer.Location);
		yyVal = null;
	  }

void case_408()
///  #line 3078 "cs-parser.jay"
{
		var lt = (Tokenizer.LocatedToken) yyVals[-1+yyTop];
		yyVal = new SimpleName (lt.Value, (TypeArguments)yyVals[0+yyTop], lt.Location);	  
	  }

void case_409()
///  #line 3082 "cs-parser.jay"
{
		var lt = (Tokenizer.LocatedToken) yyVals[-1+yyTop];
	       yyVal = new CompletionSimpleName (MemberName.MakeName (lt.Value, null), lt.Location);
	  }

void case_420()
///  #line 3123 "cs-parser.jay"
{
		yyVal = new ParenthesizedExpression ((Expression) yyVals[-1+yyTop]);
		lbag.AddLocation (yyVal, GetLocation (yyVals[-2+yyTop]), GetLocation (yyVals[0+yyTop]));
	  }

void case_422()
///  #line 3135 "cs-parser.jay"
{
		var lt = (Tokenizer.LocatedToken) yyVals[-1+yyTop];
		yyVal = new MemberAccess ((Expression) yyVals[-3+yyTop], lt.Value, (TypeArguments) yyVals[0+yyTop], lt.Location) {
			DotLocation = GetLocation (yyVals[-2+yyTop])
		};
	  }

void case_423()
///  #line 3142 "cs-parser.jay"
{
		var lt = (Tokenizer.LocatedToken) yyVals[-1+yyTop];
		yyVal = new MemberAccess ((Expression) yyVals[-3+yyTop], lt.Value, (TypeArguments) yyVals[0+yyTop], lt.Location) {
			DotLocation = GetLocation (yyVals[-2+yyTop])
		};
	  }

void case_424()
///  #line 3149 "cs-parser.jay"
{
		var lt = (Tokenizer.LocatedToken) yyVals[-1+yyTop];
		yyVal = new MemberAccess (new BaseThis (GetLocation (yyVals[-3+yyTop])), lt.Value, (TypeArguments) yyVals[0+yyTop], lt.Location) {
			DotLocation = GetLocation (yyVals[-2+yyTop])
		};
	  }

void case_425()
///  #line 3156 "cs-parser.jay"
{
		var lt1 = (Tokenizer.LocatedToken) yyVals[-2+yyTop];
		var lt2 = (Tokenizer.LocatedToken) yyVals[-1+yyTop];

		yyVal = new QualifiedAliasMember (lt1.Value, lt2.Value, (TypeArguments) yyVals[0+yyTop], lt1.Location);
		lbag.AddLocation (yyVal, GetLocation (yyVals[-1+yyTop]));
	  }

void case_427()
///  #line 3166 "cs-parser.jay"
{
		var lt = (Tokenizer.LocatedToken) yyVals[-1+yyTop];
		yyVal = new CompletionMemberAccess ((Expression) yyVals[-3+yyTop], lt.Value, lt.Location);
	  }

void case_429()
///  #line 3174 "cs-parser.jay"
{
		var lt = (Tokenizer.LocatedToken) yyVals[-1+yyTop];
		yyVal = new CompletionMemberAccess ((Expression) yyVals[-3+yyTop], lt.Value, lt.Location);
 	  }

void case_430()
///  #line 3182 "cs-parser.jay"
{
		yyVal = new Invocation ((Expression) yyVals[-3+yyTop], (Arguments) yyVals[-1+yyTop]);
		lbag.AddLocation (yyVal, GetLocation (yyVals[-2+yyTop]), GetLocation (yyVals[0+yyTop]));
	  }

void case_433()
///  #line 3195 "cs-parser.jay"
{
	  	if (yyVals[-1+yyTop] == null) {
	  		yyVal = CollectionOrObjectInitializers.Empty;
	  		/* TODO: lbag*/
	  	} else {
	  		yyVal = new CollectionOrObjectInitializers ((List<Expression>) yyVals[-1+yyTop], GetLocation (yyVals[-2+yyTop]));
	  		lbag.AddLocation (yyVal, GetLocation (yyVals[-2+yyTop]), GetLocation (yyVals[0+yyTop]));
	  	}
	  }

void case_434()
///  #line 3205 "cs-parser.jay"
{
	  	yyVal = new CollectionOrObjectInitializers ((List<Expression>) yyVals[-2+yyTop], GetLocation (yyVals[-3+yyTop]));
	  	lbag.AddLocation (yyVal, GetLocation (yyVals[-3+yyTop]), GetLocation (yyVals[-1+yyTop]), GetLocation (yyVals[0+yyTop]));
	  }

void case_437()
///  #line 3221 "cs-parser.jay"
{
	  	var a = new List<Expression> ();
	  	a.Add ((Expression) yyVals[0+yyTop]);
	  	yyVal = a;
	  }

void case_438()
///  #line 3227 "cs-parser.jay"
{
	  	var a = (List<Expression>)yyVals[-2+yyTop];
	  	a.Add ((Expression) yyVals[0+yyTop]);
		lbag.AppendTo (a, GetLocation (yyVals[-1+yyTop]));
	  	yyVal = a;
	  }

void case_439()
///  #line 3233 "cs-parser.jay"
{
	  	Error_SyntaxError (yyToken);
		yyVal = yyVals[-1+yyTop];
	  }

void case_440()
///  #line 3241 "cs-parser.jay"
{
	  	var lt = (Tokenizer.LocatedToken) yyVals[-2+yyTop];
	  	yyVal = new ElementInitializer (lt.Value, (Expression)yyVals[0+yyTop], lt.Location);
	  	lbag.AddLocation (yyVal, GetLocation (yyVals[-1+yyTop]));
	  }

void case_442()
///  #line 3250 "cs-parser.jay"
{
		CompletionSimpleName csn = yyVals[-1+yyTop] as CompletionSimpleName;
		if (csn == null)
			yyVal = new CollectionElementInitializer ((Expression)yyVals[-1+yyTop]);
		else
			yyVal = new CompletionElementInitializer (csn.Prefix, csn.Location);
	  }

void case_443()
///  #line 3258 "cs-parser.jay"
{
		if (yyVals[-1+yyTop] == null)
			yyVal = null;
		else
	  		yyVal = new CollectionElementInitializer ((List<Expression>)yyVals[-1+yyTop], GetLocation (yyVals[-2+yyTop]));
	  }

void case_444()
///  #line 3265 "cs-parser.jay"
{
	  	report.Error (1920, GetLocation (yyVals[-1+yyTop]), "An element initializer cannot be empty");
		yyVal = null;
	  }

void case_449()
///  #line 3283 "cs-parser.jay"
{ 
		Arguments list = new Arguments (4);
		list.Add ((Argument) yyVals[0+yyTop]);
		yyVal = list;
	  }

void case_450()
///  #line 3289 "cs-parser.jay"
{
		Arguments list = (Arguments) yyVals[-2+yyTop];
		if (list [list.Count - 1] is NamedArgument)
			Error_NamedArgumentExpected ((NamedArgument) list [list.Count - 1]);
		
		list.Add ((Argument) yyVals[0+yyTop]);
		lbag.AppendTo (list, GetLocation (yyVals[-1+yyTop]));
		yyVal = list;
	  }

void case_451()
///  #line 3299 "cs-parser.jay"
{
		Arguments list = (Arguments) yyVals[-2+yyTop];
		NamedArgument a = (NamedArgument) yyVals[0+yyTop];
		for (int i = 0; i < list.Count; ++i) {
			NamedArgument na = list [i] as NamedArgument;
			if (na != null && na.Name == a.Name)
				report.Error (1740, na.Location, "Named argument `{0}' specified multiple times",
					na.Name);
		}
		
		list.Add (a);
		lbag.AppendTo (list, GetLocation (yyVals[-1+yyTop]));
		yyVal = list;
	  }

void case_452()
///  #line 3314 "cs-parser.jay"
{
	  	report.Error (839, GetLocation (yyVals[0+yyTop]), "An argument is missing");
	  	yyVal = yyVals[-1+yyTop];
	  }

void case_453()
///  #line 3319 "cs-parser.jay"
{
	  	report.Error (839, GetLocation (yyVals[-1+yyTop]), "An argument is missing");
	  	yyVal = null;
	  }

void case_458()
///  #line 3340 "cs-parser.jay"
{ 
		yyVal = new Argument ((Expression) yyVals[0+yyTop], Argument.AType.Ref);
		lbag.AddLocation (yyVal, GetLocation (yyVals[-1+yyTop]));
	  }

void case_459()
///  #line 3345 "cs-parser.jay"
{ 
		yyVal = new Argument ((Expression) yyVals[0+yyTop], Argument.AType.Out);
		lbag.AddLocation (yyVal, GetLocation (yyVals[-1+yyTop]));
	  }

void case_460()
///  #line 3350 "cs-parser.jay"
{
		yyVal = new Argument (new Arglist ((Arguments) yyVals[-1+yyTop], GetLocation (yyVals[-3+yyTop])));
		lbag.AddLocation (yyVal, GetLocation (yyVals[-2+yyTop]), GetLocation (yyVals[0+yyTop]));
	  }

void case_461()
///  #line 3355 "cs-parser.jay"
{
		yyVal = new Argument (new Arglist (GetLocation (yyVals[-2+yyTop])));
		lbag.AddLocation (yyVal, GetLocation (yyVals[-1+yyTop]), GetLocation (yyVals[0+yyTop]));
	  }

void case_463()
///  #line 3367 "cs-parser.jay"
{
		yyVal = new ElementAccess ((Expression) yyVals[-3+yyTop], (Arguments) yyVals[-1+yyTop], GetLocation (yyVals[-2+yyTop]));
		lbag.AddLocation (yyVal, GetLocation (yyVals[0+yyTop]));
	  }

void case_466()
///  #line 3383 "cs-parser.jay"
{
		var list = new List<Expression> (4);
		list.Add ((Expression) yyVals[0+yyTop]);
		yyVal = list;
	  }

void case_467()
///  #line 3389 "cs-parser.jay"
{
		var list = (List<Expression>) yyVals[-2+yyTop];
		list.Add ((Expression) yyVals[0+yyTop]);
		lbag.AppendTo (list, GetLocation (yyVals[-1+yyTop]));
		yyVal = list;
	  }

void case_468()
///  #line 3395 "cs-parser.jay"
{
	  	Error_SyntaxError (yyToken);
		yyVal = yyVals[-1+yyTop];
	  }

void case_469()
///  #line 3403 "cs-parser.jay"
{
		Arguments args = new Arguments (4);
		args.Add ((Argument) yyVals[0+yyTop]);
		yyVal = args;
	  }

void case_470()
///  #line 3409 "cs-parser.jay"
{
		Arguments args = (Arguments) yyVals[-2+yyTop];
		if (args [args.Count - 1] is NamedArgument && !(yyVals[0+yyTop] is NamedArgument))
			Error_NamedArgumentExpected ((NamedArgument) args [args.Count - 1]);
	  
		args.Add ((Argument) yyVals[0+yyTop]);
		lbag.AppendTo (args, GetLocation (yyVals[-1+yyTop]));
		yyVal = args;	  
	  }

void case_474()
///  #line 3437 "cs-parser.jay"
{
	  	yyVal = new ElementAccess (new BaseThis (GetLocation (yyVals[-3+yyTop])), (Arguments) yyVals[-1+yyTop], GetLocation (yyVals[-2+yyTop]));
		lbag.AddLocation (yyVal, GetLocation (yyVals[0+yyTop]));
	  }

void case_475()
///  #line 3442 "cs-parser.jay"
{
	  	Error_SyntaxError (yyToken);
		yyVal = new ElementAccess (null, null, GetLocation (yyVals[-1+yyTop]));
	  }

void case_478()
///  #line 3464 "cs-parser.jay"
{
		if (yyVals[0+yyTop] != null) {
			if (lang_version <= LanguageVersion.ISO_2)
				FeatureIsNotAvailable (GetLocation (yyVals[-5+yyTop]), "object initializers");
				
			yyVal = new NewInitialize ((FullNamedExpression) yyVals[-4+yyTop], (Arguments) yyVals[-2+yyTop], (CollectionOrObjectInitializers) yyVals[0+yyTop], GetLocation (yyVals[-5+yyTop]));
		} else {
			yyVal = new New ((FullNamedExpression) yyVals[-4+yyTop], (Arguments) yyVals[-2+yyTop], GetLocation (yyVals[-5+yyTop]));
		}
		
		lbag.AddLocation (yyVal, GetLocation (yyVals[-3+yyTop]), GetLocation (yyVals[-1+yyTop]));
	  }

void case_479()
///  #line 3477 "cs-parser.jay"
{
		if (lang_version <= LanguageVersion.ISO_2)
			FeatureIsNotAvailable (GetLocation (yyVals[-2+yyTop]), "collection initializers");
	  
		yyVal = new NewInitialize ((FullNamedExpression) yyVals[-1+yyTop], null, (CollectionOrObjectInitializers) yyVals[0+yyTop], GetLocation (yyVals[-2+yyTop]));
	  }

void case_480()
///  #line 3489 "cs-parser.jay"
{
		yyVal = new ArrayCreation ((FullNamedExpression) yyVals[-5+yyTop], (List<Expression>) yyVals[-3+yyTop],
				new ComposedTypeSpecifier (((List<Expression>) yyVals[-3+yyTop]).Count, GetLocation (yyVals[-4+yyTop])) {
	  				Next = (ComposedTypeSpecifier) yyVals[-1+yyTop]
			  	}, (ArrayInitializer) yyVals[0+yyTop], GetLocation (yyVals[-6+yyTop]));
		lbag.AddLocation (yyVal, GetLocation (yyVals[-4+yyTop]), GetLocation (yyVals[-2+yyTop]));
	  }

void case_481()
///  #line 3497 "cs-parser.jay"
{
	  	if (yyVals[0+yyTop] == null)
	  		report.Error (1586, GetLocation (yyVals[-3+yyTop]), "Array creation must have array size or array initializer");

		yyVal = new ArrayCreation ((FullNamedExpression) yyVals[-2+yyTop], (ComposedTypeSpecifier) yyVals[-1+yyTop], (ArrayInitializer) yyVals[0+yyTop], GetLocation (yyVals[-3+yyTop]));
	  }

void case_482()
///  #line 3504 "cs-parser.jay"
{
		if (lang_version <= LanguageVersion.ISO_2)
			FeatureIsNotAvailable (GetLocation (yyVals[-2+yyTop]), "implicitly typed arrays");
	  
		yyVal = new ImplicitlyTypedArrayCreation ((ComposedTypeSpecifier) yyVals[-1+yyTop], (ArrayInitializer) yyVals[0+yyTop], GetLocation (yyVals[-2+yyTop]));
	  }

void case_483()
///  #line 3511 "cs-parser.jay"
{
		report.Error (178, GetLocation (yyVals[-1+yyTop]), "Invalid rank specifier, expecting `,' or `]'");
		yyVal = new ArrayCreation ((FullNamedExpression) yyVals[-5+yyTop], null, GetLocation (yyVals[-6+yyTop]));
	  }

void case_484()
///  #line 3516 "cs-parser.jay"
{
		Error_SyntaxError (1526, yyToken, "Unexpected symbol");
		yyVal = new ArrayCreation ((FullNamedExpression) yyVals[-1+yyTop], null, GetLocation (yyVals[-2+yyTop]));
	  }

void case_486()
///  #line 3527 "cs-parser.jay"
{
		--lexer.parsing_type;
		yyVal = yyVals[0+yyTop];
	  }

void case_487()
///  #line 3535 "cs-parser.jay"
{
		if (lang_version <= LanguageVersion.ISO_2)
	  		FeatureIsNotAvailable (GetLocation (yyVals[-3+yyTop]), "anonymous types");

		yyVal = new NewAnonymousType ((List<AnonymousTypeParameter>) yyVals[-1+yyTop], current_container, GetLocation (yyVals[-3+yyTop]));
		
		/* TODO: lbag comma location*/
		lbag.AddLocation (yyVal, GetLocation (yyVals[-2+yyTop]), GetLocation (yyVals[0+yyTop]));
	  }

void case_492()
///  #line 3558 "cs-parser.jay"
{
	  	var a = new List<AnonymousTypeParameter> (4);
	  	a.Add ((AnonymousTypeParameter) yyVals[0+yyTop]);
	  	yyVal = a;
	  }

void case_493()
///  #line 3564 "cs-parser.jay"
{
	  	var a = (List<AnonymousTypeParameter>) yyVals[-2+yyTop];
	  	a.Add ((AnonymousTypeParameter) yyVals[0+yyTop]);
	  	lbag.AppendTo (a, GetLocation (yyVals[-1+yyTop]));

	  	yyVal = a;
	  }

void case_494()
///  #line 3575 "cs-parser.jay"
{
		var lt = (Tokenizer.LocatedToken)yyVals[-2+yyTop];
	  	yyVal = new AnonymousTypeParameter ((Expression)yyVals[0+yyTop], lt.Value, lt.Location);
	  	lbag.AddLocation (yyVal, GetLocation (yyVals[-1+yyTop]));
	  }

void case_495()
///  #line 3581 "cs-parser.jay"
{
		var lt = (Tokenizer.LocatedToken)yyVals[0+yyTop];
	  	yyVal = new AnonymousTypeParameter (new SimpleName (lt.Value, lt.Location),
	  		lt.Value, lt.Location);
	  }

void case_496()
///  #line 3587 "cs-parser.jay"
{
	  	MemberAccess ma = (MemberAccess) yyVals[0+yyTop];
	  	yyVal = new AnonymousTypeParameter (ma, ma.Name, ma.Location);
	  }

void case_497()
///  #line 3592 "cs-parser.jay"
{
		report.Error (746, lexer.Location,
			"Invalid anonymous type member declarator. Anonymous type members must be a member assignment, simple name or member access expression");
		yyVal = null;
	  }

void case_501()
///  #line 3607 "cs-parser.jay"
{
	  	((ComposedTypeSpecifier) yyVals[-1+yyTop]).Next = (ComposedTypeSpecifier) yyVals[0+yyTop];
	  	yyVal = yyVals[-1+yyTop];
	  }

void case_502()
///  #line 3615 "cs-parser.jay"
{
		yyVal = ComposedTypeSpecifier.CreateArrayDimension (1, GetLocation (yyVals[-1+yyTop]));
		lbag.AddLocation (yyVal, GetLocation (yyVals[0+yyTop]));
	  }

void case_503()
///  #line 3620 "cs-parser.jay"
{
		yyVal = ComposedTypeSpecifier.CreateArrayDimension ((int)yyVals[-1+yyTop], GetLocation (yyVals[-2+yyTop]));
		lbag.AddLocation (yyVal, GetLocation (yyVals[0+yyTop]));
	  }

void case_508()
///  #line 3650 "cs-parser.jay"
{
		var ai = new ArrayInitializer (0, GetLocation (yyVals[-1+yyTop]));
		ai.VariableDeclaration = current_variable;
		lbag.AddLocation (ai, GetLocation (yyVals[0+yyTop]));
		yyVal = ai;
	  }

void case_509()
///  #line 3657 "cs-parser.jay"
{
		var ai = new ArrayInitializer ((List<Expression>) yyVals[-2+yyTop], GetLocation (yyVals[-3+yyTop]));
		ai.VariableDeclaration = current_variable;
		if (yyVals[-1+yyTop] != null) {
			lbag.AddLocation (ai, GetLocation (yyVals[-1+yyTop]), GetLocation (yyVals[0+yyTop]));
		} else {
			lbag.AddLocation (ai, GetLocation (yyVals[0+yyTop]));
		}
		yyVal = ai;
	  }

void case_510()
///  #line 3671 "cs-parser.jay"
{
		var list = new List<Expression> (4);
		list.Add ((Expression) yyVals[0+yyTop]);
		yyVal = list;
	  }

void case_511()
///  #line 3677 "cs-parser.jay"
{
		var list = (List<Expression>) yyVals[-2+yyTop];
		list.Add ((Expression) yyVals[0+yyTop]);
		lbag.AppendTo (list, GetLocation (yyVals[-1+yyTop]));
		yyVal = list;
	  }

void case_513()
///  #line 3691 "cs-parser.jay"
{
	  	lexer.TypeOfParsing = false;
		yyVal = new TypeOf ((FullNamedExpression) yyVals[-1+yyTop], GetLocation (yyVals[-4+yyTop]));
		lbag.AddLocation (yyVal, GetLocation (yyVals[-2+yyTop]), GetLocation (yyVals[0+yyTop]));
	  }

void case_516()
///  #line 3702 "cs-parser.jay"
{
	 	Error_TypeExpected (lexer.Location);
	 	yyVal = null;
	 }

void case_517()
///  #line 3710 "cs-parser.jay"
{  
		var lt = (Tokenizer.LocatedToken) yyVals[-1+yyTop];

		yyVal = new SimpleName (lt.Value, (int) yyVals[0+yyTop], lt.Location);
	  }

void case_518()
///  #line 3716 "cs-parser.jay"
{
		var lt1 = (Tokenizer.LocatedToken) yyVals[-2+yyTop];
		var lt2 = (Tokenizer.LocatedToken) yyVals[-1+yyTop];

		yyVal = new QualifiedAliasMember (lt1.Value, lt2.Value, (int) yyVals[0+yyTop], lt1.Location);
		lbag.AddLocation (yyVal, GetLocation (yyVals[-1+yyTop]));
	  }

void case_519()
///  #line 3724 "cs-parser.jay"
{
		var lt = (Tokenizer.LocatedToken) yyVals[0+yyTop];
		
		yyVal = new MemberAccess ((Expression) yyVals[-2+yyTop], lt.Value, lt.Location) {
			DotLocation = GetLocation (yyVals[-1+yyTop])
		};
	  }

void case_520()
///  #line 3732 "cs-parser.jay"
{
		var lt = (Tokenizer.LocatedToken) yyVals[-1+yyTop];
		
		yyVal = new MemberAccess ((Expression) yyVals[-3+yyTop], lt.Value, (int) yyVals[0+yyTop], lt.Location) {
			DotLocation = GetLocation (yyVals[-2+yyTop])
		};
	  }

void case_521()
///  #line 3740 "cs-parser.jay"
{
		var te = ((MemberName) yyVals[-3+yyTop]).GetTypeExpression ();
		if (te.HasTypeArguments)
			Error_TypeExpected (GetLocation (yyVals[0+yyTop]));

		var lt = (Tokenizer.LocatedToken) yyVals[-1+yyTop];
		yyVal = new MemberAccess (te, lt.Value, (int) yyVals[0+yyTop], lt.Location) {
			DotLocation = GetLocation (yyVals[-2+yyTop])
		};
	  }

void case_522()
///  #line 3754 "cs-parser.jay"
{
		if (lang_version < LanguageVersion.ISO_2)
			FeatureIsNotAvailable (GetLocation (yyVals[0+yyTop]), "generics");

		yyVal = yyVals[0+yyTop];
	  }

void case_523()
///  #line 3764 "cs-parser.jay"
{
		var lt = (Tokenizer.LocatedToken) yyVals[-1+yyTop];
		if (lang_version == LanguageVersion.ISO_1)
			FeatureIsNotAvailable (lt.Location, "namespace alias qualifier");

		yyVal = lt;		
	  }

void case_524()
///  #line 3775 "cs-parser.jay"
{ 
		yyVal = new SizeOf ((Expression) yyVals[-1+yyTop], GetLocation (yyVals[-3+yyTop]));
		lbag.AddLocation (yyVal, GetLocation (yyVals[-2+yyTop]), GetLocation (yyVals[0+yyTop]));
	  }

void case_525()
///  #line 3783 "cs-parser.jay"
{
		yyVal = new CheckedExpr ((Expression) yyVals[-1+yyTop], GetLocation (yyVals[-3+yyTop]));
		lbag.AddLocation (yyVal, GetLocation (yyVals[-2+yyTop]), GetLocation (yyVals[0+yyTop]));
	  }

void case_526()
///  #line 3791 "cs-parser.jay"
{
		yyVal = new UnCheckedExpr ((Expression) yyVals[-1+yyTop], GetLocation (yyVals[-3+yyTop]));
		lbag.AddLocation (yyVal, GetLocation (yyVals[-2+yyTop]), GetLocation (yyVals[0+yyTop]));
	  }

void case_527()
///  #line 3799 "cs-parser.jay"
{
		var lt = (Tokenizer.LocatedToken) yyVals[-1+yyTop];
		yyVal = new MemberAccess (new Indirection ((Expression) yyVals[-3+yyTop], GetLocation (yyVals[-2+yyTop])), lt.Value, (TypeArguments) yyVals[0+yyTop], lt.Location);
	  }

void case_529()
///  #line 3811 "cs-parser.jay"
{
		yyVal = end_anonymous ((ParametersBlock) yyVals[0+yyTop]);
		if ((ParametersCompiled) yyVals[-2+yyTop] != ParametersCompiled.Undefined) {
			lbag.AddLocation (yyVal, GetLocation (yyVals[-3+yyTop]), savedOpenLocation, savedCloseLocation);
		} else {
			lbag.AddLocation (yyVal, GetLocation (yyVals[-3+yyTop]));
		}
	  }

void case_531()
///  #line 3824 "cs-parser.jay"
{
		yyVal = end_anonymous ((ParametersBlock) yyVals[0+yyTop]);
		
		if ((ParametersCompiled) yyVals[-2+yyTop] != ParametersCompiled.Undefined) {
			lbag.AddLocation (yyVal, GetLocation (yyVals[-4+yyTop]), GetLocation (yyVals[-3+yyTop]), savedOpenLocation, savedCloseLocation);
		} else {
			lbag.AddLocation (yyVal, GetLocation (yyVals[-4+yyTop]), GetLocation (yyVals[-3+yyTop]));
		}
	  }

void case_535()
///  #line 3849 "cs-parser.jay"
{
		valid_param_mod = 0;
	  	yyVal = yyVals[-1+yyTop];
	  	savedOpenLocation = GetLocation (yyVals[-3+yyTop]);
	  	savedCloseLocation = GetLocation (yyVals[-2+yyTop]);
	  }

void case_536()
///  #line 3859 "cs-parser.jay"
{
		if (lang_version < LanguageVersion.ISO_2)
			FeatureIsNotAvailable (GetLocation (yyVals[-3+yyTop]), "default value expression");

		yyVal = new DefaultValueExpression ((Expression) yyVals[-1+yyTop], GetLocation (yyVals[-3+yyTop]));
		lbag.AddLocation (yyVal, GetLocation (yyVals[-2+yyTop]), GetLocation (yyVals[0+yyTop]));
	  }

void case_540()
///  #line 3879 "cs-parser.jay"
{
		yyVal = new Cast ((FullNamedExpression) yyVals[-2+yyTop], (Expression) yyVals[0+yyTop], GetLocation (yyVals[-3+yyTop]));
		lbag.AddLocation (yyVal, GetLocation (yyVals[-1+yyTop]));
	  }

void case_541()
///  #line 3884 "cs-parser.jay"
{
		if (!async_block) {
			report.Error (1992, GetLocation (yyVals[-1+yyTop]),
				"The `await' operator can only be used when its containing method or lambda expression is marked with the `async' modifier");
		} else {
			current_block.Explicit.RegisterAsyncAwait ();
		}
		
		yyVal = new Await ((Expression) yyVals[0+yyTop], GetLocation (yyVals[-1+yyTop]));
	  }

void case_550()
///  #line 3931 "cs-parser.jay"
{
		yyVal = new Binary (Binary.Operator.Multiply, 
			         (Expression) yyVals[-2+yyTop], (Expression) yyVals[0+yyTop], GetLocation (yyVals[-1+yyTop]));
	  }

void case_551()
///  #line 3936 "cs-parser.jay"
{
		yyVal = new Binary (Binary.Operator.Division, 
			         (Expression) yyVals[-2+yyTop], (Expression) yyVals[0+yyTop], GetLocation (yyVals[-1+yyTop]));
	  }

void case_552()
///  #line 3941 "cs-parser.jay"
{
		yyVal = new Binary (Binary.Operator.Modulus, 
			         (Expression) yyVals[-2+yyTop], (Expression) yyVals[0+yyTop], GetLocation (yyVals[-1+yyTop]));
	  }

void case_554()
///  #line 3950 "cs-parser.jay"
{
		yyVal = new Binary (Binary.Operator.Addition, 
			         (Expression) yyVals[-2+yyTop], (Expression) yyVals[0+yyTop], GetLocation (yyVals[-1+yyTop]));
	  }

void case_556()
///  #line 3959 "cs-parser.jay"
{
	  	/* Shift/Reduce conflict*/
		yyVal = new Binary (Binary.Operator.Subtraction, (Expression) yyVals[-2+yyTop], (Expression) yyVals[0+yyTop], GetLocation (yyVals[-1+yyTop]));
  	  }

void case_560()
///  #line 3976 "cs-parser.jay"
{
		yyVal = new Binary (Binary.Operator.LeftShift, 
			         (Expression) yyVals[-2+yyTop], (Expression) yyVals[0+yyTop], GetLocation (yyVals[-1+yyTop]));
	  }

void case_561()
///  #line 3981 "cs-parser.jay"
{
		yyVal = new Binary (Binary.Operator.RightShift, 
			         (Expression) yyVals[-2+yyTop], (Expression) yyVals[0+yyTop], GetLocation (yyVals[-1+yyTop]));
	  }

void case_563()
///  #line 3990 "cs-parser.jay"
{
		yyVal = new Binary (Binary.Operator.LessThan, 
			         (Expression) yyVals[-2+yyTop], (Expression) yyVals[0+yyTop], GetLocation (yyVals[-1+yyTop]));
	  }

void case_564()
///  #line 3995 "cs-parser.jay"
{
		yyVal = new Binary (Binary.Operator.GreaterThan, 
			         (Expression) yyVals[-2+yyTop], (Expression) yyVals[0+yyTop], GetLocation (yyVals[-1+yyTop]));
	  }

void case_565()
///  #line 4000 "cs-parser.jay"
{
		yyVal = new Binary (Binary.Operator.LessThanOrEqual, 
			         (Expression) yyVals[-2+yyTop], (Expression) yyVals[0+yyTop], GetLocation (yyVals[-1+yyTop]));
	  }

void case_566()
///  #line 4005 "cs-parser.jay"
{
		yyVal = new Binary (Binary.Operator.GreaterThanOrEqual, 
			         (Expression) yyVals[-2+yyTop], (Expression) yyVals[0+yyTop], GetLocation (yyVals[-1+yyTop]));
	  }

void case_568()
///  #line 4014 "cs-parser.jay"
{
		yyVal = new Binary (Binary.Operator.Equality, 
			         (Expression) yyVals[-2+yyTop], (Expression) yyVals[0+yyTop], GetLocation (yyVals[-1+yyTop]));
	  }

void case_569()
///  #line 4019 "cs-parser.jay"
{
		yyVal = new Binary (Binary.Operator.Inequality, 
			         (Expression) yyVals[-2+yyTop], (Expression) yyVals[0+yyTop], GetLocation (yyVals[-1+yyTop]));
	  }

void case_571()
///  #line 4028 "cs-parser.jay"
{
		yyVal = new Binary (Binary.Operator.BitwiseAnd, 
			         (Expression) yyVals[-2+yyTop], (Expression) yyVals[0+yyTop], GetLocation (yyVals[-1+yyTop]));
	  }

void case_573()
///  #line 4037 "cs-parser.jay"
{
		yyVal = new Binary (Binary.Operator.ExclusiveOr, 
			         (Expression) yyVals[-2+yyTop], (Expression) yyVals[0+yyTop], GetLocation (yyVals[-1+yyTop]));
	  }

void case_575()
///  #line 4046 "cs-parser.jay"
{
		yyVal = new Binary (Binary.Operator.BitwiseOr, 
			         (Expression) yyVals[-2+yyTop], (Expression) yyVals[0+yyTop], GetLocation (yyVals[-1+yyTop]));
	  }

void case_577()
///  #line 4055 "cs-parser.jay"
{
		yyVal = new Binary (Binary.Operator.LogicalAnd, 
			         (Expression) yyVals[-2+yyTop], (Expression) yyVals[0+yyTop], GetLocation (yyVals[-1+yyTop]));
	  }

void case_579()
///  #line 4064 "cs-parser.jay"
{
		yyVal = new Binary (Binary.Operator.LogicalOr, 
			         (Expression) yyVals[-2+yyTop], (Expression) yyVals[0+yyTop], GetLocation (yyVals[-1+yyTop]));
	  }

void case_581()
///  #line 4073 "cs-parser.jay"
{
		if (lang_version < LanguageVersion.ISO_2)
			FeatureIsNotAvailable (GetLocation (yyVals[-1+yyTop]), "null coalescing operator");
			
		yyVal = new Nullable.NullCoalescingOperator ((Expression) yyVals[-2+yyTop], (Expression) yyVals[0+yyTop], GetLocation (yyVals[-1+yyTop]));
	  }

void case_583()
///  #line 4084 "cs-parser.jay"
{
		yyVal = new Conditional (new BooleanExpression ((Expression) yyVals[-4+yyTop]), (Expression) yyVals[-2+yyTop], (Expression) yyVals[0+yyTop], GetLocation (yyVals[-3+yyTop]));
		lbag.AddLocation (yyVal, GetLocation (yyVals[-1+yyTop]));
	  }

void case_585()
///  #line 4096 "cs-parser.jay"
{
		yyVal = new CompoundAssign (
			Binary.Operator.Multiply, (Expression) yyVals[-2+yyTop], (Expression) yyVals[0+yyTop], GetLocation (yyVals[-1+yyTop]));
	  }

void case_586()
///  #line 4101 "cs-parser.jay"
{
		yyVal = new CompoundAssign (
			Binary.Operator.Division, (Expression) yyVals[-2+yyTop], (Expression) yyVals[0+yyTop], GetLocation (yyVals[-1+yyTop]));
	  }

void case_587()
///  #line 4106 "cs-parser.jay"
{
		yyVal = new CompoundAssign (
			Binary.Operator.Modulus, (Expression) yyVals[-2+yyTop], (Expression) yyVals[0+yyTop], GetLocation (yyVals[-1+yyTop]));
	  }

void case_588()
///  #line 4111 "cs-parser.jay"
{
		yyVal = new CompoundAssign (
			Binary.Operator.Addition, (Expression) yyVals[-2+yyTop], (Expression) yyVals[0+yyTop], GetLocation (yyVals[-1+yyTop]));
	  }

void case_589()
///  #line 4116 "cs-parser.jay"
{
		yyVal = new CompoundAssign (
			Binary.Operator.Subtraction, (Expression) yyVals[-2+yyTop], (Expression) yyVals[0+yyTop], GetLocation (yyVals[-1+yyTop]));
	  }

void case_590()
///  #line 4121 "cs-parser.jay"
{
		yyVal = new CompoundAssign (
			Binary.Operator.LeftShift, (Expression) yyVals[-2+yyTop], (Expression) yyVals[0+yyTop], GetLocation (yyVals[-1+yyTop]));
	  }

void case_591()
///  #line 4126 "cs-parser.jay"
{
		yyVal = new CompoundAssign (
			Binary.Operator.RightShift, (Expression) yyVals[-2+yyTop], (Expression) yyVals[0+yyTop], GetLocation (yyVals[-1+yyTop]));
	  }

void case_592()
///  #line 4131 "cs-parser.jay"
{
		yyVal = new CompoundAssign (
			Binary.Operator.BitwiseAnd, (Expression) yyVals[-2+yyTop], (Expression) yyVals[0+yyTop], GetLocation (yyVals[-1+yyTop]));
	  }

void case_593()
///  #line 4136 "cs-parser.jay"
{
		yyVal = new CompoundAssign (
			Binary.Operator.BitwiseOr, (Expression) yyVals[-2+yyTop], (Expression) yyVals[0+yyTop], GetLocation (yyVals[-1+yyTop]));
	  }

void case_594()
///  #line 4141 "cs-parser.jay"
{
		yyVal = new CompoundAssign (
			Binary.Operator.ExclusiveOr, (Expression) yyVals[-2+yyTop], (Expression) yyVals[0+yyTop], GetLocation (yyVals[-1+yyTop]));
	  }

void case_595()
///  #line 4149 "cs-parser.jay"
{
		var pars = new List<Parameter> (4);
		pars.Add ((Parameter) yyVals[0+yyTop]);
		parameterListCommas.Clear ();
		yyVal = pars;
	  }

void case_596()
///  #line 4156 "cs-parser.jay"
{
		var pars = (List<Parameter>) yyVals[-2+yyTop];
		Parameter p = (Parameter)yyVals[0+yyTop];
		if (pars[0].GetType () != p.GetType ()) {
			report.Error (748, p.Location, "All lambda parameters must be typed either explicitly or implicitly");
		}
		
		pars.Add (p);
		parameterListCommas.Add (GetLocation (yyVals[-1+yyTop]));

		yyVal = pars;
	  }

void case_597()
///  #line 4172 "cs-parser.jay"
{
		var lt = (Tokenizer.LocatedToken) yyVals[0+yyTop];

		yyVal = new Parameter ((FullNamedExpression) yyVals[-1+yyTop], lt.Value, (Parameter.Modifier) yyVals[-2+yyTop], null, lt.Location);
	  }

void case_598()
///  #line 4178 "cs-parser.jay"
{
		var lt = (Tokenizer.LocatedToken) yyVals[0+yyTop];

		yyVal = new Parameter ((FullNamedExpression) yyVals[-1+yyTop], lt.Value, Parameter.Modifier.NONE, null, lt.Location);
	  }

void case_599()
///  #line 4184 "cs-parser.jay"
{
	  	var lt = (Tokenizer.LocatedToken) yyVals[0+yyTop];
		yyVal = new ImplicitLambdaParameter (lt.Value, lt.Location);
	  }

void case_601()
///  #line 4192 "cs-parser.jay"
{ 
		var pars_list = (List<Parameter>) yyVals[0+yyTop];
		yyVal = new ParametersCompiled (pars_list.ToArray ());
		lbag.AddLocation (yyVal, parameterListCommas);
	  }

void case_605()
///  #line 4209 "cs-parser.jay"
{
		Block b = end_block (lexer.Location);
		b.IsCompilerGenerated = true;
		b.AddStatement (new ContextualReturn ((Expression) yyVals[0+yyTop]));
		yyVal = b;
	  }

void case_607()
///  #line 4220 "cs-parser.jay"
{
		Error_SyntaxError (yyToken);	
		yyVal = EmptyExpression.Null;
	  }

void case_608()
///  #line 4228 "cs-parser.jay"
{
		var lt = (Tokenizer.LocatedToken) yyVals[-1+yyTop];	
		Parameter p = new ImplicitLambdaParameter (lt.Value, lt.Location);
		start_anonymous (true, new ParametersCompiled (p), false, lt.Location);
	  }

void case_609()
///  #line 4234 "cs-parser.jay"
{
		yyVal = end_anonymous ((ParametersBlock) yyVals[0+yyTop]);
		lbag.AddLocation (yyVal, GetLocation (yyVals[-2+yyTop]));
	  }

void case_610()
///  #line 4239 "cs-parser.jay"
{
		var lt = (Tokenizer.LocatedToken) yyVals[-1+yyTop];
		Parameter p = new ImplicitLambdaParameter (lt.Value, lt.Location);
		start_anonymous (true, new ParametersCompiled (p), true, lt.Location);
	  }

void case_611()
///  #line 4245 "cs-parser.jay"
{
		yyVal = end_anonymous ((ParametersBlock) yyVals[0+yyTop]);
		lbag.AddLocation (yyVal, GetLocation (yyVals[-4+yyTop]), GetLocation (yyVals[-2+yyTop]));
	  }

void case_613()
///  #line 4254 "cs-parser.jay"
{
	  	valid_param_mod = 0;
		start_anonymous (true, (ParametersCompiled) yyVals[-2+yyTop], false, GetLocation (yyVals[-4+yyTop]));
	  }

void case_614()
///  #line 4259 "cs-parser.jay"
{
		yyVal = end_anonymous ((ParametersBlock) yyVals[0+yyTop]);
		lbag.AddLocation (yyVal, GetLocation (yyVals[-6+yyTop]), GetLocation (yyVals[-3+yyTop]), GetLocation (yyVals[-2+yyTop]));
	  }

void case_616()
///  #line 4268 "cs-parser.jay"
{
	  	valid_param_mod = 0;
		start_anonymous (true, (ParametersCompiled) yyVals[-2+yyTop], true, GetLocation (yyVals[-5+yyTop]));
	  }

void case_617()
///  #line 4273 "cs-parser.jay"
{
		yyVal = end_anonymous ((ParametersBlock) yyVals[0+yyTop]);
		lbag.AddLocation (yyVal, GetLocation (yyVals[-7+yyTop]), GetLocation (yyVals[-6+yyTop]), GetLocation (yyVals[-3+yyTop]), GetLocation (yyVals[-2+yyTop]));
	  }

void case_624()
///  #line 4296 "cs-parser.jay"
{
		yyVal = new RefValueExpr ((Expression) yyVals[-3+yyTop], (FullNamedExpression) yyVals[-1+yyTop], GetLocation (yyVals[-5+yyTop]));
		lbag.AddLocation (yyVal, GetLocation (yyVals[-4+yyTop]), GetLocation (yyVals[-2+yyTop]), GetLocation (yyVals[0+yyTop]));
	  }

void case_625()
///  #line 4301 "cs-parser.jay"
{
		yyVal = new RefTypeExpr ((Expression) yyVals[-1+yyTop], GetLocation (yyVals[-3+yyTop]));
		lbag.AddLocation (yyVal, GetLocation (yyVals[-2+yyTop]), GetLocation (yyVals[0+yyTop]));
	  }

void case_626()
///  #line 4306 "cs-parser.jay"
{
		yyVal = new MakeRefExpr ((Expression) yyVals[-1+yyTop], GetLocation (yyVals[-3+yyTop]));
		lbag.AddLocation (yyVal, GetLocation (yyVals[-2+yyTop]), GetLocation (yyVals[0+yyTop]));	  
	  }

void case_630()
///  #line 4335 "cs-parser.jay"
{
		MemberName name = MakeName ((MemberName) yyVals[0+yyTop]);
		Class c = new Class (current_namespace, current_class, name, (Modifiers) yyVals[-4+yyTop], (Attributes) yyVals[-5+yyTop]);
		if (((c.ModFlags & Modifiers.STATIC) != 0) && lang_version == LanguageVersion.ISO_1) {
			FeatureIsNotAvailable (c.Location, "static classes");
		}
		
		push_current_class (c, yyVals[-3+yyTop]);
		lbag.AddMember (current_class, GetModifierLocations (), GetLocation (yyVals[-2+yyTop]));
	  }

void case_631()
///  #line 4347 "cs-parser.jay"
{
		lexer.ConstraintsParsing = false;

		current_class.SetParameterInfo ((List<Constraints>) yyVals[0+yyTop]);

		if (doc_support) {
			current_container.DocComment = Lexer.consume_doc_comment ();
			Lexer.doc_state = XmlCommentState.Allowed;
		}
		
		lexer.parsing_modifiers = true;
	  }

void case_632()
///  #line 4360 "cs-parser.jay"
{
		--lexer.parsing_declaration;
		if (doc_support)
			Lexer.doc_state = XmlCommentState.Allowed;
	  }

void case_633()
///  #line 4366 "cs-parser.jay"
{
		lbag.AppendToMember (current_class, GetLocation (yyVals[-4+yyTop]), GetLocation (yyVals[-2+yyTop]));
		if (yyVals[0+yyTop] != null) 
			current_class.OptionalSemicolon = GetLocation (yyVals[0+yyTop]);
		yyVal = pop_current_class ();
	  }

void case_636()
///  #line 4383 "cs-parser.jay"
{
	    mod_locations = null;
		yyVal = ModifierNone;
		lexer.parsing_modifiers = false;
	  }

void case_639()
///  #line 4397 "cs-parser.jay"
{ 
		var m1 = (Modifiers) yyVals[-1+yyTop];
		var m2 = (Modifiers) yyVals[0+yyTop];

		if ((m1 & m2) != 0) {
			report.Error (1004, lexer.Location - ModifiersExtensions.Name (m2).Length,
				"Duplicate `{0}' modifier", ModifiersExtensions.Name (m2));
		} else if ((m2 & Modifiers.AccessibilityMask) != 0 && (m1 & Modifiers.AccessibilityMask) != 0 &&
			((m2 | m1 & Modifiers.AccessibilityMask) != (Modifiers.PROTECTED | Modifiers.INTERNAL))) {
			report.Error (107, lexer.Location - ModifiersExtensions.Name (m2).Length,
				"More than one protection modifier specified");
		}
		
		yyVal = m1 | m2;
	  }

void case_640()
///  #line 4416 "cs-parser.jay"
{
		yyVal = Modifiers.NEW;
		StoreModifierLocation (yyVal, GetLocation (yyVals[0+yyTop]));
		
		if (current_container == module)
			report.Error (1530, GetLocation (yyVals[0+yyTop]), "Keyword `new' is not allowed on namespace elements");
	  }

void case_641()
///  #line 4424 "cs-parser.jay"
{
		yyVal = Modifiers.PUBLIC;
		StoreModifierLocation (yyVal, GetLocation (yyVals[0+yyTop]));
	  }

void case_642()
///  #line 4429 "cs-parser.jay"
{
		yyVal = Modifiers.PROTECTED;
		StoreModifierLocation (yyVal, GetLocation (yyVals[0+yyTop]));
	  }

void case_643()
///  #line 4434 "cs-parser.jay"
{
		yyVal = Modifiers.INTERNAL;
		StoreModifierLocation (yyVal, GetLocation (yyVals[0+yyTop]));
	  }

void case_644()
///  #line 4439 "cs-parser.jay"
{
		yyVal = Modifiers.PRIVATE;
		StoreModifierLocation (yyVal, GetLocation (yyVals[0+yyTop]));
	  }

void case_645()
///  #line 4444 "cs-parser.jay"
{
		yyVal = Modifiers.ABSTRACT;
		StoreModifierLocation (yyVal, GetLocation (yyVals[0+yyTop]));
	  }

void case_646()
///  #line 4449 "cs-parser.jay"
{
		yyVal = Modifiers.SEALED;
		StoreModifierLocation (yyVal, GetLocation (yyVals[0+yyTop]));
	  }

void case_647()
///  #line 4454 "cs-parser.jay"
{
		yyVal = Modifiers.STATIC;
		StoreModifierLocation (yyVal, GetLocation (yyVals[0+yyTop]));
	  }

void case_648()
///  #line 4459 "cs-parser.jay"
{
		yyVal = Modifiers.READONLY;
		StoreModifierLocation (yyVal, GetLocation (yyVals[0+yyTop]));
	  }

void case_649()
///  #line 4464 "cs-parser.jay"
{
		yyVal = Modifiers.VIRTUAL;
		StoreModifierLocation (yyVal, GetLocation (yyVals[0+yyTop]));
	  }

void case_650()
///  #line 4469 "cs-parser.jay"
{
		yyVal = Modifiers.OVERRIDE;
		StoreModifierLocation (yyVal, GetLocation (yyVals[0+yyTop]));
	  }

void case_651()
///  #line 4474 "cs-parser.jay"
{
		yyVal = Modifiers.EXTERN;
		StoreModifierLocation (yyVal, GetLocation (yyVals[0+yyTop]));
	  }

void case_652()
///  #line 4479 "cs-parser.jay"
{
		yyVal = Modifiers.VOLATILE;
		StoreModifierLocation (yyVal, GetLocation (yyVals[0+yyTop]));
	  }

void case_653()
///  #line 4484 "cs-parser.jay"
{
		yyVal = Modifiers.UNSAFE;
		StoreModifierLocation (yyVal, GetLocation (yyVals[0+yyTop]));
		if (!settings.Unsafe)
			Error_UnsafeCodeNotAllowed (GetLocation (yyVals[0+yyTop]));
	  }

void case_654()
///  #line 4491 "cs-parser.jay"
{
		yyVal = Modifiers.ASYNC;
		StoreModifierLocation (yyVal, GetLocation (yyVals[0+yyTop]));
	  }

void case_656()
///  #line 4500 "cs-parser.jay"
{
		lbag.AppendToMember (current_class, GetLocation (yyVals[-1+yyTop]));
		current_container.AddBasesForPart (current_class, (List<FullNamedExpression>) yyVals[0+yyTop]);
	 }

void case_659()
///  #line 4513 "cs-parser.jay"
{
		Error_SyntaxError (yyToken);
		yyVal = null;
	 }

void case_660()
///  #line 4521 "cs-parser.jay"
{
		var constraints = new List<Constraints> (1);
		constraints.Add ((Constraints) yyVals[0+yyTop]);
		yyVal = constraints;
	  }

void case_661()
///  #line 4527 "cs-parser.jay"
{
		var constraints = (List<Constraints>) yyVals[-1+yyTop];
		Constraints new_constraint = (Constraints)yyVals[0+yyTop];

		foreach (Constraints c in constraints) {
			if (new_constraint.TypeParameter.Value == c.TypeParameter.Value) {
				report.Error (409, new_constraint.Location,
					"A constraint clause has already been specified for type parameter `{0}'",
					new_constraint.TypeParameter.Value);
			}
		}

		constraints.Add (new_constraint);
		yyVal = constraints;
	  }

void case_662()
///  #line 4546 "cs-parser.jay"
{
		var lt = (Tokenizer.LocatedToken) yyVals[-2+yyTop];
		yyVal = new Constraints (new SimpleMemberName (lt.Value, lt.Location), (List<FullNamedExpression>) yyVals[0+yyTop], GetLocation (yyVals[-3+yyTop]));
		lbag.AddLocation (yyVal, GetLocation (yyVals[-1+yyTop]));
	  }

void case_663()
///  #line 4555 "cs-parser.jay"
{
		var constraints = new List<FullNamedExpression> (1);
		constraints.Add ((FullNamedExpression) yyVals[0+yyTop]);
		yyVal = constraints;
	  }

void case_664()
///  #line 4561 "cs-parser.jay"
{
		var constraints = (List<FullNamedExpression>) yyVals[-2+yyTop];
		var prev = constraints [constraints.Count - 1] as SpecialContraintExpr;
		if (prev != null && (prev.Constraint & SpecialConstraint.Constructor) != 0) {			
			report.Error (401, GetLocation (yyVals[-1+yyTop]), "The `new()' constraint must be the last constraint specified");
		}
		
		prev = yyVals[0+yyTop] as SpecialContraintExpr;
		if (prev != null) {
			if ((prev.Constraint & (SpecialConstraint.Class | SpecialConstraint.Struct)) != 0) {
				report.Error (449, prev.Location, "The `class' or `struct' constraint must be the first constraint specified");			
			} else {
			 	prev = constraints [0] as SpecialContraintExpr;
			 	if (prev != null && (prev.Constraint & SpecialConstraint.Struct) != 0) {			
					report.Error (451, GetLocation (yyVals[0+yyTop]), "The `new()' constraint cannot be used with the `struct' constraint");
				}
			}
		}

		constraints.Add ((FullNamedExpression) yyVals[0+yyTop]);
		lbag.AppendTo (constraints, GetLocation (yyVals[-1+yyTop]));
		yyVal = constraints;
	  }

void case_665()
///  #line 4588 "cs-parser.jay"
{
		if (yyVals[0+yyTop] is ComposedCast)
			report.Error (706, GetLocation (yyVals[0+yyTop]), "Invalid constraint type `{0}'", ((ComposedCast)yyVals[0+yyTop]).GetSignatureForError ());
	  
	  	yyVal = yyVals[0+yyTop];
	  }

void case_666()
///  #line 4595 "cs-parser.jay"
{
		yyVal = new SpecialContraintExpr (SpecialConstraint.Constructor, GetLocation (yyVals[-2+yyTop]));
		lbag.AddLocation (yyVal, GetLocation (yyVals[-1+yyTop]), GetLocation (yyVals[0+yyTop]));
	  }

void case_670()
///  #line 4615 "cs-parser.jay"
{
		if (lang_version <= LanguageVersion.V_3)
			FeatureIsNotAvailable (lexer.Location, "generic type variance");

		yyVal = yyVals[0+yyTop];
	  }

void case_671()
///  #line 4625 "cs-parser.jay"
{
		yyVal = Variance.Covariant;
		savedLocation = GetLocation (yyVals[0+yyTop]);
	  }

void case_672()
///  #line 4630 "cs-parser.jay"
{
		yyVal = Variance.Contravariant;
		savedLocation = GetLocation (yyVals[0+yyTop]);
	  }

void case_673()
///  #line 4651 "cs-parser.jay"
{
		++lexer.parsing_block;
		start_block (GetLocation (yyVals[0+yyTop]));
	  }

void case_675()
///  #line 4663 "cs-parser.jay"
{
	 	--lexer.parsing_block;
		yyVal = end_block (GetLocation (yyVals[0+yyTop]));
	  }

void case_676()
///  #line 4668 "cs-parser.jay"
{
	 	--lexer.parsing_block;
		yyVal = end_block (lexer.Location);
	  }

void case_677()
///  #line 4677 "cs-parser.jay"
{
		++lexer.parsing_block;
		current_block.StartLocation = GetLocation (yyVals[0+yyTop]);
	  }

void case_678()
///  #line 4682 "cs-parser.jay"
{
		--lexer.parsing_block;
		yyVal = end_block (GetLocation (yyVals[0+yyTop]));
	  }

void case_686()
///  #line 4710 "cs-parser.jay"
{
		Error_SyntaxError (yyToken);
		var lt =(Tokenizer.LocatedToken) yyVals[-1+yyTop];
		var sn = new SimpleName (lt.Value, lt.Location);
		current_block.AddStatement(new StatementErrorExpression (sn));
		yyVal = null;
	}

void case_687()
///  #line 4719 "cs-parser.jay"
{
		Error_SyntaxError (yyToken);
		yyVal = null;
	  }

void case_720()
///  #line 4783 "cs-parser.jay"
{
		  report.Error (1023, GetLocation (yyVals[0+yyTop]), "An embedded statement may not be a declaration or labeled statement");
		  yyVal = null;
	  }

void case_721()
///  #line 4788 "cs-parser.jay"
{
		  report.Error (1023, GetLocation (yyVals[0+yyTop]), "An embedded statement may not be a declaration or labeled statement");
		  yyVal = null;
	  }

void case_722()
///  #line 4793 "cs-parser.jay"
{
		Error_SyntaxError (yyToken);
		yyVal = new EmptyStatement (GetLocation (yyVals[0+yyTop]));
	  }

void case_723()
///  #line 4801 "cs-parser.jay"
{
		/* Uses lexer.Location because semicolon location is not kept in quick mode*/
		yyVal = new EmptyStatement (lexer.Location);
	  }

void case_724()
///  #line 4809 "cs-parser.jay"
{
		var lt = (Tokenizer.LocatedToken) yyVals[-1+yyTop];
		LabeledStatement labeled = new LabeledStatement (lt.Value, current_block, lt.Location);
		lbag.AddLocation (labeled, GetLocation (yyVals[0+yyTop]));
		current_block.AddLabel (labeled);
		current_block.AddStatement (labeled);
	  }

void case_727()
///  #line 4822 "cs-parser.jay"
{
		if (yyVals[-1+yyTop] is VarExpr)
			yyVals[-1+yyTop] = new SimpleName ("var", ((VarExpr) yyVals[-1+yyTop]).Location);
	  
		yyVal = new ComposedCast ((FullNamedExpression) yyVals[-1+yyTop], (ComposedTypeSpecifier) yyVals[0+yyTop]);
	  }

void case_728()
///  #line 4838 "cs-parser.jay"
{ 
		/* Ok, the above "primary_expression" is there to get rid of*/
		/* both reduce/reduce and shift/reduces in the grammar, it should*/
		/* really just be "type_name".  If you use type_name, a reduce/reduce*/
		/* creeps up.  If you use namespace_or_type_name (which is all we need*/
		/* really) two shift/reduces appear.*/
		/* */

		/* So the super-trick is that primary_expression*/
		/* can only be either a SimpleName or a MemberAccess. */
		/* The MemberAccess case arises when you have a fully qualified type-name like :*/
		/* Foo.Bar.Blah i;*/
		/* SimpleName is when you have*/
		/* Blah i;*/
		
		Expression expr = (Expression) yyVals[-1+yyTop];
		if (yyVals[0+yyTop] == null) {
			SimpleName sn = expr as SimpleName;
			if (sn != null && sn.Name == "var")
				yyVal = new VarExpr (sn.Location);
			else
				yyVal = yyVals[-1+yyTop];
		} else if (expr is ATypeNameExpression) {
			yyVal = new ComposedCast ((ATypeNameExpression)expr, (ComposedTypeSpecifier) yyVals[0+yyTop]);
		} else {
			Error_ExpectingTypeName (expr);
			yyVal = null;
		}
	  }

void case_729()
///  #line 4868 "cs-parser.jay"
{
		ATypeNameExpression expr = yyVals[-1+yyTop] as ATypeNameExpression;

		if (expr != null) {
			yyVal = new ComposedCast (expr, (ComposedTypeSpecifier) yyVals[0+yyTop]);
		} else {
			Error_ExpectingTypeName ((Expression)yyVals[-1+yyTop]);
			yyVal = expr;
		}
	  }

void case_730()
///  #line 4879 "cs-parser.jay"
{
		if (yyVals[0+yyTop] == null)
			yyVal = yyVals[-1+yyTop];
		else
			yyVal = new ComposedCast ((FullNamedExpression) yyVals[-1+yyTop], (ComposedTypeSpecifier) yyVals[0+yyTop]);
	  }

void case_733()
///  #line 4894 "cs-parser.jay"
{
		Expression.Error_VoidInvalidInTheContext (GetLocation (yyVals[0+yyTop]), report);
		yyVal = new TypeExpression (compiler.BuiltinTypes.Void, GetLocation (yyVals[0+yyTop]));
	  }

void case_735()
///  #line 4903 "cs-parser.jay"
{
	  	((ComposedTypeSpecifier) yyVals[-1+yyTop]).Next = (ComposedTypeSpecifier) yyVals[0+yyTop];
	  	yyVal = yyVals[-1+yyTop];
	  }

void case_738()
///  #line 4919 "cs-parser.jay"
{
		if (async_block) {
			report.Error (4003, GetLocation (yyVals[0+yyTop]), "`await' cannot be used as an identifier within an async method or lambda expression");
			yyVal = Tokenizer.LocatedToken.Create ("await", GetLocation (yyVals[0+yyTop]));
		}
	  }

void case_739()
///  #line 4929 "cs-parser.jay"
{
		var lt = (Tokenizer.LocatedToken) yyVals[0+yyTop];
		var li = new LocalVariable (current_block, lt.Value, lt.Location);
		current_block.AddLocalName (li);
		current_variable = new BlockVariableDeclaration ((FullNamedExpression) yyVals[-1+yyTop], li);
	  }

void case_740()
///  #line 4936 "cs-parser.jay"
{
		yyVal = current_variable;
		current_variable = null;
		lbag.AppendTo (yyVal, GetLocation (yyVals[0+yyTop]));
	  }

void case_741()
///  #line 4942 "cs-parser.jay"
{
		var lt = (Tokenizer.LocatedToken) yyVals[0+yyTop];
		var li = new LocalVariable (current_block, lt.Value, LocalVariable.Flags.Constant, lt.Location);
		current_block.AddLocalName (li);
		current_variable = new BlockConstantDeclaration ((FullNamedExpression) yyVals[-1+yyTop], li);
	  }

void case_742()
///  #line 4949 "cs-parser.jay"
{
		if (current_variable.Initializer != null) {
			lbag.AddLocation (current_variable, GetLocation (yyVals[-6+yyTop]), savedLocation, GetLocation (yyVals[0+yyTop]));
		} else {
			lbag.AddLocation (current_variable, GetLocation (yyVals[-6+yyTop]), GetLocation (yyVals[0+yyTop]));
		}
		yyVal = current_variable;;
		current_variable = null;
	  }

void case_744()
///  #line 4963 "cs-parser.jay"
{
		current_variable.Initializer = (Expression) yyVals[0+yyTop];
		lbag.AppendTo (current_variable, GetLocation (yyVals[-1+yyTop]));
	  }

void case_745()
///  #line 4968 "cs-parser.jay"
{
	  	if (yyToken == Token.OPEN_BRACKET_EXPR) {
			report.Error (650, lexer.Location,
				"Syntax error, bad array declarator. To declare a managed array the rank specifier precedes the variable's identifier. To declare a fixed size buffer field, use the fixed keyword before the field type");
			current_variable.Initializer = ErrorExpression.Create (650, lexer.Location,
					"Syntax error, bad array declarator. To declare a managed array the rank specifier precedes the variable's identifier. To declare a fixed size buffer field, use the fixed keyword before the field type");
		} else {
			Error_SyntaxError (yyToken);
			current_variable.Initializer = ErrorExpression.Create (0, lexer.Location,
					"Syntax error");
		}
		lbag.AppendTo (current_variable, GetLocation (yyVals[-1+yyTop]));
	  }

void case_746()
///  #line 4982 "cs-parser.jay"
{
	  	if (yyToken == Token.OPEN_BRACKET_EXPR) {
			report.Error (650, lexer.Location,
				"Syntax error, bad array declarator. To declare a managed array the rank specifier precedes the variable's identifier. To declare a fixed size buffer field, use the fixed keyword before the field type");
		} else {
			Error_SyntaxError (yyToken);
		}
	  }

void case_750()
///  #line 5000 "cs-parser.jay"
{
		foreach (var d in current_variable.Declarators) {
			if (d.Initializer == null)
				Error_MissingInitializer (d.Variable.Location);
		}
	  }

void case_753()
///  #line 5015 "cs-parser.jay"
{
		var lt = (Tokenizer.LocatedToken) yyVals[0+yyTop];	  
		var li = new LocalVariable (current_variable.Variable, lt.Value, lt.Location);
		var d = new BlockVariableDeclaration.Declarator (li, null);
		current_variable.AddDeclarator (d);
		current_block.AddLocalName (li);
	  	lbag.AddLocation (d, GetLocation (yyVals[-1+yyTop]));
	  }

void case_754()
///  #line 5024 "cs-parser.jay"
{
		var lt = (Tokenizer.LocatedToken) yyVals[-2+yyTop];	  
		var li = new LocalVariable (current_variable.Variable, lt.Value, lt.Location);
		var d = new BlockVariableDeclaration.Declarator (li, (Expression) yyVals[0+yyTop]);
		current_variable.AddDeclarator (d);
		current_block.AddLocalName (li);
	  	lbag.AddLocation (d, GetLocation (yyVals[-3+yyTop]), GetLocation (yyVals[-1+yyTop]));
	  }

void case_756()
///  #line 5040 "cs-parser.jay"
{
		savedLocation = GetLocation (yyVals[-1+yyTop]);
		current_variable.Initializer = (Expression) yyVals[0+yyTop];
	  }

void case_761()
///  #line 5058 "cs-parser.jay"
{
		var lt = (Tokenizer.LocatedToken) yyVals[-2+yyTop];	  
		var li = new LocalVariable (current_block, lt.Value, LocalVariable.Flags.Constant, lt.Location);
		var d = new BlockVariableDeclaration.Declarator (li, (Expression) yyVals[0+yyTop]);
		current_variable.AddDeclarator (d);
		current_block.AddLocalName (li);
	  	lbag.AddLocation (d, GetLocation (yyVals[-3+yyTop]), GetLocation (yyVals[-1+yyTop]));
	  }

void case_763()
///  #line 5071 "cs-parser.jay"
{
		yyVal = new StackAlloc ((Expression) yyVals[-3+yyTop], (Expression) yyVals[-1+yyTop], GetLocation (yyVals[-4+yyTop]));
		lbag.AddLocation (yyVal, GetLocation (yyVals[-2+yyTop]), GetLocation (yyVals[0+yyTop]));
	  }

void case_764()
///  #line 5076 "cs-parser.jay"
{
		report.Error (1575, GetLocation (yyVals[-1+yyTop]), "A stackalloc expression requires [] after type");
		yyVal = new StackAlloc ((Expression) yyVals[0+yyTop], null, GetLocation (yyVals[-1+yyTop]));		
	  }

void case_765()
///  #line 5084 "cs-parser.jay"
{
		yyVal = yyVals[-1+yyTop];
		lbag.AddStatement (yyVal, GetLocation (yyVals[0+yyTop]));
	  }

void case_769()
///  #line 5102 "cs-parser.jay"
{
		ExpressionStatement s = yyVals[0+yyTop] as ExpressionStatement;
		if (s == null) {
			Expression.Error_InvalidExpressionStatement (report, GetLocation (yyVals[0+yyTop]));
			yyVal = new InvalidStatementExpression (yyVals[0+yyTop] as Expression);
		} else {
			yyVal = new StatementExpression (s);
		}
	  }

void case_770()
///  #line 5115 "cs-parser.jay"
{
		Expression expr = (Expression) yyVals[0+yyTop];
		ExpressionStatement s;

	        s = new OptionalAssign (new SimpleName ("$retval", lexer.Location), expr, lexer.Location);
		yyVal = new StatementExpression (s);
	  }

void case_771()
///  #line 5123 "cs-parser.jay"
{
		Error_SyntaxError (yyToken);
		yyVal = new EmptyStatement (GetLocation (yyVals[0+yyTop]));
	  }

void case_774()
///  #line 5137 "cs-parser.jay"
{ 
		if (yyVals[0+yyTop] is EmptyStatement)
			Warning_EmptyStatement (GetLocation (yyVals[0+yyTop]));
		
		yyVal = new If ((BooleanExpression) yyVals[-2+yyTop], (Statement) yyVals[0+yyTop], GetLocation (yyVals[-4+yyTop]));
		lbag.AddStatement (yyVal, GetLocation (yyVals[-3+yyTop]), GetLocation (yyVals[-1+yyTop]));
	  }

void case_775()
///  #line 5146 "cs-parser.jay"
{
		yyVal = new If ((BooleanExpression) yyVals[-4+yyTop], (Statement) yyVals[-2+yyTop], (Statement) yyVals[0+yyTop], GetLocation (yyVals[-6+yyTop]));
		lbag.AddStatement (yyVal, GetLocation (yyVals[-5+yyTop]), GetLocation (yyVals[-3+yyTop]), GetLocation (yyVals[-1+yyTop]));
		
		if (yyVals[-2+yyTop] is EmptyStatement)
			Warning_EmptyStatement (GetLocation (yyVals[-2+yyTop]));
		if (yyVals[0+yyTop] is EmptyStatement)
			Warning_EmptyStatement (GetLocation (yyVals[0+yyTop]));
	  }

void case_776()
///  #line 5155 "cs-parser.jay"
{
		var eloc = GetLocation (yyVals[-1+yyTop]);
		report.Error (1026, eloc, "Expected a ')'");
		yyVal = new If ((BooleanExpression) yyVals[-1+yyTop], null, GetLocation (yyVals[-3+yyTop]));
		lbag.AddStatement (yyVal, GetLocation (yyVals[-2+yyTop]));
	  }

void case_778()
///  #line 5169 "cs-parser.jay"
{
		yyVal = new Switch ((Expression) yyVals[-5+yyTop], (ExplicitBlock) current_block.Explicit, (List<SwitchSection>) yyVals[-1+yyTop], GetLocation (yyVals[-7+yyTop]));	
		end_block (GetLocation (yyVals[0+yyTop]));
		lbag.AddStatement (yyVal, GetLocation (yyVals[-6+yyTop]), GetLocation (yyVals[-4+yyTop]), GetLocation (yyVals[-3+yyTop]), GetLocation (yyVals[0+yyTop]));
	  }

void case_779()
///  #line 5178 "cs-parser.jay"
{
		report.Warning (1522, 1, current_block.StartLocation, "Empty switch block"); 
		yyVal = new List<SwitchSection> ();
	  }

void case_781()
///  #line 5187 "cs-parser.jay"
{
		var sections = new List<SwitchSection> (4);

		sections.Add ((SwitchSection) yyVals[0+yyTop]);
		yyVal = sections;
	  }

void case_782()
///  #line 5194 "cs-parser.jay"
{
		var sections = (List<SwitchSection>) yyVals[-1+yyTop];

		sections.Add ((SwitchSection) yyVals[0+yyTop]);
		yyVal = sections;
	  }

void case_783()
///  #line 5201 "cs-parser.jay"
{
		Error_SyntaxError (yyToken);	  
		yyVal = new List<SwitchSection> ();
	  }

void case_786()
///  #line 5220 "cs-parser.jay"
{
		var labels = new List<SwitchLabel> (2);

		labels.Add ((SwitchLabel) yyVals[0+yyTop]);
		yyVal = labels;
	  }

void case_787()
///  #line 5227 "cs-parser.jay"
{
		var labels = (List<SwitchLabel>) (yyVals[-1+yyTop]);
		labels.Add ((SwitchLabel) yyVals[0+yyTop]);

		yyVal = labels;
	  }

void case_788()
///  #line 5237 "cs-parser.jay"
{
	 	yyVal = new SwitchLabel ((Expression) yyVals[-1+yyTop], GetLocation (yyVals[-2+yyTop]));
	 	lbag.AddLocation (yyVal, GetLocation (yyVals[0+yyTop]));
	 }

void case_794()
///  #line 5256 "cs-parser.jay"
{
		if (yyVals[0+yyTop] is EmptyStatement && lexer.peek_token () == Token.OPEN_BRACE)
			Warning_EmptyStatement (GetLocation (yyVals[0+yyTop]));
	  
		yyVal = new While ((BooleanExpression) yyVals[-2+yyTop], (Statement) yyVals[0+yyTop], GetLocation (yyVals[-4+yyTop]));
		lbag.AddStatement (yyVal, GetLocation (yyVals[-3+yyTop]), GetLocation (yyVals[-1+yyTop]));
	  }

void case_795()
///  #line 5268 "cs-parser.jay"
{
		yyVal = new Do ((Statement) yyVals[-5+yyTop], (BooleanExpression) yyVals[-2+yyTop], GetLocation (yyVals[-6+yyTop]));
		lbag.AddStatement (yyVal, GetLocation (yyVals[-4+yyTop]), GetLocation (yyVals[-3+yyTop]), GetLocation (yyVals[-1+yyTop]), GetLocation (yyVals[0+yyTop]));
	  }

void case_796()
///  #line 5273 "cs-parser.jay"
{
		var loc = GetLocation (yyVals[-2+yyTop]);
		report.Error (-100, loc, "Expected `while'");
		yyVal = new Do ((Statement) yyVals[-1+yyTop], null, loc);
	  }

void case_797()
///  #line 5280 "cs-parser.jay"
{
		yyVal = new Do ((Statement) yyVals[-4+yyTop], (BooleanExpression) yyVals[-1+yyTop], GetLocation (yyVals[-5+yyTop]));
		lbag.AddStatement (yyVal, GetLocation (yyVals[-3+yyTop]), GetLocation (yyVals[-2+yyTop]));
	  }

void case_798()
///  #line 5288 "cs-parser.jay"
{
		start_block (GetLocation (yyVals[0+yyTop]));
		current_block.IsCompilerGenerated = true;
		For f = new For (GetLocation (yyVals[-1+yyTop]));
		current_block.AddStatement (f);
		yyVal = f;
	  }

void case_803()
///  #line 5316 "cs-parser.jay"
{
		if (yyVals[0+yyTop] is EmptyStatement && lexer.peek_token () == Token.OPEN_BRACE)
			Warning_EmptyStatement (GetLocation (yyVals[0+yyTop]));
	  
		For f = ((For) yyVals[-10+yyTop]);
		f.Statement = (Statement) yyVals[0+yyTop];
		lbag.AddStatement (f, current_block.StartLocation, GetLocation (yyVals[-8+yyTop]), GetLocation (yyVals[-5+yyTop]), GetLocation (yyVals[-2+yyTop]));

		yyVal = end_block (GetLocation (yyVals[-8+yyTop]));
	  }

void case_804()
///  #line 5327 "cs-parser.jay"
{
		Error_SyntaxError (yyToken);
		yyVal = end_block (current_block.StartLocation);
	  }

void case_807()
///  #line 5340 "cs-parser.jay"
{
		var lt = (Tokenizer.LocatedToken) yyVals[0+yyTop];
		var li = new LocalVariable (current_block, lt.Value, lt.Location);
		current_block.AddLocalName (li);
		current_variable = new BlockVariableDeclaration ((FullNamedExpression) yyVals[-1+yyTop], li);
	  }

void case_808()
///  #line 5347 "cs-parser.jay"
{
		yyVal = current_variable;
		current_variable = null;
	  }

void case_816()
///  #line 5371 "cs-parser.jay"
{
	  	var sl = yyVals[-2+yyTop] as StatementList;
	  	if (sl == null) {
	  		sl = new StatementList ((Statement) yyVals[-2+yyTop], (Statement) yyVals[0+yyTop]);
			lbag.AddStatement (sl, GetLocation (yyVals[-1+yyTop]));
	  	} else {
	  		sl.Add ((Statement) yyVals[0+yyTop]);
	  		lbag.AppendTo (sl, GetLocation (yyVals[-1+yyTop]));
	  	}
	  		
		yyVal = sl;
	  }

void case_817()
///  #line 5387 "cs-parser.jay"
{
		report.Error (230, GetLocation (yyVals[-5+yyTop]), "Type and identifier are both required in a foreach statement");
		yyVal = null;
	  }

void case_818()
///  #line 5392 "cs-parser.jay"
{
		start_block (GetLocation (yyVals[-5+yyTop]));
		current_block.IsCompilerGenerated = true;
		var lt = (Tokenizer.LocatedToken) yyVals[-3+yyTop];
		var li = new LocalVariable (current_block, lt.Value, LocalVariable.Flags.ForeachVariable | LocalVariable.Flags.Used, lt.Location);
		current_block.AddLocalName (li);
		yyVal = li;
	  }

void case_819()
///  #line 5401 "cs-parser.jay"
{
		if (yyVals[0+yyTop] is EmptyStatement && lexer.peek_token () == Token.OPEN_BRACE)
			Warning_EmptyStatement (GetLocation (yyVals[0+yyTop]));
	  
		Foreach f = new Foreach ((Expression) yyVals[-6+yyTop], (LocalVariable) yyVals[-1+yyTop], (Expression) yyVals[-3+yyTop], (Statement) yyVals[0+yyTop], GetLocation (yyVals[-8+yyTop]));
		current_block.AddStatement (f);
		
		lbag.AddStatement (f, GetLocation (yyVals[-7+yyTop]), GetLocation (yyVals[-4+yyTop]), GetLocation (yyVals[-2+yyTop]));
		yyVal = end_block (GetLocation (yyVals[-2+yyTop]));
	  }

void case_820()
///  #line 5412 "cs-parser.jay"
{
		start_block (GetLocation (yyVals[-3+yyTop]));
		current_block.IsCompilerGenerated = true;
		var lt = yyVals[-1+yyTop] as Tokenizer.LocatedToken;
		var li = lt != null ? new LocalVariable (current_block, lt.Value, LocalVariable.Flags.ForeachVariable | LocalVariable.Flags.Used, lt.Location) : null;
		
		Foreach f = new Foreach ((Expression) yyVals[-2+yyTop], li, null, null, GetLocation (yyVals[-4+yyTop]));
		current_block.AddStatement (f);
		
		lbag.AddStatement (f, GetLocation (yyVals[-3+yyTop]));
		yyVal = end_block (GetLocation (yyVals[0+yyTop]));
	  }

void case_821()
///  #line 5425 "cs-parser.jay"
{
		Foreach f = new Foreach ((Expression) yyVals[-1+yyTop], null, null, null, GetLocation (yyVals[-3+yyTop]));
		current_block.AddStatement (f);
		
		lbag.AddStatement (f, GetLocation (yyVals[-2+yyTop]));
		yyVal = f;
	  }

void case_828()
///  #line 5445 "cs-parser.jay"
{
		yyVal = new Break (GetLocation (yyVals[-1+yyTop]));
		lbag.AddStatement (yyVal, GetLocation (yyVals[0+yyTop]));
	  }

void case_829()
///  #line 5453 "cs-parser.jay"
{
		yyVal = new Continue (GetLocation (yyVals[-1+yyTop]));
		lbag.AddStatement (yyVal, GetLocation (yyVals[0+yyTop]));
	  }

void case_830()
///  #line 5461 "cs-parser.jay"
{
		var lt = (Tokenizer.LocatedToken) yyVals[-1+yyTop];
		yyVal = new Goto (lt.Value, GetLocation (yyVals[-2+yyTop]));
		lbag.AddStatement (yyVal, GetLocation (yyVals[-1+yyTop]), GetLocation (yyVals[0+yyTop]));
	  }

void case_831()
///  #line 5467 "cs-parser.jay"
{
		yyVal = new GotoCase ((Expression) yyVals[-1+yyTop], GetLocation (yyVals[-3+yyTop]));
		lbag.AddStatement (yyVal, GetLocation (yyVals[-2+yyTop]), GetLocation (yyVals[0+yyTop]));
	  }

void case_832()
///  #line 5472 "cs-parser.jay"
{
		yyVal = new GotoDefault (GetLocation (yyVals[-2+yyTop]));
		lbag.AddStatement (yyVal, GetLocation (yyVals[-1+yyTop]), GetLocation (yyVals[0+yyTop]));
	  }

void case_833()
///  #line 5480 "cs-parser.jay"
{
		yyVal = new Return ((Expression) yyVals[-1+yyTop], GetLocation (yyVals[-2+yyTop]));
		lbag.AddStatement (yyVal, GetLocation (yyVals[0+yyTop]));
	  }

void case_834()
///  #line 5488 "cs-parser.jay"
{
		yyVal = new Throw ((Expression) yyVals[-1+yyTop], GetLocation (yyVals[-2+yyTop]));
		lbag.AddStatement (yyVal, GetLocation (yyVals[0+yyTop]));
	  }

void case_835()
///  #line 5496 "cs-parser.jay"
{
		var lt = (Tokenizer.LocatedToken) yyVals[-3+yyTop];
		string s = lt.Value;
		if (s != "yield"){
			report.Error (1003, lt.Location, "; expected");
		} else if (yyVals[-1+yyTop] == null) {
			report.Error (1627, GetLocation (yyVals[0+yyTop]), "Expression expected after yield return");
		} else if (lang_version == LanguageVersion.ISO_1){
			FeatureIsNotAvailable (lt.Location, "iterators");
		}
		
		current_block.Explicit.RegisterIteratorYield ();
		yyVal = new Yield ((Expression) yyVals[-1+yyTop], lt.Location);
		lbag.AddStatement (yyVal, GetLocation (yyVals[-2+yyTop]), GetLocation (yyVals[0+yyTop]));
	  }

void case_836()
///  #line 5512 "cs-parser.jay"
{
		var lt = (Tokenizer.LocatedToken) yyVals[-2+yyTop];
		string s = lt.Value;
		if (s != "yield"){
			report.Error (1003, lt.Location, "; expected");
		} else if (lang_version == LanguageVersion.ISO_1){
			FeatureIsNotAvailable (lt.Location, "iterators");
		}
		
		current_block.Explicit.RegisterIteratorYield ();
		yyVal = new YieldBreak (lt.Location);
		lbag.AddStatement (yyVal, GetLocation (yyVals[-1+yyTop]), GetLocation (yyVals[0+yyTop]));
	  }

void case_840()
///  #line 5538 "cs-parser.jay"
{
		yyVal = new TryFinally ((Statement) yyVals[-2+yyTop], (Block) yyVals[0+yyTop], GetLocation (yyVals[-3+yyTop]));
		lbag.AddStatement (yyVal, GetLocation (yyVals[-1+yyTop]));
	  }

void case_841()
///  #line 5543 "cs-parser.jay"
{
		yyVal = new TryFinally (new TryCatch ((Block) yyVals[-3+yyTop], (List<Catch>) yyVals[-2+yyTop], GetLocation (yyVals[-4+yyTop]), true), (Block) yyVals[0+yyTop], GetLocation (yyVals[-4+yyTop]));
		lbag.AddStatement (yyVal, GetLocation (yyVals[-1+yyTop]));
	  }

void case_842()
///  #line 5548 "cs-parser.jay"
{
		report.Error (1524, GetLocation (yyVals[-2+yyTop]), "Expected catch or finally");
		yyVal = new TryCatch ((Block) yyVals[-1+yyTop], null, GetLocation (yyVals[-2+yyTop]), false);
	  }

void case_843()
///  #line 5556 "cs-parser.jay"
{
		var l = new List<Catch> (2);

		l.Add ((Catch) yyVals[0+yyTop]);
		yyVal = l;
	  }

void case_844()
///  #line 5563 "cs-parser.jay"
{
		var l = (List<Catch>) yyVals[-1+yyTop];
		
		Catch c = (Catch) yyVals[0+yyTop];
		if (l [0].IsGeneral) {
			report.Error (1017, c.loc, "Try statement already has an empty catch block");
		} else {
			if (c.IsGeneral)
				l.Insert (0, c);
			else
				l.Add (c);
		}
		
		yyVal = l;
	  }

void case_848()
///  #line 5591 "cs-parser.jay"
{
		start_block (GetLocation (yyVals[-3+yyTop]));
		var c = new Catch (current_block, GetLocation (yyVals[-4+yyTop]));
		c.TypeExpression = (FullNamedExpression) yyVals[-2+yyTop];

		if (yyVals[-1+yyTop] != null) {
			var lt = (Tokenizer.LocatedToken) yyVals[-1+yyTop];
			c.Variable = new LocalVariable (current_block, lt.Value, lt.Location);
			current_block.AddLocalName (c.Variable);
		}

		lbag.AddLocation (c, GetLocation (yyVals[-3+yyTop]), GetLocation (yyVals[0+yyTop]));
		yyVal = c;
	  }

void case_850()
///  #line 5610 "cs-parser.jay"
{
	  	if (yyToken == Token.CLOSE_PARENS) {
			report.Error (1015, lexer.Location,
				"A type that derives from `System.Exception', `object', or `string' expected");
		} else {
			Error_SyntaxError (yyToken);
		}
		
		yyVal = new Catch (null, GetLocation (yyVals[-2+yyTop]));
	  }

void case_853()
///  #line 5638 "cs-parser.jay"
{
		if (!settings.Unsafe)
			Error_UnsafeCodeNotAllowed (GetLocation (yyVals[0+yyTop]));
	  }

void case_855()
///  #line 5648 "cs-parser.jay"
{
		if (yyVals[0+yyTop] is EmptyStatement && lexer.peek_token () == Token.OPEN_BRACE)
			Warning_EmptyStatement (GetLocation (yyVals[0+yyTop]));
	  
		yyVal = new Lock ((Expression) yyVals[-2+yyTop], (Statement) yyVals[0+yyTop], GetLocation (yyVals[-4+yyTop]));
		lbag.AddStatement (yyVal, GetLocation (yyVals[-3+yyTop]), GetLocation (yyVals[-1+yyTop]));
	  }

void case_856()
///  #line 5659 "cs-parser.jay"
{
	    start_block (GetLocation (yyVals[-2+yyTop]));
	    
		current_block.IsCompilerGenerated = true;
		var lt = (Tokenizer.LocatedToken) yyVals[0+yyTop];
		var li = new LocalVariable (current_block, lt.Value, LocalVariable.Flags.FixedVariable | LocalVariable.Flags.Used, lt.Location);
		current_block.AddLocalName (li);
		current_variable = new Fixed.VariableDeclaration ((FullNamedExpression) yyVals[-1+yyTop], li);
	  }

void case_857()
///  #line 5669 "cs-parser.jay"
{
		yyVal = current_variable;
		current_variable = null;
	  }

void case_858()
///  #line 5674 "cs-parser.jay"
{
		if (yyVals[0+yyTop] is EmptyStatement && lexer.peek_token () == Token.OPEN_BRACE)
			Warning_EmptyStatement (GetLocation (yyVals[0+yyTop]));
	  
		Fixed f = new Fixed ((Fixed.VariableDeclaration) yyVals[-1+yyTop], (Statement) yyVals[0+yyTop], GetLocation (yyVals[-9+yyTop]));
		current_block.AddStatement (f);
		lbag.AddStatement (f, GetLocation (yyVals[-8+yyTop]), GetLocation (yyVals[-2+yyTop]));
		yyVal = end_block (GetLocation (yyVals[-2+yyTop]));
	  }

void case_859()
///  #line 5687 "cs-parser.jay"
{
	    start_block (GetLocation (yyVals[-2+yyTop]));
	    
		current_block.IsCompilerGenerated = true;
		var lt = (Tokenizer.LocatedToken) yyVals[0+yyTop];
		var li = new LocalVariable (current_block, lt.Value, LocalVariable.Flags.UsingVariable | LocalVariable.Flags.Used, lt.Location);
		current_block.AddLocalName (li);
		current_variable = new Using.VariableDeclaration ((FullNamedExpression) yyVals[-1+yyTop], li);
	  }

void case_860()
///  #line 5697 "cs-parser.jay"
{
		yyVal = current_variable;	  
		current_variable = null;
	  }

void case_861()
///  #line 5702 "cs-parser.jay"
{
		if (yyVals[0+yyTop] is EmptyStatement && lexer.peek_token () == Token.OPEN_BRACE)
			Warning_EmptyStatement (GetLocation (yyVals[0+yyTop]));
	  
		Using u = new Using ((Using.VariableDeclaration) yyVals[-1+yyTop], (Statement) yyVals[0+yyTop], GetLocation (yyVals[-8+yyTop]));
		lbag.AddStatement (u, GetLocation (yyVals[-7+yyTop]), GetLocation (yyVals[-2+yyTop]));
		current_block.AddStatement (u);
		yyVal = end_block (GetLocation (yyVals[-2+yyTop]));
	  }

void case_862()
///  #line 5712 "cs-parser.jay"
{
		if (yyVals[0+yyTop] is EmptyStatement && lexer.peek_token () == Token.OPEN_BRACE)
			Warning_EmptyStatement (GetLocation (yyVals[0+yyTop]));
	  
		Using u = new Using ((Expression) yyVals[-2+yyTop], (Statement) yyVals[0+yyTop], GetLocation (yyVals[-4+yyTop]));
		lbag.AddStatement (u, GetLocation (yyVals[-3+yyTop]), GetLocation (yyVals[-1+yyTop]));
		yyVal = u;
	  }

void case_864()
///  #line 5725 "cs-parser.jay"
{
		/* It has to be here for the parent to safely restore artificial block*/
	  	Error_SyntaxError (yyToken);
	  }

void case_866()
///  #line 5737 "cs-parser.jay"
{
		current_variable.Initializer = (Expression) yyVals[0+yyTop];
		lbag.AppendTo (current_variable, GetLocation (yyVals[-1+yyTop]));
		yyVal = current_variable;
	  }

void case_867()
///  #line 5749 "cs-parser.jay"
{
		lexer.query_parsing = false;
			
		Linq.AQueryClause from = yyVals[-1+yyTop] as Linq.AQueryClause;
			
		from.Tail.Next = (Linq.AQueryClause)yyVals[0+yyTop];
		yyVal = from;
		
		current_block.SetEndLocation (lexer.Location);
		current_block = current_block.Parent;
	  }

void case_868()
///  #line 5761 "cs-parser.jay"
{
		Linq.AQueryClause from = yyVals[-1+yyTop] as Linq.AQueryClause;
			
		from.Tail.Next = (Linq.AQueryClause)yyVals[0+yyTop];
		yyVal = from;
		
		current_block.SetEndLocation (lexer.Location);
		current_block = current_block.Parent;
	  }

void case_869()
///  #line 5772 "cs-parser.jay"
{
	        lexer.query_parsing = false;
		yyVal = yyVals[-1+yyTop];

		current_block.SetEndLocation (lexer.Location);
		current_block = current_block.Parent;
	  }

void case_870()
///  #line 5779 "cs-parser.jay"
{
	        yyVal = yyVals[-1+yyTop];
		current_block.SetEndLocation (lexer.Location);
		current_block = current_block.Parent;
	  }

void case_871()
///  #line 5788 "cs-parser.jay"
{
		current_block = new Linq.QueryBlock (current_block, lexer.Location);
	  
		var lt = (Tokenizer.LocatedToken) yyVals[-2+yyTop];
		var rv = new Linq.RangeVariable (lt.Value, lt.Location);
		yyVal = new Linq.QueryExpression (new Linq.QueryStartClause ((Linq.QueryBlock)current_block, (Expression)yyVals[0+yyTop], rv, GetLocation (yyVals[-3+yyTop])));
	  }

void case_872()
///  #line 5796 "cs-parser.jay"
{
		current_block = new Linq.QueryBlock (current_block, lexer.Location);
	  
		var lt = (Tokenizer.LocatedToken) yyVals[-2+yyTop];
		var rv = new Linq.RangeVariable (lt.Value, lt.Location);
		yyVal = new Linq.QueryExpression (
			new Linq.QueryStartClause ((Linq.QueryBlock)current_block, (Expression)yyVals[0+yyTop], rv, GetLocation (yyVals[-4+yyTop])) {
				IdentifierType = (FullNamedExpression)yyVals[-3+yyTop]
			}
		);
	  }

void case_873()
///  #line 5811 "cs-parser.jay"
{
		current_block = new Linq.QueryBlock (current_block, lexer.Location);
	  
		var lt = (Tokenizer.LocatedToken) yyVals[-2+yyTop];
		var rv = new Linq.RangeVariable (lt.Value, lt.Location);
		yyVal = new Linq.QueryExpression (new Linq.QueryStartClause ((Linq.QueryBlock)current_block, (Expression)yyVals[0+yyTop], rv, GetLocation (yyVals[-3+yyTop])));
	  }

void case_874()
///  #line 5819 "cs-parser.jay"
{
		current_block = new Linq.QueryBlock (current_block, lexer.Location);
	  
		var lt = (Tokenizer.LocatedToken) yyVals[-2+yyTop];
		var rv = new Linq.RangeVariable (lt.Value, lt.Location);
		yyVal = new Linq.QueryExpression (
			new Linq.QueryStartClause ((Linq.QueryBlock)current_block, (Expression)yyVals[0+yyTop], rv, GetLocation (yyVals[-4+yyTop])) {
				IdentifierType = (FullNamedExpression)yyVals[-3+yyTop]
			}
		);
	  }

void case_876()
///  #line 5838 "cs-parser.jay"
{
		var lt = (Tokenizer.LocatedToken) yyVals[-3+yyTop];
		var sn = new Linq.RangeVariable (lt.Value, lt.Location);
		yyVal = new Linq.SelectMany ((Linq.QueryBlock)current_block, sn, (Expression)yyVals[0+yyTop], GetLocation (yyVals[-4+yyTop]));
		
		current_block.SetEndLocation (lexer.Location);
		current_block = current_block.Parent;
		
		((Linq.QueryBlock)current_block).AddRangeVariable (sn);
	  }

void case_878()
///  #line 5853 "cs-parser.jay"
{
		var lt = (Tokenizer.LocatedToken) yyVals[-3+yyTop];
		var sn = new Linq.RangeVariable (lt.Value, lt.Location);

		yyVal = new Linq.SelectMany ((Linq.QueryBlock)current_block, sn, (Expression)yyVals[0+yyTop], GetLocation (yyVals[-5+yyTop])) {
			IdentifierType = (FullNamedExpression)yyVals[-4+yyTop]
		};
		
		current_block.SetEndLocation (lexer.Location);
		current_block = current_block.Parent;
		
		((Linq.QueryBlock)current_block).AddRangeVariable (sn);
	  }

void case_879()
///  #line 5870 "cs-parser.jay"
{
	  	Linq.AQueryClause head = (Linq.AQueryClause)yyVals[-1+yyTop];
		
		if (yyVals[0+yyTop] != null)
			head.Next = (Linq.AQueryClause)yyVals[0+yyTop];
				
		if (yyVals[-2+yyTop] != null) {
			Linq.AQueryClause clause = (Linq.AQueryClause)yyVals[-2+yyTop];
			clause.Tail.Next = head;
			head = clause;
		}
		
		yyVal = head;
	  }

void case_881()
///  #line 5886 "cs-parser.jay"
{
		Error_SyntaxError (yyToken);
		yyVal = null;
	  }

void case_883()
///  #line 5898 "cs-parser.jay"
{
		yyVal = new Linq.Select ((Linq.QueryBlock)current_block, (Expression)yyVals[0+yyTop], GetLocation (yyVals[-2+yyTop]));

		current_block.SetEndLocation (lexer.Location);
		current_block = current_block.Parent;
	  }

void case_884()
///  #line 5905 "cs-parser.jay"
{
	  	if (linq_clause_blocks == null)
	  		linq_clause_blocks = new Stack<Linq.QueryBlock> ();
	  		
	  	current_block = new Linq.QueryBlock (current_block, lexer.Location);
	  	linq_clause_blocks.Push ((Linq.QueryBlock)current_block);
	  }

void case_885()
///  #line 5913 "cs-parser.jay"
{
		current_block.SetEndLocation (lexer.Location);
		current_block = current_block.Parent;
	  
		current_block = new Linq.QueryBlock (current_block, lexer.Location);
	  }

void case_886()
///  #line 5920 "cs-parser.jay"
{
		yyVal = new Linq.GroupBy ((Linq.QueryBlock)current_block, (Expression)yyVals[-3+yyTop], linq_clause_blocks.Pop (), (Expression)yyVals[0+yyTop], GetLocation (yyVals[-5+yyTop]));
		lbag.AddLocation (yyVal, GetLocation (yyVals[-1+yyTop]));
		
		current_block.SetEndLocation (lexer.Location);
		current_block = current_block.Parent;
	  }

void case_890()
///  #line 5937 "cs-parser.jay"
{
		((Linq.AQueryClause)yyVals[-1+yyTop]).Tail.Next = (Linq.AQueryClause)yyVals[0+yyTop];
		yyVal = yyVals[-1+yyTop];
	  }

void case_897()
///  #line 5957 "cs-parser.jay"
{
		var lt = (Tokenizer.LocatedToken) yyVals[-3+yyTop];
		var sn = new Linq.RangeVariable (lt.Value, lt.Location);
	  	yyVal = new Linq.Let ((Linq.QueryBlock) current_block, sn, (Expression)yyVals[0+yyTop], GetLocation (yyVals[-4+yyTop]));
		lbag.AddLocation (yyVal, GetLocation (yyVals[-2+yyTop]));
		
		current_block.SetEndLocation (lexer.Location);
		current_block = current_block.Parent;
		
		((Linq.QueryBlock)current_block).AddRangeVariable (sn);
	  }

void case_899()
///  #line 5976 "cs-parser.jay"
{
		yyVal = new Linq.Where ((Linq.QueryBlock)current_block, (Expression)yyVals[0+yyTop], GetLocation (yyVals[-2+yyTop]));

		current_block.SetEndLocation (lexer.Location);
		current_block = current_block.Parent;
	  }

void case_900()
///  #line 5986 "cs-parser.jay"
{
		if (linq_clause_blocks == null)
			linq_clause_blocks = new Stack<Linq.QueryBlock> ();
	  		
		current_block = new Linq.QueryBlock (current_block, lexer.Location);
		linq_clause_blocks.Push ((Linq.QueryBlock) current_block);
	  }

void case_901()
///  #line 5994 "cs-parser.jay"
{
		current_block.SetEndLocation (lexer.Location);
		current_block = current_block.Parent;

		current_block = new Linq.QueryBlock (current_block, lexer.Location);
		linq_clause_blocks.Push ((Linq.QueryBlock) current_block);
	  }

void case_902()
///  #line 6002 "cs-parser.jay"
{
		current_block.AddStatement (new ContextualReturn ((Expression) yyVals[-1+yyTop]));
		current_block.SetEndLocation (lexer.Location);
		current_block = current_block.Parent;

		current_block = new Linq.QueryBlock (current_block, lexer.Location);
	  }

void case_903()
///  #line 6010 "cs-parser.jay"
{
		current_block.AddStatement (new ContextualReturn ((Expression) yyVals[-1+yyTop]));
		current_block.SetEndLocation (lexer.Location);
	  
		var outer_selector = linq_clause_blocks.Pop ();
		var block = linq_clause_blocks.Pop ();

		var lt = (Tokenizer.LocatedToken) yyVals[-10+yyTop];	
		var sn = new Linq.RangeVariable (lt.Value, lt.Location);
		Linq.RangeVariable into;
		
		if (yyVals[0+yyTop] == null) {
			into = sn;
	  		yyVal = new Linq.Join (block, sn, (Expression)yyVals[-7+yyTop], outer_selector, (Linq.QueryBlock) current_block, GetLocation (yyVals[-11+yyTop]));
			lbag.AddLocation (yyVal, GetLocation (yyVals[-9+yyTop]), GetLocation (yyVals[-6+yyTop]), GetLocation (yyVals[-3+yyTop]));
		} else {
			/**/
			/* Set equals right side parent to beginning of linq query, it is not accessible therefore cannot cause name collisions*/
			/**/
			var parent = block.Parent;
			while (parent is Linq.QueryBlock) {
				parent = parent.Parent;
			}
			current_block.Parent = parent;
			
			((Linq.QueryBlock)current_block).AddRangeVariable (sn);
		
			lt = (Tokenizer.LocatedToken) yyVals[0+yyTop];
			into = new Linq.RangeVariable (lt.Value, lt.Location);

			yyVal = new Linq.GroupJoin (block, sn, (Expression)yyVals[-7+yyTop], outer_selector, (Linq.QueryBlock) current_block, into, GetLocation (yyVals[-11+yyTop]));	
			lbag.AddLocation (yyVal, GetLocation (yyVals[-9+yyTop]), GetLocation (yyVals[-6+yyTop]), GetLocation (yyVals[-3+yyTop]), GetLocation (yyVals[0+yyTop]));
		}

		current_block = block.Parent;
		((Linq.QueryBlock)current_block).AddRangeVariable (into);
	  }

void case_904()
///  #line 6048 "cs-parser.jay"
{
		if (linq_clause_blocks == null)
			linq_clause_blocks = new Stack<Linq.QueryBlock> ();
	  		
		current_block = new Linq.QueryBlock (current_block, lexer.Location);
		linq_clause_blocks.Push ((Linq.QueryBlock) current_block);
	  }

void case_905()
///  #line 6056 "cs-parser.jay"
{
		current_block.SetEndLocation (lexer.Location);
		current_block = current_block.Parent;

		current_block = new Linq.QueryBlock (current_block, lexer.Location);
		linq_clause_blocks.Push ((Linq.QueryBlock) current_block);
	  }

void case_906()
///  #line 6064 "cs-parser.jay"
{
		current_block.AddStatement (new ContextualReturn ((Expression) yyVals[-1+yyTop]));
		current_block.SetEndLocation (lexer.Location);
		current_block = current_block.Parent;

		current_block = new Linq.QueryBlock (current_block, lexer.Location);
	  }

void case_907()
///  #line 6072 "cs-parser.jay"
{
		current_block.AddStatement (new ContextualReturn ((Expression) yyVals[-1+yyTop]));
		current_block.SetEndLocation (lexer.Location);
	  
		var outer_selector = linq_clause_blocks.Pop ();
		var block = linq_clause_blocks.Pop ();
		
		var lt = (Tokenizer.LocatedToken) yyVals[-10+yyTop];
		var sn = new Linq.RangeVariable (lt.Value, lt.Location);
		Linq.RangeVariable into;
		
		if (yyVals[0+yyTop] == null) {
			into = sn;		
	  		yyVal = new Linq.Join (block, sn, (Expression)yyVals[-7+yyTop], outer_selector, (Linq.QueryBlock) current_block, GetLocation (yyVals[-12+yyTop])) {
	  			IdentifierType = (FullNamedExpression)yyVals[-11+yyTop]
	  		};
		} else {
			/**/
			/* Set equals right side parent to beginning of linq query, it is not accessible therefore cannot cause name collisions*/
			/**/
			var parent = block.Parent;
			while (parent is Linq.QueryBlock) {
				parent = parent.Parent;
			}
			current_block.Parent = parent;
		
			((Linq.QueryBlock)current_block).AddRangeVariable (sn);
		
			lt = (Tokenizer.LocatedToken) yyVals[0+yyTop];
			into = new Linq.RangeVariable (lt.Value, lt.Location); /* TODO:*/
			
			yyVal = new Linq.GroupJoin (block, sn, (Expression)yyVals[-7+yyTop], outer_selector, (Linq.QueryBlock) current_block, into, GetLocation (yyVals[-12+yyTop])) {
	  			IdentifierType = (FullNamedExpression)yyVals[-11+yyTop]
	  		};			
		}
		
		current_block = block.Parent;
		((Linq.QueryBlock)current_block).AddRangeVariable (into);		
	  }

void case_911()
///  #line 6127 "cs-parser.jay"
{
		current_block.SetEndLocation (lexer.Location);
		current_block = current_block.Parent;
	  
		yyVal = yyVals[0+yyTop];
	  }

void case_913()
///  #line 6138 "cs-parser.jay"
{
		current_block.SetEndLocation (lexer.Location);
		current_block = current_block.Parent;
	  
		current_block = new Linq.QueryBlock (current_block, lexer.Location);
	  }

void case_914()
///  #line 6145 "cs-parser.jay"
{
		((Linq.AQueryClause)yyVals[-3+yyTop]).Next = (Linq.AQueryClause)yyVals[0+yyTop];
		yyVal = yyVals[-3+yyTop];
	  }

void case_916()
///  #line 6154 "cs-parser.jay"
{
		current_block.SetEndLocation (lexer.Location);
		current_block = current_block.Parent;
	  
		current_block = new Linq.QueryBlock ((Linq.QueryBlock) current_block, lexer.Location);	 
	 }

void case_917()
///  #line 6161 "cs-parser.jay"
{
		((Linq.AQueryClause)yyVals[-3+yyTop]).Tail.Next = (Linq.AQueryClause)yyVals[0+yyTop];
		yyVal = yyVals[-3+yyTop];
	 }

void case_919()
///  #line 6173 "cs-parser.jay"
{
		yyVal = new Linq.OrderByAscending ((Linq.QueryBlock) current_block, (Expression)yyVals[-1+yyTop]);	
		lbag.AddLocation (yyVal, GetLocation (yyVals[0+yyTop]));
	  }

void case_920()
///  #line 6178 "cs-parser.jay"
{
		yyVal = new Linq.OrderByDescending ((Linq.QueryBlock) current_block, (Expression)yyVals[-1+yyTop]);	
		lbag.AddLocation (yyVal, GetLocation (yyVals[0+yyTop]));
	  }

void case_922()
///  #line 6190 "cs-parser.jay"
{
		yyVal = new Linq.ThenByAscending ((Linq.QueryBlock) current_block, (Expression)yyVals[-1+yyTop]);	
		lbag.AddLocation (yyVal, GetLocation (yyVals[0+yyTop]));
	  }

void case_923()
///  #line 6195 "cs-parser.jay"
{
		yyVal = new Linq.ThenByDescending ((Linq.QueryBlock) current_block, (Expression)yyVals[-1+yyTop]);	
		lbag.AddLocation (yyVal, GetLocation (yyVals[0+yyTop]));
	  }

void case_925()
///  #line 6205 "cs-parser.jay"
{
		/* query continuation block is not linked with query block but with block*/
		/* before. This means each query can use same range variable names for*/
		/* different identifiers.*/

		current_block.SetEndLocation (GetLocation (yyVals[-1+yyTop]));
		current_block = current_block.Parent;
	
		current_block = new Linq.QueryBlock (current_block, lexer.Location);
		
		if (linq_clause_blocks == null)
			linq_clause_blocks = new Stack<Linq.QueryBlock> ();
	  		
		linq_clause_blocks.Push ((Linq.QueryBlock) current_block);		
	  }

void case_926()
///  #line 6221 "cs-parser.jay"
{
		var current_block = linq_clause_blocks.Pop ();	  
		var lt = (Tokenizer.LocatedToken) yyVals[-2+yyTop];
		var rv = new Linq.RangeVariable (lt.Value, lt.Location);
  		yyVal = new Linq.QueryStartClause ((Linq.QueryBlock)current_block, null, rv, GetLocation (yyVals[-3+yyTop])) {
  			next = (Linq.AQueryClause)yyVals[0+yyTop]
  		};
	  }

void case_929()
///  #line 6248 "cs-parser.jay"
{ 
		current_container = new Class (current_namespace, current_class, new MemberName ("<InteractiveExpressionClass>"), Modifiers.PUBLIC, null);
		current_class = current_container;

		/* (ref object retval)*/
		Parameter [] mpar = new Parameter [1];
		mpar [0] = new Parameter (new TypeExpression (compiler.BuiltinTypes.Object, Location.Null), "$retval", Parameter.Modifier.REF, null, Location.Null);

		ParametersCompiled pars = new ParametersCompiled (mpar);
		var mods = Modifiers.PUBLIC | Modifiers.STATIC;
		if (settings.Unsafe)
			mods |= Modifiers.UNSAFE;

		current_local_parameters = pars;
		Method method = new Method (
			current_class,
			null, /* generic*/
			new TypeExpression (compiler.BuiltinTypes.Void, Location.Null),
			mods,
			new MemberName ("Host"),
			pars,
			null /* attributes */);
			
		current_container.AddMethod (method);			

		oob_stack.Push (method);
		++lexer.parsing_block;
		start_block (lexer.Location);
	  }

void case_930()
///  #line 6278 "cs-parser.jay"
{
		--lexer.parsing_block;
		Method method = (Method) oob_stack.Pop ();

		method.Block = (ToplevelBlock) end_block(lexer.Location);

		InteractiveResult = (Class) pop_current_class ();
		current_local_parameters = null;
	  }

void case_940()
///  #line 6321 "cs-parser.jay"
{
		module.DocumentationBuilder.ParsedBuiltinType = (TypeExpression)yyVals[-1+yyTop];
		module.DocumentationBuilder.ParsedParameters = (List<DocumentationParameter>)yyVals[0+yyTop];
		yyVal = null;
	  }

void case_941()
///  #line 6327 "cs-parser.jay"
{
		module.DocumentationBuilder.ParsedBuiltinType = (TypeExpression)yyVals[-3+yyTop];
		module.DocumentationBuilder.ParsedParameters = (List<DocumentationParameter>)yyVals[0+yyTop];
		var lt = (Tokenizer.LocatedToken) yyVals[-1+yyTop];
		yyVal = new MemberName (lt.Value);
	  }

void case_944()
///  #line 6342 "cs-parser.jay"
{
		module.DocumentationBuilder.ParsedParameters = (List<DocumentationParameter>)yyVals[-1+yyTop];
		yyVal = new MemberName ((MemberName) yyVals[-6+yyTop], new MemberName (MemberCache.IndexerNameAlias));
	  }

void case_945()
///  #line 6347 "cs-parser.jay"
{
		var p = (List<DocumentationParameter>)yyVals[0+yyTop] ?? new List<DocumentationParameter> (1);
		p.Add (new DocumentationParameter ((FullNamedExpression) yyVals[-1+yyTop]));
		module.DocumentationBuilder.ParsedParameters = p;
		module.DocumentationBuilder.ParsedOperator = Operator.OpType.Explicit;
		yyVal = null;
	  }

void case_946()
///  #line 6355 "cs-parser.jay"
{
		var p = (List<DocumentationParameter>)yyVals[0+yyTop] ?? new List<DocumentationParameter> (1);
		p.Add (new DocumentationParameter ((FullNamedExpression) yyVals[-1+yyTop]));
		module.DocumentationBuilder.ParsedParameters = p;
		module.DocumentationBuilder.ParsedOperator = Operator.OpType.Implicit;
		yyVal = null;
	  }

void case_947()
///  #line 6363 "cs-parser.jay"
{
		var p = (List<DocumentationParameter>)yyVals[0+yyTop] ?? new List<DocumentationParameter> (1);
		module.DocumentationBuilder.ParsedParameters = p;
		module.DocumentationBuilder.ParsedOperator = (Operator.OpType) yyVals[-1+yyTop];
		yyVal = null;
	  }

void case_955()
///  #line 6401 "cs-parser.jay"
{
		var parameters = new List<DocumentationParameter> ();
		parameters.Add ((DocumentationParameter) yyVals[0+yyTop]);
		yyVal = parameters;
	  }

void case_956()
///  #line 6407 "cs-parser.jay"
{
		var parameters = yyVals[-2+yyTop] as List<DocumentationParameter>;
		parameters.Add ((DocumentationParameter) yyVals[0+yyTop]);
		yyVal = parameters;
	  }

void case_957()
///  #line 6416 "cs-parser.jay"
{
		if (yyVals[-1+yyTop] != null)
			yyVal = new DocumentationParameter ((Parameter.Modifier) yyVals[-1+yyTop], (FullNamedExpression) yyVals[0+yyTop]);
		else
			yyVal = new DocumentationParameter ((FullNamedExpression) yyVals[0+yyTop]);
	  }

///  #line default
   static readonly short [] yyLhs  = {              -1,
    0,    4,    0,    0,    1,    1,    1,    1,    2,    2,
   11,   11,   12,   12,   13,   13,   14,   14,   15,   15,
   16,   21,   22,   19,   20,   20,   20,   24,   24,   25,
   25,   18,    7,    7,    6,    6,   23,   23,    8,    8,
   26,   26,   27,   27,   27,   27,   27,    9,    9,   10,
   10,   35,   33,   38,   34,   34,   36,   36,   36,   36,
   37,   37,   42,   39,   40,   41,   41,   43,   43,   43,
   43,   43,   44,   44,   48,   45,   47,   50,   50,   50,
   51,   51,   52,   52,   53,   53,   53,   53,   53,   53,
   53,   53,   53,   53,   53,   66,   68,   70,   71,   72,
   29,   29,   75,   54,   76,   76,   77,   77,   78,   80,
   74,   74,   79,   79,   85,   55,   89,   55,   55,   84,
   92,   84,   86,   86,   93,   93,   94,   95,   94,   90,
   90,   96,   96,   97,   98,   88,   88,   91,   91,   91,
  101,   56,  104,  105,   99,  106,  107,  108,   99,   99,
  100,  100,  103,  103,  111,  111,  111,  111,  111,  111,
  111,  111,  111,  111,  112,  112,  115,  115,  115,  118,
  115,  116,  116,  119,  119,  120,  120,  120,  113,  113,
  113,  121,  121,  121,  114,  123,  125,  126,   57,  128,
  129,  130,   59,  124,  124,  124,  124,  124,  134,  131,
  135,  132,  133,  133,  133,  136,  137,  138,  140,   30,
   30,  139,  139,  141,  141,  142,  142,  142,  142,  142,
  142,  142,  142,  142,  145,   60,  144,  144,  146,  146,
  149,  143,  143,  148,  148,  148,  148,  148,  148,  148,
  148,  148,  148,  148,  148,  148,  148,  148,  148,  148,
  148,  148,  148,  148,  148,  151,  150,  152,  150,  150,
  150,   61,  155,  157,  153,  154,  154,  156,  156,  161,
  159,  162,  159,  159,  163,   62,  165,   58,  168,  169,
   58,  164,  171,  164,  166,  166,  172,  172,  173,  174,
  173,  175,  170,  167,  167,  167,  167,  167,  179,  176,
  180,  177,  178,  178,  182,  184,  185,   31,  181,  181,
  181,  183,  183,  183,  186,  186,  187,  188,  187,  189,
  190,  191,   32,  192,  192,   17,   17,  193,  193,  196,
  195,  195,  195,  197,  197,  199,   65,  122,  102,  102,
  127,  127,  200,  200,  200,  198,  198,  201,  201,  202,
  202,  204,  204,   83,   73,   73,   87,   87,  117,  117,
  147,  147,  205,  205,  205,  205,  205,  209,  209,  210,
  210,  208,  208,  208,  208,  208,  208,  208,  211,  211,
  211,  211,  211,  211,  211,  211,  211,  212,  212,  212,
  212,  212,  212,  212,  212,  212,  212,  212,  212,  212,
  212,  212,  212,  212,  212,  212,  212,  213,  213,  213,
  214,  214,  214,  234,  234,  235,  235,  236,  236,  216,
  216,  233,  233,  233,  233,  233,  233,  233,  233,  218,
  237,  237,  238,  238,  239,  239,  241,  241,  241,  242,
  242,  242,  242,  242,  243,  243,  160,  160,  247,  247,
  247,  247,  247,  249,  249,  248,  248,  250,  250,  250,
  250,  251,  219,  219,  219,  246,  246,  246,  252,  252,
  253,  253,  220,  221,  221,  222,  223,  224,  224,  215,
  215,  215,  215,  215,  258,  254,  225,  259,  259,  260,
  260,  261,  261,  262,  262,  262,  262,  255,  255,  206,
  206,  257,  257,  263,  263,  256,  256,   82,   82,  264,
  264,  265,  226,  266,  266,  266,  267,  267,  267,  267,
  267,  268,  194,  227,  228,  229,  230,  270,  231,  271,
  231,  269,  269,  273,  272,  217,  274,  274,  274,  274,
  274,  275,  275,  275,  275,  275,  275,  275,  276,  276,
  276,  276,  277,  277,  277,  277,  277,  277,  278,  278,
  278,  279,  279,  279,  279,  279,  280,  280,  280,  281,
  281,  282,  282,  283,  283,  284,  284,  285,  285,  286,
  286,  287,  287,  288,  288,  288,  288,  288,  288,  288,
  288,  288,  288,  288,  289,  289,  290,  290,  290,  291,
  291,  292,  292,  295,  293,  294,  294,  297,  296,  298,
  296,  299,  300,  296,  301,  302,  296,   46,   46,  244,
  244,  244,  244,  232,  232,  232,   81,  304,  305,  306,
  307,  308,   28,   64,   64,   63,   63,  109,  109,  309,
  309,  309,  309,  309,  309,  309,  309,  309,  309,  309,
  309,  309,  309,  309,   67,   67,   69,   69,   69,  310,
  310,  311,  312,  312,  313,  313,  313,  313,  203,  203,
  314,  314,  316,  110,  317,  317,  318,  158,  315,  315,
  319,  319,  320,  320,  320,  320,  320,  324,  324,  325,
  325,  325,  322,  322,  322,  322,  322,  322,  322,  322,
  322,  322,  322,  322,  322,  326,  326,  326,  326,  326,
  326,  326,  326,  326,  326,  326,  326,  326,  340,  340,
  340,  340,  327,  341,  323,  342,  342,  343,  343,  343,
  343,  343,  343,  207,  207,  344,   49,   49,  346,  321,
  349,  321,  345,  345,  345,  345,  347,  347,  353,  353,
  352,  352,  354,  354,  348,  348,  350,  350,  355,  355,
  356,  351,  351,  351,  328,  328,  339,  339,  357,  358,
  358,  329,  329,  359,  359,  359,  362,  360,  361,  361,
  363,  363,  363,  366,  364,  365,  365,  367,  367,  330,
  330,  330,  330,  368,  369,  369,  369,  373,  370,  376,
  378,  379,  372,  372,  374,  374,  381,  380,  380,  375,
  375,  377,  377,  383,  382,  382,  371,  384,  371,  371,
  371,  331,  331,  331,  331,  331,  331,  385,  386,  387,
  387,  387,  388,  389,  390,  390,  391,  391,  332,  332,
  332,  332,  392,  392,  394,  394,  393,  395,  393,  393,
  333,  334,  396,  337,  335,  398,  399,  338,  401,  402,
  336,  336,  400,  400,  397,  397,  303,  303,  303,  303,
  403,  403,  405,  405,  407,  406,  408,  406,  404,  404,
  404,  412,  410,  413,  414,  410,  409,  409,  415,  415,
  416,  416,  416,  416,  416,  421,  417,  422,  418,  423,
  424,  425,  419,  427,  428,  429,  419,  426,  426,  431,
  420,  430,  434,  430,  433,  436,  433,  432,  432,  432,
  435,  435,  435,  411,  437,  411,    3,    3,  438,    3,
    3,  439,  439,  245,  245,  240,  240,    5,  440,  440,
  440,  440,  444,  440,  440,  440,  440,  441,  441,  442,
  445,  442,  443,  443,  446,  446,  447,
  };
   static readonly short [] yyLen = {           2,
    2,    0,    3,    1,    2,    4,    3,    1,    0,    1,
    1,    2,    4,    2,    1,    2,    1,    1,    5,    2,
    3,    0,    0,   11,    1,    3,    1,    0,    1,    0,
    1,    1,    0,    1,    0,    1,    0,    1,    1,    2,
    1,    1,    1,    1,    1,    1,    1,    0,    1,    1,
    2,    0,    3,    0,    6,    3,    1,    1,    1,    1,
    1,    3,    0,    3,    1,    0,    3,    0,    1,    1,
    3,    3,    1,    1,    0,    4,    4,    0,    1,    1,
    0,    1,    1,    2,    1,    1,    1,    1,    1,    1,
    1,    1,    1,    1,    1,    0,    0,    0,    0,    0,
   16,    5,    0,    9,    0,    1,    1,    2,    3,    0,
    3,    1,    1,    1,    0,    8,    0,    9,    6,    0,
    0,    3,    0,    1,    1,    2,    2,    0,    5,    0,
    1,    1,    2,    3,    0,    4,    2,    1,    1,    1,
    0,    3,    0,    0,   10,    0,    0,    0,   12,    8,
    1,    1,    0,    1,    1,    3,    3,    3,    5,    3,
    5,    1,    1,    1,    1,    3,    4,    6,    4,    0,
    7,    0,    1,    1,    2,    1,    1,    1,    4,    6,
    4,    1,    2,    2,    1,    0,    0,    0,   10,    0,
    0,    0,   13,    1,    2,    1,    2,    1,    0,    5,
    0,    5,    1,    1,    1,    0,    0,    0,    0,   15,
    5,    0,    1,    1,    2,    1,    1,    1,    1,    1,
    1,    1,    1,    1,    0,    5,    1,    1,    1,    1,
    0,    7,    1,    1,    1,    1,    1,    1,    1,    1,
    1,    1,    1,    1,    1,    1,    1,    1,    1,    1,
    1,    1,    1,    1,    1,    0,    7,    0,    7,    2,
    2,    2,    0,    0,    9,    1,    1,    0,    1,    0,
    6,    0,    6,    1,    0,    8,    0,    9,    0,    0,
   10,    0,    0,    3,    0,    1,    1,    2,    2,    0,
    5,    0,    2,    2,    2,    1,    1,    1,    0,    5,
    0,    5,    1,    1,    0,    0,    0,   12,    0,    2,
    2,    0,    1,    2,    1,    3,    2,    0,    5,    0,
    0,    0,   13,    0,    1,    1,    3,    1,    4,    2,
    0,    3,    2,    1,    3,    0,    3,    1,    1,    3,
    1,    2,    3,    4,    4,    0,    3,    1,    3,    3,
    1,    1,    1,    1,    1,    1,    1,    1,    1,    1,
    1,    2,    2,    2,    2,    2,    2,    1,    3,    1,
    1,    1,    1,    1,    1,    1,    1,    1,    1,    1,
    1,    1,    1,    1,    1,    1,    1,    1,    1,    1,
    1,    1,    1,    1,    1,    1,    1,    1,    1,    1,
    1,    1,    1,    1,    1,    1,    1,    2,    2,    1,
    1,    1,    1,    1,    1,    1,    1,    1,    1,    3,
    3,    4,    4,    4,    3,    3,    4,    3,    4,    4,
    0,    1,    3,    4,    0,    1,    1,    3,    2,    3,
    1,    2,    3,    2,    1,    1,    0,    1,    1,    3,
    3,    2,    2,    1,    1,    1,    1,    2,    2,    4,
    3,    1,    4,    4,    3,    1,    3,    2,    1,    3,
    1,    1,    1,    4,    3,    2,    2,    6,    3,    7,
    4,    3,    7,    3,    0,    2,    4,    1,    2,    0,
    1,    1,    3,    3,    1,    1,    1,    0,    1,    1,
    2,    2,    3,    1,    2,    0,    1,    2,    4,    1,
    3,    0,    5,    1,    1,    1,    2,    3,    3,    4,
    4,    1,    2,    4,    4,    4,    4,    0,    4,    0,
    5,    0,    1,    0,    4,    4,    1,    2,    2,    4,
    2,    1,    2,    2,    2,    2,    2,    2,    1,    3,
    3,    3,    1,    3,    3,    3,    3,    3,    1,    3,
    3,    1,    3,    3,    3,    3,    1,    3,    3,    1,
    3,    1,    3,    1,    3,    1,    3,    1,    3,    1,
    3,    1,    5,    3,    3,    3,    3,    3,    3,    3,
    3,    3,    3,    3,    1,    3,    3,    2,    1,    0,
    1,    1,    1,    0,    2,    1,    1,    0,    4,    0,
    5,    0,    0,    7,    0,    0,    8,    1,    1,    1,
    1,    1,    1,    6,    4,    4,    1,    1,    0,    0,
    0,    0,   15,    0,    1,    0,    1,    1,    2,    1,
    1,    1,    1,    1,    1,    1,    1,    1,    1,    1,
    1,    1,    1,    1,    0,    2,    0,    1,    1,    1,
    2,    4,    1,    3,    1,    3,    1,    1,    0,    1,
    1,    1,    0,    4,    1,    1,    0,    4,    0,    1,
    1,    2,    1,    1,    1,    2,    1,    1,    2,    1,
    1,    1,    1,    1,    1,    1,    1,    1,    1,    1,
    1,    1,    1,    1,    1,    1,    1,    1,    1,    1,
    1,    1,    1,    1,    1,    1,    1,    1,    1,    1,
    1,    1,    1,    0,    4,    1,    2,    2,    2,    2,
    2,    2,    1,    1,    2,    1,    1,    1,    0,    6,
    0,    7,    0,    2,    2,    1,    0,    1,    0,    1,
    1,    2,    2,    4,    0,    2,    0,    1,    1,    2,
    4,    1,    5,    2,    2,    2,    2,    2,    1,    1,
    1,    1,    1,    5,    7,    4,    0,    8,    0,    1,
    1,    2,    1,    0,    3,    1,    2,    3,    1,    1,
    1,    1,    1,    5,    7,    3,    6,    0,    4,    0,
    0,    0,   10,    1,    0,    1,    0,    5,    1,    0,
    1,    0,    1,    1,    1,    3,    6,    0,    9,    5,
    4,    1,    1,    1,    1,    1,    1,    2,    2,    3,
    4,    3,    3,    3,    4,    3,    0,    1,    3,    4,
    5,    3,    1,    2,    0,    1,    2,    0,    7,    3,
    2,    2,    0,    3,    5,    0,    0,   10,    0,    0,
    9,    5,    2,    1,    0,    2,    2,    2,    2,    2,
    4,    5,    4,    5,    0,    5,    0,    6,    3,    2,
    1,    0,    3,    0,    0,    6,    0,    1,    1,    2,
    1,    1,    1,    1,    1,    0,    5,    0,    3,    0,
    0,    0,   12,    0,    0,    0,   13,    0,    2,    0,
    3,    1,    0,    4,    1,    0,    4,    1,    2,    2,
    1,    2,    2,    0,    0,    4,    2,    3,    0,    4,
    2,    2,    3,    0,    1,    1,    1,    2,    2,    2,
    4,    3,    0,    7,    4,    4,    3,    1,    3,    0,
    0,    4,    0,    1,    1,    3,    2,
  };
   static readonly short [] yyDefRed = {            0,
    8,    0,    0,    0,    0,    0,    0,    0,    2,    4,
    0,    0,   11,   14,    0,  927,    0,    0,  931,    0,
    0,   15,   17,   18,  374,  380,  387,  375,  377,    0,
  376,    0,  383,  385,  372,    0,  379,  381,  373,  384,
  386,  382,  336,  948,    0,  378,  938,    0,   10,    1,
    0,    0,    0,   12,    0,  771,    0,    0,    0,    0,
    0,    0,    0,    0,  415,    0,    0,    0,    0,    0,
    0,    0,  413,    0,    0,    0,  473,    0,  414,    0,
  512,    0,  853,    0,    0,    0,  623,    0,    0,    0,
    0,    0,    0,    0,  673,    0,  723,    0,    0,    0,
    0,    0,    0,    0,    0,  412,    0,  612,    0,  770,
    0,  706,    0,    0,    0,    0,  389,  390,    0,  392,
  393,  394,  395,  396,  397,  398,  399,  400,  401,  402,
  403,  404,  405,  406,  407,  410,  411,  619,  542,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,  620,  618,  621,  622,  690,  692,    0,  688,  691,
  707,  709,  710,  711,  712,  713,  714,  715,  716,  717,
  718,  708,    0,    0,    0,  772,  773,  790,  791,  792,
  793,  822,  823,  824,  825,  826,  827,    0,    0,    0,
   20,    0,    0,    0,  326,    0,  328,  935,   16,  928,
    0,    0,  239,  238,  235,  240,  241,  234,  253,  252,
  245,  246,  242,  244,  243,  247,  236,  237,  248,  249,
  255,  254,  250,  251,    0,    0,  951,    0,  940,    0,
  939,    3,   52,    0,    0,    0,   42,   39,   41,   43,
   44,   45,   46,   47,   50,   13,    0,    0,    0,  828,
  416,  417,  851,    0,    0,    0,    0,    0,    0,  391,
    0,  829,    0,  534,  528,  533,  722,  769,  693,  720,
  719,  721,  694,  695,  696,  697,  698,  699,  700,  701,
  702,  703,  704,  705,    0,    0,    0,  798,    0,    0,
    0,  738,  737,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,  838,    0,  388,    0,    0,    0,    0,
    0,    0,  852,    0,    0,    0,  736,  732,    0,    0,
    0,    0,    0,    0,    0,  355,    0,    0,    0,    0,
    0,    0,    0,    0,  615,    0,  541,    0,    0,  539,
  543,  544,  538,  548,  547,  545,  546,  608,  523,    0,
  409,  408,    0,    0,    0,    0,    0,  724,    0,  325,
    0,  730,  731,    0,  476,  477,    0,    0,    0,  728,
  729,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,  930,  689,  739,  727,    0,
  767,  768,  881,  898,    0,    0,    0,  910,  869,  867,
  891,    0,    0,  889,  892,  893,  894,  895,  870,  868,
    0,    0,    0,  330,    0,   21,    0,    0,    0,  947,
    0,  337,    0,    0,    0,  949,    0,    0,   40,  645,
  651,  643,    0,  640,  650,  644,  642,  641,  648,  646,
  647,  653,  649,  652,  654,    0,    0,  638,   51,  475,
    0,  471,  472,    0,    0,  469,    0,  741,    0,    0,
    0,  796,    0,  765,  766,    0,    0,    0,  627,    0,
  832,  830,  628,    0,    0,  497,    0,    0,    0,  488,
    0,  492,  502,  504,    0,  484,    0,    0,    0,    0,
    0,  479,    0,  482,    0,  486,  357,  833,    0,    0,
  834,  842,    0,    0,    0,  843,    0,    0,  854,    0,
    0,  735,    0,  367,  363,  364,    0,    0,  362,  365,
  366,    0,    0,    0,  549,    0,    0,  530,    0,  610,
  687,    0,    0,    0,  681,  683,  684,  685,  420,  421,
    0,  333,  334,    0,  177,  176,  178,    0,    0,    0,
    0,  359,    0,  595,    0,    0,  836,    0,    0,  425,
    0,  428,    0,  426,    0,  465,    0,    0,    0,    0,
    0,  454,  457,    0,    0,  449,  456,  455,    0,  584,
  585,  586,  587,  588,  589,  590,  591,  592,  594,  593,
  550,  552,  551,  557,  558,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
  581,    0,    0,  501,    0,    0,    0,    0,    0,    0,
    0,  882,  884,  880,    0,  890,    0,    0,  327,  945,
  946,  351,    0,    0,  348,    0,    0,  174,    0,    0,
  955,  941,  943,   60,   58,   59,    0,    0,   53,    0,
    0,   61,   63,   27,   25,    0,    0,    0,  635,    0,
  639,  424,    0,  474,    0,  525,    0,  536,  164,  185,
    0,    0,  154,    0,    0,    0,  165,  529,    0,  856,
  804,    0,  815,  799,    0,  806,    0,  821,    0,    0,
  831,  776,    0,    0,    0,  487,    0,  503,  505,    0,
    0,  441,    0,    0,  437,    0,    0,  466,    0,  507,
  481,    0,  140,  508,  138,  139,  510,    0,  524,    0,
  847,    0,  840,    0,  844,  516,    0,    0,    0,  352,
    0,  514,    0,    0,  526,    0,  859,    0,  873,    0,
  871,    0,    0,  625,  626,    0,    0,    0,  686,  675,
  676,  674,  682,  603,  609,  602,    0,    0,  332,  598,
    0,    0,    0,  540,  835,  725,  429,  423,  427,  422,
  527,  464,  463,  462,  459,  458,    0,  453,  418,  419,
  430,    0,    0,  746,    0,    0,  899,  875,    0,  900,
    0,  896,    0,  911,    0,    0,    0,    0,  879,   19,
  329,  672,  671,    0,  670,    0,  347,  957,  175,  952,
    0,    0,   54,    0,    0,    0,    0,    0,    0,  354,
    0,  629,    0,    0,   80,   79,    0,  470,    0,    0,
    0,    0,    0,  535,    0,    0,    0,    0,    0,  807,
  800,    0,    0,  820,    0,    0,  855,  494,  493,  444,
    0,    0,  936,  937,  433,  439,    0,  442,    0,  468,
    0,    0,    0,    0,    0,  777,  850,    0,  841,    0,
  522,  517,    0,    0,  513,    0,  862,    0,  794,  874,
  872,    0,  531,    0,  611,  607,  606,  605,  335,  597,
  596,  613,  461,    0,  451,  450,  583,  140,    0,  762,
  744,    0,    0,    0,  751,    0,  877,    0,  904,    0,
  919,  920,  913,  883,  885,  925,  350,  349,  956,    0,
    0,   62,   56,    0,   64,   26,   23,    0,    0,  305,
    0,  211,    0,  102,    0,   77,  756,  113,  114,    0,
    0,    0,  759,  183,  184,    0,    0,    0,    0,  157,
  166,  158,  160,  797,    0,    0,    0,    0,    0,  816,
  817,    0,    0,  443,  445,  446,  440,  434,  438,    0,
  499,    0,  467,  478,  432,  511,  509,    0,  846,    0,
    0,    0,  518,    0,  864,    0,    0,  624,  616,    0,
  460,    0,    0,  740,  752,  876,    0,    0,    0,  897,
    0,    0,    0,  944,    0,    0,    0,   69,   70,   73,
   74,    0,  320,  311,  310,    0,  630,  207,   97,    0,
  742,  760,  169,    0,  181,    0,    0,    0,  795,  866,
    0,    0,    0,  811,    0,  818,  775,  483,  480,  783,
    0,  789,    0,    0,  781,    0,  786,  848,  521,  520,
  863,  860,    0,  614,    0,    0,  878,  901,    0,    0,
    0,  915,    0,  926,    0,   75,   67,    0,    0,    0,
  306,    0,    0,    0,    0,    0,  170,    0,  161,  159,
  857,  808,  801,    0,    0,  778,  782,    0,  787,    0,
    0,  617,    0,  754,    0,  905,  922,  923,  916,  886,
   55,    0,   71,   72,    0,    0,    0,    0,    0,    0,
    0,  761,  168,    0,  180,    0,    0,  819,  788,    0,
  677,  849,  861,  763,    0,    0,    0,   76,    0,    0,
  321,    0,  307,    0,  315,  371,  370,    0,  368,  659,
    0,  631,    0,  660,  208,   98,  171,  858,    0,    0,
  813,    0,  902,    0,  917,    0,    0,    0,    0,    0,
    0,    0,    0,  661,    0,    0,  802,    0,    0,  906,
   29,   24,  322,    0,    0,  316,  369,    0,    0,    0,
   99,    0,  678,    0,    0,    0,    0,  308,  667,    0,
  668,  665,    0,  663,   95,    0,   94,    0,    0,   83,
   85,   86,   87,   88,   89,   90,   91,   92,   93,  141,
    0,    0,  224,  216,  217,  218,  219,  220,  221,  222,
  223,    0,    0,  214,    0,  803,    0,  903,    0,  323,
  319,    0,    0,    0,  632,   84,    0,  267,  262,  266,
    0,  209,  215,    0,  909,  907,  666,  664,    0,    0,
    0,    0,    0,    0,    0,  275,    0,    0,  225,    0,
    0,  233,    0,  152,  142,  151,    0,  100,    0,    0,
  261,    0,    0,  260,    0,  146,    0,    0,  341,    0,
  339,    0,    0,  186,    0,    0,    0,    0,    0,  633,
  210,    0,  103,    0,  338,    0,    0,    0,    0,  117,
    0,    0,    0,    0,    0,    0,  143,    0,    0,  190,
    0,  342,    0,  228,  227,  226,    0,  101,    0,  279,
    0,  258,  119,    0,  256,    0,    0,    0,  121,    0,
  343,    0,    0,  187,    0,    0,    0,  340,  231,  112,
  110,    0,    0,  283,    0,    0,    0,    0,    0,  147,
    0,  264,    0,    0,    0,    0,  125,    0,    0,    0,
    0,  344,  345,    0,    0,    0,    0,    0,  107,  298,
    0,  280,    0,    0,  292,    0,    0,    0,  287,    0,
  137,    0,    0,    0,    0,  132,    0,    0,  276,    0,
  122,    0,  116,  126,  144,  150,  198,    0,  188,    0,
    0,    0,    0,  111,    0,  104,  108,    0,    0,    0,
  294,    0,  295,  284,    0,    0,  278,  288,  259,    0,
    0,  118,  133,  257,    0,  274,    0,  265,  269,  128,
    0,    0,    0,  195,  197,  191,  232,  109,  299,  301,
  281,    0,    0,  293,  290,  136,  134,  148,    0,    0,
    0,  145,  199,  201,  189,    0,    0,    0,  292,    0,
  270,  272,  129,    0,    0,  192,  303,  304,  300,  302,
  291,  149,    0,    0,  205,  204,  203,  200,  202,    0,
    0,    0,  193,  271,  273,
  };
  protected static readonly short [] yyDgoto  = {             7,
    8,   50,    9,   51,   10,   11,   52,  234,  691,  432,
   12,   13,   53,   22,   23,   24,  323,  194,  237,  676,
  838, 1032, 1150, 1487,  835,  238,  239,  240,  241,  242,
  243,  244,  245,  669,  447,  670,  671,  941,  672,  673,
  945,  836, 1027, 1028, 1029,  268,  593, 1122,  111,  847,
 1218, 1219, 1220, 1221, 1222, 1223, 1224, 1225, 1226, 1227,
 1228, 1229,  466,  680, 1301,  955, 1129, 1094, 1162, 1186,
 1245, 1312, 1157, 1362, 1339, 1387, 1388, 1389,  957, 1385,
  958,  736, 1278, 1350, 1325, 1375,  516, 1368, 1344, 1404,
  920, 1373, 1376, 1377, 1471, 1405, 1406, 1402, 1230, 1285,
 1257, 1302,  692, 1352, 1451, 1322, 1408, 1480,  467,  269,
  693,  694,  695,  696,  697,  656,  570, 1134,  657,  658,
  853, 1304, 1329, 1419, 1380, 1453, 1305, 1355, 1476, 1500,
 1420, 1421, 1498, 1484, 1485,  953, 1093, 1185, 1242, 1287,
 1243, 1244, 1279, 1336, 1308, 1280,  326,  225, 1384, 1282,
 1369, 1366, 1231, 1259, 1298, 1448, 1410, 1142, 1449,  594,
 1493, 1494, 1297, 1365, 1341, 1397, 1392, 1363, 1429, 1434,
 1395, 1398, 1399, 1479, 1435, 1393, 1394, 1489, 1477, 1478,
  950, 1036, 1153, 1127, 1179, 1154, 1155, 1194, 1090, 1177,
 1206,  535,  195,  113,  352,  197,  564,  442,  226, 1317,
  654,  655,  824,  840,  327,  409,  534,  305, 1158, 1159,
   46,  115,  306,  117,  118,  119,  120,  121,  122,  123,
  124,  125,  126,  127,  128,  129,  130,  131,  132,  133,
  134,  135,  136,  137,  254,  801,  994,  512,  723,  875,
  724,  725,  987,  138,  200,  729,  595,  596,  597,  598,
  795,  475,  476,  299,  992,  731,  410,  301,  499,  500,
  501,  502,  505,  738,  312,  753,  754,  892,  265,  481,
  766,  266,  480,  139,  140,  141,  142,  143,  144,  145,
  146,  147,  148,  149,  150,  151,  152,  153,  573,  574,
  575,  775,  776,  908,  777,  154,  561,  768,  353, 1010,
  549, 1073,  155,  494,  951, 1092, 1183, 1283,  468, 1163,
 1164, 1213, 1214,  825,  553,  338,  772, 1172,  554,  555,
  270,  271,  272,  158,  159,  160,  273,  274,  275,  276,
  277,  278,  279,  280,  281,  282,  283,  284,  172,  285,
  579,  173,  174,  319,  806,  633,  923,  850,  687,  961,
  921,  924, 1052,  925,  962,  963,  286,  175,  176,  177,
 1063,  998, 1064, 1065, 1066, 1108, 1067,  178,  179,  180,
  181,  704,  487,  705, 1055,  979, 1169, 1137, 1202,  706,
  978,  707, 1171, 1104,  182,  183,  184,  185,  186,  187,
  307,  525,  526, 1000, 1110,  315,  977,  859, 1136, 1007,
  898, 1111,  188,  420,  189,  421,  926, 1017,  422,  645,
  819,  816,  817, 1022,  423,  424,  425,  426,  427,  428,
  930,  635,  928, 1115, 1189, 1248, 1019, 1146, 1205,  814,
  641,  815, 1081, 1021, 1082, 1147, 1023,   17,   19,   47,
   48,  229,  659,  832,  443,  660,  661,
  };
  protected static readonly short [] yySindex = {          -54,
    0, -193, -117,  -72,  106,12157,    0,  257,    0,    0,
  106,  -72,    0,    0,  180,    0, 6802,  106,    0, -188,
 -231,    0,    0,    0,    0,    0,    0,    0,    0,  235,
    0,  301,    0,    0,    0, 1660,    0,    0,    0,    0,
    0,    0,    0,    0,  549,    0,    0,  559,    0,    0,
  257,  267,  106,    0,  291,    0,  243,  315, -259,11657,
  330,  232,  375, 6959,    0,  232,  232,  232, -163,  232,
  232,   12,    0,10760,  232,  232,    0,10760,    0,  374,
    0, -259,    0,  232,  351,  232,    0,11736,11736,  390,
  232,  232,    4,11440,    0,10760,    0,11440,11440,11440,
11440,11440,11440,11440,11440,    0, -237,    0, 8508,    0,
  179,    0,  371,  421,  533,  340,    0,    0,  441,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0, 1089,
  668,   50,   20,  462,  463,  508,  458,  518,  544,  408,
  543,    0,    0,    0,    0,    0,    0, 3463,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0, -262,  589, -266,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0, -168, -158,  267,
    0,  336,  574,  600,    0,  572,    0,    0,    0,    0,
 8508, 8508,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,  660,  584,    0,  621,    0, -260,
    0,    0,    0,  267,12653,  267,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,  806,  647,10896,    0,
    0,    0,    0,10760,  232,  232,  819,  367,  533,    0,
 -262,    0, 8508,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,  157,  158,11657,    0, 8508,10760,
  714,    0,    0,  733,10760,10760, 9119, -235,  -84,  759,
 8665,11440, -237,    0,  745,    0,  756, 8508,10760,  761,
 -118,  232,    0,10760,  374,10216,    0,    0,  351,10760,
  351, -264,  480,  855, -262,    0,  589,  340,  861, -262,
10760,10760,10760,  375,    0,  820,    0, 7116, -263,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0, 4346,
    0,    0,12086, -264,  768,  784,10760,    0,  788,    0,
  233,    0,    0,  461,    0,    0,  793, 8648, 9944,    0,
    0,11440,10760,10760,10760,10760,10760,10760,10760,10760,
10760,10760,10760,11440,11440,11440, 8508, 8508,11440,11440,
11440,11440,11440,11440,11440,11440,11440,11440,11440,11440,
11440,11440,11440,11440,10760,    0,    0,    0,    0,  589,
    0,    0,    0,    0,11736,11736, -262,    0,    0,    0,
    0,   97,  640,    0,    0,    0,    0,    0,    0,    0,
  267,  267,  807,    0,  811,    0,  788,  660,  660,    0,
   73,    0,  618,  660,  837,    0, -183,12653,    0,    0,
    0,    0, -174,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0, -145,12696,    0,    0,    0,
  788,    0,    0,  849,  445,    0,  854,    0,  859, -140,
  374,    0,  232,    0,    0, -262, 8196, -201,    0,  856,
    0,    0,    0, -156,  865,    0,  320,    0,  866,    0,
  864,    0,    0,    0,  465,    0, 8332,  664,10760,  759,
 9944,    0, 7587,    0,  351,    0,    0,    0,  867,  868,
    0,    0, -259,  374,  373,    0, 4187,  871,    0,  873,
 -262,    0,  874,    0,    0,    0,10760,  953,    0,    0,
    0,10760,  956,  877,    0,  881,  884,    0,12086,    0,
    0, -186, -221, 7116,    0,    0,    0,    0,    0,    0,
  374,    0,    0,  131,    0,    0,    0,  351, -264, -262,
 8822,    0,  885,    0,  889,11440,    0,  886, 7116,    0,
  349,    0,  354,    0,  788,    0,   13,10760,10760,  893,
 1009,    0,    0, -255,  895,    0,    0,    0,  668,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,  668,  668,   50,   50,   20,
   20,   20,   20,  462,  462,  463,  508,  458,  518,  544,
    0,  892, -198,    0,10760,  976, -262,  977, -262,  896,
10760,    0,    0,    0,  917,    0,  386,  788,    0,    0,
    0,    0,  517,  230,    0, 8822,  618,    0,  904,  905,
    0,    0,    0,    0,    0,    0, -264,  574,    0,  906,
  912,    0,    0,    0,    0,  916, 8979,  872,    0,  250,
    0,    0,  582,    0,10896,    0,  913,    0,    0,    0,
  577,  924,    0,  923,  930,  931,    0,    0,10760,    0,
    0, -262,    0,    0,  935,    0,  932,    0,10760, -119,
    0,    0, 6959, 6959, 8491,    0, 9119,    0,    0,10352,
  236,    0,  -55, -112,    0,  878,  890,    0,   28,    0,
    0,  943,    0,    0,    0,    0,    0,  942,    0,  950,
    0, 4505,    0,  374,    0,    0,  351,  511,  897,    0,
  168,    0,  947,  949,    0, 6959,    0, 6959,    0,10760,
    0,10760, 8508,    0,    0,  374,  948,  374,    0,    0,
    0,    0,    0,    0,    0,    0, 8805, 8508,    0,    0,
 -262,12086,  980,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0, 9808,    0,    0,    0,
    0,10080,10760,    0, 7744,  958,    0,    0, 1040,    0,
 1041,    0,  614,    0,  962,10760,10760, -262,    0,    0,
    0,    0,    0,  919,    0,   73,    0,    0,    0,    0,
  618,  618,    0,  807,  970,  971,  926,  974,  872,    0,
  972,    0, 1090, 1092,    0,    0,10760,    0,10488,  981,
  577, 8822, 8508,    0,  274, 1098, 1101, -126,  964,    0,
    0,10760,  990,    0,10760, 1083,    0,    0,    0,    0,
 -164,10624,    0,    0,    0,    0, 7880,    0, 1108,    0,
  589,10760,  998, 8491,  999,    0,    0, -262,    0,  203,
    0,    0,  788,  897,    0, -262,    0, -167,    0,    0,
    0, 1000,    0, 1027,    0,    0,    0,    0,    0,    0,
    0,    0,    0,  578,    0,    0,    0,    0, 8665,    0,
    0, -262,  995,  958,    0,10760,    0,10760,    0,10760,
    0,    0,    0,    0,    0,    0,    0,    0,    0, 1003,
  807,    0,    0,11032,    0,    0,    0, 1004, 7604,    0,
  872,    0,  872,    0,  872,    0,    0,    0,    0, -262,
 1005,  981,    0,    0,    0, -169, -166, 1002, 1008,    0,
    0,    0,    0,    0, 1010, 8491,  958, -198,10760,    0,
    0, 1012, 6959,    0,    0,    0,    0,    0,    0, 1017,
    0,  759,    0,    0,    0,    0,    0, -202,    0, 1016,
  788,  897,    0,  897,    0,  958, 1018,    0,    0,  374,
    0,  951, 1013,    0,    0,    0,10760, 1042,10760,    0,
10760, 1044,  256,    0,  912,  279,  613,    0,    0,    0,
    0,  -72,    0,    0,    0, 1026,    0,    0,    0, 1015,
    0,    0,    0,  455,    0, 1029, 1155, 1156,    0,    0,
  958, 1045,  958,    0, 1037,    0,    0,    0,    0,    0,
10760,    0, 1048, -209,    0, -209,    0,    0,    0,    0,
    0,    0,  374,    0,10760, 8039,    0,    0, 1072,  692,
 1050,    0,10760,    0, 1052,    0,    0,11032,  106, -140,
    0, 1051, 1051, 1051,10488, 1063,    0,10760,    0,    0,
    0,    0,    0, 6959, 1060,    0,    0, 7116,    0, 1069,
 6959,    0, 1068,    0,10760,    0,    0,    0,    0,    0,
    0,10760,    0,    0,  267, 1067,  267, 7761, -101, -101,
 -101,    0,    0,10760,    0, 6959,10760,    0,    0, 7116,
    0,    0,    0,    0, 1093,10760,10760,    0,  267, 1074,
    0, 1028,    0, 1070,    0,    0,    0, 1071,    0,    0,
 1030,    0, 1111,    0,    0,    0,    0,    0, 1075,  932,
    0, 7116,    0, 1103,    0, 1076, -101,    0, 1085,  267,
 7761, 1080, 1091,    0, 1094, 1095,    0, 1088,10760,    0,
    0,    0,    0, 1078, 1076,    0,    0, 1269,   79,  267,
    0, 6959,    0, 1104,10760, 1081,10760,    0,    0, 1096,
    0,    0, 1099,    0,    0,12696,    0, 1106,   79,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
  416,12696,    0,    0,    0,    0,    0,    0,    0,    0,
    0, 1107,  267,    0,   79,    0, -262,    0, 1104,    0,
    0, 1097, 1269,11902,    0,    0,  466,    0,    0,    0,
11934,    0,    0, 1110,    0,    0,    0,    0, 8508, 8508,
  370, 8665,  385,  351, 1126,    0, -264, 9744,    0, 1161,
    0,    0, 1076,    0,    0,    0, 1076,    0, 1061, 1062,
    0, 8508, -165,    0, 8508,    0, 1065, 1113,    0, -264,
    0, 1114, 9155,    0, 1112, 1087, -251,  575, 1660,    0,
    0, 1076,    0, -264,    0, 1119, 1105, 1116, 1125,    0,
 1117, 1062, 1135, -140, 1143, 1154,    0, 1157, 1162,    0,
  788,    0,  777,    0,    0,    0, 1159,    0, -154,    0,
 1150,    0,    0, 1163,    0, 1164, 1165, 1166,    0, 1168,
    0, -140, -140,    0, -140, 1160, 1170,    0,    0,    0,
    0, 1171,   81,    0, 1173, -140, 1293, 1176, -140,    0,
  466,    0, 8491, 1133, 1177, 1168,    0, 1180, 1183,   94,
 1186,    0,    0, -140,10488, 1140, 1184, 1171,    0,    0,
12696,    0,  267,  267,    0, 1141, 1185, 1173,    0, 1190,
    0,10760, 1145, 1189, 1176,    0, 1194, -140,    0, -153,
    0, 1191,    0,    0,    0,    0,    0,12696,    0,   94,
   94, 1200, 1202,    0, -154,    0,    0,   62, 1208,12696,
    0,12696,    0,    0, 8491, 1196,    0,    0,    0, 1209,
 1163,    0,    0,    0, 1211,    0,  154,    0,    0,    0,
 -101,  829, 1212,    0,    0,    0,    0,    0,    0,    0,
    0, 1266, 1321,    0,    0,    0,    0,    0, 1215, 1217,
 8491,    0,    0,    0,    0,   94,  579,  579,    0, -101,
    0,    0,    0,   19,   19,    0,    0,    0,    0,    0,
    0,    0, 9944, 9944,    0,    0,    0,    0,    0, 1221,
 1223, 1224,    0,    0,    0,
  };
  protected static readonly short [] yyRindex = {         2890,
    0,    0, 7273, 2890,    0,    0,    0, 1593,    0,    0,
 3117,  991,    0,    0,    0,    0,    0, 3117,    0,    0,
   52,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0, 1597,    0,    0, 1597,    0,    0,
 1593, 3180, 2959,    0,    0,    0,    0,    0,    0,    0,
    0,    0, 1231,    0,    0,    0,    0,    0,    0,    0,
    0,12176,    0, 1225,    0,    0,    0, 1225,    0,    0,
    0,    0,    0,    0,  140,    0,    0,    0,    0,    0,
    0,    0,    0,  192,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0, 4726,    0,    0,    0,
    0,    0,    0,  166, 4663, 3939,    0,    0, 4504,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0, 4819,
 4923, 5231, 5435, 5775, 5979, 6115, 6251, 6387, 6523, 1053,
 1538,    0,    0,    0,    0,    0,    0,   52,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,  207,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,  846,  846, 3227,
    0,  541, 1226,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0, 1597,  218,    0,    0,    0,    0,
    0,    0,    0, 3274,  246, 3341,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0, 3550,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0, 1236,    0,    0,    0,
    0,    0, 3550,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0, 2270,    0,
 1607,  118, 2400,    0,    0,    0, 2547, 2400,    0,    0,
    0,    0,    0, 1231,    0,    0,    0,   92,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0, 1233, 1284,    0,    0, 1225,    0, 3550,    0,
    0,    0,    0,    0,    0,    0,    0,    0, -226,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0, 1430,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,  121,    0,    0,    0,    0,    0,    0,    0,
 3384, 2714,    0,    0,    0,    0, 2117, 1597, 1597,    0,
 -199,    0, 8056, 1597, 1608,    0,    0,  159,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,  284,11589,    0,    0,    0,
 3550,    0,    0,    0,    0,    0,    0,    0,    0,11978,
    0,    0,    0,    0,    0,    0, 1234,    0,    0,    0,
    0,    0,    0,    0,    0,    0,  862,  957,    0,    0,
 1241,    0,    0,    0,    0,    0,   99,    0,    0, 4027,
 1239,    0,    0,    0,  263,    0,    0,    0,    0,    0,
    0,    0,    0,    0, 1821,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0, 1233,    0,
    0, 6642,    0,  115,    0,    0,    0,    0,    0,    0,
 8962,    0,    0,    0,    0,    0,    0, -127,  655,    0,
    0,    0, 1240,    0,    0,    0,    0,    0,    0,    0,
 3550,    0, 3550,    0, 4186,    0,    0,    0,    0,  181,
    0,    0,    0,    0,  -28,    0,    0,    0, 4991,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0, 5095, 5163, 5299, 5367, 5503,
 5571, 5639, 5707, 5843, 5911, 6047, 6183, 6319, 6455, 6579,
    0,    0,  700,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0, 3830,    0,    0, 2117,    0,    0,
    0,    0, 1195,    0,    0,    0,12230,    0,    0,  707,
    0,    0,    0,    0,    0,    0,  678,  698,    0,    0,
 1244,    0,    0,    0,    0, 1248,    0,    0,    0,    0,
    0,    0,11168,    0,    0,    0,  721,    0,    0,    0,
12247,    0,    0,  736,  742,  751,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0, 1242,    0,    0,    0,
    0,    0,    0,    0,    0,    0, 1249,    0,    0,    0,
 3616,    0,    0,  116,    0,  -74, 3709,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0, 1251,    0,    0,
    0,    0,    0,    0,    0,    0,   82,  144,    0,    0,
    0,    0,    0, 1250,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0, 8962,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,  190,    0,    0,    0, 1245,    0,    0,    0,    0,
    0,    0,  464,    0,  452,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0, -199,    0,    0,    0,    0,
12247, 8213,    0, 1246,    0,  712,    0,    0,    0,    0,
 1253,    0, 1206, 1227,    0,    0,    0,    0,    0, 1267,
12301,    0,    0,    0,12054,    0,    0,    0,  780,    0,
    0,    0,    0,    0,    0, 1988,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
 3868,    0, 4345, 1258,    0,    0,    0, 1255,    0,    0,
    0,    0,  343,    0,    0,    0,    0,  780,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0, 1268,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,  787,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0, 1270,    0,    0,    0,    0,    0,  794,  798,    0,
    0,    0,    0,    0,    0,    0, 1257,  700, 1272,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0, 4027,    0,    0,    0,    0,    0, 1274,    0,    0,
  343,    0,    0,  825,    0, 1257,    0,    0,    0, 8962,
    0,  631,  645,    0,    0,    0,    0,    0,    0,    0,
    0,    0,  123,    0, 1244, 9156,    0,    0,    0,    0,
    0,12343,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,  719,    0,  730,    0,    0,    0,    0,
 1277,    0, 1245,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0, 1282,    0, 7430,    0,    0,    0,    0,
    0,    0, 8962,    0,    0,    0,    0,    0,    0,  566,
  525,    0,    0,    0,    0,    0,    0,    0,12419,11978,
    0,  113,  113,  113,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,12462,    0,   59,    0, 1294, 1294,
 1294,    0,    0,    0,    0,    0, 1299,    0,    0, -172,
    0,    0,    0,    0,    0,    0,    0,    0,12505,    0,
    0,    0,    0, 1304,    0,    0,    0,  122,    0,    0,
    0,    0,  581,    0,    0,    0,    0,    0,    0, 1303,
    0, 1308,    0,    0,    0, 3028, 1301,  560,    0,  210,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0, 2827,    0,    0,    0, 9310, 9508,
    0,    0,    0,  628,    0,    0,    0,    0,    0,    0,
    0,    0,  366,    0,    0,11760,    0,    0, 9409,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,11828,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0, 9602,    0, 9310,    0,    0,    0,  628,    0,
    0,    0,    0,  284,    0,    0,    0,    0,    0,    0,
  284,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0, 4660,  324,    0, 9644,    0,    0,    0,
 9714,    0, 2827,    0,    0,    0, 2827,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,  669,
    0, 1322,    0,    0,    0,    0,    0,    0,    0,    0,
    0, 2827,    0,  760,    0,  649,    0,    0,    0,    0,
    0,    0,    0,11978,  797,    0,    0,    0,    0,    0,
 1315,    0,  772,    0,    0,    0,    0,    0,    0,    0,
  800,    0,    0,    0,    0,    0,    0,    0,    0, 1314,
    0,11978,11978,    0,12010,    0,    0,    0,    0,    0,
    0, 1317, 3070,    0, 1319,11978,11304, 1320,11978,    0,
    0,    0,    0,    0,    0, 1323,    0,    0,    0,12623,
    0,    0,    0,11978,    0,    0,    0, 1325,    0,    0,
  128,    0,12547,12585,    0,    0,    0, 1327,    0,    0,
    0,    0,    0,    0, 1329,    0,    0,11978,    0,  597,
    0,  808,    0,    0,    0,    0,    0,  840,    0,  411,
  699,    0,    0,    0,    0,    0,    0,    0,    0, 1383,
    0, 1443,    0,    0,    0,  813,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
  599,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,12623, 9676,11476,    0,  599,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0, 1239, 1239,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,
  };
  protected static readonly short [] yyGindex = {            0,
    0, 1657,    0,    0,    0,    1,   -9, -176,  -42, 1658,
    0, 1697, 1706,  624,    0,    0,   -3,    0,    0,    0,
    0,    0,    0, -850, -694, -210, -349,    0,    0,    0,
    0,    0, -191,    0,    0,    0,  771,    0,  882,    0,
    0,    0,    0,  627,  629,  -17, -234,    0,  -62,    0,
  473,    0,  500, -858, -595, -569, -567, -544, -504, -491,
 -486,    0,-1160,    0,   10,    0,  114,    0,-1083,    0,
    0,    0,  145,  295,    0,    0,    0,  333,-1055,    0,
 -271, -298, 1046,    0,    0,    0, -882,  281,    0,    0,
 -500,    0,    0,  350,    0,    0,  325,    0,    0,  358,
    0, -516, -867,    0,    0,    0,    0,    0,  447,  -13,
    0,    0,  879,  883,  888, 1054, -528,    0,    0, -275,
  898,  443,    0, -838,    0,    0,    0,    0,    0,    0,
    0,    0,  251,    0,    0,    0,    0,    0,    0,    0,
    0,  501,    0,    0,    0,    0, -239,  437,    0,    0,
    0,    0,    0,    0,    0,    0,    0,  519,    0, -505,
    0,    0,    0,    0,    0,    0,    0,    0,    0,  269,
    0,    0,  355,    0,    0,  360,  359,  278,    0,    0,
    0,    0,    0,    0,    0,    0,  580,    0,    0,    0,
    0,  -52,    0,  -12, -189,    0,    0,  424,    0,  481,
    0,  936,    0, 1237, -290, -267,  -63,  812,    0,  586,
    0,  -30,   11,    0,    0,   21,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0, -261,    0,  587,    0,    0,    8,    0,    0,
    0,  894,    0, -296, -125, 1038,  968,    0,  973,    0,
 1181, 1404, 1102,    0,    0,  781, 1716,    0,    0,    0,
    0, 1073,    0,    0,    0,    0,    0, -685, 1455,    0,
    0,    0,    0,    0, 1118,  382,  818,  785,  817, 1392,
 1396, 1398, 1405, 1407,    0, 1440,    0,    0,    0, 1079,
 1296, -730,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0, -299,    0,    0,    0,    0, -447,    0,
  684,    0,  596,    0,  679,    0,    0,    0,  744, -529,
  -16, -315,   -5,    0, 1696,    0,   66,    0,   68,  101,
  132,  135,  143,  147,  156,  162,  163,  176,    0, -672,
    0,  -21,    0,    0,  880,    0,  802,    0,    0,    0,
  786, -241,  850, -844,    0,  902, -453,    0,    0,    0,
    0,    0,    0,  796,    0,    0,  791,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,  728,    0,    0,    0,    0,    0,    0,    0,    0,
  -29,    0, 1341,    0,    0,    0,  969,    0,    0,    0,
    0,    0,    0, -171,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0, 1445,    0,    0,    0,    0,
    0,    0,    0,    0,    0,  622,    0,    0,    0,    0,
    0,    0,    0,    0,  723,    0,    0,    0,    0,    0,
    0,  -19, 1043,    0,    0,    0, 1047,
  };
  protected static readonly short [] yyTable = {           110,
  156,  514,  434,  112,   18,  732,  294,  196,  190,  235,
  517,  157,  737,  431,  473,   44,  193,  430,  490,  681,
  533,  318,  557,  449,  773,  324,  329,  116,  231,  259,
  336,  510,  406,  703,  544,  498, 1012,  905,  261, 1132,
  866,  867,  781,  885,  469,  253, 1165, 1166,  310,  786,
  363,  934,  371, 1060,  708, 1254,  304,  804, 1061,  539,
  304,  362,   14,  370,  445, 1061,  311,  191,  313,  769,
  116, 1261,  664, 1332,  116,  196,  196,  571,  339, 1015,
  260,  674,  161,  897,  162,  899, 1043,  413, 1005, 1045,
 1319,  880,  709, 1193,   48,  785,  196,  413,  665,  712,
  292, 1360, 1446,   20,  290,  348,   48,   95,  559,  411,
  408,  251,  291,  572,  260,  689,  799,  163,  260,  260,
  260,  260,  260,  260,  260,  260,  349,  828,  360,  974,
  666,  677,  434,  473,  503,  678,  864,  522,  504,   16,
  110,  156,  634,  876,  112,  447,  770,  235,  164,  350,
  523,  165,  157,  349, 1160,  293,  348,   43,    6,  166,
  252,  292,  412,  167,  434,  560, 1333,  524,  116,  580,
  414,  506,  168,  800,  865,  415,  350,  416,  169,  170,
  414,  934,  805,  417,  418,  415,  474,  416,  196,  196,
  351,  448,  171,  417,  418,  785,  679,  198,  478,  292,
  690,    1,  447,  984,  349,  440, 1015,  771, 1003,  882,
  726,  730, 1062,  976,  868,  713,  293,  346,   48, 1062,
  449, 1447, 1126,  161,   15,  162, 1361,  350,  233,  192,
 1428,  472,  325,  330,  667,  360,  477, 1161,  557,  446,
  469,  351,  360,  675,  360,  975,  360,  649, 1044,  785,
  196, 1046, 1320,  355,  293,  532,  259, 1452,  163,  536,
  419,  877,  538,  557,  541,  486,  259,  543,  792, 1462,
  429, 1463,  489,  571, 1495,  540,  196,  493,  495, 1074,
  334,  682,  507,  880,  508,    2,  251,  750,  196,  164,
  360,  520,  165,  934,  531,  196,  528,  116,  530,  934,
  166,  529,  493,    2,  167,  474,  474,  260,  934,  572,
 1057,  387,  873,  168,  546,  547, 1069,  260, 1070,  169,
  170,  556,  260,  966, 1459, 1192,  116,  578,  652, 1424,
 1085,  572,  558,  171, 1215,  252, 1390,  196,  509,  304,
  196, 1234, 1112,  448, 1208,  438,  439,  388,  116, 1417,
  472,  592,  636,  638,  640,  600,  601,  602,  603,  604,
  605,  606,  607,  608,  609,  610,  292, 1472,  655,    3,
    4,    5,    6,  874,  196,  196, 1460,  656,  297,  434,
  298,  829,  793,  996, 1234,   95,  685,  632,  235, 1293,
  636,  788,  260,  790, 1496,  791, 1492,  881,  653,  858,
  448,  882,  196,  196,  260,  260,  260,  479,  980,  260,
  260,  737,  482,  391,  392,    6,  572, 1469,  650,  651,
  196,  293,  335,  700,  662,  710,  312,  389,  390,  647,
  636, 1138, 1310,  488,  196,  636, 1311,  750, 1143,  636,
   20,  233,  636,  668,  356,  642,  643,  233,  474,  233,
  473,  655,  519,  353,  636,  498, 1348,  738,  821,  679,
  656, 1338,  233, 1168,  749,  702,  435,  698,  757,  888,
  888,  887,  887,  346,  331, 1050,   48,  434, 1470,  655,
  331,  636,  680,  436, 1378, 1379,  331, 1381,  656,  331,
  331,  728,  357,  592,  563,  735,  483,  116, 1400,  353,
  636, 1407,  733,  331,  778,  738,  571,  780,  733,  741,
  743,  413,  324,   49,  751,  324, 1423,  636,  358,  759,
  679,  842,  636,  748,  761,  644,  636,  435,  324, 1246,
  292,  614,  615,  484,  324,  331,  196,  556,  737,  201,
 1445,  636,  572,  680,  436,  843,  779,  774,  558,  888,
  959,  887,  623,  358,  623,  634,  346,  733,  196,  637,
  639,  452,  556,  452,  116,  292,  738,  915,  636,  726,
  794,  794,  844,  558,  809,  964,  811,  314,  348,  634,
  726, 1454, 1455,  324,  346,  893,  485,  636,  346,  116,
  346,  346,  346,  346,  414,  635,  260,   55,  346,  415,
  878,  416,  251,  826, 1235,  202,  634,  417,  418,  623,
  773,  247,  572,  991,  968,  248,  872,  807,  452,  635,
 1001,  348,  474,  813,  726, 1291,  349,   48,  517,  358,
 1236,  358, 1237,  358,  358,  233,  358, 1486,  358,  860,
 1294,  523,  233,  196,  199,  827,  635, 1235,  263,  350,
  581,  252,  287,  288,  289, 1238,  295,  296,  744, 1086,
  582,  308,  309,  351,  196,  249,  246,  472,  314,  349,
  316,   48,  320, 1236, 1292, 1237,  199,  332,  333, 1054,
  358,  493,  358,  703,  536,  358,  259,  841,  894, 1295,
  250,  863,  350,  730,   48, 1239,  360,  735, 1238,  331,
  715,  369,  728,  649,  662,  262,  351,   48, 1240, 1031,
  349,  331,   48, 1241,  331,  331,  433,   48,  910,   48,
   48,   48,   48,  116,  116,  317,  349,   48,  331,  196,
  889,   48,  662,  350,  474, 1051,  317,  260, 1239,  474,
   95,  662,  900,   48,  901,  264,   48,  351,   48,  350,
  196, 1240,  903,  599,  774,  936, 1241,  349,  435,  907,
  331,  820,  350,  737, 1051,  196,  116,  350,  116,  196,
  616,  617,   48, 1315,   48,   48,  787,  360,  194,  592,
  350,  789, 1141,  653,  592,  917, 1328,  735,  359, 1105,
  912, 1258,  557,  361,  351,  912,  959,  912,  934,  935,
  912,  912,  918,  912,  912, 1346,  317,  918,  403,  918,
  822,  821,  918,  918,  684,  918,  918,   45,  685,  372,
  404,  196,  823, 1096,  557,  999, 1135, 1002,  114,  956,
  668,  489,   95, 1004,  718, 1097,  360,  918,  719,  196,
  196, 1284,  314,  393,  394,  369,  400,  982,  948, 1217,
 1233, 1084,  435, 1031,  985,  681,  557,  395,  396, 1013,
  397,  398, 1167,  914,  993,  317,  735,  360,  914, 1217,
  914,  114, 1411,  914,  914,  114,  914,  914,  583,  986,
  912,  474,  565,  890,  851,  511,  888,  845,  584,  566,
  995,  399,  918, 1233,  846, 1217,  317, 1040,  527,  328,
  328,  567,  401,  251,  921,  364,  196,  902, 1016,  921,
 1018,  921, 1020,  331,  921,  921,  331,  921,  921,  227,
  328,  228,  909,  565,  365,  366, 1030,  317,  196,  227,
  566,  230,  405,  317, 1464, 1251,  196,  668,  449,  921,
  318,   95,  567,  402,  367,   95,  435,  658, 1149, 1011,
 1334,  802,  252,  914, 1191,  368,  658,  298,  735,   48,
 1037,  493, 1038,  268, 1039,  657,  908,  931,  932,  114,
 1483,  908,  268,  908,  657,  436,  908,  908,  414,  908,
  908,  517,   48,  415, 1087,  416, 1088, 1501, 1502,  437,
   36,  417,  418,  116,  921,   48,  774,  967,  441, 1077,
   48, 1079,  764, 1080,  764,   48,  764,   48,   48,   48,
   48,  331,  328,  328, 1281,   48,  753,  331,  753,   48,
  753, 1281,  277,  331,  277,  474,  599,  331,  599,  277,
  227,   48, 1089,  727,   48,  336,   48,  504,  444,  336,
  331,  331,  115,  489,  115, 1117, 1118,  331,  331,  115,
  331,  331,   57,  384,  385,  386,  908, 1113,  735,  774,
   48,  470,   48,   48,  471, 1120,  196,   65,   65,  699,
 1030,   65,  331,  743,  328,  743,  954,  489,  954, 1125,
  489,   66,  235,  336, 1152,   66,  959,  434,  167,  491,
  167,  556,  167, 1035,  755,  334,  755, 1145,  114,  179,
  328,  179,  558,  179, 1148,  155,  235,  155,  492,  742,
 1326,  162,  328,  162,  116,  196,  489,  361,  116,  328,
  163,  116,  163,  556, 1326,  513,  336,  114, 1174, 1080,
  336,  518,  331,  336,  558,  336,  521, 1152,  346,  576,
  336, 1356,  346, 1357,  331,  346,  116,  346,  537,  114,
  116,  865,  346,  865,  542,  556, 1216, 1232,   68,  577,
   68,  328,  550,  185,  328,  185,  558,  156,  196,  156,
  120, 1204,  120,  282,  336,  282, 1216,  620,  621,  622,
  623,  127,  116,  127, 1265,  196,  289, 1249,  289,  489,
  350,  441, 1473, 1474,  887,  887,  519,  519,  328,  328,
 1232,  350, 1216,  636,  636,  663, 1130, 1131,  618,  619,
  585,  337,  116,  624,  625,  340,  341,  342,  343,  344,
  345,  346,  347,  683,  354,  686,  328,  328,  648,  495,
  688,  711,  331,  716,  331,  495,  714,  717,  739,  740,
  196,  196,  755, 1286,  756,  758,  760,   36,  196,  762,
  763,   36,  764,  331,  331,  765,  196,  196,  782,  196,
  783,  785,   36,  797,  798, 1306,  803,   36,  802,  808,
  810,   36,  818,  331,   36,  830,  812, 1306,  831,  196,
  833,  331,  196,  331,  331,  834,   36,   36,  837,   43,
 1306,   36,   36,  849, 1335,  854,  855,   36,  114,   36,
   36,   36,   36,  856,  857,  862,  198,   36,  580, 1306,
  861,   36,  879,   36,  883,  884,  886,  891,  895,  904,
 1391,  896,  912,   36,  496,   36,   36,  410,   36,  410,
  496,  922,   36,  927,  929,  933,  937, 1418,  328,  943,
  947,  944, 1212,  946,  976,  952,  949,  954,  410,  410,
 1430, 1432,   36,  972,  960,  735,  973, 1286,   36,   36,
  328,  981,  983,  990,  507,  114,  997,  489,  410, 1009,
 1014, 1008, 1024, 1075, 1033, 1047,  410, 1418, 1418,  410,
 1041, 1048,  328, 1056, 1440, 1049, 1058, 1068, 1078, 1072,
  114,  580, 1091, 1076, 1083, 1095,  580, 1212,  580,  580,
  580,  580,  580,  580,  580,  580,  580,  580,  580, 1098,
 1099, 1100, 1103, 1289, 1290, 1106, 1101,  735, 1116,  337,
  580, 1121,  580, 1119,  580, 1128,  580,  580,  580,  500,
  474,  474, 1133, 1418, 1139, 1141, 1318, 1144, 1151, 1321,
 1173, 1176,  580, 1180, 1181, 1178, 1187, 1182,  545, 1161,
 1190, 1191, 1195,  735, 1198, 1203, 1250, 1199, 1207, 1247,
 1200, 1201, 1296, 1488, 1488, 1309, 1252,  328, 1267,  373,
 1497, 1497, 1253, 1255, 1262,  592,  592, 1288, 1313, 1314,
 1330,  580, 1323, 1324, 1327, 1340, 1342, 1345,  328,  545,
  374,  375,  376,  377,  378,  379,  380,  381,  382,  383,
 1343,  611,  612,  613, 1331, 1347,  545,  545,  545,  545,
  545,  545,  545,  545,  545,  545,  545,  545,  545,  545,
  545,  545, 1333, 1349,  114,  114, 1351, 1353, 1354, 1359,
 1364, 1367, 1382,   25, 1370,   26, 1371, 1372,   27,  331,
 1209, 1374, 1383,   28, 1386,  331, 1396,   29, 1401, 1403,
 1412, 1415, 1413,  328, 1416, 1422,   31, 1425, 1436, 1426,
 1437, 1439, 1441,   33, 1442, 1444, 1456,  114,   34,  114,
 1210, 1450,   35, 1457,  328, 1461, 1465,  331, 1466, 1475,
 1460,  331, 1468, 1459,   37, 1481,   38, 1482, 1503,  328,
   39, 1211,    9,  328, 1504, 1505,  950,  532,   40,   41,
  837,   32,   42,  490,  600,  321,  356,  942,  491,  805,
  447,  601,  669,   30,   22,   31,  489,  809,   30,  309,
  747,  515,  331,  206,  545,   31,  845,  331,  749,  331,
  331,  331,  331,  331,  331,  331,  331,  331,  331,  331,
  331,  779,  757,  748,   96,  758,  331,  810,  750,  780,
  331,  331,  331,  331,  331,  331,  331,  331,  331,  331,
  657,  331,  331,  328,  328,  331,  331,  331,  331,  331,
  812,  313,  331,  331,  814,  679,  657,  331,  331,  331,
  331,  331,  331,  331,  331,  500,  354,  331,  338,  123,
  500,  500,  105,  784,  285,  130,  331,  636,  124,  331,
  106,  331,  286,  331,  131,  636,  331,  232,   54,  236,
   21, 1025,  331,  500, 1123,  942, 1124, 1264, 1256, 1458,
 1427, 1467,  839,  500, 1303, 1414,  500,  500, 1409, 1443,
  328,  500, 1316,  969,  500, 1499,  500,  970,  500,  500,
  500,  500,  971, 1263,  852, 1337,  500, 1491,  965, 1260,
  500, 1431, 1438, 1433,  500, 1490, 1358,  871, 1307, 1196,
  328,  938,  500,  752,  914,  500, 1197,  500,  500,  796,
  989,  587, 1059,  500,  916,  500,  500,  500,  500,  500,
  500,  500,  500,  500,  500,  500,  848,  300,  548,  869,
  626,  500,  500,  582,  114,  627,  500,  500,  628,  500,
  500,  500,  500,  500,  500,  500,  629,  500,  500,  630,
  500,  500,  500,  500,  500,  500,  500,  500,  500,  500,
  839,  500,  500,  500,  500,  500,  500,  500,  500,  500,
  500,  500,  500,  500,  500,  500,  500,  500,  500,  500,
  500,  500,  500,  631,  767,  500, 1184,  500, 1268,  500,
 1188, 1140,  500,  407, 1102, 1071, 1109, 1053,  500, 1107,
  911, 1114,  356, 1042, 1170,  745, 1006,  646,  356, 1175,
 1266,    0,    0,    0,  940,    0,  582,  939,    0,    0,
    0,  582,    0,  582,  582,  582,  582,  582,  582,  582,
  582,  582,  582,  582,    0,    0,    0,    0,    0,    0,
  356,    0,    0,    0,  356,  582,    0,  582,    0,  582,
    0,  582,  582,  582,    0,  114,    0,    0,    0,  114,
    0,    0,  114,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,  328,
    0,    0,    0,    0,  203,  356,    0,  114,    0,    0,
  356,  114,  356,  356,  356,  356,  356,  356,  356,  356,
  356,  356,  356,    0,    0,    0,  582,    0,    0,  356,
    0,    0,    0,  356,  356,    0,  356,  356,  356,    0,
  356,  356,  356,  114,  356,  356,  204,  774,  356,  356,
  356,  356,  328,    0,  545,  356,  356,    0,    0,    0,
  356,  356,  356,  356,  356,  356,  356,  356,    0,  328,
    0,    0,    0,  114,    0,    0,    0,    0,    0,  356,
    0,    0,  356,    0,  356,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,  356,  205,  206,  207,  208,
    0,  209,  210,  211,  212,  213,  214,  215,  216,    0,
    0,  217,  218,  219,  220,  221,  222,  223,  224,    0,
    0,    0,    0,    0,  328,  328,    0,    0,    0,    0,
    0,    0,  328,    0,    0,    0,  839,  839,    0,    0,
  328,  328,    0,  328,  839,  839,  839,  839,  839,    0,
  839,  839,    0,  839,  839,  839,  839,  839,  839,  839,
  839,    0,    0,  328,    0,  839,  328,  839,  839,  839,
  839,  839,  839,    0,    0,  839,  331,    0,    0,  839,
  839,    0,  839,  839,  839,    0,    0,    0,    0,    0,
    0,    0,    0,    0,  839,    0,  839,    0,  839,  839,
    0,    0,  839,    0,  839,  839,  839,  839,  839,  839,
  839,  839,  839,  839,  839,  839,    0,  839,    0,    0,
  839,  839,    0,    0,  839,  839,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,  839,
  839,  839,  839,  839,    0,    0,    0,  839,  839,    0,
    0,  839,    0,    0,    0,    0,  839,  839,  839,  839,
  839,    0,    0,    0,  839,    0,  839,    0,    0,    0,
    0,    0,  839,  839,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,  839,  839,  839,
  839,    0,  839,  774,  774,    0,    0,    0,    0,  839,
    0,  774,  774,  774,  774,  774,    0,  774,  774,    0,
  774,  774,  774,  774,  774,  774,  774,    0,    0,  734,
    0,    0,  774,    0,  774,  774,  774,  774,  774,  774,
    0,    0,  774,    0,    0,    0,  774,  774,    0,  774,
  774,  774,    0,    0,    0,    0,    0,    0,    0,    0,
    0,  774,    0,  774,    0,  774,  774,    0,    0,  774,
    0,  774,  774,  774,  774,  774,  774,  774,  774,  774,
  774,  774,  774,    0,  774,    0,    0,  774,  774,    0,
    0,  774,  774,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,  774,  774,  774,  774,
  774,    0,    0,    0,  774,  774,    0,    0,  774,    0,
    0,    0,    0,  774,  774,  774,  774,  774,    0,    0,
    0,  774,  331,  774,    0,    0,    0,  331,  331,  774,
  774,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,  324,
  331,    0,    0,    0,  774,  774,  774,  774,    0,  774,
  331,    0,    0,  331,  331,    0,  774,    0,  331,    0,
    0,  331,    0,  331,    0,  331,  331,  331,  331,    0,
    0,    0,    0,  331,    0,    0,    0,  331,    0,    0,
    0,  331,    0,    0,    0,    0,    0,    0,    0,  331,
    0,    0,  331,    0,  331,  331,    0,    0,    0,    0,
  331,    0,  331,  331,  331,  331,  331,  331,  331,  331,
  331,  331,  331,  331,    0,    0,    0,    0,  331,  331,
    0,    0,    0,  331,  331,  331,  331,  331,  331,  331,
  331,  331,  331,    0,  331,  331,    0,    0,  331,  331,
  331,  331,  331,    0,    0,  331,  331,    0,    0,    0,
  331,  331,  331,  331,  331,  331,  331,  331,    0,    0,
    0,    0,    0,    0,    0,  734,    0,    0,    0,  331,
  734,  734,  331,    0,  331,    0,  331,    0,    0,  331,
    0,    0,    0,    0,    0,  331,  361,    0,    0,    0,
    0,    0,    0,  734,    0,    0,    0,    0,    0,    0,
    0,    0,    0,  734,    0,    0,  734,  734,    0,    0,
    0,  734,    0,    0,  734,    0,  734,    0,  734,  734,
  734,  734,    0,    0,    0,    0,  734,    0,    0,    0,
  734,    0,    0,    0,  734,    0,    0,    0,    0,    0,
    0,    0,  734,    0,    0,  734,    0,  734,  734,    0,
    0,    0,    0,  734,    0,  734,  734,  734,  734,  734,
  734,  734,  734,  734,  734,  734,    0,    0,    0,    0,
    0,  734,  734,    0,    0,    0,  734,  734,  734,  734,
  734,  734,    0,  734,  734,  734,    0,  734,  734,    0,
    0,  734,  734,  734,  734,  324,    0,    0,  734,  734,
  324,  324,    0,  734,  734,  734,  734,  734,  734,  734,
  734,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,  734,  324,    0,  734,    0,  734,    0,  734,
    0,    0,  734,  324,    0,    0,  324,  324,  734,    0,
    0,  324,    0,    0,  324,    0,  324,    0,  324,  324,
  324,  324,    0,   49,    0,    0,  324,    0,    0,    0,
  324,    0,    0,    0,  324,    0,    0,    0,    0,    0,
    0,    0,  324,    0,    0,  324,    0,  324,  324,    0,
    0,    0,    0,  324,    0,  324,  324,  324,  324,  324,
  324,  324,  324,  324,  324,  324,    0,    0,    0,    0,
    0,  324,  324,    0,    0,    0,  324,  324,  324,  324,
  324,  324,    0,  324,  324,  324,    0,  324,  324,    0,
    0,  324,  324,  324,  324,    0,    0,    0,  324,  324,
    0,    0,    0,  324,  324,  324,  324,  324,  324,  324,
  324,    0,  361,    0,    0,    0,    0,  361,  361,    0,
    0,    0,  324,    0,    0,  324,    0,  324,    0,  324,
    0,    0,  324,    0,    0,    0,   28,    0,  324,    0,
  361,    0,    0,    0,    0,    0,    0,    0,    0,    0,
  361,    0,    0,  361,  361,    0,    0,    0,  361,    0,
    0,  361,    0,  361,    0,  361,  361,  361,  361,    0,
    0,    0,    0,  361,    0,    0,    0,  361,    0,    0,
    0,  361,    0,    0,    0,    0,    0,    0,    0,  361,
    0,    0,  361,    0,  361,  361,    0,    0,    0,   35,
  361,    0,  361,  361,  361,  361,  361,  361,  361,  361,
  361,  361,  361,    0,    0,    0,    0,    0,  361,  361,
    0,    0,    0,  361,  361,    0,  361,  361,  361,    0,
  361,  361,  361,    0,  361,  361,    0,    0,  361,  361,
  361,  361,    0,    0,    0,  361,  361,    0,    0,    0,
  361,  361,  361,  361,  361,  361,  361,  361,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,   34,  361,
    0,    0,  361,    0,  361,    0,    0,    0,    0,    0,
   49,    0,    0,    0,   49,  361,   49,    0,   49,    0,
   49,    0,    0,   49,    0,   49,   49,    0,   49,    0,
   49,    0,   49,    0,   49,   49,   49,   49,    0,    0,
   49,   49,    0,    0,    0,    0,   49,   49,   49,   49,
   49,    0,    0,   49,   49,   49,    0,   49,    0,   49,
   49,   49,   49,   49,   49,   49,   49,   28,   49,   49,
   49,   49,    0,    0,   49,   49,   49,    0,   49,    0,
    0,    0,    0,   49,   49,    0,   49,   49,    0,   49,
   49,   49,    0,    0,    0,   49,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,   49,    0,   49,   49,    0,
    0,    0,   28,   28,    0,    0,    0,   28,    0,    0,
   49,   28,    0,   28,    0,    0,   28,    0,   28,   28,
    0,   28,    0,   28,    0,   28,    0,   28,   28,   28,
   28,    0,    0,   28,   28,    0,   33,    0,    0,   28,
    0,   28,   28,   28,    0,    0,   28,   28,   28,    0,
   28,   49,    0,   28,    0,   28,   28,   28,   28,    0,
    0,    0,   28,   28,   28,    0,   35,   28,   28,   28,
   35,    0,    0,    0,    0,    0,   28,   28,    0,   28,
   28,   35,   28,   28,   28,    0,   35,    0,   28,    0,
   35,    0,    0,   35,    0,    0,    0,    0,    0,    5,
    0,    0,    0,    0,    0,   35,   35,    0,   28,    0,
   35,   35,    0,    0,   28,   28,   35,    0,   35,   35,
   35,   35,    0,   28,    0,    0,   35,    0,    0,    0,
   35,    0,   35,    0,    0,   34,    0,    0,    0,   34,
    0,    0,   35,    0,   35,   35,  932,   35,    0,    0,
   34,   35,    0,    0,    0,   34,    0,    0,    0,   34,
    0,    0,   34,    0,   28,    0,    0,    0,    0,    0,
    0,   35,    0,    0,   34,   34,    0,    0,   35,   34,
   34,    0,    0,    0,    0,   34,    0,   34,   34,   34,
   34,    0,    0,   48,    0,   34,    0,    0,    0,   34,
    0,   34,    0,    0,   28,    0,    0,    0,   28,    0,
    0,   34,    0,    0,   34,    0,   34,    0,    0,   28,
   34,    0,    0,    0,   28,    0,    0,    0,   28,    0,
    0,   28,    0,    0,    0,    0,    0,    0,    0,    0,
   34,    0,    0,   28,   28,    0,   34,   34,   28,   28,
   48,    0,   48,    0,   28,    0,   28,   28,   28,   28,
    7,    0,    0,    0,   28,    0,    0,    0,   28,    0,
   28,    0,    0,   48,    0,    0,    0,    0,    0,    0,
   28,    0,    0,   28,    0,   28,   48,    0,    0,   28,
    0,   48,    0,   33,    0,    0,   48,   33,   48,   48,
   48,   48,    0,  933,   48,    0,   48,    0,   33,   28,
   48,    0,    0,   33,    0,   28,   28,   33,    0,    0,
   33,    0,   48,    0,    0,   48,    0,   48,    0,    0,
    0,    0,   33,   33,    0,    0,    0,   33,   33,    0,
    0,    0,    0,   33,    0,   33,   33,   33,   33,    0,
    0,   48,    0,   33,    0,    0,    5,   33,    0,   33,
   48,    0,    0,    0,    0,    0,    0,    0,    0,   33,
    0,   48,   33,    0,   33,    0,   48,    0,   33,    0,
   48,    0,    0,   48,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,   48,   48,    0,   33,    0,
   48,   48,    0,  932,    0,   33,   48,   48,   48,   48,
   48,   48,    0,    0,    0,    0,   48,    0,   48,    0,
   48,    0,   48,   48,    0,    0,    0,   48,    0,    0,
   48,    0,   48,    0,    0,   48,    0,   48,    0,    0,
    0,   48,   48,   48,    0,    0,    0,   48,   48,    0,
   48,    0,    0,   48,   48,   48,   48,   48,   48,    0,
    0,   48,    0,   48,    0,   48,    0,   48,    0,   48,
   48,    0,    0,    0,   48,    0,    0,   48,    0,   48,
    0,    0,   48,    0,   48,    0,    0,    0,   48,   48,
   48,    0,    0,    0,   48,   48,    0,    0,    0,    0,
   48,    0,   48,   48,   48,   48,    0,    0,   48,    0,
   48,    0,    0,    0,   48,    0,   48,    7,    0,    0,
    0,   49,    0,    0,    0,    0,   48,    0,    0,   48,
    0,   48,   49,    0,    0,   48,    0,   49,    0,    0,
    0,   49,    0,    0,   49,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,   48,   49,   49,    0,    0,
  933,   49,   49,    0,   48,    0,    0,   49,    0,   49,
   49,   49,   49,    0,    0,   48,    0,   49,    0,    0,
   48,   49,    0,   49,   48,    0,    0,   48,    0,    0,
    0,    0,    0,   49,    0,    0,   49,    0,   49,   48,
   48,    0,   49,    0,   48,   48,    0,    0,    0,    0,
   48,    0,   48,   48,   48,   48,    0,    0,    0,    0,
   48,    0,   49,    0,   48,    0,   48,    0,    0,    0,
    0,    0,    0,    0,    0,    0,   48,    0,   56,   48,
    0,   48,    0,    0,    0,   48,   57,   25,   58,   26,
    0,    0,   27,   59,    0,   60,   61,   28,   62,   63,
   64,   29,    0,    0,    0,   48,    0,   65,    0,   66,
   31,   67,   68,   69,   70,    0,    0,   33,    0,    0,
    0,   71,   34,    0,   72,   73,   35,    0,    0,    0,
    0,    0,    0,    0,    0,    0,   74,    0,   37,    0,
   38,   75,    0,    0,   39,    0,   76,   77,   78,   79,
   80,   81,   40,   41,   82,   83,   42,   84,    0,   85,
    0,    0,   86,   87,    0,  331,   88,   89,    0,    0,
    0,  331,    0,    0,    0,    0,    0,    0,    0,    0,
    0,   90,   91,   92,   93,   94,    0,    0,    0,   95,
    0,    0,    0,   96,    0,    0,    0,    0,   97,   98,
   99,  100,  101,    0,    0,    0,  102,  331,  103,    0,
    0,    0,    0,    0,  104,  105,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,  331,    0,    0,    0,    0,    0,  331,    0,  106,
  107,  108,  109,    0,    0,    0,    0,    0,  331,    0,
    0,  198,    0,  331,    0,  331,  331,  331,  331,  331,
  331,  331,  331,  331,  331,  331,  331,    0,    0,    0,
    0,    0,  331,  331,    0,    0,    0,  331,  331,  331,
  331,  331,  331,  331,  331,  331,    0,  331,  331,    0,
  331,  331,  331,  331,  331,  331,  331,  331,  331,  331,
    0,  331,  331,  331,  331,  331,  331,  331,  331,  331,
  331,  331,  331,  331,  331,  331,  331,  331,  331,  331,
  331,  331,  331,    0,  502,    0,    0,  331,    0,  331,
  502,    0,  331,    0,    0,    0,    0,    0,  331,    0,
    0,    0,    0,  331,    0,    0,  331,    0,  331,  331,
    0,    0,    0,  331,  331,    0,    0,  331,  331,  331,
  331,  331,  331,  331,  331,  331,  502,  331,  331,  331,
  331,  331,  331,  331,  331,  331,  331,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,  331,  331,    0,
    0,    0,    0,    0,    0,  331,    0,    0,  331,    0,
    0,    0,    0,    0,  331,    0,    0,  502,    0,    0,
    0,    0,  502,    0,  502,  502,  502,  502,  502,  502,
  502,  502,  502,  502,  502,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,  502,  502,  502,  502,  502,
  502,  502,  502,  502,  502,  924,  502,  502,    0,  502,
  502,  502,  502,  502,  502,  502,  502,  502,  502,    0,
  502,  502,  502,  502,  502,  502,  502,  502,  502,  502,
  502,  502,  502,  502,  502,  502,  502,  502,  502,  502,
  502,  502,    0,  498,    0,    0,    0,    0,  502,  498,
    0,    0,    0,    0,    0,    0,    0,  502,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,  498,    0,    0,  924,    0,
    0,    0,    0,  924,    0,  924,  924,  924,  924,  924,
  924,  924,  924,  924,  924,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,  924,    0,  924,
  388,  924,    0,  924,  924,  924,  498,    0,    0,    0,
    0,  498,    0,  498,  498,  498,  498,  498,  498,  498,
  498,  498,  498,  498,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,  498,  498,  388,  498,  498,  498,
  498,  498,  498,  498,    0,  498,  498,    0,  498,  498,
  498,  498,  498,  498,  498,  498,  498,  498,  924,  498,
  498,  498,  498,  498,  498,  498,  498,  498,  498,  498,
  498,  498,  498,  498,  498,  498,  498,  498,  498,  498,
  498,    0,  506,    0,    0,    0,    0,  498,  506,    0,
  498,    0,    0,    0,    0,    0,  498,    0,    0,    0,
    0,  324,    0,    0,    0,    0,    0,  324,    0,  388,
  388,  388,  388,    0,  388,    0,  388,  388,    0,  388,
  388,  388,  388,  388,  506,  388,  388,  388,  388,    0,
  388,  388,  388,  388,  388,  388,  388,  388,  388,  388,
  388,  388,  388,  388,  388,  388,  388,  388,  388,  388,
  388,  388,    0,    0,    0,    0,  324,    0,  388,    0,
    0,  388,    0,    0,    0,  506,    0,  388,    0,    0,
  506,    0,  506,  506,  506,  506,  506,  506,  506,  506,
  506,  506,  506,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,  506,    0,  506,  506,  506,  506,
  506,  506,  506,    0,  506,  506,    0,  506,  506,  506,
  506,  506,  506,  506,  506,  506,  506,    0,  506,  506,
  506,  506,  506,  506,  506,  506,  506,  506,  506,  506,
  506,  506,  506,  506,  506,  506,  506,  506,  506,  506,
    0,  331,  746,    0,    0,    0,  506,  331,    0,  506,
    0,   25,    0,   26,    0,  506,   27,    0,    0,    0,
    0,   28,    0,    0,    0,   29,    0,    0,    0,    0,
    0,    0,    0,    0,   31,    0,    0,    0,    0,    0,
    0,   33,    0,  331,    0,    0,   34,    0,    0,    0,
   35,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,   37,    0,   38,    0,    0,    0,   39,    0,
    0,    0,    0,    0,    0,    0,   40,   41,    0,    0,
   42,    0,    0,  747,  331,    0,    0,    0,    0,  331,
    0,  331,  331,  331,  331,  331,  331,  331,  331,  331,
  331,  331,    0,    0,    0,    0,    0,    0,    0,  292,
    0,    0,    0,  331,    0,  331,  331,  331,  331,  331,
  331,  331,    0,  331,  331,    0,  331,  331,  331,  331,
  331,  331,  331,  331,  331,  331,    0,  331,  331,  331,
  331,  331,  331,  331,  331,  331,  331,  331,  331,  331,
  331,  331,  331,  331,  331,  331,  331,  331,  331,    0,
  431,  562,    0,    0,  322,  331,  431,    0,  331,    0,
   25,    0,   26,    0,  331,   27,    0,    0,    0,    0,
   28,    0,    0,    0,   29,    0,    0,    0,    0,    0,
    0,    0,    0,   31,    0,    0,    0,    0,    0,    0,
   33,    0,  431,    0,    0,   34,    0,    0,    0,   35,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,   37,    0,   38,    0,    0,    0,   39,    0,    0,
    0,    0,    0,    0,    0,   40,   41,    0,    0,   42,
    0,    0,  321,  431,    0,    0,    0,    0,  431,    0,
  431,  431,  431,  431,  431,  431,  431,  431,  431,  431,
  431,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,  431,    0,  431,  431,  431,  431,  431,  431,
  431,    0,  431,  431,    0,  431,  431,  431,  431,  431,
  431,  431,  431,  431,  431,    0,  431,  431,  431,  431,
  431,  431,  431,  431,  431,  431,  431,  431,  431,  431,
  431,  431,  431,  431,  431,  431,  431,  431,    0,  391,
  887,    0,    0,  354,  431,  391,    0,  431,    0,   25,
    0,   26,    0,  431,   27,    0,    0,    0,    0,   28,
    0,    0,    0,   29,    0,    0,    0,    0,    0,    0,
    0,    0,   31,    0,    0,    0,    0,    0,    0,   33,
    0,  391,    0,    0,   34,    0,    0,    0,   35,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
   37,    0,   38,    0,    0,    0,   39,    0,    0,    0,
    0,    0,    0,    0,   40,   41,    0,    0,   42,    0,
    0,  321,  391,    0,    0,    0,    0,  391,    0,  391,
  391,  391,  391,  391,  391,  391,  391,  391,  391,  391,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,  391,    0,  391,  391,  391,  391,  391,  391,  391,
    0,  391,    0,    0,  391,  391,  391,  391,  391,  391,
  391,  391,  391,  391,    0,  391,  391,  391,  391,  391,
  391,  391,  391,  391,  391,  391,  391,  391,  391,  391,
  391,  391,  391,  391,  391,  391,  391,    0,  537,    0,
  353,    0,  354,  391,  537,    0,  391,    0,    0,    0,
    0,    0,  391,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,  353,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,  353,    0,    0,    0,
  537,  353,    0,    0,  230,    0,  353,    0,  353,  353,
  353,  353,    0,    0,    0,    0,  353,    0,    0,    0,
  353,  331,    0,    0,  353,    0,    0,  331,    0,    0,
    0,  737,  353,    0,    0,  353,    0,  353,    0,    0,
    0,  537,    0,    0,    0,    0,  537,    0,  537,  537,
  537,  537,  537,  537,  537,  537,  537,  537,  537,    0,
    0,  353,    0,  331,    0,    0,    0,    0,    0,    0,
  537,    0,  537,    0,  537,    0,  537,  537,  537,  737,
  537,  537,    0,  537,  537,  537,  537,  537,  537,  537,
  537,  537,  537,    0,    0,    0,  537,  537,  537,  537,
  537,  537,  537,  537,  537,  537,  537,  537,  537,  537,
  537,  537,  537,  537,  549,  537,    0,  353,    0,    0,
  549,    0,  331,    0,    0,    0,    0,    0,  331,    0,
    0,  537,    0,    0,  331,  331,  331,  331,  331,  331,
  737,  331,    0,  331,  331,    0,  331,  331,  331,  331,
  331,  331,  331,  331,  331,  331,  549,  331,  331,  331,
  331,  331,  331,  331,  331,  331,  331,  331,  331,  331,
  331,  331,  331,  331,  331,  331,  331,  331,  331,    0,
    0,    0,    0,  331,    0,  331,    0,    0,  331,    0,
    0,    0,    0,    0,  331,    0,    0,  549,    0,    0,
    0,    0,  549,    0,  549,  549,  549,  549,  549,  549,
  549,  549,  549,  549,  549,    0,    0,    0,  553,    0,
    0,    0,    0,    0,  553,    0,  549,    0,  549,    0,
  549,    0,  549,  549,  549,    0,  549,  549,    0,    0,
  549,  549,  549,  549,  549,  549,  549,  549,  549,    0,
    0,    0,  549,  549,  549,  549,  549,  549,  549,  549,
  553,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,  549,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,  556,  549,    0,    0,
    0,    0,  556,    0,    0,    0,    0,    0,    0,    0,
    0,  553,    0,    0,    0,    0,  553,    0,  553,  553,
  553,  553,  553,  553,  553,  553,  553,  553,  553,    0,
    0,    0,    0,    0,    0,    0,    0,    0,  556,    0,
  553,    0,  553,    0,  553,    0,  553,  553,  553,    0,
  553,  553,    0,    0,  553,  553,  553,  553,    0,    0,
    0,  553,  553,    0,    0,    0,  553,  553,  553,  553,
  553,  553,  553,  553,    0,    0,    0,    0,    0,  556,
    0,    0,    0,    0,  556,  553,  556,  556,  556,  556,
  556,  556,  556,  556,  556,  556,  556,    0,    0,    0,
  554,  553,    0,    0,    0,    0,  554,    0,  556,    0,
  556,    0,  556,    0,  556,  556,  556,    0,  556,  556,
    0,    0,  556,  556,  556,  556,    0,    0,    0,  556,
  556,    0,    0,    0,  556,  556,  556,  556,  556,  556,
  556,  556,  554,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,  556,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,  555,  556,
    0,    0,    0,    0,  555,    0,    0,    0,    0,    0,
    0,    0,    0,  554,    0,    0,    0,    0,  554,    0,
  554,  554,  554,  554,  554,  554,  554,  554,  554,  554,
  554,    0,    0,    0,    0,    0,    0,    0,    0,    0,
  555,    0,  554,    0,  554,    0,  554,    0,  554,  554,
  554,    0,  554,  554,    0,    0,  554,  554,  554,  554,
    0,    0,    0,  554,  554,    0,  559,    0,  554,  554,
  554,  554,  554,  554,  554,  554,    0,    0,    0,    0,
    0,  555,    0,    0,    0,    0,  555,  554,  555,  555,
  555,  555,  555,  555,  555,  555,  555,  555,  555,    0,
    0,    0,    0,  554,    0,    0,    0,    0,    0,    0,
  555,    0,  555,    0,  555,    0,  555,  555,  555,    0,
  555,  555,    0,    0,  555,  555,  555,  555,    0,    0,
    0,  555,  555,    0,  560,    0,  555,  555,  555,  555,
  555,  555,  555,  555,    0,    0,    0,    0,    0,  559,
    0,    0,    0,    0,  559,  555,  559,  559,  559,  559,
  559,  559,  559,  559,  559,  559,  559,    0,    0,    0,
    0,  555,    0,    0,    0,    0,    0,    0,  559,    0,
  559,    0,  559,    0,  559,  559,  559,    0,    0,    0,
    0,    0,  559,  559,  559,  559,    0,    0,    0,  559,
  559,    0,  561,    0,  559,  559,  559,  559,  559,  559,
  559,  559,    0,    0,    0,    0,    0,  560,    0,    0,
    0,    0,  560,  559,  560,  560,  560,  560,  560,  560,
  560,  560,  560,  560,  560,    0,    0,    0,    0,  559,
    0,    0,    0,    0,    0,    0,  560,    0,  560,    0,
  560,    0,  560,  560,  560,    0,    0,    0,    0,    0,
  560,  560,  560,  560,    0,    0,    0,  560,  560,    0,
  562,    0,  560,  560,  560,  560,  560,  560,  560,  560,
    0,    0,    0,    0,    0,  561,    0,    0,    0,    0,
  561,  560,  561,  561,  561,  561,  561,  561,  561,  561,
  561,  561,  561,    0,    0,    0,    0,  560,    0,    0,
    0,    0,    0,    0,  561,    0,  561,    0,  561,    0,
  561,  561,  561,    0,    0,    0,    0,    0,  561,  561,
  561,  561,    0,    0,    0,  561,  561,    0,  563,    0,
  561,  561,  561,  561,  561,  561,  561,  561,    0,    0,
    0,    0,    0,  562,    0,    0,    0,    0,  562,  561,
  562,  562,  562,  562,  562,  562,  562,  562,  562,  562,
  562,    0,    0,    0,    0,  561,    0,    0,    0,    0,
    0,    0,  562,    0,  562,    0,  562,    0,  562,  562,
  562,    0,    0,    0,    0,    0,  562,  562,  562,  562,
    0,    0,    0,  562,  562,    0,  564,    0,    0,    0,
  562,  562,  562,  562,  562,  562,    0,    0,    0,    0,
    0,  563,    0,    0,    0,    0,  563,  562,  563,  563,
  563,  563,  563,  563,  563,  563,  563,  563,  563,    0,
    0,    0,    0,  562,    0,    0,    0,    0,    0,    0,
  563,    0,  563,    0,  563,    0,  563,  563,  563,    0,
    0,    0,    0,    0,  563,  563,  563,  563,    0,    0,
    0,  563,  563,    0,  565,    0,    0,    0,  563,  563,
  563,  563,  563,  563,    0,    0,    0,    0,    0,  564,
    0,    0,    0,    0,  564,  563,  564,  564,  564,  564,
  564,  564,  564,  564,  564,  564,  564,    0,    0,    0,
    0,  563,    0,    0,    0,    0,    0,    0,  564,    0,
  564,    0,  564,    0,  564,  564,  564,    0,    0,    0,
    0,    0,  564,  564,  564,  564,    0,    0,    0,  564,
  564,    0,  566,    0,    0,    0,  564,  564,  564,  564,
  564,  564,    0,    0,    0,    0,    0,  565,    0,    0,
    0,    0,  565,  564,  565,  565,  565,  565,  565,  565,
  565,  565,  565,  565,  565,    0,    0,    0,    0,  564,
    0,    0,    0,    0,    0,    0,  565,    0,  565,    0,
  565,    0,  565,  565,  565,    0,    0,    0,    0,    0,
  565,  565,  565,  565,    0,    0,    0,  565,  565,    0,
  567,    0,    0,    0,  565,  565,  565,  565,  565,  565,
    0,    0,    0,    0,    0,  566,    0,    0,    0,    0,
  566,  565,  566,  566,  566,  566,  566,  566,  566,  566,
  566,  566,  566,    0,    0,    0,    0,  565,    0,    0,
    0,    0,    0,    0,  566,    0,  566,    0,  566,    0,
  566,  566,  566,    0,    0,    0,    0,    0,  566,  566,
  566,  566,    0,    0,    0,  566,  566,    0,  568,    0,
    0,    0,  566,  566,  566,  566,  566,  566,    0,    0,
    0,    0,    0,  567,    0,    0,    0,    0,  567,  566,
  567,  567,  567,  567,  567,  567,  567,  567,  567,  567,
  567,    0,    0,    0,    0,  566,    0,    0,    0,    0,
    0,    0,  567,    0,  567,    0,  567,    0,  567,  567,
  567,    0,    0,    0,    0,    0,    0,    0,  567,  567,
    0,    0,    0,  567,  567,    0,  569,    0,    0,    0,
    0,    0,  567,  567,  567,  567,    0,    0,    0,    0,
    0,  568,    0,    0,    0,    0,  568,  567,  568,  568,
  568,  568,  568,  568,  568,  568,  568,  568,  568,    0,
    0,    0,    0,  567,    0,    0,    0,    0,    0,    0,
  568,    0,  568,    0,  568,    0,  568,  568,  568,    0,
    0,    0,    0,    0,    0,    0,  568,  568,    0,    0,
    0,  568,  568,    0,  570,    0,    0,    0,    0,    0,
  568,  568,  568,  568,    0,    0,    0,    0,    0,  569,
    0,    0,    0,    0,  569,  568,  569,  569,  569,  569,
  569,  569,  569,  569,  569,  569,  569,    0,    0,    0,
    0,  568,    0,    0,    0,    0,    0,    0,  569,    0,
  569,    0,  569,    0,  569,  569,  569,    0,    0,    0,
    0,    0,    0,    0,  569,  569,    0,    0,    0,  569,
  569,    0,  571,    0,    0,    0,    0,    0,  569,  569,
  569,  569,    0,    0,    0,    0,    0,  570,    0,    0,
    0,    0,  570,  569,  570,  570,  570,  570,  570,  570,
  570,  570,  570,  570,  570,    0,    0,    0,    0,  569,
    0,    0,    0,    0,    0,    0,  570,    0,  570,    0,
  570,    0,  570,  570,  570,    0,    0,    0,    0,    0,
    0,    0,  570,  570,    0,    0,    0,  570,  570,    0,
  572,    0,    0,    0,    0,    0,    0,    0,  570,  570,
    0,    0,    0,    0,    0,  571,    0,    0,    0,    0,
  571,  570,  571,  571,  571,  571,  571,  571,  571,  571,
  571,  571,  571,    0,    0,    0,    0,  570,    0,    0,
    0,    0,    0,    0,  571,    0,  571,    0,  571,    0,
  571,  571,  571,    0,    0,    0,    0,    0,    0,    0,
  571,  571,    0,    0,    0,  571,  571,    0,  573,    0,
    0,    0,    0,    0,    0,    0,  571,  571,    0,    0,
    0,    0,    0,  572,    0,    0,    0,    0,  572,  571,
  572,  572,  572,  572,  572,  572,  572,  572,  572,  572,
  572,    0,    0,    0,    0,  571,    0,    0,    0,    0,
    0,    0,  572,    0,  572,    0,  572,    0,  572,  572,
  572,    0,    0,    0,    0,    0,    0,    0,    0,  572,
    0,    0,    0,  572,  572,    0,  574,    0,    0,    0,
    0,    0,    0,    0,  572,  572,    0,    0,    0,    0,
    0,  573,    0,    0,    0,    0,  573,  572,  573,  573,
  573,  573,  573,  573,  573,  573,  573,  573,  573,    0,
    0,    0,    0,  572,    0,    0,    0,    0,    0,    0,
  573,    0,  573,    0,  573,    0,  573,  573,  573,    0,
    0,    0,    0,    0,    0,    0,    0,  573,    0,    0,
    0,  573,  573,    0,  575,    0,    0,    0,    0,    0,
    0,    0,  573,  573,    0,    0,    0,    0,    0,  574,
    0,    0,    0,    0,  574,  573,  574,  574,  574,  574,
  574,  574,  574,  574,  574,  574,  574,    0,    0,    0,
    0,  573,    0,    0,    0,    0,    0,    0,  574,    0,
  574,    0,  574,    0,  574,  574,  574,    0,    0,    0,
    0,    0,    0,    0,    0,  574,    0,    0,    0,    0,
  574,    0,  576,    0,    0,    0,    0,    0,    0,    0,
  574,  574,    0,    0,    0,    0,    0,  575,    0,    0,
    0,    0,  575,  574,  575,  575,  575,  575,  575,  575,
  575,  575,  575,  575,  575,    0,    0,    0,    0,  574,
    0,    0,    0,    0,    0,    0,  575,    0,  575,    0,
  575,    0,  575,  575,  575,    0,    0,    0,    0,    0,
    0,    0,    0,  575,    0,    0,    0,    0,  575,    0,
  577,    0,    0,    0,    0,    0,    0,    0,  575,  575,
    0,    0,    0,    0,    0,  576,    0,    0,    0,    0,
  576,  575,  576,  576,  576,  576,  576,  576,  576,  576,
  576,  576,  576,    0,    0,    0,    0,  575,    0,    0,
    0,    0,    0,    0,  576,    0,  576,    0,  576,    0,
  576,  576,  576,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,  576,    0,  578,    0,
    0,    0,    0,    0,    0,    0,  576,  576,    0,    0,
    0,    0,    0,  577,    0,    0,    0,    0,  577,  576,
  577,  577,  577,  577,  577,  577,  577,  577,  577,  577,
  577,    0,    0,    0,    0,  576,    0,    0,    0,    0,
    0,    0,  577,    0,  577,    0,  577,    0,  577,  577,
  577,    0,    0,    0,  579,    0,    0,    0,    0,    0,
    0,    0,    0,    0,  577,    0,    0,    0,    0,    0,
    0,    0,    0,    0,  577,  577,    0,    0,    0,    0,
    0,  578,    0,    0,    0,    0,  578,  577,  578,  578,
  578,  578,  578,  578,  578,  578,  578,  578,  578,    0,
    0,    0,    0,  577,    0,    0,    0,    0,    0,    0,
  578,    0,  578,    0,  578,    0,  578,  578,  578,    0,
    0,    0,    0,  331,    0,    0,    0,  737,    0,    0,
    0,    0,  578,    0,    0,    0,    0,  579,    0,    0,
    0,    0,  579,  578,  579,  579,  579,  579,  579,  579,
  579,  579,  579,  579,  579,  578,    0,    0,    0,  331,
    0,    0,    0,    0,    0,    0,  579,    0,  579,    0,
  579,  578,  579,  579,  579,  737,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,  579,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,  579,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,  579,    0,    0,    0,    0,    0,    0,  331,    0,
    0,    0,    0,    0,  331,    0,    0,  579,    0,    0,
  331,    0,  331,    0,  331,    0,  737,  331,    0,  331,
  331,    0,  331,  331,  331,  331,  331,  331,  331,  331,
  331,  331,    0,  331,  331,  331,  331,  331,  331,  331,
  331,  331,  331,  331,  331,  331,  331,  331,  331,  331,
  331,  331,  331,  331,  331,    0,    0,   56,    0,  331,
    0,  331,    0,    0,  331,   57,   25,   58,   26,    0,
  331,   27,   59,    0,   60,   61,   28,   62,   63,   64,
   29,    0,    0,    0,    0,    0,   65,    0,   66,   31,
   67,   68,   69,   70,    0,    0,   33,    0,    0,    0,
   71,   34,    0,   72,   73,   35,    0,    0,    0,    0,
    0,    0,    0,    0,    0,   74,    0,   37,    0,   38,
   75,    0,    0,   39,    0,   76,   77,   78,   79,   80,
   81,   40,   41,   82,   83,   42,   84,    0,   85,    0,
    0,   86,   87,    0,    0,   88,   89,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
   90,   91,   92,   93,   94,    0,    0,    0,   95,    0,
    0,    0,   96,    0,    0,    0,    0,   97,   98,   99,
  100,  101,    0,    0,    0,  102,    0,  103,    0,    0,
    0,    0,    0,  104,  105,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,  267,    0,    0,    0,  106,  107,
  108,  109,   57,   25,   58,   26,    0,    0,   27,   59,
    0,   60,   61,   28,   62,   63,   64,   29,    0,    0,
    0,    0,    0,   65,    0,   66,   31,   67,   68,   69,
   70,    0,    0,   33,    0,    0,    0,   71,   34,    0,
   72,   73,   35,    0,    0,    0,    0,    0,    0,    0,
    0,    0,   74,    0,   37,    0,   38,   75,    0,    0,
   39,    0,   76,   77,   78,   79,   80,   81,   40,   41,
   82,   83,   42,   84,    0,   85,    0,    0,   86,   87,
    0,    0,   88,   89,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,   90,   91,   92,
   93,   94,    0,    0,    0,   95,    0,    0,    0,   96,
    0,    0,    0,    0,   97,   98,   99,  100,  101,    0,
    0,    0,  102,    0,  103,    0,    0,    0,    0,    0,
  104,  105,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,  551,    0,    0,    0,  106,  107,  108,  109,   57,
   25,   58,   26,    0,    0,   27,   59,    0,   60,   61,
   28,   62,   63,   64,   29,    0,    0,    0,    0,    0,
   65,    0,   66,   31,   67,   68,   69,   70,    0,    0,
   33,    0,    0,    0,   71,   34,    0,   72,   73,   35,
    0,    0,    0,    0,    0,    0,    0,    0,    0,   74,
    0,   37,    0,   38,   75,    0,    0,   39,    0,   76,
   77,   78,   79,   80,   81,   40,   41,   82,   83,   42,
   84,    0,   85,    0,    0,   86,   87,    0,    0,   88,
   89,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,   90,   91,   92,   93,   94,    0,
    0,    0,   95,    0,    0,    0,   96,    0,    0,    0,
    0,   97,   98,   99,  100,  101,    0,    0,    0,  102,
    0,  103,    0,    0,    0,    0,    0,  104,  105,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,  929,    0,
    0,    0,  106,  552,  108,  109,  929,  929,  929,  929,
    0,    0,  929,  929,    0,  929,  929,  929,  929,  929,
  929,  929,    0,    0,    0,    0,    0,  929,    0,  929,
  929,  929,  929,  929,  929,    0,    0,  929,    0,    0,
    0,  929,  929,    0,  929,  929,  929,    0,    0,    0,
    0,    0,    0,    0,    0,    0,  929,    0,  929,    0,
  929,  929,    0,    0,  929,    0,  929,  929,  929,  929,
  929,  929,  929,  929,  929,  929,  929,  929,    0,  929,
    0,    0,  929,  929,    0,    0,  929,  929,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,  929,  929,  929,  929,  929,    0,    0,    0,  929,
    0,    0,    0,  929,    0,    0,    0,    0,  929,  929,
  929,  929,  929,    0,    0,    0,  929,    0,  929,    0,
    0,    0,    0,    0,  929,  929,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,  784,    0,    0,    0,  929,
  929,  929,  929,  784,  784,  784,  784,    0,    0,  784,
  784,    0,  784,  784,  784,  784,  784,  784,  784,    0,
    0,    0,    0,    0,  784,    0,  784,  784,  784,  784,
  784,  784,    0,    0,  784,    0,    0,    0,  784,  784,
    0,  784,  784,  784,    0,    0,    0,    0,    0,    0,
    0,    0,    0,  784,    0,  784,    0,  784,  784,    0,
    0,  784,    0,  784,  784,  784,  784,  784,  784,  784,
  784,  784,  784,  784,  784,    0,  784,    0,    0,  784,
  784,    0,    0,  784,  784,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,  784,  784,
  784,  784,  784,    0,    0,    0,  784,    0,    0,    0,
  784,    0,    0,    0,    0,  784,  784,  784,  784,  784,
    0,    0,    0,  784,    0,  784,    0,    0,    0,    0,
    0,  784,  784,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,  733,    0,    0,    0,  784,  784,  784,  784,
   57,   25,    0,   26,    0,    0,   27,  255,    0, 1034,
    0,   28,   62,   63,    0,   29,    0,    0,   25,    0,
   26,   65,    0,   27,   31,    0,    0,    0,   28,    0,
    0,   33,   29,    0,    0,    0,   34,    0,   72,   73,
   35,   31,    0,    0,    0,    0,    0,    0,   33,    0,
    0,    0,   37,   34,   38,   75,    0,   35,   39,    0,
    0,   77,    0,   79,    0,   81,   40,   41,  256,   37,
   42,   38,    0,    0,    0,   39,    0,   87,    0,    0,
   88,   89,    0,   40,   41,    0,    0,   42,    0,    0,
  321,    0,    0,    0,    0,   90,   91,   92,   93,  302,
    0,    0,    0,  513,  734,    0,    0,   96,    0,    0,
    0,    0,    0,   98,   99,  100,  101,    0,    0,    0,
  102,    0,  103,    0,    0,    0,    0,    0,  104,  105,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,  918,
    0,    0,    0,  106,  303,  108,  109,   57,   25,    0,
   26,    0,    0,   27,  255,    0, 1156,    0,   28,   62,
   63,  354,   29,    0,    0,   25,    0,   26,   65,    0,
   27,   31,    0,    0,    0,   28,    0,    0,   33,   29,
    0,    0,    0,   34,    0,   72,   73,   35,   31,    0,
    0,    0,    0,    0,    0,   33,    0,    0,    0,   37,
   34,   38,   75,  919,   35,   39,    0,    0,   77,    0,
   79,    0,   81,   40,   41,  256,   37,   42,   38,    0,
    0,    0,   39,    0,   87,    0,    0,   88,   89,    0,
   40,   41,    0,    0,   42,    0,    0,  321,    0,    0,
    0,    0,   90,   91,   92,   93,  302,    0,    0,    0,
  513,    0,    0,    0,   96,    0,    0,    0,    0,    0,
   98,   99,  100,  101,    0,    0,    0,  102,    0,  103,
    0,    0,    0,    0,    0,  104,  105,    0,    0,    0,
    0,    0,    0,   57,   25,    0,   26,    0,    0,   27,
  255,    0,    0,    0,   28,   62,   63,    0,   29,    0,
  106,  303,  108,  109,   65,    0,    0,   31,    0,    0,
    0,    0,    0,    0,   33,    0,    0,    0,  354,   34,
    0,   72,   73,   35,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,   37,    0,   38,   75,    0,
    0,   39,    0,    0,   77,    0,   79,    0,   81,   40,
   41,  256,    0,   42,    0,    0,    0,    0,    0,    0,
   87,    0,    0,   88,   89,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,   90,   91,
   92,   93,  302,    0,    0,    0,  720,  988,    0,    0,
   96,    0,    0,    0,    0,    0,   98,   99,  100,  101,
    0,    0,    0,  102,    0,  103,    0,    0,    0,    0,
    0,  104,  105,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,  733,    0,  106,  721,  108,  109,
    0,    0,   57,   25,    0,   26,    0,  722,   27,  255,
    0,    0,    0,   28,   62,   63,    0,   29,    0,    0,
  172,    0,  172,   65,    0,  172,   31,    0,    0,    0,
  172,    0,    0,   33,  172,    0,    0,    0,   34,    0,
   72,   73,   35,  172,    0,    0,    0,    0,    0,    0,
  172,    0,    0,    0,   37,  172,   38,   75,  919,  172,
   39,    0,    0,   77,    0,   79,    0,   81,   40,   41,
  256,  172,   42,  172,    0,    0,    0,  172,    0,   87,
    0,    0,   88,   89,    0,  172,  172,    0,    0,  172,
    0,    0,  172,    0,    0,    0,    0,   90,   91,   92,
   93,  302,    0,    0,    0,  513,    0,    0,    0,   96,
    0,    0,    0,    0,    0,   98,   99,  100,  101,    0,
    0,    0,  102,    0,  103,    0,    0,  953,    0,    0,
  104,  105,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,  701,    0,    0,    0,  106,  303,  108,  109,   57,
   25,    0,   26,    0,    0,   27,  255,    0,    0,    0,
   28,   62,   63,  172,   29,    0,    0,  172,    0,  172,
   65,    0,  172,   31,    0,    0,    0,  172,    0,    0,
   33,  172,    0,    0,    0,   34,    0,   72,   73,   35,
  172,    0,    0,    0,    0,    0,    0,  172,    0,    0,
    0,   37,  172,   38,   75,    0,  172,   39,    0,    0,
   77,    0,   79,    0,   81,   40,   41,  256,  172,   42,
  172,    0,   85,    0,  172,    0,   87,    0,    0,   88,
   89,    0,  172,  172,    0,    0,  172,    0,    0,  172,
    0,    0,    0,    0,   90,   91,   92,   93,  302,    0,
    0,    0,    0,    0,    0,    0,   96,    0,    0,    0,
    0,    0,   98,   99,  100,  101,    0,    0,    0,  102,
    0,  103,  953,    0,    0,    0,    0,  104,  105,    0,
    0,    0,    0,    0,    0,   57,   25,    0,   26,    0,
    0,   27,  255,    0,    0,    0,   28,   62,   63,    0,
   29,    0,  106,  303,  108,  109,   65,    0,    0,   31,
    0,    0,    0,    0,    0,    0,   33,    0,    0,    0,
  172,   34,    0,   72,   73,   35,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,   37,    0,   38,
   75,    0,    0,   39,    0,    0,   77,    0,   79,    0,
   81,   40,   41,  256,    0,   42,    0,    0,    0,    0,
    0,    0,   87,    0,    0,   88,   89,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
   90,   91,   92,   93,  302,    0,    0,    0,  720,    0,
    0,    0,   96,    0,    0,    0,    0,    0,   98,   99,
  100,  101,    0,    0,    0,  102,    0,  103,    0,    0,
    0,    0,    0,  104,  105,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,  733,    0,  106,  721,
  108,  109,    0,    0,   57,   25,    0,   26,    0,  722,
   27,  255,    0,    0,    0,   28,   62,   63,    0,   29,
    0,    0,   25,    0,   26,   65,    0,   27,   31,    0,
    0,    0,   28,    0,    0,   33,   29,    0,    0,    0,
   34,    0,   72,   73,   35,   31,    0,    0,    0,    0,
    0,    0,   33,    0,    0,    0,   37,   34,   38,   75,
    0,   35,   39,    0,    0,   77,    0,   79,    0,   81,
   40,   41,  256,   37,   42,   38,    0,    0,    0,   39,
    0,   87,    0,    0,   88,   89,    0,   40,   41,    0,
    0,   42,    0,    0,  321,    0,    0,    0,    0,   90,
   91,   92,   93,  302,    0,    0,    0,  513,    0,    0,
    0,   96,    0,    0,    0,    0,    0,   98,   99,  100,
  101,    0,    0,    0,  102,    0,  103,    0,    0,    0,
    0,    0,  104,  105,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,  586,    0,    0,    0,  106,  303,  108,
  109,   57,   25,    0,   26,    0,    0,   27,  255,    0,
    0,    0,   28,   62,   63,  354,   29,    0,    0,   25,
    0,   26,   65,    0,   27,   31,    0,    0,    0,   28,
    0,    0,   33,   29,    0,    0,    0,   34,    0,   72,
   73,   35,   31,    0,    0,    0,    0,    0,    0,   33,
    0,    0,    0,   37,   34,   38,   75,    0,   35,   39,
    0,    0,   77,    0,   79,    0,   81,   40,   41,  256,
   37,   42,   38,    0,    0,    0,   39,    0,   87,    0,
    0,   88,   89,    0,   40,   41,    0,    0,   42,    0,
    0,  515,    0,    0,    0,    0,   90,   91,   92,   93,
   94,    0,    0,    0,    0,    0,    0,    0,   96,    0,
    0,    0,    0,    0,   98,   99,  100,  101,    0,    0,
    0,  102,    0,  103,    0,    0,    0,    0,    0,  104,
  105,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
  906,    0,    0,    0,  106,  107,  108,  109,   57,   25,
    0,   26,    0,    0,   27,  255,    0,    0,    0,   28,
   62,   63,  354,   29,    0,    0,   25,    0,   26,   65,
    0,   27,   31,    0,    0,    0,   28,    0,    0,   33,
   29,    0,    0,    0,   34,    0,   72,   73,   35,   31,
    0,    0,    0,    0,    0,    0,   33,    0,    0,    0,
   37,   34,   38,   75,    0,   35,   39,    0,    0,   77,
    0,   79,    0,   81,   40,   41,  256,   37,   42,   38,
    0,    0,    0,   39,    0,   87,    0,    0,   88,   89,
    0,   40,   41,    0,    0,   42,    0,    0,  568,    0,
    0,    0,    0,   90,   91,   92,   93,  302,    0,    0,
    0,    0,    0,    0,    0,   96,    0,    0,    0,    0,
    0,   98,   99,  100,  101,    0,    0,    0,  102,    0,
  103,    0,    0,    0,    0,    0,  104,  105,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,  604,    0,    0,
    0,  106,  303,  108,  109,  604,  604,    0,  604,    0,
    0,  604,  604,    0,    0,    0,  604,  604,  604,  354,
  604,    0,    0,   25,    0,   26,  604,    0,   27,  604,
    0,    0,    0,   28,    0,    0,  604,   29,    0,    0,
    0,  604,    0,  604,  604,  604,   31,    0,    0,    0,
    0,    0,    0,   33,    0,    0,    0,  604,   34,  604,
  604,    0,   35,  604,    0,    0,  604,    0,  604,    0,
  604,  604,  604,  604,   37,  604,   38,    0,    0,    0,
   39,    0,  604,    0,    0,  604,  604,    0,   40,   41,
    0,    0,   42,    0,    0,  747,    0,    0,    0,    0,
  604,  604,  604,  604,  604,    0,    0,    0,    0,    0,
    0,    0,  604,    0,    0,    0,    0,    0,  604,  604,
  604,  604,    0,    0,    0,  604,    0,  604,    0,    0,
    0,    0,    0,  604,  604,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,  496,    0,    0,    0,  604,  604,
  604,  604,   57,   25,    0,   26,    0,    0,   27,  255,
    0,    0,    0,   28,   62,   63,  354,   29,    0,    0,
    0,    0,    0,   65,    0,    0,   31,    0,    0,    0,
    0,    0,    0,   33,    0,  450,    0,  331,   34,    0,
   72,   73,   35,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,   37,    0,   38,   75,  451,    0,
   39,    0,    0,   77,    0,   79,    0,   81,   40,   41,
  256,  452,   42,  331,    0,    0,  454,    0,    0,    0,
    0,  455,    0,  456,  457,  458,  459,    0,    0,    0,
    0,  460,    0,    0,    0,  461,    0,   90,   91,   92,
  257,    0,    0,    0,    0,    0,    0,  462,    0,   96,
  463,    0,  464,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,  465,    0,    0,    0,
    0,    0,    0,    0,    0,    0,  331,  331,  331,  331,
  737,    0,    0,  331,  331,  106,  497,  331,  331,  331,
  331,  331,  331,  331,  331,  331,    0,  331,  331,  331,
  331,  331,  331,  331,  331,  331,  331,  331,  331,  331,
  331,  331,  331,  331,  331,  331,  331,  331,  331,    0,
   48,    0, 1314,    0,   48,  331,   48,    0,  331,   48,
    0,   48,   48,    0,   48,    0,   48,    0,   48,    0,
   48,   48,   48,   48,    0,    0,   48,   48,    0,    0,
    0,    0,   48,    0,   48,   48,   48,    0,    0,   48,
    0,   48,    0,   48,    0,    0,   48,    0,   48,   48,
   48,   48,    0,    0,    0,   48,   48,   48,    0,    0,
   48,   48,   48,    0,    0,    0,    0,    0,    0,   48,
   48,    0,   48,   48,    0,   48,   48,   48,    0,    0,
    0,   48,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,   48,
    0,   48,    0,   48,    0,   48,    0,   81,   48,    0,
   48,   48,    0,   48,    0,   48,   48,   48,    0,   48,
   48,   48,   48,    0,    0,   48,   48,    0,    0,    0,
    0,   48,    0,   48,   48,   48,    0,    0,   48,    0,
   48,    0,   48,    0,    0,   48,    0,   48,   48,   48,
   48,    0,    0,    0,   48,   48,   48,   48,    0,   48,
   48,   48,    0,    0,    0,    0,    0,    0,   48,   48,
    0,   48,   48,    0,   48,   48,   48,    0,    0,    0,
   48,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,   48,    0,
   48,    0,   48,    0,   48,    0,   82,   48,    0,   48,
   48,    0,   48,    0,   48,   48,   48,    0,   48,   48,
   48,   48,    0,    0,   48,   48,    0,    0,    0,    0,
   48,    0,   48,   48,   48,    0,    0,   48,    0,   48,
    0,   48,    0,    0,   48,    0,   48,   48,   48,   48,
    0,    0,    0,   48,   48,   48,   48,    0,   48,   48,
   48,    0,    0,    0,    0,    0,    0,   48,   48,    0,
   48,   48,    0,   48,   48,   48,    0,    0,    0,   48,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,   48,    0,    0,    0,   48,    0,   48,   48,
    0,   48,    0,   48,   48,  212,   48,    0,   48,    0,
   48,    0,   48,   48,   48,   48,    0,    0,   48,   48,
    0,    0,    0,    0,   48,    0,   48,   48,   48,    0,
    0,   48,    0,   48,  331,   48,    0,    0,   48,    0,
   48,   48,   48,   48,    0,    0,    0,   48,   48,   48,
    0,    0,   48,   48,   48,   48,    0,  331,    0,    0,
    0,   48,   48,    0,   48,   48,   28,   48,   48,   48,
  331,    0,    0,   48,    0,  331,    0,    0,  331,    0,
  331,    0,  331,  331,  331,  331,    0,    0,    0,   28,
  331,    0,    0,   48,  331,    0,    0,    0,  331,  213,
    0,    0,   28,    0,  352,    0,  331,   28,    0,  331,
    0,  331,   28,    0,   28,   28,   28,   28,    0,    0,
   28,    0,   28,    0,    0,    0,   28,  352,    0,    0,
  331,    0,    0,    0,  450,  331,    0,    0,   28,    0,
  352,   28,  331,   28,  263,  352,  331,    0,  229,   48,
  352,    0,  352,  352,  352,  352,    0,  451,    0,  331,
  352,    0,    0,    0,  352,    0,    0,   28,  352,    0,
  452,    0,    0,   28,   28,  454,  352,    0,    0,  352,
  455,  352,  456,  457,  458,  459,    0,    0,    0,    0,
  460,  331,    0,    0,  461,    0,    0,    0, 1299,    0,
    0,   57,   25,    0,   26,  352,  462,   27,  255,  463,
    0,  464,   28,   62,   63,    0,   29,    0,    0,    0,
    0,    0,   65,    0,    0,   31,    0,    0,    0,    0,
    0,    0,   33,    0,    0,  465,    0,   34,    0,   72,
   73,   35,    0,  588,    0,    0,    0,    0,    0,    0,
  589,    0,    0,   37,    0,   38,   75,    0,    0,   39,
    0,  352,   77,    0,   79,    0,   81,   40,   41,  256,
    0,   42,    0,    0,    0,    0,    0,    0,  590,    0,
    0,   88,   89,    0,    0,    0,    0,    0,    0,    0,
    0, 1300,    0,    0,    0,    0,   90,   91,   92,   93,
   94,    0,    0,    0,    0,    0,    0,    0,   96,  913,
    0,  591,    0,    0,   98,   99,  100,  101,    0,    0,
    0,  102,    0,  103,    0,    0,    0,    0,    0,  104,
  105,    0,    0,    0,    0,    0,    0,   57,   25,    0,
   26,    0,    0,   27,  255,    0,    0,    0,   28,   62,
   63,    0,   29,    0,  106,  107,  108,  109,   65,    0,
    0,   31,    0,    0,    0,    0,    0,    0,   33,    0,
    0,    0,    0,   34,    0,   72,   73,   35,    0,  588,
    0,    0,    0,    0,    0,    0,  589,    0,    0,   37,
    0,   38,   75,    0,    0,   39,    0,    0,   77,    0,
   79,    0,   81,   40,   41,  256,    0,   42,    0,    0,
    0,    0,    0,    0,  590,    0,    0,   88,   89,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,   90,   91,   92,   93,   94,    0,    0,    0,
    0,    0,    0,    0,   96,    0,    0,  591,    0,    0,
   98,   99,  100,  101,    0,    0,    0,  102,    0,  103,
    0,    0,    0,    0,    0,  104,  105,    0,    0,    0,
    0,    0,    0,   57,   25,    0,   26,    0,    0,   27,
  255,    0,    0,    0,   28,   62,   63,    0,   29,    0,
  106,  107,  108,  109,   65,    0,    0,   31,    0,    0,
    0,    0,    0,    0,   33,    0,    0,    0,    0,   34,
    0,   72,   73,   35,    0,  588,    0,    0,    0,    0,
    0,    0,  589,    0,    0,   37,    0,   38,   75,    0,
    0,   39,    0,    0,   77,    0,   79,    0,   81,   40,
   41,  256,    0,   42,    0,    0,    0,    0,    0,    0,
  590,    0,    0,   88,   89,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,   90,   91,
   92,   93,   94,    0,    0,    0,    0,    0,    0,    0,
   96,    0,    0,    0,    0,    0,   98,   99,  100,  101,
    0,    0,    0,  102,    0,  103,    0,    0,    0,    0,
    0,  104,  105,    0,    0,    0,    0,    0,    0,   57,
   25,    0,   26,    0,    0,   27,  255,    0,    0,    0,
   28,   62,   63,    0,   29,    0,  106,  107,  108,  109,
   65,    0,    0,   31,    0,    0,    0,    0,    0,    0,
   33,    0,    0,    0,    0,   34,    0,   72,   73,   35,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,   37,    0,   38,   75,    0,    0,   39,    0,    0,
   77,    0,   79,    0,   81,   40,   41,  256,    0,   42,
    0,    0,   85,    0,    0,    0,   87,    0,    0,   88,
   89,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,   90,   91,   92,   93,  302,    0,
    0,    0,    0,    0,    0,    0,   96,    0,    0,    0,
    0,    0,   98,   99,  100,  101,    0,    0,    0,  102,
    0,  103,    0,    0,    0,    0,    0,  104,  105,    0,
    0,    0,    0,    0,    0,   57,   25,    0,   26,    0,
    0,   27,  255,    0,    0,    0,   28,   62,   63,    0,
   29,    0,  106,  303,  108,  109,   65,    0,    0,   31,
    0,    0,    0,    0,    0,    0,   33,    0,    0,    0,
    0,   34,    0,   72,   73,   35,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,   37,    0,   38,
   75,    0,    0,   39,    0,    0,   77,    0,   79,    0,
   81,   40,   41,  256,    0,   42,    0,    0,    0,    0,
    0,    0,   87,    0,    0,   88,   89,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
   90,   91,   92,   93,  302,    0,    0,    0,    0,  870,
    0,    0,   96,    0,    0,    0,    0,    0,   98,   99,
  100,  101,    0,    0,    0,  102,    0,  103,    0,    0,
    0,    0,    0,  104,  105,    0,    0,    0,    0,    0,
    0,   57,   25,    0,   26,    0,    0,   27,  255,    0,
    0,    0,   28,   62,   63,    0,   29,    0,  106,  303,
  108,  109,   65,    0,    0,   31,    0,    0,    0,    0,
    0,    0,   33,    0,    0,    0,    0,   34,    0,   72,
   73,   35,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,   37,    0,   38,   75,    0,    0,   39,
    0,    0,   77,    0,   79,    0,   81,   40,   41,  256,
    0,   42,    0,    0,    0,    0,    0,    0,   87,    0,
    0,   88,   89,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,   90,   91,   92,   93,
  302,    0,    0,    0,  513,    0,    0,    0,   96,    0,
    0,    0,    0,    0,   98,   99,  100,  101,    0,    0,
    0,  102,    0,  103,    0,    0,    0,    0,    0,  104,
  105,    0,    0,    0,    0,    0,    0,   57,   25,    0,
   26,    0,    0,   27,  255,    0,    0,    0,   28,   62,
   63,    0,   29,    0,  106,  303,  108,  109,   65,    0,
    0,   31,    0,    0,    0,    0,    0,    0,   33,    0,
    0,    0,    0,   34,    0,   72,   73,   35,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,   37,
    0,   38,   75,    0,    0,   39,    0,    0,   77,    0,
   79,    0,   81,   40,   41,  256,    0,   42,    0,    0,
    0,    0,    0,    0,   87,    0,    0,   88,   89,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,   90,   91,   92,   93,  302,    0,    0,    0,
  507,    0,    0,    0,   96,    0,    0,    0,    0,    0,
   98,   99,  100,  101,    0,    0,    0,  102,    0,  103,
    0,    0,    0,    0,    0,  104,  105,    0,    0,    0,
    0,    0,    0,   57,   25,    0,   26,    0,    0,   27,
  255,    0,    0,    0,   28,   62,   63,    0,   29,    0,
  106,  303,  108,  109,   65,    0,    0,   31,    0,    0,
    0,    0,    0,    0,   33,    0,    0,    0,    0,   34,
    0,   72,   73,   35,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,   37,    0,   38,   75,    0,
    0,   39,    0,    0,   77,    0,   79,    0,   81,   40,
   41,  256,    0,   42,    0,    0,    0,    0,    0,    0,
   87,    0,    0,   88,   89,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,   90,   91,
   92,   93,  302,    0,    0,    0,    0,    0,    0,    0,
   96,    0,    0,    0,    0,    0,   98,   99,  100,  101,
    0,    0,    0,  102,    0,  103,    0,    0,    0,    0,
    0,  104,  105,    0,    0,    0,    0,    0,    0,   57,
   25,    0,   26,    0,    0,   27,  255,    0,    0,    0,
   28,   62,   63,    0,   29,    0,  106,  303,  108,  109,
   65,    0,    0,   31,    0,    0,    0,    0,    0,    0,
   33,    0,    0,    0,    0,   34,    0,   72,   73,   35,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,   37,    0,   38,   75,    0,    0,   39,    0,    0,
   77,    0,   79,    0,   81,   40,   41,  256,    0,   42,
    0,    0,    0,    0,    0,    0,   87,    0,    0,   88,
   89,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,   90,   91,   92,   93,   94,    0,
    0,    0,    0,    0,    0,    0,   96,    0,    0,    0,
    0,    0,   98,   99,  100,  101,    0,    0,    0,  102,
    0,  103,    0,    0,    0,    0,    0,  104,  105,    0,
    0,    0,    0,    0,    0,   57,   25,    0,   26,    0,
    0,   27,  255,    0,    0,    0,   28,   62,   63,    0,
   29,    0,  106,  107,  108,  109,   65,    0,    0,   31,
    0,    0,    0,    0,    0,    0,   33,    0,    0,    0,
    0,   34,    0,   72,   73,   35,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,   37,    0,   38,
   75,    0,    0,   39,    0,    0,   77,    0,   79,    0,
   81,   40,   41,  256,    0,   42,    0,    0,    0,    0,
    0,    0,   87,    0,    0,   88,   89,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
   90,   91,   92,   93,   94,    0,    0,    0,    0,    0,
    0,    0,   96,    0,    0,    0,    0,    0,   98,   99,
  100,  101,    0,    0,    0,  102,    0,  103,    0,    0,
    0,    0,    0,  104,  105,    0,    0,    0,    0,    0,
    0,   78,   78,    0,   78,    0,    0,   78,   78,    0,
    0,    0,   78,   78,   78,    0,   78,    0,  106, 1026,
  108,  109,   78,    0,    0,   78,    0,    0,    0,    0,
    0,    0,   78,    0,    0,    0,    0,   78,    0,   78,
   78,   78,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,   78,    0,   78,   78,    0,    0,   78,
    0,    0,   78,    0,   78,    0,   78,   78,   78,   78,
    0,   78,    0,    0,    0,    0,    0,    0,   78,    0,
    0,   78,   78,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,   78,   78,   78,   78,
   78,    0,    0,    0,    0,    0,    0,    0,   78,    0,
    0,    0,    0,    0,   78,   78,   78,   78,    0,    0,
    0,   78,    0,   78,    0,    0,    0,    0,    0,   78,
   78,    0,    0,    0,    0,    0,    0,  135,  135,    0,
  135,    0,    0,  135,  135,    0,    0,    0,  135,  135,
  135,    0,  135,    0,   78,   78,   78,   78,  135,    0,
    0,  135,    0,    0,    0,    0,    0,    0,  135,    0,
    0,    0,    0,  135,    0,  135,  135,  135,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,  135,
    0,  135,  135,    0,    0,  135,    0,    0,  135,    0,
  135,    0,  135,  135,  135,  135,    0,  135,    0,    0,
    0,    0,    0,    0,  135,    0,    0,  135,  135,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,  135,  135,  135,  135,  135,    0,    0,    0,
    0,    0,    0,    0,  135,    0,    0,    0,    0,    0,
  135,  135,  135,  135,    0,    0,    0,  135,    0,  135,
    0,    0,    0,    0,    0,  135,  135,    0,    0,    0,
    0,    0,    0,   57,   25,    0,   26,    0,    0,   27,
  255,    0,    0,    0,   28,   62,   63,    0,   29,    0,
  135,  135,  135,  135,   65,    0,    0,   31,    0,    0,
    0,    0,    0,    0,   33,    0,   28,    0,   28,   34,
    0,   72,   73,   35,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,   37,    0,   38,   75,   28,
    0,   39,    0,    0,   77,    0,   79,    0,   81,   40,
   41,  256,   28,   42,    0,    0,    0,   28,    0,    0,
    0,    0,   28,    0,   28,   28,   28,   28,    0,    0,
    0,    0,   28,    0,    0,    0,   28,    0,   90,   91,
   92,  257,  302,    0,    0,    0,    0,    0,   28,    0,
   96,   28,    0,   28,    0,    0,   98,   99,  100,  101,
    0,    0,    0,  102,    0,  103,    0,    0,    0,    0,
    0,  104,  105,    0,    0,    0,    0,   28,    0,    0,
    0,    0,    0,   28,   28,    0,    0,    0,    0,    0,
    0,  637,    0,  637,    0,  637,  106,  258,  637,  109,
  637,  637,    0,  637,    0,  637,    0,  637,    0,  637,
  637,  637,    0,    0,    0,  637,  637,    0,    0,    0,
    0,  637,    0,  637,  637,    0,    0,    0,  637,    0,
    0,    0,  637,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,  637,  637,    0,  637,    0,    0,    0,
  637,  637,    0,    0,    0,    0,    0,    0,  637,  637,
   57,   25,  637,   26,    0,  637,   27,  255,    0,    0,
  637,   28,   62,   63,    0,   29,    0,    0,    0,    0,
    0,   65,    0,    0,   31,    0,    0,    0,    0,    0,
    0,   33,  637,  637,    0,    0,   34,    0,   72,   73,
   35,    0,    0,    0,    0,  637,    0,    0,    0,    0,
    0,    0,   37,    0,   38,   75,    0,    0,   39,    0,
    0,   77,    0,   79,    0,   81,   40,   41,  256,    0,
   42,    0,    0,   85,    0,    0,    0,    0,    0,    0,
   25,    0,   26,    0,    0,   27,  637,    0,    0,    0,
   28,    0,    0,    0,   29,   90,   91,   92,  257,    0,
    0,    0,    0,   31,  636,    0,  636,   96,    0,  636,
   33,  636,  636,    0,  636,   34,  636,    0,  636,   35,
  636,  636,  636,    0,    0,    0,  636,  636,    0,    0,
    0,   37,  636,   38,  636,  636,    0,   39,    0,  636,
    0,    0,    0,  636,    0,   40,   41,    0,    0,   42,
    0,    0,  321,  106,  258,  636,    0,  636,    0,    0,
    0,  636,  636,    0,    0,    0,    0,    0,    0,  636,
  636,    0,  636,  636,  636,    0,  636,  636,  292,  636,
  636,  636,  636,    0,  636,    0,  636,    0,  636,  636,
  636,    0,    0,    0,  636,  636,    0,    0,    0,    0,
  636,    0,  636,  636,    0,    0,    0,  636,    0,    0,
    0,  636,    0,    0,    0,    0,  636,    0,    0,    0,
    0,    0,    0,  636,    0,  636,    0,    0,    0,  636,
  636,    0,    0,  322,    0,    0,    0,  636,  636,    0,
    0,  636,    0,    0,  636,    0,   25,    0,   26,  636,
    0,   27,    0,    0, 1269,    0,   28,  636,  677,    0,
   29,    0,  678, 1270, 1271,    0,    0,    0, 1272,   31,
    0,    0,    0,    0, 1273,    0,   33,    0,   25,    0,
   26,   34,    0,   27,    0,   35, 1269,    0,   28,    0,
  677,    0,   29,    0,  678, 1270, 1271,   37,    0,   38,
 1272,   31,    0,   39,    0,    0, 1273,    0,   33,    0,
    0,   40,   41,   34,    0,   42,    0,   35, 1274,    0,
    0,    0,   48, 1275,   48,  636,    0,   48,    0,   37,
    0,   38,   48,    0,    0,   39,   48,    0,    0,    0,
    0,    0,    0,   40,   41,   48,    0,   42,    0,    0,
 1274,    0,   48,    0,   48, 1275,   48,   48, 1276,   48,
    0,   48,    0,   48,   48,   48,    0,    0,   48,    0,
   48,    0,    0,   48,    0,   48,    0,   48,    0,   48,
    0,    0,   48,    0,   48,    0,    0,   48,   48,   48,
    0,   48,    0,   48,   48,   48,    0,   48,   48, 1277,
   48,    0,   48,   48,    0,   48,    0,   48,   48,    0,
    0,   48,   48,    0,   48,    0,    0,    0,    0,   48,
   48,   48,    0,   48,    0,    0,   48,    0,   48,  153,
   25, 1277,   26,   48,    0,   27,    0,   48,    0,   48,
   28,   48,    0,    0,   29,    0,   48,    0,    0,   48,
    0,   48,    0,   31,    0,   48,    0,    0,   48,  153,
   33,    0,    0,   48,   48,   34,    0,   48,    0,   35,
   48,  565,    0,    0,    0,   48,    0,    0,  566,    0,
    0,   37,    0,   38,    0,    0,    0,   39,    0,    0,
  567,    0,    0,    0,    0,   40,   41,    0,    0,   42,
    0,   25,  568,   26,    0,    0,   27,   48,    0,    0,
    0,   28,    0,    0,    0,   29,    0,    0,    0,   30,
  485,    0,  485,    0,   31,  485,    0,    0,    0,   32,
  485,   33,    0,    0,  485,    0,   34,    0,    0,    0,
   35,   36,    0,  485,    0,    0,    0,    0,    0,    0,
  485,   48,   37,    0,   38,  485,    0,    0,   39,  485,
    0,    0,    0,    0,    0,    0,   40,   41,    0,    0,
   42,  485,    0,  485,  173,    0,  173,  485,    0,  173,
    0,    0,    0,  569,  173,  485,  485,    0,  173,  485,
    0,  172,  485,  172,    0,    0,  172,  173,    0,    0,
    0,  172,    0,    0,  173,  172,    0,    0,    0,  173,
    0,    0,    0,  173,  172,    0,    0,    0,    0,    0,
    0,  172,    0,    0,    0,  173,  172,  173,    0,    0,
  172,  173,    0,    0,    0,    0,    0,    0,    0,  173,
  173,    0,  172,  173,  172,  182,  173,  182,  172,    0,
  182,    0,    0,    0,   43,  182,  172,  172,    0,  182,
  172,    0,    0,  172,    0,    0,    0,    0,  182,    0,
    0,    0,    0,  485,    0,  182,    0,    0,    0,    0,
  182,    0,    0,   35,  182,    0,    0,    0,    0,    0,
    0,    0,    0,    0,   35,    0,  182,    0,  182,   35,
    0,    0,  182,   35,    0,    0,   35,    0,    0,    0,
  182,  182,    0,    0,  182,    0,    0,  182,   35,   35,
    0,    0,    0,   35,   35,    0,    0,  173,    0,   35,
    0,   35,   35,   35,   35,    0,    0,    0,    0,   35,
    0,    0,    0,   35,  172,   35,    0,    0,    0,    0,
    0,    0,    0,    0,    0,   35,    0,   35,   35,   33,
   35,    0,    0,    0,   35,    0,    0,    0,    0,    0,
   33,    0,    0,    0,    0,   33,    0,    0,    0,   33,
    0,    0,   33,    0,   35,    0,    0,    0,    0,    0,
   35,   35,    0,    0,   33,   33,    0,    0,  182,   33,
   33,    0,   48,    0,    0,   33,    0,   33,   33,   33,
   33,    0,    0,   48,    0,   33,    0,    0,   48,   33,
    0,   33,   48,    0,    0,   48,    0,    0,    0,    0,
    0,   33,    0,    0,   33,    0,   33,   48,   48,    0,
   33,    0,   48,   48,    0,   48,    0,    0,   48,    0,
   48,   48,   48,   48,    0,    0,   48,    0,   48,    0,
   33,   48,   48,    0,   48,   48,   33,   33,   48,    0,
    0,    0,    0,    0,   48,    0,    0,   48,    0,   48,
   48,   48,    0,   48,    0,   48,   48,   48,    0,    0,
    0,   48,    0,   48,   48,   48,   48,    0,    0,    0,
    0,   48,    0,   48,    0,   48,    0,   48,    0,   37,
   48,    0,    0,    0,    0,    0,    0,   48,    0,    0,
   48,    0,   48,   48,    0,   48,   48,   48,   48,    0,
    0,    0,    0,   48,    0,   48,   48,   48,   48,    0,
    0,   48,    0,   48,    0,    0,   48,   48,   48,    0,
    0,    0,   38,    0,    0,    0,    0,    0,    0,   48,
    0,   48,   48,   48,   48,    0,   48,    0,    0,    0,
    0,   48,    0,   48,   48,   48,   48,    0,    0,    0,
    0,   48,    0,    0,    0,   48,   48,    0,   48,    0,
    0,    0,    0,  450,  296,    0,    0,   48,    0,   48,
   48,    0,   48,    0,   48,    0,    0,    0,    0,   48,
    0,   48,   48,   48,   48,    0,  451,    0,    0,   48,
    0,    0,    0,   48,    0,    0,   48,    0,    0,  452,
    0,    0,  297,  453,  454,   48,  450,    0,   48,  455,
   48,  456,  457,  458,  459,    0,    0,    0,    0,  460,
    0,    0,    0,  461,    0,    0,    0,    0,    0,  451,
    0,    0,    0,    0,   48,  462,   48,   48,  463,    0,
  464,    0,  452,    0,    0,    0,    0,  454,    0,    0,
    0,    0,  455,    0,  456,  457,  458,  459,    0,    0,
    0,    0,  460,    0,  465,    0,  461,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,  462,    0,
    0,  463,    0,  464,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,    0,    0,    0,
    0,    0,    0,    0,    0,    0,    0,  465,
  };
  protected static readonly short [] yyCheck = {            17,
   17,  300,  192,   17,    4,  511,   69,   20,   18,   52,
  301,   17,  513,  190,  249,    6,   20,  189,  290,  467,
  320,   85,  338,  234,  554,   88,   89,   17,   48,   60,
   93,  299,  158,  487,  331,  297,  919,  768,   60, 1095,
  713,  714,  571,  738,  236,   59, 1130, 1131,   78,  579,
  114,    0,  116,  256,  256, 1216,   74,  256,  268,  327,
   78,  114,  256,  116,  325,  268,   80,  256,   82,  256,
   60, 1232,  256,  325,   64,   88,   89,  353,   96,  924,
   60,  256,   17,  756,   17,  758,  256,  256,  256,  256,
  256,  256,  294, 1177,  294,  268,  109,  256,  282,  256,
  363,  256,  256,  335,  268,  343,  306,  367,  372,  376,
  173,  371,  276,  353,   94,  256,  372,   17,   98,   99,
  100,  101,  102,  103,  104,  105,  391,  656,  256,  256,
  314,  277,  322,  368,  370,  281,  256,  256,  374,  257,
  158,  158,  410,  256,  158,  372,  368,  190,   17,  414,
  269,   17,  158,  391,  256,  418,  343,  418,    0,   17,
  420,  363,  429,   17,  354,  429,  418,  286,  158,  359,
  339,  256,   17,  429,  294,  344,  414,  346,   17,   17,
  339,  256,  381,  352,  353,  344,  249,  346,  201,  202,
  428,  234,   17,  352,  353,  368,  342,  429,  261,  363,
  341,  256,  429,  368,  391,  225, 1051,  429,  894,  374,
  507,  510,  422,  381,  715,  372,  418,    0,  418,  422,
  431,  375, 1090,  158,  418,  158,  381,  414,  369,  418,
 1391,  249,   88,   89,  418,  363,  254,  339,  554,  230,
  432,  428,  370,  418,  372,  372,  374,  437,  418,  422,
  263,  418,  418,  109,  418,  319,  287, 1418,  158,  323,
  429,  374,  325,  579,  328,  287,  297,  330,  256, 1430,
  429, 1432,  290,  549,  256,  328,  289,  295,  296, 1010,
  277,  471,  367,  256,  369,  358,  371,  527,  301,  158,
  418,  309,  158,  368,  316,  308,  314,  287,  316,  374,
  158,  315,  320,  358,  158,  368,  369,  287,  257,  549,
  983,  262,  368,  158,  332,  333, 1002,  297, 1004,  158,
  158,  338,  302,  852,  263, 1176,  316,  357,  256, 1385,
 1025,  571,  338,  158,  256,  420,  256,  350,  423,  357,
  353, 1200, 1073,  372, 1195,  201,  202,  298,  338,  256,
  368,  369,  415,  416,  417,  373,  374,  375,  376,  377,
  378,  379,  380,  381,  382,  383,  363, 1451,  256,  424,
  425,  426,  427,  429,  387,  388,  315,  256,  367,  569,
  369,  657,  370,  884, 1243,  367,  374,  405,  431, 1272,
  263,  581,  372,  583,  376,  585, 1480,  370,  441,  699,
  429,  374,  415,  416,  384,  385,  386,  263,  862,  389,
  390,  294,  256,  394,  395,  257,  656,  264,  438,  439,
  433,  418,  419,  486,  444,  488,  368,  378,  379,  433,
  272, 1104, 1283,  289,  447,  277, 1287,  677, 1111,  281,
  335,  369,  315,  447,  266,  349,  350,  369,  511,  369,
  685,  339,  308,  372,  296,  717, 1324,  266,  648,  368,
  339, 1312,  369, 1136,  527,  487,  368,  481,  531,  349,
  350,  349,  350,  256,  357,  976,  418,  667,  325,  367,
  363,  323,  368,  368, 1352, 1353,  369, 1355,  367,  372,
  373,  509,  314,  511,  350,  513,  340,  487, 1366,  418,
  342, 1369,  363,  386,  374,  314,  782,  570,  369,  523,
  524,  256,  369,  257,  527,  372, 1384,  272,  256,  537,
  429,  272,  277,  527,  542,  429,  281,  429,  363, 1202,
  363,  387,  388,  376,  369,  418,  549,  554,  421,  305,
 1408,  296,  782,  429,  429,  296,  416,  561,  554,  429,
  849,  429,  372,  375,  374,  272,  339,  418,  571,  415,
  416,  372,  579,  374,  554,  363,  375,  802,  323,  363,
  588,  589,  323,  579,  637,  851,  639,  368,  343,  296,
  877, 1420, 1421,  418,  367,  418,  429,  342,  371,  579,
  373,  374,  375,  376,  339,  272,  576,  418,  381,  344,
  726,  346,  371,  374, 1200,  305,  323,  352,  353,  429,
 1140,  369,  852,  881,  341,  373,  381,  635,  429,  296,
  418,  343,  685,  641,  418,  256,  391,  418,  919,  367,
 1200,  369, 1200,  371,  372,  369,  374, 1476,  376,  702,
  256,  269,  369,  656,   21,  416,  323, 1243,   62,  414,
  418,  420,   66,   67,   68, 1200,   70,   71,  286,  381,
  428,   75,   76,  428,  677,  423,  376,  685,   82,  391,
   84,  261,   86, 1243,  305, 1243,   53,   91,   92,  979,
  418,  699,  420, 1137,  748,  423,  717,  678,  751,  305,
  376,  709,  414,  992,  284, 1200,  357,  715, 1243,  357,
  381,  115,  720,  893,  339,  376,  428,  297, 1200,  944,
  391,  369,  302, 1200,  372,  373,  381,  307,  781,  309,
  310,  311,  312,  713,  714,  386,  391,  317,  386,  742,
  744,  321,  367,  414,  797,  977,  386,  717, 1243,  802,
  367,  376,  760,  333,  762,  371,  336,  428,  338,  414,
  763, 1243,  766,  372,  768,  818, 1243,  391,  373,  777,
  371,  376,  414,  421, 1006,  778,  756,  414,  758,  782,
  389,  390,  362, 1290,  364,  365,  428,  357,  368,  797,
  414,  428,  367,  826,  802,  803, 1303,  805,  418, 1061,
  339,  376, 1108,  373,  428,  344, 1095,  346,  816,  817,
  349,  350,  339,  352,  353, 1322,  386,  344,  401,  346,
  294, 1001,  349,  350,  370,  352,  353,    6,  374,  379,
  413,  834,  306,  369, 1140,  888, 1098,  890,   17,  847,
  834,  849,  367,  896,  370,  381,  357,  374,  374,  852,
  853,  376,  256,  382,  383,  259,  389,  865,  839, 1199,
 1200, 1023,  373, 1088,  872, 1303, 1172,  396,  397,  922,
  398,  399, 1134,  339,  882,  386,  884,  357,  344, 1219,
  346,   60, 1373,  349,  350,   64,  352,  353,  418,  872,
  429,  944,  306,  373,  308,  299,  742,  306,  428,  313,
  883,  384,  429, 1243,  313, 1245,  386,  960,  312,   88,
   89,  325,  385,  371,  339,  373,  919,  763,  926,  344,
  928,  346,  930,  373,  349,  350,  376,  352,  353,  371,
  109,  373,  778,  306,  392,  393,  944,  368,  941,  371,
  313,  373,  390,  374, 1435, 1207,  949,  941, 1149,  374,
  381,  367,  325,  400,  412,  367,  373,  367, 1125,  372,
  376,  374,  420,  429,  376,  423,  376,  369,  976,  261,
  951,  979,  953,  367,  955,  367,  339,  354,  355,  158,
 1471,  344,  376,  346,  376,  376,  349,  350,  339,  352,
  353, 1272,  284,  344,  372,  346,  374, 1493, 1494,  418,
    0,  352,  353,  983,  429,  297, 1010,  853,  415, 1017,
  302, 1019,  372, 1021,  374,  307,  376,  309,  310,  311,
  312,  357,  201,  202, 1254,  317,  372,  363,  374,  321,
  376, 1261,  374,  369,  376, 1088,  372,  373,  374,  381,
  371,  333, 1032,  370,  336,  367,  338,  374,  418,  371,
  386,  373,  374, 1061,  376,  354,  355,  370,  371,  381,
  373,  374,  375,  386,  387,  388,  429, 1075, 1076, 1073,
  362,  256,  364,  365,  418, 1083,  368,  370,  371,  483,
 1088,  374,  418,  374,  263,  376,  370, 1095,  372, 1089,
 1098,  370, 1125,  415, 1127,  374, 1385, 1277,  370,  376,
  372, 1108,  374,  949,  374,  277,  376, 1115,  287,  370,
  289,  372, 1108,  374, 1122,  370, 1149,  372,  376,  523,
 1300,  370,  301,  372, 1104, 1128, 1134,  373, 1108,  308,
  370, 1111,  372, 1140, 1314,  367,  367,  316, 1146, 1147,
  371,  376,  373,  374, 1140,  376,  376, 1180,  367,  372,
  381, 1331,  371, 1333,  373,  374, 1136,  376,  294,  338,
 1140,  372,  381,  374,  294, 1172, 1199, 1200,  372,  376,
  374,  350,  343,  370,  353,  372, 1172,  370, 1181,  372,
  374, 1189,  376,  374,  415,  376, 1219,  393,  394,  395,
  396,  374, 1172,  376, 1247, 1198,  374, 1205,  376, 1207,
  414,  415,  364,  365,  349,  350,  372,  373,  387,  388,
 1243,  414, 1245,  364,  365,  369, 1093, 1094,  391,  392,
  418,   94, 1202,  397,  398,   98,   99,  100,  101,  102,
  103,  104,  105,  375,  418,  372,  415,  416,  418,  368,
  372,  376,  371,  368,  373,  374,  372,  374,  372,  372,
 1253, 1254,  372, 1257,  372,  372,  294,  257, 1261,  294,
  374,  261,  372,  392,  393,  372, 1269, 1270,  374, 1272,
  372,  376,  272,  371,  256, 1278,  375,  277,  374,  294,
  294,  281,  356,  412,  284,  372,  381, 1290,  374, 1292,
  375,  420, 1295,    0,  423,  374,  296,  297,  373,  418,
 1303,  301,  302,  381, 1308,  372,  374,  307,  487,  309,
  310,  311,  312,  374,  374,  374,  429,  317,  256, 1322,
  376,  321,  423,  323,  372,  374,  367,  421,  372,  372,
 1363,  373,  343,  333,  368,  335,  336,  371,  338,  373,
  374,  374,  342,  294,  294,  374,  418, 1380,  527,  370,
  367,  371, 1198,  418,  381,  256,  375,  256,  392,  393,
 1393, 1394,  362,  256,  374, 1373,  256, 1371,  368,  369,
  549,  372,  280,  256,  367,  554,  368, 1385,  412,  343,
  376,  372,  370,  423,  371,  374,  420, 1420, 1421,  423,
  376,  374,  571,  372, 1402,  376,  370,  372,  347,  372,
  579,  339,  367,  381,  351,  381,  344, 1253,  346,  347,
  348,  349,  350,  351,  352,  353,  354,  355,  356,  381,
  256,  256,  376, 1269, 1270,  368,  372, 1435,  347,  302,
  368,  370,  370,  374,  372,  375,  374,  375,  376,    0,
 1493, 1494,  370, 1476,  375,  367, 1292,  370,  372, 1295,
  348,  368,  390,  374,  374,  418,  372,  418,  331,  339,
  348,  376,  368, 1471,  375,  368,  376,  367,  381,  356,
  367,  367,  337, 1477, 1478,  305,  371,  656,  372,  381,
 1484, 1485,  374,  368,  368, 1493, 1494,  368,  418,  418,
  369,  429,  418,  371,  371,  367,  371,  371,  677,  372,
  402,  403,  404,  405,  406,  407,  408,  409,  410,  411,
  376,  384,  385,  386,  418,  371,  389,  390,  391,  392,
  393,  394,  395,  396,  397,  398,  399,  400,  401,  402,
  403,  404,  418,  381,  713,  714,  373,  371,  367,  371,
  381,  369,  373,  265,  371,  267,  372,  372,  270,  256,
  272,  374,  373,  275,  374,  262,  374,  279,  256,  374,
  418,  372,  376,  742,  372,  370,  288,  418,  418,  376,
  376,  372,  418,  295,  376,  372,  367,  756,  300,  758,
  302,  381,  304,  372,  763,  368,  381,  294,  370,  368,
  315,  298,  372,  263,  316,  371,  318,  371,  368,  778,
  322,  323,    0,  782,  372,  372,    0,  367,  330,  331,
  376,  376,  334,  368,  372,  337,    0,    0,  368,  376,
  372,  372,  418,  370,  367,  370,  368,  376,  368,  367,
  376,  372,  339,  418,  507,  368,  372,  344,  372,  346,
  347,  348,  349,  350,  351,  352,  353,  354,  355,  356,
  357,  368,  376,  376,  418,  376,  363,  376,  372,  368,
  367,  368,  369,  370,  371,  372,  373,  374,  375,  376,
  367,  378,  379,  852,  853,  382,  383,  384,  385,  386,
  372,  368,  389,  390,  372,  368,  376,  394,  395,  396,
  397,  398,  399,  400,  401,  256,  418,  373,  367,  376,
  261,  262,  376,  576,  376,  376,  413,  315,  376,  416,
  376,  418,  376,  420,  376,  263,  423,   51,   12,   52,
    5,  941,  429,  284, 1088,  834, 1088, 1245, 1219, 1425,
 1388, 1441,  677,  294, 1278, 1376,  297,  298, 1371, 1405,
  919,  302, 1290,  855,  305, 1485,  307,  855,  309,  310,
  311,  312,  855, 1243,  691, 1309,  317, 1479,  851, 1231,
  321, 1393, 1398, 1394,  325, 1478, 1333,  720, 1278, 1180,
  949,  826,  333,  527,  797,  336, 1181,  338,  339,  589,
  877,  368,  992,  344,  802,  346,  347,  348,  349,  350,
  351,  352,  353,  354,  355,  356,  685,   72,  334,  717,
  399,  362,  363,  256,  983,  400,  367,  368,  401,  370,
  371,  372,  373,  374,  375,  376,  402,  378,  379,  403,
  381,  382,  383,  384,  385,  386,  387,  388,  389,  390,
    0,  392,  393,  394,  395,  396,  397,  398,  399,  400,
  401,  402,  403,  404,  405,  406,  407,  408,  409,  410,
  411,  412,  413,  404,  549,  416, 1163,  418, 1253,  420,
 1172, 1108,  423,  158, 1053, 1006, 1066,  978,  429, 1064,
  782, 1076,  256,  962, 1137,  525,  898,  423,  262, 1147,
 1249,   -1,   -1,   -1,  832,   -1,  339,  831,   -1,   -1,
   -1,  344,   -1,  346,  347,  348,  349,  350,  351,  352,
  353,  354,  355,  356,   -1,   -1,   -1,   -1,   -1,   -1,
  294,   -1,   -1,   -1,  298,  368,   -1,  370,   -1,  372,
   -1,  374,  375,  376,   -1, 1104,   -1,   -1,   -1, 1108,
   -1,   -1, 1111,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1, 1128,
   -1,   -1,   -1,   -1,  285,  339,   -1, 1136,   -1,   -1,
  344, 1140,  346,  347,  348,  349,  350,  351,  352,  353,
  354,  355,  356,   -1,   -1,   -1,  429,   -1,   -1,  363,
   -1,   -1,   -1,  367,  368,   -1,  370,  371,  372,   -1,
  374,  375,  376, 1172,  378,  379,  327,    0,  382,  383,
  384,  385, 1181,   -1,  877,  389,  390,   -1,   -1,   -1,
  394,  395,  396,  397,  398,  399,  400,  401,   -1, 1198,
   -1,   -1,   -1, 1202,   -1,   -1,   -1,   -1,   -1,  413,
   -1,   -1,  416,   -1,  418,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  429,  377,  378,  379,  380,
   -1,  382,  383,  384,  385,  386,  387,  388,  389,   -1,
   -1,  392,  393,  394,  395,  396,  397,  398,  399,   -1,
   -1,   -1,   -1,   -1, 1253, 1254,   -1,   -1,   -1,   -1,
   -1,   -1, 1261,   -1,   -1,   -1,  256,  257,   -1,   -1,
 1269, 1270,   -1, 1272,  264,  265,  266,  267,  268,   -1,
  270,  271,   -1,  273,  274,  275,  276,  277,  278,  279,
  280,   -1,   -1, 1292,   -1,  285, 1295,  287,  288,  289,
  290,  291,  292,   -1,   -1,  295,    0,   -1,   -1,  299,
  300,   -1,  302,  303,  304,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  314,   -1,  316,   -1,  318,  319,
   -1,   -1,  322,   -1,  324,  325,  326,  327,  328,  329,
  330,  331,  332,  333,  334,  335,   -1,  337,   -1,   -1,
  340,  341,   -1,   -1,  344,  345,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  359,
  360,  361,  362,  363,   -1,   -1,   -1,  367,  368,   -1,
   -1,  371,   -1,   -1,   -1,   -1,  376,  377,  378,  379,
  380,   -1,   -1,   -1,  384,   -1,  386,   -1,   -1,   -1,
   -1,   -1,  392,  393,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  417,  418,  419,
  420,   -1,  422,  256,  257,   -1,   -1,   -1,   -1,  429,
   -1,  264,  265,  266,  267,  268,   -1,  270,  271,   -1,
  273,  274,  275,  276,  277,  278,  279,   -1,   -1,    0,
   -1,   -1,  285,   -1,  287,  288,  289,  290,  291,  292,
   -1,   -1,  295,   -1,   -1,   -1,  299,  300,   -1,  302,
  303,  304,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  314,   -1,  316,   -1,  318,  319,   -1,   -1,  322,
   -1,  324,  325,  326,  327,  328,  329,  330,  331,  332,
  333,  334,  335,   -1,  337,   -1,   -1,  340,  341,   -1,
   -1,  344,  345,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  359,  360,  361,  362,
  363,   -1,   -1,   -1,  367,  368,   -1,   -1,  371,   -1,
   -1,   -1,   -1,  376,  377,  378,  379,  380,   -1,   -1,
   -1,  384,  256,  386,   -1,   -1,   -1,  261,  262,  392,
  393,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,    0,
  284,   -1,   -1,   -1,  417,  418,  419,  420,   -1,  422,
  294,   -1,   -1,  297,  298,   -1,  429,   -1,  302,   -1,
   -1,  305,   -1,  307,   -1,  309,  310,  311,  312,   -1,
   -1,   -1,   -1,  317,   -1,   -1,   -1,  321,   -1,   -1,
   -1,  325,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  333,
   -1,   -1,  336,   -1,  338,  339,   -1,   -1,   -1,   -1,
  344,   -1,  346,  347,  348,  349,  350,  351,  352,  353,
  354,  355,  356,  357,   -1,   -1,   -1,   -1,  362,  363,
   -1,   -1,   -1,  367,  368,  369,  370,  371,  372,  373,
  374,  375,  376,   -1,  378,  379,   -1,   -1,  382,  383,
  384,  385,  386,   -1,   -1,  389,  390,   -1,   -1,   -1,
  394,  395,  396,  397,  398,  399,  400,  401,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  256,   -1,   -1,   -1,  413,
  261,  262,  416,   -1,  418,   -1,  420,   -1,   -1,  423,
   -1,   -1,   -1,   -1,   -1,  429,    0,   -1,   -1,   -1,
   -1,   -1,   -1,  284,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  294,   -1,   -1,  297,  298,   -1,   -1,
   -1,  302,   -1,   -1,  305,   -1,  307,   -1,  309,  310,
  311,  312,   -1,   -1,   -1,   -1,  317,   -1,   -1,   -1,
  321,   -1,   -1,   -1,  325,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  333,   -1,   -1,  336,   -1,  338,  339,   -1,
   -1,   -1,   -1,  344,   -1,  346,  347,  348,  349,  350,
  351,  352,  353,  354,  355,  356,   -1,   -1,   -1,   -1,
   -1,  362,  363,   -1,   -1,   -1,  367,  368,  369,  370,
  371,  372,   -1,  374,  375,  376,   -1,  378,  379,   -1,
   -1,  382,  383,  384,  385,  256,   -1,   -1,  389,  390,
  261,  262,   -1,  394,  395,  396,  397,  398,  399,  400,
  401,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  413,  284,   -1,  416,   -1,  418,   -1,  420,
   -1,   -1,  423,  294,   -1,   -1,  297,  298,  429,   -1,
   -1,  302,   -1,   -1,  305,   -1,  307,   -1,  309,  310,
  311,  312,   -1,    0,   -1,   -1,  317,   -1,   -1,   -1,
  321,   -1,   -1,   -1,  325,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  333,   -1,   -1,  336,   -1,  338,  339,   -1,
   -1,   -1,   -1,  344,   -1,  346,  347,  348,  349,  350,
  351,  352,  353,  354,  355,  356,   -1,   -1,   -1,   -1,
   -1,  362,  363,   -1,   -1,   -1,  367,  368,  369,  370,
  371,  372,   -1,  374,  375,  376,   -1,  378,  379,   -1,
   -1,  382,  383,  384,  385,   -1,   -1,   -1,  389,  390,
   -1,   -1,   -1,  394,  395,  396,  397,  398,  399,  400,
  401,   -1,  256,   -1,   -1,   -1,   -1,  261,  262,   -1,
   -1,   -1,  413,   -1,   -1,  416,   -1,  418,   -1,  420,
   -1,   -1,  423,   -1,   -1,   -1,    0,   -1,  429,   -1,
  284,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  294,   -1,   -1,  297,  298,   -1,   -1,   -1,  302,   -1,
   -1,  305,   -1,  307,   -1,  309,  310,  311,  312,   -1,
   -1,   -1,   -1,  317,   -1,   -1,   -1,  321,   -1,   -1,
   -1,  325,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  333,
   -1,   -1,  336,   -1,  338,  339,   -1,   -1,   -1,    0,
  344,   -1,  346,  347,  348,  349,  350,  351,  352,  353,
  354,  355,  356,   -1,   -1,   -1,   -1,   -1,  362,  363,
   -1,   -1,   -1,  367,  368,   -1,  370,  371,  372,   -1,
  374,  375,  376,   -1,  378,  379,   -1,   -1,  382,  383,
  384,  385,   -1,   -1,   -1,  389,  390,   -1,   -1,   -1,
  394,  395,  396,  397,  398,  399,  400,  401,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,    0,  413,
   -1,   -1,  416,   -1,  418,   -1,   -1,   -1,   -1,   -1,
  257,   -1,   -1,   -1,  261,  429,  263,   -1,  265,   -1,
  267,   -1,   -1,  270,   -1,  272,  273,   -1,  275,   -1,
  277,   -1,  279,   -1,  281,  282,  283,  284,   -1,   -1,
  287,  288,   -1,   -1,   -1,   -1,  293,  294,  295,  296,
  297,   -1,   -1,  300,  301,  302,   -1,  304,   -1,  306,
  307,  308,  309,  310,  311,  312,  313,    0,  315,  316,
  317,  318,   -1,   -1,  321,  322,  323,   -1,  325,   -1,
   -1,   -1,   -1,  330,  331,   -1,  333,  334,   -1,  336,
  337,  338,   -1,   -1,   -1,  342,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  362,   -1,  364,  365,   -1,
   -1,   -1,  256,  257,   -1,   -1,   -1,  261,   -1,   -1,
  377,  265,   -1,  267,   -1,   -1,  270,   -1,  272,  273,
   -1,  275,   -1,  277,   -1,  279,   -1,  281,  282,  283,
  284,   -1,   -1,  287,  288,   -1,    0,   -1,   -1,  293,
   -1,  295,  296,  297,   -1,   -1,  300,  301,  302,   -1,
  304,  418,   -1,  307,   -1,  309,  310,  311,  312,   -1,
   -1,   -1,  316,  317,  318,   -1,  257,  321,  322,  323,
  261,   -1,   -1,   -1,   -1,   -1,  330,  331,   -1,  333,
  334,  272,  336,  337,  338,   -1,  277,   -1,  342,   -1,
  281,   -1,   -1,  284,   -1,   -1,   -1,   -1,   -1,    0,
   -1,   -1,   -1,   -1,   -1,  296,  297,   -1,  362,   -1,
  301,  302,   -1,   -1,  368,  369,  307,   -1,  309,  310,
  311,  312,   -1,  377,   -1,   -1,  317,   -1,   -1,   -1,
  321,   -1,  323,   -1,   -1,  257,   -1,   -1,   -1,  261,
   -1,   -1,  333,   -1,  335,  336,    0,  338,   -1,   -1,
  272,  342,   -1,   -1,   -1,  277,   -1,   -1,   -1,  281,
   -1,   -1,  284,   -1,  418,   -1,   -1,   -1,   -1,   -1,
   -1,  362,   -1,   -1,  296,  297,   -1,   -1,  369,  301,
  302,   -1,   -1,   -1,   -1,  307,   -1,  309,  310,  311,
  312,   -1,   -1,    0,   -1,  317,   -1,   -1,   -1,  321,
   -1,  323,   -1,   -1,  257,   -1,   -1,   -1,  261,   -1,
   -1,  333,   -1,   -1,  336,   -1,  338,   -1,   -1,  272,
  342,   -1,   -1,   -1,  277,   -1,   -1,   -1,  281,   -1,
   -1,  284,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  362,   -1,   -1,  296,  297,   -1,  368,  369,  301,  302,
  261,   -1,  263,   -1,  307,   -1,  309,  310,  311,  312,
    0,   -1,   -1,   -1,  317,   -1,   -1,   -1,  321,   -1,
  323,   -1,   -1,  284,   -1,   -1,   -1,   -1,   -1,   -1,
  333,   -1,   -1,  336,   -1,  338,  297,   -1,   -1,  342,
   -1,  302,   -1,  257,   -1,   -1,  307,  261,  309,  310,
  311,  312,   -1,    0,  315,   -1,  317,   -1,  272,  362,
  321,   -1,   -1,  277,   -1,  368,  369,  281,   -1,   -1,
  284,   -1,  333,   -1,   -1,  336,   -1,  338,   -1,   -1,
   -1,   -1,  296,  297,   -1,   -1,   -1,  301,  302,   -1,
   -1,   -1,   -1,  307,   -1,  309,  310,  311,  312,   -1,
   -1,  362,   -1,  317,   -1,   -1,  257,  321,   -1,  323,
  261,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  333,
   -1,  272,  336,   -1,  338,   -1,  277,   -1,  342,   -1,
  281,   -1,   -1,  284,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  296,  297,   -1,  362,   -1,
  301,  302,   -1,  257,   -1,  369,  307,  261,  309,  310,
  311,  312,   -1,   -1,   -1,   -1,  317,   -1,  272,   -1,
  321,   -1,  323,  277,   -1,   -1,   -1,  281,   -1,   -1,
  284,   -1,  333,   -1,   -1,  336,   -1,  338,   -1,   -1,
   -1,  342,  296,  297,   -1,   -1,   -1,  301,  302,   -1,
  257,   -1,   -1,  307,  261,  309,  310,  311,  312,   -1,
   -1,  362,   -1,  317,   -1,  272,   -1,  321,   -1,  323,
  277,   -1,   -1,   -1,  281,   -1,   -1,  284,   -1,  333,
   -1,   -1,  336,   -1,  338,   -1,   -1,   -1,  342,  296,
  297,   -1,   -1,   -1,  301,  302,   -1,   -1,   -1,   -1,
  307,   -1,  309,  310,  311,  312,   -1,   -1,  362,   -1,
  317,   -1,   -1,   -1,  321,   -1,  323,  257,   -1,   -1,
   -1,  261,   -1,   -1,   -1,   -1,  333,   -1,   -1,  336,
   -1,  338,  272,   -1,   -1,  342,   -1,  277,   -1,   -1,
   -1,  281,   -1,   -1,  284,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  362,  296,  297,   -1,   -1,
  257,  301,  302,   -1,  261,   -1,   -1,  307,   -1,  309,
  310,  311,  312,   -1,   -1,  272,   -1,  317,   -1,   -1,
  277,  321,   -1,  323,  281,   -1,   -1,  284,   -1,   -1,
   -1,   -1,   -1,  333,   -1,   -1,  336,   -1,  338,  296,
  297,   -1,  342,   -1,  301,  302,   -1,   -1,   -1,   -1,
  307,   -1,  309,  310,  311,  312,   -1,   -1,   -1,   -1,
  317,   -1,  362,   -1,  321,   -1,  323,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  333,   -1,  256,  336,
   -1,  338,   -1,   -1,   -1,  342,  264,  265,  266,  267,
   -1,   -1,  270,  271,   -1,  273,  274,  275,  276,  277,
  278,  279,   -1,   -1,   -1,  362,   -1,  285,   -1,  287,
  288,  289,  290,  291,  292,   -1,   -1,  295,   -1,   -1,
   -1,  299,  300,   -1,  302,  303,  304,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  314,   -1,  316,   -1,
  318,  319,   -1,   -1,  322,   -1,  324,  325,  326,  327,
  328,  329,  330,  331,  332,  333,  334,  335,   -1,  337,
   -1,   -1,  340,  341,   -1,  256,  344,  345,   -1,   -1,
   -1,  262,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  359,  360,  361,  362,  363,   -1,   -1,   -1,  367,
   -1,   -1,   -1,  371,   -1,   -1,   -1,   -1,  376,  377,
  378,  379,  380,   -1,   -1,   -1,  384,  298,  386,   -1,
   -1,   -1,   -1,   -1,  392,  393,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  256,   -1,   -1,   -1,   -1,   -1,  262,   -1,  417,
  418,  419,  420,   -1,   -1,   -1,   -1,   -1,  339,   -1,
   -1,  429,   -1,  344,   -1,  346,  347,  348,  349,  350,
  351,  352,  353,  354,  355,  356,  357,   -1,   -1,   -1,
   -1,   -1,  363,  298,   -1,   -1,   -1,  368,  369,  370,
  371,  372,  373,  374,  375,  376,   -1,  378,  379,   -1,
  381,  382,  383,  384,  385,  386,  387,  388,  389,  390,
   -1,  392,  393,  394,  395,  396,  397,  398,  399,  400,
  401,  402,  403,  404,  405,  406,  407,  408,  409,  410,
  411,  412,  413,   -1,  256,   -1,   -1,  418,   -1,  420,
  262,   -1,  423,   -1,   -1,   -1,   -1,   -1,  429,   -1,
   -1,   -1,   -1,  368,   -1,   -1,  371,   -1,  373,  374,
   -1,   -1,   -1,  378,  379,   -1,   -1,  382,  383,  384,
  385,  386,  387,  388,  389,  390,  298,  392,  393,  394,
  395,  396,  397,  398,  399,  400,  401,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  412,  413,   -1,
   -1,   -1,   -1,   -1,   -1,  420,   -1,   -1,  423,   -1,
   -1,   -1,   -1,   -1,  429,   -1,   -1,  339,   -1,   -1,
   -1,   -1,  344,   -1,  346,  347,  348,  349,  350,  351,
  352,  353,  354,  355,  356,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  367,  368,  369,  370,  371,
  372,  373,  374,  375,  376,  256,  378,  379,   -1,  381,
  382,  383,  384,  385,  386,  387,  388,  389,  390,   -1,
  392,  393,  394,  395,  396,  397,  398,  399,  400,  401,
  402,  403,  404,  405,  406,  407,  408,  409,  410,  411,
  412,  413,   -1,  256,   -1,   -1,   -1,   -1,  420,  262,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  429,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  298,   -1,   -1,  339,   -1,
   -1,   -1,   -1,  344,   -1,  346,  347,  348,  349,  350,
  351,  352,  353,  354,  355,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  368,   -1,  370,
  262,  372,   -1,  374,  375,  376,  339,   -1,   -1,   -1,
   -1,  344,   -1,  346,  347,  348,  349,  350,  351,  352,
  353,  354,  355,  356,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  367,  368,  298,  370,  371,  372,
  373,  374,  375,  376,   -1,  378,  379,   -1,  381,  382,
  383,  384,  385,  386,  387,  388,  389,  390,  429,  392,
  393,  394,  395,  396,  397,  398,  399,  400,  401,  402,
  403,  404,  405,  406,  407,  408,  409,  410,  411,  412,
  413,   -1,  256,   -1,   -1,   -1,   -1,  420,  262,   -1,
  423,   -1,   -1,   -1,   -1,   -1,  429,   -1,   -1,   -1,
   -1,  363,   -1,   -1,   -1,   -1,   -1,  369,   -1,  371,
  372,  373,  374,   -1,  376,   -1,  378,  379,   -1,  381,
  382,  383,  384,  385,  298,  387,  388,  389,  390,   -1,
  392,  393,  394,  395,  396,  397,  398,  399,  400,  401,
  402,  403,  404,  405,  406,  407,  408,  409,  410,  411,
  412,  413,   -1,   -1,   -1,   -1,  418,   -1,  420,   -1,
   -1,  423,   -1,   -1,   -1,  339,   -1,  429,   -1,   -1,
  344,   -1,  346,  347,  348,  349,  350,  351,  352,  353,
  354,  355,  356,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  368,   -1,  370,  371,  372,  373,
  374,  375,  376,   -1,  378,  379,   -1,  381,  382,  383,
  384,  385,  386,  387,  388,  389,  390,   -1,  392,  393,
  394,  395,  396,  397,  398,  399,  400,  401,  402,  403,
  404,  405,  406,  407,  408,  409,  410,  411,  412,  413,
   -1,  256,  256,   -1,   -1,   -1,  420,  262,   -1,  423,
   -1,  265,   -1,  267,   -1,  429,  270,   -1,   -1,   -1,
   -1,  275,   -1,   -1,   -1,  279,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  288,   -1,   -1,   -1,   -1,   -1,
   -1,  295,   -1,  298,   -1,   -1,  300,   -1,   -1,   -1,
  304,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  316,   -1,  318,   -1,   -1,   -1,  322,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  330,  331,   -1,   -1,
  334,   -1,   -1,  337,  339,   -1,   -1,   -1,   -1,  344,
   -1,  346,  347,  348,  349,  350,  351,  352,  353,  354,
  355,  356,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  363,
   -1,   -1,   -1,  368,   -1,  370,  371,  372,  373,  374,
  375,  376,   -1,  378,  379,   -1,  381,  382,  383,  384,
  385,  386,  387,  388,  389,  390,   -1,  392,  393,  394,
  395,  396,  397,  398,  399,  400,  401,  402,  403,  404,
  405,  406,  407,  408,  409,  410,  411,  412,  413,   -1,
  256,  256,   -1,   -1,  418,  420,  262,   -1,  423,   -1,
  265,   -1,  267,   -1,  429,  270,   -1,   -1,   -1,   -1,
  275,   -1,   -1,   -1,  279,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  288,   -1,   -1,   -1,   -1,   -1,   -1,
  295,   -1,  298,   -1,   -1,  300,   -1,   -1,   -1,  304,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  316,   -1,  318,   -1,   -1,   -1,  322,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  330,  331,   -1,   -1,  334,
   -1,   -1,  337,  339,   -1,   -1,   -1,   -1,  344,   -1,
  346,  347,  348,  349,  350,  351,  352,  353,  354,  355,
  356,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  368,   -1,  370,  371,  372,  373,  374,  375,
  376,   -1,  378,  379,   -1,  381,  382,  383,  384,  385,
  386,  387,  388,  389,  390,   -1,  392,  393,  394,  395,
  396,  397,  398,  399,  400,  401,  402,  403,  404,  405,
  406,  407,  408,  409,  410,  411,  412,  413,   -1,  256,
  256,   -1,   -1,  418,  420,  262,   -1,  423,   -1,  265,
   -1,  267,   -1,  429,  270,   -1,   -1,   -1,   -1,  275,
   -1,   -1,   -1,  279,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  288,   -1,   -1,   -1,   -1,   -1,   -1,  295,
   -1,  298,   -1,   -1,  300,   -1,   -1,   -1,  304,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  316,   -1,  318,   -1,   -1,   -1,  322,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  330,  331,   -1,   -1,  334,   -1,
   -1,  337,  339,   -1,   -1,   -1,   -1,  344,   -1,  346,
  347,  348,  349,  350,  351,  352,  353,  354,  355,  356,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  368,   -1,  370,  371,  372,  373,  374,  375,  376,
   -1,  378,   -1,   -1,  381,  382,  383,  384,  385,  386,
  387,  388,  389,  390,   -1,  392,  393,  394,  395,  396,
  397,  398,  399,  400,  401,  402,  403,  404,  405,  406,
  407,  408,  409,  410,  411,  412,  413,   -1,  256,   -1,
  261,   -1,  418,  420,  262,   -1,  423,   -1,   -1,   -1,
   -1,   -1,  429,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  284,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  297,   -1,   -1,   -1,
  298,  302,   -1,   -1,  305,   -1,  307,   -1,  309,  310,
  311,  312,   -1,   -1,   -1,   -1,  317,   -1,   -1,   -1,
  321,  256,   -1,   -1,  325,   -1,   -1,  262,   -1,   -1,
   -1,  266,  333,   -1,   -1,  336,   -1,  338,   -1,   -1,
   -1,  339,   -1,   -1,   -1,   -1,  344,   -1,  346,  347,
  348,  349,  350,  351,  352,  353,  354,  355,  356,   -1,
   -1,  362,   -1,  298,   -1,   -1,   -1,   -1,   -1,   -1,
  368,   -1,  370,   -1,  372,   -1,  374,  375,  376,  314,
  378,  379,   -1,  381,  382,  383,  384,  385,  386,  387,
  388,  389,  390,   -1,   -1,   -1,  394,  395,  396,  397,
  398,  399,  400,  401,  402,  403,  404,  405,  406,  407,
  408,  409,  410,  411,  256,  413,   -1,  418,   -1,   -1,
  262,   -1,  357,   -1,   -1,   -1,   -1,   -1,  363,   -1,
   -1,  429,   -1,   -1,  369,  370,  371,  372,  373,  374,
  375,  376,   -1,  378,  379,   -1,  381,  382,  383,  384,
  385,  386,  387,  388,  389,  390,  298,  392,  393,  394,
  395,  396,  397,  398,  399,  400,  401,  402,  403,  404,
  405,  406,  407,  408,  409,  410,  411,  412,  413,   -1,
   -1,   -1,   -1,  418,   -1,  420,   -1,   -1,  423,   -1,
   -1,   -1,   -1,   -1,  429,   -1,   -1,  339,   -1,   -1,
   -1,   -1,  344,   -1,  346,  347,  348,  349,  350,  351,
  352,  353,  354,  355,  356,   -1,   -1,   -1,  256,   -1,
   -1,   -1,   -1,   -1,  262,   -1,  368,   -1,  370,   -1,
  372,   -1,  374,  375,  376,   -1,  378,  379,   -1,   -1,
  382,  383,  384,  385,  386,  387,  388,  389,  390,   -1,
   -1,   -1,  394,  395,  396,  397,  398,  399,  400,  401,
  298,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  413,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  256,  429,   -1,   -1,
   -1,   -1,  262,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  339,   -1,   -1,   -1,   -1,  344,   -1,  346,  347,
  348,  349,  350,  351,  352,  353,  354,  355,  356,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  298,   -1,
  368,   -1,  370,   -1,  372,   -1,  374,  375,  376,   -1,
  378,  379,   -1,   -1,  382,  383,  384,  385,   -1,   -1,
   -1,  389,  390,   -1,   -1,   -1,  394,  395,  396,  397,
  398,  399,  400,  401,   -1,   -1,   -1,   -1,   -1,  339,
   -1,   -1,   -1,   -1,  344,  413,  346,  347,  348,  349,
  350,  351,  352,  353,  354,  355,  356,   -1,   -1,   -1,
  256,  429,   -1,   -1,   -1,   -1,  262,   -1,  368,   -1,
  370,   -1,  372,   -1,  374,  375,  376,   -1,  378,  379,
   -1,   -1,  382,  383,  384,  385,   -1,   -1,   -1,  389,
  390,   -1,   -1,   -1,  394,  395,  396,  397,  398,  399,
  400,  401,  298,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  413,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  256,  429,
   -1,   -1,   -1,   -1,  262,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  339,   -1,   -1,   -1,   -1,  344,   -1,
  346,  347,  348,  349,  350,  351,  352,  353,  354,  355,
  356,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  298,   -1,  368,   -1,  370,   -1,  372,   -1,  374,  375,
  376,   -1,  378,  379,   -1,   -1,  382,  383,  384,  385,
   -1,   -1,   -1,  389,  390,   -1,  256,   -1,  394,  395,
  396,  397,  398,  399,  400,  401,   -1,   -1,   -1,   -1,
   -1,  339,   -1,   -1,   -1,   -1,  344,  413,  346,  347,
  348,  349,  350,  351,  352,  353,  354,  355,  356,   -1,
   -1,   -1,   -1,  429,   -1,   -1,   -1,   -1,   -1,   -1,
  368,   -1,  370,   -1,  372,   -1,  374,  375,  376,   -1,
  378,  379,   -1,   -1,  382,  383,  384,  385,   -1,   -1,
   -1,  389,  390,   -1,  256,   -1,  394,  395,  396,  397,
  398,  399,  400,  401,   -1,   -1,   -1,   -1,   -1,  339,
   -1,   -1,   -1,   -1,  344,  413,  346,  347,  348,  349,
  350,  351,  352,  353,  354,  355,  356,   -1,   -1,   -1,
   -1,  429,   -1,   -1,   -1,   -1,   -1,   -1,  368,   -1,
  370,   -1,  372,   -1,  374,  375,  376,   -1,   -1,   -1,
   -1,   -1,  382,  383,  384,  385,   -1,   -1,   -1,  389,
  390,   -1,  256,   -1,  394,  395,  396,  397,  398,  399,
  400,  401,   -1,   -1,   -1,   -1,   -1,  339,   -1,   -1,
   -1,   -1,  344,  413,  346,  347,  348,  349,  350,  351,
  352,  353,  354,  355,  356,   -1,   -1,   -1,   -1,  429,
   -1,   -1,   -1,   -1,   -1,   -1,  368,   -1,  370,   -1,
  372,   -1,  374,  375,  376,   -1,   -1,   -1,   -1,   -1,
  382,  383,  384,  385,   -1,   -1,   -1,  389,  390,   -1,
  256,   -1,  394,  395,  396,  397,  398,  399,  400,  401,
   -1,   -1,   -1,   -1,   -1,  339,   -1,   -1,   -1,   -1,
  344,  413,  346,  347,  348,  349,  350,  351,  352,  353,
  354,  355,  356,   -1,   -1,   -1,   -1,  429,   -1,   -1,
   -1,   -1,   -1,   -1,  368,   -1,  370,   -1,  372,   -1,
  374,  375,  376,   -1,   -1,   -1,   -1,   -1,  382,  383,
  384,  385,   -1,   -1,   -1,  389,  390,   -1,  256,   -1,
  394,  395,  396,  397,  398,  399,  400,  401,   -1,   -1,
   -1,   -1,   -1,  339,   -1,   -1,   -1,   -1,  344,  413,
  346,  347,  348,  349,  350,  351,  352,  353,  354,  355,
  356,   -1,   -1,   -1,   -1,  429,   -1,   -1,   -1,   -1,
   -1,   -1,  368,   -1,  370,   -1,  372,   -1,  374,  375,
  376,   -1,   -1,   -1,   -1,   -1,  382,  383,  384,  385,
   -1,   -1,   -1,  389,  390,   -1,  256,   -1,   -1,   -1,
  396,  397,  398,  399,  400,  401,   -1,   -1,   -1,   -1,
   -1,  339,   -1,   -1,   -1,   -1,  344,  413,  346,  347,
  348,  349,  350,  351,  352,  353,  354,  355,  356,   -1,
   -1,   -1,   -1,  429,   -1,   -1,   -1,   -1,   -1,   -1,
  368,   -1,  370,   -1,  372,   -1,  374,  375,  376,   -1,
   -1,   -1,   -1,   -1,  382,  383,  384,  385,   -1,   -1,
   -1,  389,  390,   -1,  256,   -1,   -1,   -1,  396,  397,
  398,  399,  400,  401,   -1,   -1,   -1,   -1,   -1,  339,
   -1,   -1,   -1,   -1,  344,  413,  346,  347,  348,  349,
  350,  351,  352,  353,  354,  355,  356,   -1,   -1,   -1,
   -1,  429,   -1,   -1,   -1,   -1,   -1,   -1,  368,   -1,
  370,   -1,  372,   -1,  374,  375,  376,   -1,   -1,   -1,
   -1,   -1,  382,  383,  384,  385,   -1,   -1,   -1,  389,
  390,   -1,  256,   -1,   -1,   -1,  396,  397,  398,  399,
  400,  401,   -1,   -1,   -1,   -1,   -1,  339,   -1,   -1,
   -1,   -1,  344,  413,  346,  347,  348,  349,  350,  351,
  352,  353,  354,  355,  356,   -1,   -1,   -1,   -1,  429,
   -1,   -1,   -1,   -1,   -1,   -1,  368,   -1,  370,   -1,
  372,   -1,  374,  375,  376,   -1,   -1,   -1,   -1,   -1,
  382,  383,  384,  385,   -1,   -1,   -1,  389,  390,   -1,
  256,   -1,   -1,   -1,  396,  397,  398,  399,  400,  401,
   -1,   -1,   -1,   -1,   -1,  339,   -1,   -1,   -1,   -1,
  344,  413,  346,  347,  348,  349,  350,  351,  352,  353,
  354,  355,  356,   -1,   -1,   -1,   -1,  429,   -1,   -1,
   -1,   -1,   -1,   -1,  368,   -1,  370,   -1,  372,   -1,
  374,  375,  376,   -1,   -1,   -1,   -1,   -1,  382,  383,
  384,  385,   -1,   -1,   -1,  389,  390,   -1,  256,   -1,
   -1,   -1,  396,  397,  398,  399,  400,  401,   -1,   -1,
   -1,   -1,   -1,  339,   -1,   -1,   -1,   -1,  344,  413,
  346,  347,  348,  349,  350,  351,  352,  353,  354,  355,
  356,   -1,   -1,   -1,   -1,  429,   -1,   -1,   -1,   -1,
   -1,   -1,  368,   -1,  370,   -1,  372,   -1,  374,  375,
  376,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  384,  385,
   -1,   -1,   -1,  389,  390,   -1,  256,   -1,   -1,   -1,
   -1,   -1,  398,  399,  400,  401,   -1,   -1,   -1,   -1,
   -1,  339,   -1,   -1,   -1,   -1,  344,  413,  346,  347,
  348,  349,  350,  351,  352,  353,  354,  355,  356,   -1,
   -1,   -1,   -1,  429,   -1,   -1,   -1,   -1,   -1,   -1,
  368,   -1,  370,   -1,  372,   -1,  374,  375,  376,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  384,  385,   -1,   -1,
   -1,  389,  390,   -1,  256,   -1,   -1,   -1,   -1,   -1,
  398,  399,  400,  401,   -1,   -1,   -1,   -1,   -1,  339,
   -1,   -1,   -1,   -1,  344,  413,  346,  347,  348,  349,
  350,  351,  352,  353,  354,  355,  356,   -1,   -1,   -1,
   -1,  429,   -1,   -1,   -1,   -1,   -1,   -1,  368,   -1,
  370,   -1,  372,   -1,  374,  375,  376,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  384,  385,   -1,   -1,   -1,  389,
  390,   -1,  256,   -1,   -1,   -1,   -1,   -1,  398,  399,
  400,  401,   -1,   -1,   -1,   -1,   -1,  339,   -1,   -1,
   -1,   -1,  344,  413,  346,  347,  348,  349,  350,  351,
  352,  353,  354,  355,  356,   -1,   -1,   -1,   -1,  429,
   -1,   -1,   -1,   -1,   -1,   -1,  368,   -1,  370,   -1,
  372,   -1,  374,  375,  376,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  384,  385,   -1,   -1,   -1,  389,  390,   -1,
  256,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  400,  401,
   -1,   -1,   -1,   -1,   -1,  339,   -1,   -1,   -1,   -1,
  344,  413,  346,  347,  348,  349,  350,  351,  352,  353,
  354,  355,  356,   -1,   -1,   -1,   -1,  429,   -1,   -1,
   -1,   -1,   -1,   -1,  368,   -1,  370,   -1,  372,   -1,
  374,  375,  376,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  384,  385,   -1,   -1,   -1,  389,  390,   -1,  256,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  400,  401,   -1,   -1,
   -1,   -1,   -1,  339,   -1,   -1,   -1,   -1,  344,  413,
  346,  347,  348,  349,  350,  351,  352,  353,  354,  355,
  356,   -1,   -1,   -1,   -1,  429,   -1,   -1,   -1,   -1,
   -1,   -1,  368,   -1,  370,   -1,  372,   -1,  374,  375,
  376,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  385,
   -1,   -1,   -1,  389,  390,   -1,  256,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  400,  401,   -1,   -1,   -1,   -1,
   -1,  339,   -1,   -1,   -1,   -1,  344,  413,  346,  347,
  348,  349,  350,  351,  352,  353,  354,  355,  356,   -1,
   -1,   -1,   -1,  429,   -1,   -1,   -1,   -1,   -1,   -1,
  368,   -1,  370,   -1,  372,   -1,  374,  375,  376,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  385,   -1,   -1,
   -1,  389,  390,   -1,  256,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  400,  401,   -1,   -1,   -1,   -1,   -1,  339,
   -1,   -1,   -1,   -1,  344,  413,  346,  347,  348,  349,
  350,  351,  352,  353,  354,  355,  356,   -1,   -1,   -1,
   -1,  429,   -1,   -1,   -1,   -1,   -1,   -1,  368,   -1,
  370,   -1,  372,   -1,  374,  375,  376,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  385,   -1,   -1,   -1,   -1,
  390,   -1,  256,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  400,  401,   -1,   -1,   -1,   -1,   -1,  339,   -1,   -1,
   -1,   -1,  344,  413,  346,  347,  348,  349,  350,  351,
  352,  353,  354,  355,  356,   -1,   -1,   -1,   -1,  429,
   -1,   -1,   -1,   -1,   -1,   -1,  368,   -1,  370,   -1,
  372,   -1,  374,  375,  376,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  385,   -1,   -1,   -1,   -1,  390,   -1,
  256,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  400,  401,
   -1,   -1,   -1,   -1,   -1,  339,   -1,   -1,   -1,   -1,
  344,  413,  346,  347,  348,  349,  350,  351,  352,  353,
  354,  355,  356,   -1,   -1,   -1,   -1,  429,   -1,   -1,
   -1,   -1,   -1,   -1,  368,   -1,  370,   -1,  372,   -1,
  374,  375,  376,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  390,   -1,  256,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  400,  401,   -1,   -1,
   -1,   -1,   -1,  339,   -1,   -1,   -1,   -1,  344,  413,
  346,  347,  348,  349,  350,  351,  352,  353,  354,  355,
  356,   -1,   -1,   -1,   -1,  429,   -1,   -1,   -1,   -1,
   -1,   -1,  368,   -1,  370,   -1,  372,   -1,  374,  375,
  376,   -1,   -1,   -1,  256,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  390,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  400,  401,   -1,   -1,   -1,   -1,
   -1,  339,   -1,   -1,   -1,   -1,  344,  413,  346,  347,
  348,  349,  350,  351,  352,  353,  354,  355,  356,   -1,
   -1,   -1,   -1,  429,   -1,   -1,   -1,   -1,   -1,   -1,
  368,   -1,  370,   -1,  372,   -1,  374,  375,  376,   -1,
   -1,   -1,   -1,  262,   -1,   -1,   -1,  266,   -1,   -1,
   -1,   -1,  390,   -1,   -1,   -1,   -1,  339,   -1,   -1,
   -1,   -1,  344,  401,  346,  347,  348,  349,  350,  351,
  352,  353,  354,  355,  356,  413,   -1,   -1,   -1,  298,
   -1,   -1,   -1,   -1,   -1,   -1,  368,   -1,  370,   -1,
  372,  429,  374,  375,  376,  314,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  390,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  401,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  413,   -1,   -1,   -1,   -1,   -1,   -1,  357,   -1,
   -1,   -1,   -1,   -1,  363,   -1,   -1,  429,   -1,   -1,
  369,   -1,  371,   -1,  373,   -1,  375,  376,   -1,  378,
  379,   -1,  381,  382,  383,  384,  385,  386,  387,  388,
  389,  390,   -1,  392,  393,  394,  395,  396,  397,  398,
  399,  400,  401,  402,  403,  404,  405,  406,  407,  408,
  409,  410,  411,  412,  413,   -1,   -1,  256,   -1,  418,
   -1,  420,   -1,   -1,  423,  264,  265,  266,  267,   -1,
  429,  270,  271,   -1,  273,  274,  275,  276,  277,  278,
  279,   -1,   -1,   -1,   -1,   -1,  285,   -1,  287,  288,
  289,  290,  291,  292,   -1,   -1,  295,   -1,   -1,   -1,
  299,  300,   -1,  302,  303,  304,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  314,   -1,  316,   -1,  318,
  319,   -1,   -1,  322,   -1,  324,  325,  326,  327,  328,
  329,  330,  331,  332,  333,  334,  335,   -1,  337,   -1,
   -1,  340,  341,   -1,   -1,  344,  345,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  359,  360,  361,  362,  363,   -1,   -1,   -1,  367,   -1,
   -1,   -1,  371,   -1,   -1,   -1,   -1,  376,  377,  378,
  379,  380,   -1,   -1,   -1,  384,   -1,  386,   -1,   -1,
   -1,   -1,   -1,  392,  393,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  256,   -1,   -1,   -1,  417,  418,
  419,  420,  264,  265,  266,  267,   -1,   -1,  270,  271,
   -1,  273,  274,  275,  276,  277,  278,  279,   -1,   -1,
   -1,   -1,   -1,  285,   -1,  287,  288,  289,  290,  291,
  292,   -1,   -1,  295,   -1,   -1,   -1,  299,  300,   -1,
  302,  303,  304,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  314,   -1,  316,   -1,  318,  319,   -1,   -1,
  322,   -1,  324,  325,  326,  327,  328,  329,  330,  331,
  332,  333,  334,  335,   -1,  337,   -1,   -1,  340,  341,
   -1,   -1,  344,  345,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  359,  360,  361,
  362,  363,   -1,   -1,   -1,  367,   -1,   -1,   -1,  371,
   -1,   -1,   -1,   -1,  376,  377,  378,  379,  380,   -1,
   -1,   -1,  384,   -1,  386,   -1,   -1,   -1,   -1,   -1,
  392,  393,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  256,   -1,   -1,   -1,  417,  418,  419,  420,  264,
  265,  266,  267,   -1,   -1,  270,  271,   -1,  273,  274,
  275,  276,  277,  278,  279,   -1,   -1,   -1,   -1,   -1,
  285,   -1,  287,  288,  289,  290,  291,  292,   -1,   -1,
  295,   -1,   -1,   -1,  299,  300,   -1,  302,  303,  304,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  314,
   -1,  316,   -1,  318,  319,   -1,   -1,  322,   -1,  324,
  325,  326,  327,  328,  329,  330,  331,  332,  333,  334,
  335,   -1,  337,   -1,   -1,  340,  341,   -1,   -1,  344,
  345,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  359,  360,  361,  362,  363,   -1,
   -1,   -1,  367,   -1,   -1,   -1,  371,   -1,   -1,   -1,
   -1,  376,  377,  378,  379,  380,   -1,   -1,   -1,  384,
   -1,  386,   -1,   -1,   -1,   -1,   -1,  392,  393,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  256,   -1,
   -1,   -1,  417,  418,  419,  420,  264,  265,  266,  267,
   -1,   -1,  270,  271,   -1,  273,  274,  275,  276,  277,
  278,  279,   -1,   -1,   -1,   -1,   -1,  285,   -1,  287,
  288,  289,  290,  291,  292,   -1,   -1,  295,   -1,   -1,
   -1,  299,  300,   -1,  302,  303,  304,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  314,   -1,  316,   -1,
  318,  319,   -1,   -1,  322,   -1,  324,  325,  326,  327,
  328,  329,  330,  331,  332,  333,  334,  335,   -1,  337,
   -1,   -1,  340,  341,   -1,   -1,  344,  345,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  359,  360,  361,  362,  363,   -1,   -1,   -1,  367,
   -1,   -1,   -1,  371,   -1,   -1,   -1,   -1,  376,  377,
  378,  379,  380,   -1,   -1,   -1,  384,   -1,  386,   -1,
   -1,   -1,   -1,   -1,  392,  393,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  256,   -1,   -1,   -1,  417,
  418,  419,  420,  264,  265,  266,  267,   -1,   -1,  270,
  271,   -1,  273,  274,  275,  276,  277,  278,  279,   -1,
   -1,   -1,   -1,   -1,  285,   -1,  287,  288,  289,  290,
  291,  292,   -1,   -1,  295,   -1,   -1,   -1,  299,  300,
   -1,  302,  303,  304,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  314,   -1,  316,   -1,  318,  319,   -1,
   -1,  322,   -1,  324,  325,  326,  327,  328,  329,  330,
  331,  332,  333,  334,  335,   -1,  337,   -1,   -1,  340,
  341,   -1,   -1,  344,  345,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  359,  360,
  361,  362,  363,   -1,   -1,   -1,  367,   -1,   -1,   -1,
  371,   -1,   -1,   -1,   -1,  376,  377,  378,  379,  380,
   -1,   -1,   -1,  384,   -1,  386,   -1,   -1,   -1,   -1,
   -1,  392,  393,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  256,   -1,   -1,   -1,  417,  418,  419,  420,
  264,  265,   -1,  267,   -1,   -1,  270,  271,   -1,  256,
   -1,  275,  276,  277,   -1,  279,   -1,   -1,  265,   -1,
  267,  285,   -1,  270,  288,   -1,   -1,   -1,  275,   -1,
   -1,  295,  279,   -1,   -1,   -1,  300,   -1,  302,  303,
  304,  288,   -1,   -1,   -1,   -1,   -1,   -1,  295,   -1,
   -1,   -1,  316,  300,  318,  319,   -1,  304,  322,   -1,
   -1,  325,   -1,  327,   -1,  329,  330,  331,  332,  316,
  334,  318,   -1,   -1,   -1,  322,   -1,  341,   -1,   -1,
  344,  345,   -1,  330,  331,   -1,   -1,  334,   -1,   -1,
  337,   -1,   -1,   -1,   -1,  359,  360,  361,  362,  363,
   -1,   -1,   -1,  367,  368,   -1,   -1,  371,   -1,   -1,
   -1,   -1,   -1,  377,  378,  379,  380,   -1,   -1,   -1,
  384,   -1,  386,   -1,   -1,   -1,   -1,   -1,  392,  393,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  256,
   -1,   -1,   -1,  417,  418,  419,  420,  264,  265,   -1,
  267,   -1,   -1,  270,  271,   -1,  256,   -1,  275,  276,
  277,  418,  279,   -1,   -1,  265,   -1,  267,  285,   -1,
  270,  288,   -1,   -1,   -1,  275,   -1,   -1,  295,  279,
   -1,   -1,   -1,  300,   -1,  302,  303,  304,  288,   -1,
   -1,   -1,   -1,   -1,   -1,  295,   -1,   -1,   -1,  316,
  300,  318,  319,  320,  304,  322,   -1,   -1,  325,   -1,
  327,   -1,  329,  330,  331,  332,  316,  334,  318,   -1,
   -1,   -1,  322,   -1,  341,   -1,   -1,  344,  345,   -1,
  330,  331,   -1,   -1,  334,   -1,   -1,  337,   -1,   -1,
   -1,   -1,  359,  360,  361,  362,  363,   -1,   -1,   -1,
  367,   -1,   -1,   -1,  371,   -1,   -1,   -1,   -1,   -1,
  377,  378,  379,  380,   -1,   -1,   -1,  384,   -1,  386,
   -1,   -1,   -1,   -1,   -1,  392,  393,   -1,   -1,   -1,
   -1,   -1,   -1,  264,  265,   -1,  267,   -1,   -1,  270,
  271,   -1,   -1,   -1,  275,  276,  277,   -1,  279,   -1,
  417,  418,  419,  420,  285,   -1,   -1,  288,   -1,   -1,
   -1,   -1,   -1,   -1,  295,   -1,   -1,   -1,  418,  300,
   -1,  302,  303,  304,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  316,   -1,  318,  319,   -1,
   -1,  322,   -1,   -1,  325,   -1,  327,   -1,  329,  330,
  331,  332,   -1,  334,   -1,   -1,   -1,   -1,   -1,   -1,
  341,   -1,   -1,  344,  345,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  359,  360,
  361,  362,  363,   -1,   -1,   -1,  367,  368,   -1,   -1,
  371,   -1,   -1,   -1,   -1,   -1,  377,  378,  379,  380,
   -1,   -1,   -1,  384,   -1,  386,   -1,   -1,   -1,   -1,
   -1,  392,  393,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  256,   -1,  417,  418,  419,  420,
   -1,   -1,  264,  265,   -1,  267,   -1,  428,  270,  271,
   -1,   -1,   -1,  275,  276,  277,   -1,  279,   -1,   -1,
  265,   -1,  267,  285,   -1,  270,  288,   -1,   -1,   -1,
  275,   -1,   -1,  295,  279,   -1,   -1,   -1,  300,   -1,
  302,  303,  304,  288,   -1,   -1,   -1,   -1,   -1,   -1,
  295,   -1,   -1,   -1,  316,  300,  318,  319,  320,  304,
  322,   -1,   -1,  325,   -1,  327,   -1,  329,  330,  331,
  332,  316,  334,  318,   -1,   -1,   -1,  322,   -1,  341,
   -1,   -1,  344,  345,   -1,  330,  331,   -1,   -1,  334,
   -1,   -1,  337,   -1,   -1,   -1,   -1,  359,  360,  361,
  362,  363,   -1,   -1,   -1,  367,   -1,   -1,   -1,  371,
   -1,   -1,   -1,   -1,   -1,  377,  378,  379,  380,   -1,
   -1,   -1,  384,   -1,  386,   -1,   -1,  372,   -1,   -1,
  392,  393,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  256,   -1,   -1,   -1,  417,  418,  419,  420,  264,
  265,   -1,  267,   -1,   -1,  270,  271,   -1,   -1,   -1,
  275,  276,  277,  418,  279,   -1,   -1,  265,   -1,  267,
  285,   -1,  270,  288,   -1,   -1,   -1,  275,   -1,   -1,
  295,  279,   -1,   -1,   -1,  300,   -1,  302,  303,  304,
  288,   -1,   -1,   -1,   -1,   -1,   -1,  295,   -1,   -1,
   -1,  316,  300,  318,  319,   -1,  304,  322,   -1,   -1,
  325,   -1,  327,   -1,  329,  330,  331,  332,  316,  334,
  318,   -1,  337,   -1,  322,   -1,  341,   -1,   -1,  344,
  345,   -1,  330,  331,   -1,   -1,  334,   -1,   -1,  337,
   -1,   -1,   -1,   -1,  359,  360,  361,  362,  363,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  371,   -1,   -1,   -1,
   -1,   -1,  377,  378,  379,  380,   -1,   -1,   -1,  384,
   -1,  386,  370,   -1,   -1,   -1,   -1,  392,  393,   -1,
   -1,   -1,   -1,   -1,   -1,  264,  265,   -1,  267,   -1,
   -1,  270,  271,   -1,   -1,   -1,  275,  276,  277,   -1,
  279,   -1,  417,  418,  419,  420,  285,   -1,   -1,  288,
   -1,   -1,   -1,   -1,   -1,   -1,  295,   -1,   -1,   -1,
  418,  300,   -1,  302,  303,  304,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  316,   -1,  318,
  319,   -1,   -1,  322,   -1,   -1,  325,   -1,  327,   -1,
  329,  330,  331,  332,   -1,  334,   -1,   -1,   -1,   -1,
   -1,   -1,  341,   -1,   -1,  344,  345,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  359,  360,  361,  362,  363,   -1,   -1,   -1,  367,   -1,
   -1,   -1,  371,   -1,   -1,   -1,   -1,   -1,  377,  378,
  379,  380,   -1,   -1,   -1,  384,   -1,  386,   -1,   -1,
   -1,   -1,   -1,  392,  393,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  256,   -1,  417,  418,
  419,  420,   -1,   -1,  264,  265,   -1,  267,   -1,  428,
  270,  271,   -1,   -1,   -1,  275,  276,  277,   -1,  279,
   -1,   -1,  265,   -1,  267,  285,   -1,  270,  288,   -1,
   -1,   -1,  275,   -1,   -1,  295,  279,   -1,   -1,   -1,
  300,   -1,  302,  303,  304,  288,   -1,   -1,   -1,   -1,
   -1,   -1,  295,   -1,   -1,   -1,  316,  300,  318,  319,
   -1,  304,  322,   -1,   -1,  325,   -1,  327,   -1,  329,
  330,  331,  332,  316,  334,  318,   -1,   -1,   -1,  322,
   -1,  341,   -1,   -1,  344,  345,   -1,  330,  331,   -1,
   -1,  334,   -1,   -1,  337,   -1,   -1,   -1,   -1,  359,
  360,  361,  362,  363,   -1,   -1,   -1,  367,   -1,   -1,
   -1,  371,   -1,   -1,   -1,   -1,   -1,  377,  378,  379,
  380,   -1,   -1,   -1,  384,   -1,  386,   -1,   -1,   -1,
   -1,   -1,  392,  393,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  256,   -1,   -1,   -1,  417,  418,  419,
  420,  264,  265,   -1,  267,   -1,   -1,  270,  271,   -1,
   -1,   -1,  275,  276,  277,  418,  279,   -1,   -1,  265,
   -1,  267,  285,   -1,  270,  288,   -1,   -1,   -1,  275,
   -1,   -1,  295,  279,   -1,   -1,   -1,  300,   -1,  302,
  303,  304,  288,   -1,   -1,   -1,   -1,   -1,   -1,  295,
   -1,   -1,   -1,  316,  300,  318,  319,   -1,  304,  322,
   -1,   -1,  325,   -1,  327,   -1,  329,  330,  331,  332,
  316,  334,  318,   -1,   -1,   -1,  322,   -1,  341,   -1,
   -1,  344,  345,   -1,  330,  331,   -1,   -1,  334,   -1,
   -1,  337,   -1,   -1,   -1,   -1,  359,  360,  361,  362,
  363,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  371,   -1,
   -1,   -1,   -1,   -1,  377,  378,  379,  380,   -1,   -1,
   -1,  384,   -1,  386,   -1,   -1,   -1,   -1,   -1,  392,
  393,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  256,   -1,   -1,   -1,  417,  418,  419,  420,  264,  265,
   -1,  267,   -1,   -1,  270,  271,   -1,   -1,   -1,  275,
  276,  277,  418,  279,   -1,   -1,  265,   -1,  267,  285,
   -1,  270,  288,   -1,   -1,   -1,  275,   -1,   -1,  295,
  279,   -1,   -1,   -1,  300,   -1,  302,  303,  304,  288,
   -1,   -1,   -1,   -1,   -1,   -1,  295,   -1,   -1,   -1,
  316,  300,  318,  319,   -1,  304,  322,   -1,   -1,  325,
   -1,  327,   -1,  329,  330,  331,  332,  316,  334,  318,
   -1,   -1,   -1,  322,   -1,  341,   -1,   -1,  344,  345,
   -1,  330,  331,   -1,   -1,  334,   -1,   -1,  337,   -1,
   -1,   -1,   -1,  359,  360,  361,  362,  363,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  371,   -1,   -1,   -1,   -1,
   -1,  377,  378,  379,  380,   -1,   -1,   -1,  384,   -1,
  386,   -1,   -1,   -1,   -1,   -1,  392,  393,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  256,   -1,   -1,
   -1,  417,  418,  419,  420,  264,  265,   -1,  267,   -1,
   -1,  270,  271,   -1,   -1,   -1,  275,  276,  277,  418,
  279,   -1,   -1,  265,   -1,  267,  285,   -1,  270,  288,
   -1,   -1,   -1,  275,   -1,   -1,  295,  279,   -1,   -1,
   -1,  300,   -1,  302,  303,  304,  288,   -1,   -1,   -1,
   -1,   -1,   -1,  295,   -1,   -1,   -1,  316,  300,  318,
  319,   -1,  304,  322,   -1,   -1,  325,   -1,  327,   -1,
  329,  330,  331,  332,  316,  334,  318,   -1,   -1,   -1,
  322,   -1,  341,   -1,   -1,  344,  345,   -1,  330,  331,
   -1,   -1,  334,   -1,   -1,  337,   -1,   -1,   -1,   -1,
  359,  360,  361,  362,  363,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  371,   -1,   -1,   -1,   -1,   -1,  377,  378,
  379,  380,   -1,   -1,   -1,  384,   -1,  386,   -1,   -1,
   -1,   -1,   -1,  392,  393,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  256,   -1,   -1,   -1,  417,  418,
  419,  420,  264,  265,   -1,  267,   -1,   -1,  270,  271,
   -1,   -1,   -1,  275,  276,  277,  418,  279,   -1,   -1,
   -1,   -1,   -1,  285,   -1,   -1,  288,   -1,   -1,   -1,
   -1,   -1,   -1,  295,   -1,  261,   -1,  262,  300,   -1,
  302,  303,  304,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  316,   -1,  318,  319,  284,   -1,
  322,   -1,   -1,  325,   -1,  327,   -1,  329,  330,  331,
  332,  297,  334,  298,   -1,   -1,  302,   -1,   -1,   -1,
   -1,  307,   -1,  309,  310,  311,  312,   -1,   -1,   -1,
   -1,  317,   -1,   -1,   -1,  321,   -1,  359,  360,  361,
  362,   -1,   -1,   -1,   -1,   -1,   -1,  333,   -1,  371,
  336,   -1,  338,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  362,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  371,  372,  373,  374,
  375,   -1,   -1,  378,  379,  417,  418,  382,  383,  384,
  385,  386,  387,  388,  389,  390,   -1,  392,  393,  394,
  395,  396,  397,  398,  399,  400,  401,  402,  403,  404,
  405,  406,  407,  408,  409,  410,  411,  412,  413,   -1,
  261,   -1,  418,   -1,  265,  420,  267,   -1,  423,  270,
   -1,  272,  273,   -1,  275,   -1,  277,   -1,  279,   -1,
  281,  282,  283,  284,   -1,   -1,  287,  288,   -1,   -1,
   -1,   -1,  293,   -1,  295,  296,  297,   -1,   -1,  300,
   -1,  302,   -1,  304,   -1,   -1,  307,   -1,  309,  310,
  311,  312,   -1,   -1,   -1,  316,  317,  318,   -1,   -1,
  321,  322,  323,   -1,   -1,   -1,   -1,   -1,   -1,  330,
  331,   -1,  333,  334,   -1,  336,  337,  338,   -1,   -1,
   -1,  342,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  261,
   -1,  362,   -1,  265,   -1,  267,   -1,  368,  270,   -1,
  272,  273,   -1,  275,   -1,  277,  377,  279,   -1,  281,
  282,  283,  284,   -1,   -1,  287,  288,   -1,   -1,   -1,
   -1,  293,   -1,  295,  296,  297,   -1,   -1,  300,   -1,
  302,   -1,  304,   -1,   -1,  307,   -1,  309,  310,  311,
  312,   -1,   -1,   -1,  316,  317,  318,  418,   -1,  321,
  322,  323,   -1,   -1,   -1,   -1,   -1,   -1,  330,  331,
   -1,  333,  334,   -1,  336,  337,  338,   -1,   -1,   -1,
  342,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  261,   -1,
  362,   -1,  265,   -1,  267,   -1,  368,  270,   -1,  272,
  273,   -1,  275,   -1,  277,  377,  279,   -1,  281,  282,
  283,  284,   -1,   -1,  287,  288,   -1,   -1,   -1,   -1,
  293,   -1,  295,  296,  297,   -1,   -1,  300,   -1,  302,
   -1,  304,   -1,   -1,  307,   -1,  309,  310,  311,  312,
   -1,   -1,   -1,  316,  317,  318,  418,   -1,  321,  322,
  323,   -1,   -1,   -1,   -1,   -1,   -1,  330,  331,   -1,
  333,  334,   -1,  336,  337,  338,   -1,   -1,   -1,  342,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  261,   -1,   -1,   -1,  265,   -1,  267,  362,
   -1,  270,   -1,  272,  273,  368,  275,   -1,  277,   -1,
  279,   -1,  281,  282,  283,  284,   -1,   -1,  287,  288,
   -1,   -1,   -1,   -1,  293,   -1,  295,  296,  297,   -1,
   -1,  300,   -1,  302,  261,  304,   -1,   -1,  307,   -1,
  309,  310,  311,  312,   -1,   -1,   -1,  316,  317,  318,
   -1,   -1,  321,  322,  323,  418,   -1,  284,   -1,   -1,
   -1,  330,  331,   -1,  333,  334,  261,  336,  337,  338,
  297,   -1,   -1,  342,   -1,  302,   -1,   -1,  305,   -1,
  307,   -1,  309,  310,  311,  312,   -1,   -1,   -1,  284,
  317,   -1,   -1,  362,  321,   -1,   -1,   -1,  325,  368,
   -1,   -1,  297,   -1,  261,   -1,  333,  302,   -1,  336,
   -1,  338,  307,   -1,  309,  310,  311,  312,   -1,   -1,
  315,   -1,  317,   -1,   -1,   -1,  321,  284,   -1,   -1,
  357,   -1,   -1,   -1,  261,  362,   -1,   -1,  333,   -1,
  297,  336,  369,  338,  371,  302,  373,   -1,  305,  418,
  307,   -1,  309,  310,  311,  312,   -1,  284,   -1,  386,
  317,   -1,   -1,   -1,  321,   -1,   -1,  362,  325,   -1,
  297,   -1,   -1,  368,  369,  302,  333,   -1,   -1,  336,
  307,  338,  309,  310,  311,  312,   -1,   -1,   -1,   -1,
  317,  418,   -1,   -1,  321,   -1,   -1,   -1,  325,   -1,
   -1,  264,  265,   -1,  267,  362,  333,  270,  271,  336,
   -1,  338,  275,  276,  277,   -1,  279,   -1,   -1,   -1,
   -1,   -1,  285,   -1,   -1,  288,   -1,   -1,   -1,   -1,
   -1,   -1,  295,   -1,   -1,  362,   -1,  300,   -1,  302,
  303,  304,   -1,  306,   -1,   -1,   -1,   -1,   -1,   -1,
  313,   -1,   -1,  316,   -1,  318,  319,   -1,   -1,  322,
   -1,  418,  325,   -1,  327,   -1,  329,  330,  331,  332,
   -1,  334,   -1,   -1,   -1,   -1,   -1,   -1,  341,   -1,
   -1,  344,  345,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  418,   -1,   -1,   -1,   -1,  359,  360,  361,  362,
  363,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  371,  372,
   -1,  374,   -1,   -1,  377,  378,  379,  380,   -1,   -1,
   -1,  384,   -1,  386,   -1,   -1,   -1,   -1,   -1,  392,
  393,   -1,   -1,   -1,   -1,   -1,   -1,  264,  265,   -1,
  267,   -1,   -1,  270,  271,   -1,   -1,   -1,  275,  276,
  277,   -1,  279,   -1,  417,  418,  419,  420,  285,   -1,
   -1,  288,   -1,   -1,   -1,   -1,   -1,   -1,  295,   -1,
   -1,   -1,   -1,  300,   -1,  302,  303,  304,   -1,  306,
   -1,   -1,   -1,   -1,   -1,   -1,  313,   -1,   -1,  316,
   -1,  318,  319,   -1,   -1,  322,   -1,   -1,  325,   -1,
  327,   -1,  329,  330,  331,  332,   -1,  334,   -1,   -1,
   -1,   -1,   -1,   -1,  341,   -1,   -1,  344,  345,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  359,  360,  361,  362,  363,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  371,   -1,   -1,  374,   -1,   -1,
  377,  378,  379,  380,   -1,   -1,   -1,  384,   -1,  386,
   -1,   -1,   -1,   -1,   -1,  392,  393,   -1,   -1,   -1,
   -1,   -1,   -1,  264,  265,   -1,  267,   -1,   -1,  270,
  271,   -1,   -1,   -1,  275,  276,  277,   -1,  279,   -1,
  417,  418,  419,  420,  285,   -1,   -1,  288,   -1,   -1,
   -1,   -1,   -1,   -1,  295,   -1,   -1,   -1,   -1,  300,
   -1,  302,  303,  304,   -1,  306,   -1,   -1,   -1,   -1,
   -1,   -1,  313,   -1,   -1,  316,   -1,  318,  319,   -1,
   -1,  322,   -1,   -1,  325,   -1,  327,   -1,  329,  330,
  331,  332,   -1,  334,   -1,   -1,   -1,   -1,   -1,   -1,
  341,   -1,   -1,  344,  345,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  359,  360,
  361,  362,  363,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  371,   -1,   -1,   -1,   -1,   -1,  377,  378,  379,  380,
   -1,   -1,   -1,  384,   -1,  386,   -1,   -1,   -1,   -1,
   -1,  392,  393,   -1,   -1,   -1,   -1,   -1,   -1,  264,
  265,   -1,  267,   -1,   -1,  270,  271,   -1,   -1,   -1,
  275,  276,  277,   -1,  279,   -1,  417,  418,  419,  420,
  285,   -1,   -1,  288,   -1,   -1,   -1,   -1,   -1,   -1,
  295,   -1,   -1,   -1,   -1,  300,   -1,  302,  303,  304,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  316,   -1,  318,  319,   -1,   -1,  322,   -1,   -1,
  325,   -1,  327,   -1,  329,  330,  331,  332,   -1,  334,
   -1,   -1,  337,   -1,   -1,   -1,  341,   -1,   -1,  344,
  345,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  359,  360,  361,  362,  363,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  371,   -1,   -1,   -1,
   -1,   -1,  377,  378,  379,  380,   -1,   -1,   -1,  384,
   -1,  386,   -1,   -1,   -1,   -1,   -1,  392,  393,   -1,
   -1,   -1,   -1,   -1,   -1,  264,  265,   -1,  267,   -1,
   -1,  270,  271,   -1,   -1,   -1,  275,  276,  277,   -1,
  279,   -1,  417,  418,  419,  420,  285,   -1,   -1,  288,
   -1,   -1,   -1,   -1,   -1,   -1,  295,   -1,   -1,   -1,
   -1,  300,   -1,  302,  303,  304,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  316,   -1,  318,
  319,   -1,   -1,  322,   -1,   -1,  325,   -1,  327,   -1,
  329,  330,  331,  332,   -1,  334,   -1,   -1,   -1,   -1,
   -1,   -1,  341,   -1,   -1,  344,  345,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  359,  360,  361,  362,  363,   -1,   -1,   -1,   -1,  368,
   -1,   -1,  371,   -1,   -1,   -1,   -1,   -1,  377,  378,
  379,  380,   -1,   -1,   -1,  384,   -1,  386,   -1,   -1,
   -1,   -1,   -1,  392,  393,   -1,   -1,   -1,   -1,   -1,
   -1,  264,  265,   -1,  267,   -1,   -1,  270,  271,   -1,
   -1,   -1,  275,  276,  277,   -1,  279,   -1,  417,  418,
  419,  420,  285,   -1,   -1,  288,   -1,   -1,   -1,   -1,
   -1,   -1,  295,   -1,   -1,   -1,   -1,  300,   -1,  302,
  303,  304,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  316,   -1,  318,  319,   -1,   -1,  322,
   -1,   -1,  325,   -1,  327,   -1,  329,  330,  331,  332,
   -1,  334,   -1,   -1,   -1,   -1,   -1,   -1,  341,   -1,
   -1,  344,  345,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  359,  360,  361,  362,
  363,   -1,   -1,   -1,  367,   -1,   -1,   -1,  371,   -1,
   -1,   -1,   -1,   -1,  377,  378,  379,  380,   -1,   -1,
   -1,  384,   -1,  386,   -1,   -1,   -1,   -1,   -1,  392,
  393,   -1,   -1,   -1,   -1,   -1,   -1,  264,  265,   -1,
  267,   -1,   -1,  270,  271,   -1,   -1,   -1,  275,  276,
  277,   -1,  279,   -1,  417,  418,  419,  420,  285,   -1,
   -1,  288,   -1,   -1,   -1,   -1,   -1,   -1,  295,   -1,
   -1,   -1,   -1,  300,   -1,  302,  303,  304,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  316,
   -1,  318,  319,   -1,   -1,  322,   -1,   -1,  325,   -1,
  327,   -1,  329,  330,  331,  332,   -1,  334,   -1,   -1,
   -1,   -1,   -1,   -1,  341,   -1,   -1,  344,  345,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  359,  360,  361,  362,  363,   -1,   -1,   -1,
  367,   -1,   -1,   -1,  371,   -1,   -1,   -1,   -1,   -1,
  377,  378,  379,  380,   -1,   -1,   -1,  384,   -1,  386,
   -1,   -1,   -1,   -1,   -1,  392,  393,   -1,   -1,   -1,
   -1,   -1,   -1,  264,  265,   -1,  267,   -1,   -1,  270,
  271,   -1,   -1,   -1,  275,  276,  277,   -1,  279,   -1,
  417,  418,  419,  420,  285,   -1,   -1,  288,   -1,   -1,
   -1,   -1,   -1,   -1,  295,   -1,   -1,   -1,   -1,  300,
   -1,  302,  303,  304,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  316,   -1,  318,  319,   -1,
   -1,  322,   -1,   -1,  325,   -1,  327,   -1,  329,  330,
  331,  332,   -1,  334,   -1,   -1,   -1,   -1,   -1,   -1,
  341,   -1,   -1,  344,  345,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  359,  360,
  361,  362,  363,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  371,   -1,   -1,   -1,   -1,   -1,  377,  378,  379,  380,
   -1,   -1,   -1,  384,   -1,  386,   -1,   -1,   -1,   -1,
   -1,  392,  393,   -1,   -1,   -1,   -1,   -1,   -1,  264,
  265,   -1,  267,   -1,   -1,  270,  271,   -1,   -1,   -1,
  275,  276,  277,   -1,  279,   -1,  417,  418,  419,  420,
  285,   -1,   -1,  288,   -1,   -1,   -1,   -1,   -1,   -1,
  295,   -1,   -1,   -1,   -1,  300,   -1,  302,  303,  304,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,  316,   -1,  318,  319,   -1,   -1,  322,   -1,   -1,
  325,   -1,  327,   -1,  329,  330,  331,  332,   -1,  334,
   -1,   -1,   -1,   -1,   -1,   -1,  341,   -1,   -1,  344,
  345,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  359,  360,  361,  362,  363,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  371,   -1,   -1,   -1,
   -1,   -1,  377,  378,  379,  380,   -1,   -1,   -1,  384,
   -1,  386,   -1,   -1,   -1,   -1,   -1,  392,  393,   -1,
   -1,   -1,   -1,   -1,   -1,  264,  265,   -1,  267,   -1,
   -1,  270,  271,   -1,   -1,   -1,  275,  276,  277,   -1,
  279,   -1,  417,  418,  419,  420,  285,   -1,   -1,  288,
   -1,   -1,   -1,   -1,   -1,   -1,  295,   -1,   -1,   -1,
   -1,  300,   -1,  302,  303,  304,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  316,   -1,  318,
  319,   -1,   -1,  322,   -1,   -1,  325,   -1,  327,   -1,
  329,  330,  331,  332,   -1,  334,   -1,   -1,   -1,   -1,
   -1,   -1,  341,   -1,   -1,  344,  345,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
  359,  360,  361,  362,  363,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  371,   -1,   -1,   -1,   -1,   -1,  377,  378,
  379,  380,   -1,   -1,   -1,  384,   -1,  386,   -1,   -1,
   -1,   -1,   -1,  392,  393,   -1,   -1,   -1,   -1,   -1,
   -1,  264,  265,   -1,  267,   -1,   -1,  270,  271,   -1,
   -1,   -1,  275,  276,  277,   -1,  279,   -1,  417,  418,
  419,  420,  285,   -1,   -1,  288,   -1,   -1,   -1,   -1,
   -1,   -1,  295,   -1,   -1,   -1,   -1,  300,   -1,  302,
  303,  304,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  316,   -1,  318,  319,   -1,   -1,  322,
   -1,   -1,  325,   -1,  327,   -1,  329,  330,  331,  332,
   -1,  334,   -1,   -1,   -1,   -1,   -1,   -1,  341,   -1,
   -1,  344,  345,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,  359,  360,  361,  362,
  363,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  371,   -1,
   -1,   -1,   -1,   -1,  377,  378,  379,  380,   -1,   -1,
   -1,  384,   -1,  386,   -1,   -1,   -1,   -1,   -1,  392,
  393,   -1,   -1,   -1,   -1,   -1,   -1,  264,  265,   -1,
  267,   -1,   -1,  270,  271,   -1,   -1,   -1,  275,  276,
  277,   -1,  279,   -1,  417,  418,  419,  420,  285,   -1,
   -1,  288,   -1,   -1,   -1,   -1,   -1,   -1,  295,   -1,
   -1,   -1,   -1,  300,   -1,  302,  303,  304,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  316,
   -1,  318,  319,   -1,   -1,  322,   -1,   -1,  325,   -1,
  327,   -1,  329,  330,  331,  332,   -1,  334,   -1,   -1,
   -1,   -1,   -1,   -1,  341,   -1,   -1,  344,  345,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,  359,  360,  361,  362,  363,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  371,   -1,   -1,   -1,   -1,   -1,
  377,  378,  379,  380,   -1,   -1,   -1,  384,   -1,  386,
   -1,   -1,   -1,   -1,   -1,  392,  393,   -1,   -1,   -1,
   -1,   -1,   -1,  264,  265,   -1,  267,   -1,   -1,  270,
  271,   -1,   -1,   -1,  275,  276,  277,   -1,  279,   -1,
  417,  418,  419,  420,  285,   -1,   -1,  288,   -1,   -1,
   -1,   -1,   -1,   -1,  295,   -1,  261,   -1,  263,  300,
   -1,  302,  303,  304,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  316,   -1,  318,  319,  284,
   -1,  322,   -1,   -1,  325,   -1,  327,   -1,  329,  330,
  331,  332,  297,  334,   -1,   -1,   -1,  302,   -1,   -1,
   -1,   -1,  307,   -1,  309,  310,  311,  312,   -1,   -1,
   -1,   -1,  317,   -1,   -1,   -1,  321,   -1,  359,  360,
  361,  362,  363,   -1,   -1,   -1,   -1,   -1,  333,   -1,
  371,  336,   -1,  338,   -1,   -1,  377,  378,  379,  380,
   -1,   -1,   -1,  384,   -1,  386,   -1,   -1,   -1,   -1,
   -1,  392,  393,   -1,   -1,   -1,   -1,  362,   -1,   -1,
   -1,   -1,   -1,  368,  369,   -1,   -1,   -1,   -1,   -1,
   -1,  263,   -1,  265,   -1,  267,  417,  418,  270,  420,
  272,  273,   -1,  275,   -1,  277,   -1,  279,   -1,  281,
  282,  283,   -1,   -1,   -1,  287,  288,   -1,   -1,   -1,
   -1,  293,   -1,  295,  296,   -1,   -1,   -1,  300,   -1,
   -1,   -1,  304,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,  315,  316,   -1,  318,   -1,   -1,   -1,
  322,  323,   -1,   -1,   -1,   -1,   -1,   -1,  330,  331,
  264,  265,  334,  267,   -1,  337,  270,  271,   -1,   -1,
  342,  275,  276,  277,   -1,  279,   -1,   -1,   -1,   -1,
   -1,  285,   -1,   -1,  288,   -1,   -1,   -1,   -1,   -1,
   -1,  295,  364,  365,   -1,   -1,  300,   -1,  302,  303,
  304,   -1,   -1,   -1,   -1,  377,   -1,   -1,   -1,   -1,
   -1,   -1,  316,   -1,  318,  319,   -1,   -1,  322,   -1,
   -1,  325,   -1,  327,   -1,  329,  330,  331,  332,   -1,
  334,   -1,   -1,  337,   -1,   -1,   -1,   -1,   -1,   -1,
  265,   -1,  267,   -1,   -1,  270,  418,   -1,   -1,   -1,
  275,   -1,   -1,   -1,  279,  359,  360,  361,  362,   -1,
   -1,   -1,   -1,  288,  265,   -1,  267,  371,   -1,  270,
  295,  272,  273,   -1,  275,  300,  277,   -1,  279,  304,
  281,  282,  283,   -1,   -1,   -1,  287,  288,   -1,   -1,
   -1,  316,  293,  318,  295,  296,   -1,  322,   -1,  300,
   -1,   -1,   -1,  304,   -1,  330,  331,   -1,   -1,  334,
   -1,   -1,  337,  417,  418,  316,   -1,  318,   -1,   -1,
   -1,  322,  323,   -1,   -1,   -1,   -1,   -1,   -1,  330,
  331,   -1,  265,  334,  267,   -1,  337,  270,  363,  272,
  273,  342,  275,   -1,  277,   -1,  279,   -1,  281,  282,
  283,   -1,   -1,   -1,  287,  288,   -1,   -1,   -1,   -1,
  293,   -1,  295,  296,   -1,   -1,   -1,  300,   -1,   -1,
   -1,  304,   -1,   -1,   -1,   -1,  377,   -1,   -1,   -1,
   -1,   -1,   -1,  316,   -1,  318,   -1,   -1,   -1,  322,
  323,   -1,   -1,  418,   -1,   -1,   -1,  330,  331,   -1,
   -1,  334,   -1,   -1,  337,   -1,  265,   -1,  267,  342,
   -1,  270,   -1,   -1,  273,   -1,  275,  418,  277,   -1,
  279,   -1,  281,  282,  283,   -1,   -1,   -1,  287,  288,
   -1,   -1,   -1,   -1,  293,   -1,  295,   -1,  265,   -1,
  267,  300,   -1,  270,   -1,  304,  273,   -1,  275,   -1,
  277,   -1,  279,   -1,  281,  282,  283,  316,   -1,  318,
  287,  288,   -1,  322,   -1,   -1,  293,   -1,  295,   -1,
   -1,  330,  331,  300,   -1,  334,   -1,  304,  337,   -1,
   -1,   -1,  265,  342,  267,  418,   -1,  270,   -1,  316,
   -1,  318,  275,   -1,   -1,  322,  279,   -1,   -1,   -1,
   -1,   -1,   -1,  330,  331,  288,   -1,  334,   -1,   -1,
  337,   -1,  295,   -1,  265,  342,  267,  300,  377,  270,
   -1,  304,   -1,  306,  275,  308,   -1,   -1,  279,   -1,
  313,   -1,   -1,  316,   -1,  318,   -1,  288,   -1,  322,
   -1,   -1,  325,   -1,  295,   -1,   -1,  330,  331,  300,
   -1,  334,   -1,  304,  337,  306,   -1,  308,  265,  418,
  267,   -1,  313,  270,   -1,  316,   -1,  318,  275,   -1,
   -1,  322,  279,   -1,  325,   -1,   -1,   -1,   -1,  330,
  331,  288,   -1,  334,   -1,   -1,  337,   -1,  295,  372,
  265,  418,  267,  300,   -1,  270,   -1,  304,   -1,  306,
  275,  308,   -1,   -1,  279,   -1,  313,   -1,   -1,  316,
   -1,  318,   -1,  288,   -1,  322,   -1,   -1,  325,  370,
  295,   -1,   -1,  330,  331,  300,   -1,  334,   -1,  304,
  337,  306,   -1,   -1,   -1,  418,   -1,   -1,  313,   -1,
   -1,  316,   -1,  318,   -1,   -1,   -1,  322,   -1,   -1,
  325,   -1,   -1,   -1,   -1,  330,  331,   -1,   -1,  334,
   -1,  265,  337,  267,   -1,   -1,  270,  418,   -1,   -1,
   -1,  275,   -1,   -1,   -1,  279,   -1,   -1,   -1,  283,
  265,   -1,  267,   -1,  288,  270,   -1,   -1,   -1,  293,
  275,  295,   -1,   -1,  279,   -1,  300,   -1,   -1,   -1,
  304,  305,   -1,  288,   -1,   -1,   -1,   -1,   -1,   -1,
  295,  418,  316,   -1,  318,  300,   -1,   -1,  322,  304,
   -1,   -1,   -1,   -1,   -1,   -1,  330,  331,   -1,   -1,
  334,  316,   -1,  318,  265,   -1,  267,  322,   -1,  270,
   -1,   -1,   -1,  418,  275,  330,  331,   -1,  279,  334,
   -1,  265,  337,  267,   -1,   -1,  270,  288,   -1,   -1,
   -1,  275,   -1,   -1,  295,  279,   -1,   -1,   -1,  300,
   -1,   -1,   -1,  304,  288,   -1,   -1,   -1,   -1,   -1,
   -1,  295,   -1,   -1,   -1,  316,  300,  318,   -1,   -1,
  304,  322,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  330,
  331,   -1,  316,  334,  318,  265,  337,  267,  322,   -1,
  270,   -1,   -1,   -1,  418,  275,  330,  331,   -1,  279,
  334,   -1,   -1,  337,   -1,   -1,   -1,   -1,  288,   -1,
   -1,   -1,   -1,  418,   -1,  295,   -1,   -1,   -1,   -1,
  300,   -1,   -1,  261,  304,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,  272,   -1,  316,   -1,  318,  277,
   -1,   -1,  322,  281,   -1,   -1,  284,   -1,   -1,   -1,
  330,  331,   -1,   -1,  334,   -1,   -1,  337,  296,  297,
   -1,   -1,   -1,  301,  302,   -1,   -1,  418,   -1,  307,
   -1,  309,  310,  311,  312,   -1,   -1,   -1,   -1,  317,
   -1,   -1,   -1,  321,  418,  323,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,  333,   -1,  335,  336,  261,
  338,   -1,   -1,   -1,  342,   -1,   -1,   -1,   -1,   -1,
  272,   -1,   -1,   -1,   -1,  277,   -1,   -1,   -1,  281,
   -1,   -1,  284,   -1,  362,   -1,   -1,   -1,   -1,   -1,
  368,  369,   -1,   -1,  296,  297,   -1,   -1,  418,  301,
  302,   -1,  261,   -1,   -1,  307,   -1,  309,  310,  311,
  312,   -1,   -1,  272,   -1,  317,   -1,   -1,  277,  321,
   -1,  323,  281,   -1,   -1,  284,   -1,   -1,   -1,   -1,
   -1,  333,   -1,   -1,  336,   -1,  338,  296,  297,   -1,
  342,   -1,  301,  302,   -1,  261,   -1,   -1,  307,   -1,
  309,  310,  311,  312,   -1,   -1,  272,   -1,  317,   -1,
  362,  277,  321,   -1,  323,  281,  368,  369,  284,   -1,
   -1,   -1,   -1,   -1,  333,   -1,   -1,  336,   -1,  338,
  296,  297,   -1,  342,   -1,  301,  302,  261,   -1,   -1,
   -1,  307,   -1,  309,  310,  311,  312,   -1,   -1,   -1,
   -1,  317,   -1,  362,   -1,  321,   -1,  323,   -1,  368,
  284,   -1,   -1,   -1,   -1,   -1,   -1,  333,   -1,   -1,
  336,   -1,  338,  297,   -1,  261,  342,  263,  302,   -1,
   -1,   -1,   -1,  307,   -1,  309,  310,  311,  312,   -1,
   -1,  315,   -1,  317,   -1,   -1,  362,  321,  284,   -1,
   -1,   -1,  368,   -1,   -1,   -1,   -1,   -1,   -1,  333,
   -1,  297,  336,  261,  338,   -1,  302,   -1,   -1,   -1,
   -1,  307,   -1,  309,  310,  311,  312,   -1,   -1,   -1,
   -1,  317,   -1,   -1,   -1,  321,  284,   -1,  362,   -1,
   -1,   -1,   -1,  261,  368,   -1,   -1,  333,   -1,  297,
  336,   -1,  338,   -1,  302,   -1,   -1,   -1,   -1,  307,
   -1,  309,  310,  311,  312,   -1,  284,   -1,   -1,  317,
   -1,   -1,   -1,  321,   -1,   -1,  362,   -1,   -1,  297,
   -1,   -1,  368,  301,  302,  333,  261,   -1,  336,  307,
  338,  309,  310,  311,  312,   -1,   -1,   -1,   -1,  317,
   -1,   -1,   -1,  321,   -1,   -1,   -1,   -1,   -1,  284,
   -1,   -1,   -1,   -1,  362,  333,  364,  365,  336,   -1,
  338,   -1,  297,   -1,   -1,   -1,   -1,  302,   -1,   -1,
   -1,   -1,  307,   -1,  309,  310,  311,  312,   -1,   -1,
   -1,   -1,  317,   -1,  362,   -1,  321,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,  333,   -1,
   -1,  336,   -1,  338,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,   -1,
   -1,   -1,   -1,   -1,   -1,   -1,   -1,  362,
  };

///  #line 6425 "cs-parser.jay"

// <summary>
//  A class used to hold info about an operator declarator
// </summary>
class OperatorDeclaration {
	public readonly Operator.OpType optype;
	public readonly FullNamedExpression ret_type;
	public readonly Location location;

	public OperatorDeclaration (Operator.OpType op, FullNamedExpression ret_type, Location location)
	{
		optype = op;
		this.ret_type = ret_type;
		this.location = location;
	}
}

void Error_ExpectingTypeName (Expression expr)
{
	if (expr is Invocation){
		report.Error (1002, expr.Location, "Expecting `;'");
	} else {
		Expression.Error_InvalidExpressionStatement (report, expr.Location);
	}
}

void Error_ParameterModifierNotValid (string modifier, Location loc)
{
	report.Error (631, loc, "The parameter modifier `{0}' is not valid in this context",
			              modifier);
}

void Error_DuplicateParameterModifier (Location loc, Parameter.Modifier mod)
{
	report.Error (1107, loc, "Duplicate parameter modifier `{0}'",
  		Parameter.GetModifierSignature (mod));
}

void Error_TypeExpected (Location loc)
{
	report.Error (1031, loc, "Type expected");
}

void Error_UnsafeCodeNotAllowed (Location loc)
{
	report.Error (227, loc, "Unsafe code requires the `unsafe' command line option to be specified");
}

void Warning_EmptyStatement (Location loc)
{
	report.Warning (642, 3, loc, "Possible mistaken empty statement");
}

void Error_NamedArgumentExpected (NamedArgument a)
{
	report.Error (1738, a.Location, "Named arguments must appear after the positional arguments");
}

void Error_MissingInitializer (Location loc)
{
	report.Error (210, loc, "You must provide an initializer in a fixed or using statement declaration");
}

void push_current_class (TypeContainer tc, object partial_token)
{
	if (module.Evaluator != null && current_container is ModuleContainer){
		tc.Definition.Modifiers = tc.ModFlags = (tc.ModFlags & ~Modifiers.AccessibilityMask) | Modifiers.PUBLIC;
		if (undo == null)
			undo = new Undo ();
		undo.AddTypeContainer (current_container, tc);
	}

	if (partial_token != null)
		current_container = current_container.AddPartial (tc);
	else
		current_container = current_container.AddTypeContainer (tc);

	++lexer.parsing_declaration;
	current_class = tc;
	ubag.PushTypeDeclaration (tc);
}

DeclSpace pop_current_class ()
{
	DeclSpace retval = current_class;

	current_class = current_class.Parent;
	current_container = current_class.PartialContainer;
	ubag.PopTypeDeclaration ();
	
	return retval;
}

// <summary>
//   Given the @class_name name, it creates a fully qualified name
//   based on the containing declaration space
// </summary>
MemberName
MakeName (MemberName class_name)
{
	Namespace ns = current_namespace.NS;

	if (current_container == module) {
		if (ns.Name.Length != 0)
			return new MemberName (ns.MemberName, class_name);
		else
			return class_name;
	} else {
		return new MemberName (current_container.MemberName, class_name);
	}
}

[System.Diagnostics.Conditional ("FULL_AST")]
void StoreModifierLocation (object token, Location loc)
{
	if (lbag == null)
		return;

	if (mod_locations == null)
  		mod_locations = new List<Tuple<Modifiers, Location>> ();

	mod_locations.Add (Tuple.Create ((Modifiers) token, loc));
}

List<Tuple<Modifiers, Location>> GetModifierLocations ()
{
	var result = mod_locations;
	mod_locations = null;
	return result;
}

string CheckAttributeTarget (string a, Location l)
{
	switch (a) {
	case "assembly" : case "module" : case "field" : case "method" : case "param" : case "property" : case "type" :
			return a;
	}

	report.Warning (658, 1, l,
		 "`{0}' is invalid attribute target. All attributes in this attribute section will be ignored", a);
	return string.Empty;
}

static bool IsUnaryOperator (Operator.OpType op)
{
	switch (op) {
		
	case Operator.OpType.LogicalNot: 
	case Operator.OpType.OnesComplement: 
	case Operator.OpType.Increment:
	case Operator.OpType.Decrement:
	case Operator.OpType.True: 
	case Operator.OpType.False: 
	case Operator.OpType.UnaryPlus: 
	case Operator.OpType.UnaryNegation:
		return true;
	}
	return false;
}

void syntax_error (Location l, string msg)
{
	report.Error (1003, l, "Syntax error, " + msg);
}

Tokenizer lexer;

public Tokenizer Lexer {
	get {
		return lexer;
	}
}		   

static CSharpParser ()
{
	oob_stack = new Stack<object> ();
}

public CSharpParser (SeekableStreamReader reader, CompilationSourceFile file)
	: this (reader, file, file.NamespaceContainer.Module.Compiler.Report)
{
}

public CSharpParser (SeekableStreamReader reader, CompilationSourceFile file, Report report)
{
	this.file = file;
	current_namespace = file.NamespaceContainer;
	
	this.module = current_namespace.Module;
	this.compiler = module.Compiler;
	this.settings = compiler.Settings;
	this.report = report;
	
	lang_version = settings.Version;
	doc_support = settings.DocumentationFile != null;
	current_class = current_namespace.SlaveDeclSpace;
	current_container = current_class.PartialContainer; // == RootContest.ToplevelTypes
	oob_stack.Clear ();
	lexer = new Tokenizer (reader, file, compiler);
	
	use_global_stacks = true;
}

public void parse ()
{
	eof_token = Token.EOF;
	Tokenizer.LocatedToken.Initialize ();
	
	try {
		if (yacc_verbose_flag > 1)
			yyparse (lexer, new yydebug.yyDebugSimple ());
		else
			yyparse (lexer);
			
		Tokenizer tokenizer = lexer as Tokenizer;
		tokenizer.cleanup ();		
	} catch (Exception e){
	  	if (e is yyParser.yyUnexpectedEof) {
			Error_SyntaxError (yyToken);
			UnexpectedEOF = true;
			return;
		}
			
		if (e is yyParser.yyException) {
			report.Error (-25, lexer.Location, "Parsing error");
		} else {
			// Used by compiler-tester to test internal errors
			if (yacc_verbose_flag > 0)
				throw;
		
			report.Error (589, lexer.Location, "Internal compiler error during parsing");
		}
	}
}

void CheckToken (int error, int yyToken, string msg, Location loc)
{
	if (yyToken >= Token.FIRST_KEYWORD && yyToken <= Token.LAST_KEYWORD)
		report.Error (error, loc, "{0}: `{1}' is a keyword", msg, GetTokenName (yyToken));
	else
		report.Error (error, loc, msg);
}

string ConsumeStoredComment ()
{
	string s = tmpComment;
	tmpComment = null;
	Lexer.doc_state = XmlCommentState.Allowed;
	return s;
}

void FeatureIsNotAvailable (Location loc, string feature)
{
	report.FeatureIsNotAvailable (compiler, loc, feature);
}

Location GetLocation (object obj)
{
	var lt = obj as Tokenizer.LocatedToken;
	if (lt != null)
		return lt.Location;
		
	var mn = obj as MemberName;
	if (mn != null)
		return mn.Location;
		
	var expr = obj as Expression;
	if (expr != null)
		return expr.Location;

	return lexer.Location;
}

public LocationsBag LocationsBag {
	get {
		return lbag;
	}
	set {
		lbag = value;
	}
}

public UsingsBag UsingsBag {
	get {
		return ubag;
	}
	set {
		ubag = value;
	}
}

void start_block (Location loc)
{
	if (current_block == null) {
		current_block = new ToplevelBlock (compiler, current_local_parameters, loc);
		parsing_anonymous_method = false;
	} else if (parsing_anonymous_method) {
		current_block = new ParametersBlock (current_block, current_local_parameters, loc);
		parsing_anonymous_method = false;
	} else {
		current_block = new ExplicitBlock (current_block, loc, Location.Null);
	}
}

Block
end_block (Location loc)
{
	Block retval = current_block.Explicit;
	retval.SetEndLocation (loc);
	current_block = retval.Parent;
	return retval;
}

void start_anonymous (bool isLambda, ParametersCompiled parameters, bool isAsync, Location loc)
{
	oob_stack.Push (current_anonymous_method);
	oob_stack.Push (current_local_parameters);
	oob_stack.Push (current_variable);
	oob_stack.Push (async_block);

	current_local_parameters = parameters;
	if (isLambda) {
		if (lang_version <= LanguageVersion.ISO_2)
			FeatureIsNotAvailable (loc, "lambda expressions");

		current_anonymous_method = new LambdaExpression (isAsync, loc);
	} else {
		if (lang_version == LanguageVersion.ISO_1)
			FeatureIsNotAvailable (loc, "anonymous methods");
			
		current_anonymous_method = new AnonymousMethodExpression (isAsync, loc);
	}

	async_block = isAsync;
	// Force the next block to be created as a ToplevelBlock
	parsing_anonymous_method = true;
}

/*
 * Completes the anonymous method processing, if lambda_expr is null, this
 * means that we have a Statement instead of an Expression embedded 
 */
AnonymousMethodExpression end_anonymous (ParametersBlock anon_block)
{
	AnonymousMethodExpression retval;

	if (async_block)
		anon_block.IsAsync = true;

	current_anonymous_method.Block = anon_block;
	retval = current_anonymous_method;

	async_block = (bool) oob_stack.Pop ();
	current_variable = (BlockVariableDeclaration) oob_stack.Pop ();
	current_local_parameters = (ParametersCompiled) oob_stack.Pop ();
	current_anonymous_method = (AnonymousMethodExpression) oob_stack.Pop ();

	return retval;
}

void Error_SyntaxError (int token)
{
	Error_SyntaxError (0, token, "Unexpected symbol");
}

void Error_SyntaxError (int error_code, int token, string msg)
{
	Lexer.CompleteOnEOF = false;

	// An error message has been reported by tokenizer
	if (token == Token.ERROR)
		return;

	string symbol = GetSymbolName (token);
	string expecting = GetExpecting ();
	var loc = lexer.Location - symbol.Length;
	
	if (error_code == 0) {
		if (expecting == "`identifier'") {
			if (token > Token.FIRST_KEYWORD && token < Token.LAST_KEYWORD) {
				report.Error (1041, loc, "Identifier expected, `{0}' is a keyword", symbol);
				return;
			}
			
			error_code = 1001;
			expecting = "identifier";
		} else if (expecting == "`)'") {
			error_code = 1026;
		} else {
			error_code = 1525;
		}
	}
	
	if (string.IsNullOrEmpty (expecting))
		report.Error (error_code, loc, "{1} `{0}'", symbol, msg);
	else
		report.Error (error_code, loc, "{2} `{0}', expecting {1}", symbol, expecting, msg);	  
}

string GetExpecting ()
{
	int [] tokens = yyExpectingTokens (yyExpectingState);
	var names = new List<string> (tokens.Length);
	bool has_type = false;
	bool has_identifier = false;
	for (int i = 0; i < tokens.Length; i++){
		int token = tokens [i];
		has_identifier |= token == Token.IDENTIFIER;
		
		string name = GetTokenName (token);
		if (name == "<internal>")
			continue;
			
		has_type |= name == "type";
		if (names.Contains (name))
			continue;
		
		names.Add (name);
	}

	//
	// Too many tokens to enumerate
	//
	if (names.Count > 8)
		return null;

	if (has_type && has_identifier)
		names.Remove ("identifier");

	if (names.Count == 1)
		return "`" + GetTokenName (tokens [0]) + "'";
	
	StringBuilder sb = new StringBuilder ();
	names.Sort ();
	int count = names.Count;
	for (int i = 0; i < count; i++){
		bool last = i + 1 == count;
		if (last)
			sb.Append ("or ");
		sb.Append ('`');
		sb.Append (names [i]);
		sb.Append (last ? "'" : count < 3 ? "' " : "', ");
	}
	return sb.ToString ();
}


string GetSymbolName (int token)
{
	switch (token){
	case Token.LITERAL:
		return ((Constant)lexer.Value).GetValue ().ToString ();
	case Token.IDENTIFIER:
		return ((Tokenizer.LocatedToken)lexer.Value).Value;

	case Token.BOOL:
		return "bool";
	case Token.BYTE:
		return "byte";
	case Token.CHAR:
		return "char";
	case Token.VOID:
		return "void";
	case Token.DECIMAL:
		return "decimal";
	case Token.DOUBLE:
		return "double";
	case Token.FLOAT:
		return "float";
	case Token.INT:
		return "int";
	case Token.LONG:
		return "long";
	case Token.SBYTE:
		return "sbyte";
	case Token.SHORT:
		return "short";
	case Token.STRING:
		return "string";
	case Token.UINT:
		return "uint";
	case Token.ULONG:
		return "ulong";
	case Token.USHORT:
		return "ushort";
	case Token.OBJECT:
		return "object";
		
	case Token.PLUS:
		return "+";
	case Token.UMINUS:
	case Token.MINUS:
		return "-";
	case Token.BANG:
		return "!";
	case Token.BITWISE_AND:
		return "&";
	case Token.BITWISE_OR:
		return "|";
	case Token.STAR:
		return "*";
	case Token.PERCENT:
		return "%";
	case Token.DIV:
		return "/";
	case Token.CARRET:
		return "^";
	case Token.OP_INC:
		return "++";
	case Token.OP_DEC:
		return "--";
	case Token.OP_SHIFT_LEFT:
		return "<<";
	case Token.OP_SHIFT_RIGHT:
		return ">>";
	case Token.OP_LT:
		return "<";
	case Token.OP_GT:
		return ">";
	case Token.OP_LE:
		return "<=";
	case Token.OP_GE:
		return ">=";
	case Token.OP_EQ:
		return "==";
	case Token.OP_NE:
		return "!=";
	case Token.OP_AND:
		return "&&";
	case Token.OP_OR:
		return "||";
	case Token.OP_PTR:
		return "->";
	case Token.OP_COALESCING:	
		return "??";
	case Token.OP_MULT_ASSIGN:
		return "*=";
	case Token.OP_DIV_ASSIGN:
		return "/=";
	case Token.OP_MOD_ASSIGN:
		return "%=";
	case Token.OP_ADD_ASSIGN:
		return "+=";
	case Token.OP_SUB_ASSIGN:
		return "-=";
	case Token.OP_SHIFT_LEFT_ASSIGN:
		return "<<=";
	case Token.OP_SHIFT_RIGHT_ASSIGN:
		return ">>=";
	case Token.OP_AND_ASSIGN:
		return "&=";
	case Token.OP_XOR_ASSIGN:
		return "^=";
	case Token.OP_OR_ASSIGN:
		return "|=";
	}

	return GetTokenName (token);
}

static string GetTokenName (int token)
{
	switch (token){
	case Token.ABSTRACT:
		return "abstract";
	case Token.AS:
		return "as";
	case Token.ADD:
		return "add";
	case Token.ASYNC:
		return "async";
	case Token.BASE:
		return "base";
	case Token.BREAK:
		return "break";
	case Token.CASE:
		return "case";
	case Token.CATCH:
		return "catch";
	case Token.CHECKED:
		return "checked";
	case Token.CLASS:
		return "class";
	case Token.CONST:
		return "const";
	case Token.CONTINUE:
		return "continue";
	case Token.DEFAULT:
		return "default";
	case Token.DELEGATE:
		return "delegate";
	case Token.DO:
		return "do";
	case Token.ELSE:
		return "else";
	case Token.ENUM:
		return "enum";
	case Token.EVENT:
		return "event";
	case Token.EXPLICIT:
		return "explicit";
	case Token.EXTERN:
	case Token.EXTERN_ALIAS:
		return "extern";
	case Token.FALSE:
		return "false";
	case Token.FINALLY:
		return "finally";
	case Token.FIXED:
		return "fixed";
	case Token.FOR:
		return "for";
	case Token.FOREACH:
		return "foreach";
	case Token.GOTO:
		return "goto";
	case Token.IF:
		return "if";
	case Token.IMPLICIT:
		return "implicit";
	case Token.IN:
		return "in";
	case Token.INTERFACE:
		return "interface";
	case Token.INTERNAL:
		return "internal";
	case Token.IS:
		return "is";
	case Token.LOCK:
		return "lock";
	case Token.NAMESPACE:
		return "namespace";
	case Token.NEW:
		return "new";
	case Token.NULL:
		return "null";
	case Token.OPERATOR:
		return "operator";
	case Token.OUT:
		return "out";
	case Token.OVERRIDE:
		return "override";
	case Token.PARAMS:
		return "params";
	case Token.PRIVATE:
		return "private";
	case Token.PROTECTED:
		return "protected";
	case Token.PUBLIC:
		return "public";
	case Token.READONLY:
		return "readonly";
	case Token.REF:
		return "ref";
	case Token.RETURN:
		return "return";
	case Token.REMOVE:
		return "remove";
	case Token.SEALED:
		return "sealed";
	case Token.SIZEOF:
		return "sizeof";
	case Token.STACKALLOC:
		return "stackalloc";
	case Token.STATIC:
		return "static";
	case Token.STRUCT:
		return "struct";
	case Token.SWITCH:
		return "switch";
	case Token.THIS:
		return "this";
	case Token.THROW:
		return "throw";
	case Token.TRUE:
		return "true";
	case Token.TRY:
		return "try";
	case Token.TYPEOF:
		return "typeof";
	case Token.UNCHECKED:
		return "unchecked";
	case Token.UNSAFE:
		return "unsafe";
	case Token.USING:
		return "using";
	case Token.VIRTUAL:
		return "virtual";
	case Token.VOLATILE:
		return "volatile";
	case Token.WHERE:
		return "where";
	case Token.WHILE:
		return "while";
	case Token.ARGLIST:
		return "__arglist";
	case Token.REFVALUE:
		return "__refvalue";
	case Token.REFTYPE:
		return "__reftype";
	case Token.MAKEREF:
		return "__makeref";
	case Token.PARTIAL:
		return "partial";
	case Token.ARROW:
		return "=>";
	case Token.FROM:
	case Token.FROM_FIRST:
		return "from";
	case Token.JOIN:
		return "join";
	case Token.ON:
		return "on";
	case Token.EQUALS:
		return "equals";
	case Token.SELECT:
		return "select";
	case Token.GROUP:
		return "group";
	case Token.BY:
		return "by";
	case Token.LET:
		return "let";
	case Token.ORDERBY:
		return "orderby";
	case Token.ASCENDING:
		return "ascending";
	case Token.DESCENDING:
		return "descending";
	case Token.INTO:
		return "into";
	case Token.GET:
		return "get";
	case Token.SET:
		return "set";
	case Token.OPEN_BRACE:
		return "{";
	case Token.CLOSE_BRACE:
		return "}";
	case Token.OPEN_BRACKET:
	case Token.OPEN_BRACKET_EXPR:
		return "[";
	case Token.CLOSE_BRACKET:
		return "]";
	case Token.OPEN_PARENS_CAST:
	case Token.OPEN_PARENS_LAMBDA:
	case Token.OPEN_PARENS:
		return "(";
	case Token.CLOSE_PARENS:
		return ")";
	case Token.DOT:
		return ".";
	case Token.COMMA:
		return ",";
	case Token.DEFAULT_COLON:
		return "default:";
	case Token.COLON:
		return ":";
	case Token.SEMICOLON:
		return ";";
	case Token.TILDE:
		return "~";
		
	case Token.PLUS:
	case Token.UMINUS:
	case Token.MINUS:
	case Token.BANG:
	case Token.OP_LT:
	case Token.OP_GT:
	case Token.BITWISE_AND:
	case Token.BITWISE_OR:
	case Token.STAR:
	case Token.PERCENT:
	case Token.DIV:
	case Token.CARRET:
	case Token.OP_INC:
	case Token.OP_DEC:
	case Token.OP_SHIFT_LEFT:
	case Token.OP_SHIFT_RIGHT:
	case Token.OP_LE:
	case Token.OP_GE:
	case Token.OP_EQ:
	case Token.OP_NE:
	case Token.OP_AND:
	case Token.OP_OR:
	case Token.OP_PTR:
	case Token.OP_COALESCING:	
	case Token.OP_MULT_ASSIGN:
	case Token.OP_DIV_ASSIGN:
	case Token.OP_MOD_ASSIGN:
	case Token.OP_ADD_ASSIGN:
	case Token.OP_SUB_ASSIGN:
	case Token.OP_SHIFT_LEFT_ASSIGN:
	case Token.OP_SHIFT_RIGHT_ASSIGN:
	case Token.OP_AND_ASSIGN:
	case Token.OP_XOR_ASSIGN:
	case Token.OP_OR_ASSIGN:
		return "<operator>";

	case Token.BOOL:
	case Token.BYTE:
	case Token.CHAR:
	case Token.VOID:
	case Token.DECIMAL:
	case Token.DOUBLE:
	case Token.FLOAT:
	case Token.INT:
	case Token.LONG:
	case Token.SBYTE:
	case Token.SHORT:
	case Token.STRING:
	case Token.UINT:
	case Token.ULONG:
	case Token.USHORT:
	case Token.OBJECT:
		return "type";
	
	case Token.ASSIGN:
		return "=";
	case Token.OP_GENERICS_LT:
	case Token.GENERIC_DIMENSION:
		return "<";
	case Token.OP_GENERICS_GT:
		return ">";
	case Token.INTERR:
	case Token.INTERR_NULLABLE:
		return "?";
	case Token.DOUBLE_COLON:
		return "::";
	case Token.LITERAL:
		return "value";
	case Token.IDENTIFIER:
	case Token.AWAIT:
		return "identifier";

	case Token.EOF:
		return "end-of-file";

		// All of these are internal.
	case Token.NONE:
	case Token.ERROR:
	case Token.FIRST_KEYWORD:
	case Token.EVAL_COMPILATION_UNIT_PARSER:
	case Token.EVAL_USING_DECLARATIONS_UNIT_PARSER:
	case Token.EVAL_STATEMENT_PARSER:
	case Token.LAST_KEYWORD:
	case Token.GENERATE_COMPLETION:
	case Token.COMPLETE_COMPLETION:
		return "<internal>";

		// A bit more robust.
	default:
		return yyNames [token];
        }
}

/* end end end */
}
///  #line default
namespace yydebug {
        using System;
	 internal interface yyDebug {
		 void push (int state, Object value);
		 void lex (int state, int token, string name, Object value);
		 void shift (int from, int to, int errorFlag);
		 void pop (int state);
		 void discard (int state, int token, string name, Object value);
		 void reduce (int from, int to, int rule, string text, int len);
		 void shift (int from, int to);
		 void accept (Object value);
		 void error (string message);
		 void reject ();
	 }
	 
	 class yyDebugSimple : yyDebug {
		 void println (string s){
			 Console.Error.WriteLine (s);
		 }
		 
		 public void push (int state, Object value) {
			 println ("push\tstate "+state+"\tvalue "+value);
		 }
		 
		 public void lex (int state, int token, string name, Object value) {
			 println("lex\tstate "+state+"\treading "+name+"\tvalue "+value);
		 }
		 
		 public void shift (int from, int to, int errorFlag) {
			 switch (errorFlag) {
			 default:				// normally
				 println("shift\tfrom state "+from+" to "+to);
				 break;
			 case 0: case 1: case 2:		// in error recovery
				 println("shift\tfrom state "+from+" to "+to
					     +"\t"+errorFlag+" left to recover");
				 break;
			 case 3:				// normally
				 println("shift\tfrom state "+from+" to "+to+"\ton error");
				 break;
			 }
		 }
		 
		 public void pop (int state) {
			 println("pop\tstate "+state+"\ton error");
		 }
		 
		 public void discard (int state, int token, string name, Object value) {
			 println("discard\tstate "+state+"\ttoken "+name+"\tvalue "+value);
		 }
		 
		 public void reduce (int from, int to, int rule, string text, int len) {
			 println("reduce\tstate "+from+"\tuncover "+to
				     +"\trule ("+rule+") "+text);
		 }
		 
		 public void shift (int from, int to) {
			 println("goto\tfrom state "+from+" to "+to);
		 }
		 
		 public void accept (Object value) {
			 println("accept\tvalue "+value);
		 }
		 
		 public void error (string message) {
			 println("error\t"+message);
		 }
		 
		 public void reject () {
			 println("reject");
		 }
		 
	 }
}
// %token constants
 class Token {
  public const int EOF = 257;
  public const int NONE = 258;
  public const int ERROR = 259;
  public const int FIRST_KEYWORD = 260;
  public const int ABSTRACT = 261;
  public const int AS = 262;
  public const int ADD = 263;
  public const int BASE = 264;
  public const int BOOL = 265;
  public const int BREAK = 266;
  public const int BYTE = 267;
  public const int CASE = 268;
  public const int CATCH = 269;
  public const int CHAR = 270;
  public const int CHECKED = 271;
  public const int CLASS = 272;
  public const int CONST = 273;
  public const int CONTINUE = 274;
  public const int DECIMAL = 275;
  public const int DEFAULT = 276;
  public const int DELEGATE = 277;
  public const int DO = 278;
  public const int DOUBLE = 279;
  public const int ELSE = 280;
  public const int ENUM = 281;
  public const int EVENT = 282;
  public const int EXPLICIT = 283;
  public const int EXTERN = 284;
  public const int FALSE = 285;
  public const int FINALLY = 286;
  public const int FIXED = 287;
  public const int FLOAT = 288;
  public const int FOR = 289;
  public const int FOREACH = 290;
  public const int GOTO = 291;
  public const int IF = 292;
  public const int IMPLICIT = 293;
  public const int IN = 294;
  public const int INT = 295;
  public const int INTERFACE = 296;
  public const int INTERNAL = 297;
  public const int IS = 298;
  public const int LOCK = 299;
  public const int LONG = 300;
  public const int NAMESPACE = 301;
  public const int NEW = 302;
  public const int NULL = 303;
  public const int OBJECT = 304;
  public const int OPERATOR = 305;
  public const int OUT = 306;
  public const int OVERRIDE = 307;
  public const int PARAMS = 308;
  public const int PRIVATE = 309;
  public const int PROTECTED = 310;
  public const int PUBLIC = 311;
  public const int READONLY = 312;
  public const int REF = 313;
  public const int RETURN = 314;
  public const int REMOVE = 315;
  public const int SBYTE = 316;
  public const int SEALED = 317;
  public const int SHORT = 318;
  public const int SIZEOF = 319;
  public const int STACKALLOC = 320;
  public const int STATIC = 321;
  public const int STRING = 322;
  public const int STRUCT = 323;
  public const int SWITCH = 324;
  public const int THIS = 325;
  public const int THROW = 326;
  public const int TRUE = 327;
  public const int TRY = 328;
  public const int TYPEOF = 329;
  public const int UINT = 330;
  public const int ULONG = 331;
  public const int UNCHECKED = 332;
  public const int UNSAFE = 333;
  public const int USHORT = 334;
  public const int USING = 335;
  public const int VIRTUAL = 336;
  public const int VOID = 337;
  public const int VOLATILE = 338;
  public const int WHERE = 339;
  public const int WHILE = 340;
  public const int ARGLIST = 341;
  public const int PARTIAL = 342;
  public const int ARROW = 343;
  public const int FROM = 344;
  public const int FROM_FIRST = 345;
  public const int JOIN = 346;
  public const int ON = 347;
  public const int EQUALS = 348;
  public const int SELECT = 349;
  public const int GROUP = 350;
  public const int BY = 351;
  public const int LET = 352;
  public const int ORDERBY = 353;
  public const int ASCENDING = 354;
  public const int DESCENDING = 355;
  public const int INTO = 356;
  public const int INTERR_NULLABLE = 357;
  public const int EXTERN_ALIAS = 358;
  public const int REFVALUE = 359;
  public const int REFTYPE = 360;
  public const int MAKEREF = 361;
  public const int ASYNC = 362;
  public const int AWAIT = 363;
  public const int GET = 364;
  public const int SET = 365;
  public const int LAST_KEYWORD = 366;
  public const int OPEN_BRACE = 367;
  public const int CLOSE_BRACE = 368;
  public const int OPEN_BRACKET = 369;
  public const int CLOSE_BRACKET = 370;
  public const int OPEN_PARENS = 371;
  public const int CLOSE_PARENS = 372;
  public const int DOT = 373;
  public const int COMMA = 374;
  public const int COLON = 375;
  public const int SEMICOLON = 376;
  public const int TILDE = 377;
  public const int PLUS = 378;
  public const int MINUS = 379;
  public const int BANG = 380;
  public const int ASSIGN = 381;
  public const int OP_LT = 382;
  public const int OP_GT = 383;
  public const int BITWISE_AND = 384;
  public const int BITWISE_OR = 385;
  public const int STAR = 386;
  public const int PERCENT = 387;
  public const int DIV = 388;
  public const int CARRET = 389;
  public const int INTERR = 390;
  public const int DOUBLE_COLON = 391;
  public const int OP_INC = 392;
  public const int OP_DEC = 393;
  public const int OP_SHIFT_LEFT = 394;
  public const int OP_SHIFT_RIGHT = 395;
  public const int OP_LE = 396;
  public const int OP_GE = 397;
  public const int OP_EQ = 398;
  public const int OP_NE = 399;
  public const int OP_AND = 400;
  public const int OP_OR = 401;
  public const int OP_MULT_ASSIGN = 402;
  public const int OP_DIV_ASSIGN = 403;
  public const int OP_MOD_ASSIGN = 404;
  public const int OP_ADD_ASSIGN = 405;
  public const int OP_SUB_ASSIGN = 406;
  public const int OP_SHIFT_LEFT_ASSIGN = 407;
  public const int OP_SHIFT_RIGHT_ASSIGN = 408;
  public const int OP_AND_ASSIGN = 409;
  public const int OP_XOR_ASSIGN = 410;
  public const int OP_OR_ASSIGN = 411;
  public const int OP_PTR = 412;
  public const int OP_COALESCING = 413;
  public const int OP_GENERICS_LT = 414;
  public const int OP_GENERICS_LT_DECL = 415;
  public const int OP_GENERICS_GT = 416;
  public const int LITERAL = 417;
  public const int IDENTIFIER = 418;
  public const int OPEN_PARENS_LAMBDA = 419;
  public const int OPEN_PARENS_CAST = 420;
  public const int GENERIC_DIMENSION = 421;
  public const int DEFAULT_COLON = 422;
  public const int OPEN_BRACKET_EXPR = 423;
  public const int EVAL_STATEMENT_PARSER = 424;
  public const int EVAL_COMPILATION_UNIT_PARSER = 425;
  public const int EVAL_USING_DECLARATIONS_UNIT_PARSER = 426;
  public const int DOC_SEE = 427;
  public const int GENERATE_COMPLETION = 428;
  public const int COMPLETE_COMPLETION = 429;
  public const int UMINUS = 430;
  public const int yyErrorCode = 256;
 }
 namespace yyParser {
  using System;
  /** thrown for irrecoverable syntax errors and stack overflow.
    */
  internal class yyException : System.Exception {
    public yyException (string message) : base (message) {
    }
  }
  internal class yyUnexpectedEof : yyException {
    public yyUnexpectedEof (string message) : base (message) {
    }
    public yyUnexpectedEof () : base ("") {
    }
  }

  /** must be implemented by a scanner object to supply input to the parser.
    */
  internal interface yyInput {
    /** move on to next token.
        @return false if positioned beyond tokens.
        @throws IOException on input error.
      */
    bool advance (); // throws java.io.IOException;
    /** classifies current token.
        Should not be called if advance() returned false.
        @return current %token or single character.
      */
    int token ();
    /** associated with current token.
        Should not be called if advance() returned false.
        @return value for token().
      */
    Object value ();
  }
 }
} // close outermost namespace, that MUST HAVE BEEN opened in the prolog
